/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "backend-vdcs.h"

#include <ds/asio/retryer.h>
#include <ds/elapsed.h>
#include <ds/json.h>
#include <ds/locale/ds-locale.h>
#include <ds/log.h>

#include <boost/algorithm/string.hpp>
#include <boost/range/adaptor/map.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <boost/serialization/strong_typedef.hpp>

#include "businterface.h"
#include "ds485types.h"
#include "event.h"
#include "handler/system_states.h"
#include "http_client.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "model/state.h"
#include "propertysystem_common_paths.h"
#include "scanner/vdcdataquery.h"

static ds::LogChannel dsLogChannel("backendVdcs");

using namespace std::literals;
namespace dss {

std::ostream& operator<<(std::ostream& ostream, const BackendVdcId& x) {
  return ostream << static_cast<const std::string&>(x);
}

std::ostream& operator<<(std::ostream& ostream, const BackendVdcDeviceId& x) {
  return ostream << static_cast<const std::string&>(x);
}

static BackendVdcId HEKATRON{"hekatron"};

class BackendVdcsImpl final : public ds::asio::Resource<>,
                              public BackendVdcs,
                              private PropertyListener,
                              public RegisteredEventInterpreterPlugin,
                              public StructureQueryBusInterface,
                              public StructureModifyingBusInterface {
public:
  struct VdcDevice {
    VdsdSpec_t spec;
  };

  struct Vdc {
    VdcSpec spec;
    std::map<BackendVdcDeviceId, VdcDevice> devices;
    devid_t lastShortAddress;

    Vdc() : spec{}, devices{}, lastShortAddress{} {}
  };

  BackendVdcsImpl(ds::asio::ResourceBase& parent, Apartment& apartment, EventInterpreter& eventInterpreter,
                  SimpleHttpClient& httpClient)
      : Resource<>{parent, "backendVdcs"},
        RegisteredEventInterpreterPlugin("BackendVdcs", eventInterpreter),
        m_apartment(apartment),
        m_httpClient(httpClient) {
    DS_INFO_ENTER("");
    m_apartment.getBusInterface()->add(*this);
    m_enabledNode = propertySystem().getProperty(pp_websvc_mshub_active);
    m_enabledNode->addListener(this);

    eventInterpreter.subscribe(*this, EventName::StateChange);

    DS_DEBUG_LEAVE("");
  }

  ~BackendVdcsImpl() {
    DS_INFO_ENTER("");
    m_enabledNode->removeListener(this);
    DS_DEBUG_LEAVE("");
  }

  void propertyChanged(PropertyNodePtr, PropertyNodePtr node) override {
    // TODO(soon): thread synchronization
    DS_DEBUG_ENTER(node->getName());
    if (node == m_enabledNode) {
      reregister();
    }
  }

  void handleEvent(Event& event, const EventSubscription&) override {
    // TODO(soon): thread synchronization
    DS_DEBUG_ENTER(event);

    // notify hekatron vdc on active -> inactive || active -> activeMuted transition
    //
    // TODO(someday): Give cloud backendVdcs option to subscribe
    // to events and move this logic there.
    try {
      if (m_vdcs.find(HEKATRON) != m_vdcs.end() && event.getName() == EventName::StateChange &&
          event.getPropertyByName("statename") == StateName::Fire) {
        const auto& oldValue = event.getPropertyByName("oldvalue");
        if (oldValue == "1") {
          DS_INFO("Notify hekatron mute fire.", event);
          http(HttpRequest{.url = ds::str("/v1/cloudvdc/vdcs/hekatron/action/muteFire"), .type = HttpMethod::POST});
        }
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to notify hekatron fire mute", e);
    }
  }

  void notify(const std::string& path, ResourceNotification&& notification) override {
    // TODO(soon): thread synchronization
    DS_INFO_ENTER(path);
    DS_REQUIRE(hasBackend(), "Dropping notification to vDC. Backend not ready yet.");

    std::vector<std::string> pathParts;
    boost::split(pathParts, path, boost::is_any_of("/"));
    DS_REQUIRE(pathParts.size() > 0);
    pathParts.erase(pathParts.begin());

    if (pathParts.size() >= 2 && pathParts.at(0) == "vdcs") {
      const auto& vdcId = BackendVdcId{pathParts.at(1)};
      if (pathParts.size() == 2) {
        notifyVdc(vdcId, std::move(notification));
        return;
      } else if (pathParts.size() >= 4 && pathParts.at(2) == "devices") {
        const auto& vdcDeviceId = BackendVdcDeviceId{pathParts.at(3)};
        if (pathParts.size() == 4) {
          notifyVdcDevice(vdcId, vdcDeviceId, std::move(notification));
          return;
        }
      }
    }
    DS_FAIL_REQUIRE("Unsupported notify path.", path);
  }

  void notifyVdc(const BackendVdcId& vdcId, ResourceNotification&& notification) {
    DS_DEBUG_ENTER(vdcId);
    if (auto put = std::get_if<PutResource>(&notification)) {
      putVdc(vdcId, ds::fromJson<VdcSpec>(put->value));
      return;
    } else if (std::get_if<DeleteResource>(&notification)) {
      deleteVdc(vdcId);
      return;
    } else if (std::get_if<PatchResource>(&notification)) {
      DS_FAIL_REQUIRE("PATCH vdc not supported (yet).");
    }
    DS_FAIL_REQUIRE("Unsupported notify.", vdcId);
  }

  void notifyVdcDevice(const BackendVdcId& vdcId, const BackendVdcDeviceId& id, ResourceNotification&& notification) {
    DS_DEBUG_ENTER(vdcId, id);
    if (auto put = std::get_if<PutResource>(&notification)) {
      putVdcDevice(vdcId, id, ds::fromJson<VdsdSpec_t>(put->value));
      return;
    } else if (std::get_if<DeleteResource>(&notification)) {
      deleteVdcDevice(vdcId, id);
      return;
    } else if (std::get_if<PatchResource>(&notification)) {
      DS_FAIL_REQUIRE("PATCH vdc device not supported (yet).");
    }
    DS_FAIL_REQUIRE("Unsupported notify.", vdcId, id);
  }

  void putVdc(const BackendVdcId& id, VdcSpec&& specIn) {
    DS_DEBUG_ENTER(id, specIn);
    DS_NOTICE("Put vdc", id, specIn);
    specIn.implementationId = static_cast<const std::string&>(id);
    auto& vdc = m_vdcs[id];
    m_dsuidToVdc.erase(vdc.spec.dSUID);
    vdc.spec = std::move(specIn);
    m_dsuidToVdc[vdc.spec.dSUID] = id;
    const auto& spec = vdc.spec;
    auto dsm = m_apartment.allocateDSMeter(str2dsuid(spec.dSUID));
    dsm->setBusMemberType(BusMember_backendVdc);
    dsm->setCapability_HasDevices(true);
    dsm->setCapability_HasMetering(spec.capabilities.metering);
    dsm->setCapability_HasBlinking(spec.capabilities.identification);
    dsm->setCapability_HasDynamicDefinitions(spec.capabilities.dynamicDefinitions);
    dsm->setHardwareVersion(spec.hardwareVersion);
    dsm->setHardwareName(spec.model);
    dsm->setSoftwareVersion(spec.modelVersion);
    dsm->setVdcModelUID(spec.modelUID);
    dsm->setVdcHardwareGuid(spec.hardwareGuid);
    dsm->setVdcHardwareModelGuid(spec.hardwareModelGuid);
    dsm->setVdcVendorGuid(spec.vendorGuid);
    dsm->setVdcOemGuid(spec.oemGuid);
    dsm->setVdcOemModelGuid(spec.oemModelGuid);
    dsm->setVdcConfigURL(spec.configURL);
    if (dsm->getName().empty()) {
      dsm->setName(spec.name);
    }
    dsm->setVdcImplementationId(spec.implementationId);
    dsm->setAuthorized(spec.authorized);
    dsm->setIsPresent(true);
    dsm->setIsConnected(true);
    dsm->setIsValid(true);
  }

  void deleteVdc(const BackendVdcId& id) {
    DS_DEBUG_ENTER(id);
    DS_NOTICE("Delete vdc", id);
    const auto& vdcIt = m_vdcs.find(id);
    if (vdcIt != m_vdcs.end()) {
      const auto& vdc = vdcIt->second;
      const auto& spec = vdc.spec;
      if (const auto& dsm = m_apartment.tryGetDSMeterByDSID(str2dsuid(spec.dSUID))) {
        dsm->setIsPresent(false);
      }
      m_vdcs.erase(vdcIt);
    } else {
      DS_WARNING("Vdc to delete not found.", id);
    }
  }

  Vdc& vdc(const BackendVdcId& vdcId) {
    const auto& it = m_vdcs.find(vdcId);
    DS_REQUIRE(it != m_vdcs.end(), "Unknown vdc.", vdcId);
    return it->second;
  }

  void putVdcDevice(const BackendVdcId& vdcId, const BackendVdcDeviceId& id, VdsdSpec_t&& specIn) {
    DS_DEBUG_ENTER(vdcId, id, specIn);
    auto& vdc = this->vdc(vdcId);

    const bool wasReady = vdc.devices.find(id) != vdc.devices.end();
    auto& device = vdc.devices[id];
    m_deviceDsuidToVdc.erase(device.spec.dSUID);
    device.spec = specIn;
    m_deviceDsuidToVdc[device.spec.dSUID] = vdcId;
    const auto& spec = device.spec;
    DS_NOTICE("Put vdc device", vdcId, id, spec);
    const auto& dsm = m_apartment.getDSMeterByDSID(str2dsuid(vdc.spec.dSUID));
    auto dsuid = str2dsuid(spec.dSUID);
    auto modelDevice = m_apartment.tryGetDeviceByDSID(dsuid);
    bool isAdded = false;
    if (!modelDevice) {
      modelDevice = m_apartment.allocateDevice(str2dsuid(spec.dSUID));
      isAdded = true;
    }
    bool vdcDataQueryNeeded = modelDevice->getVdsdSpec().oemModelGuid.empty() && !specIn.oemModelGuid.empty();
    modelDevice->setVdsdSpec(std::move(specIn));
    modelDevice->setVdcDevice(true);
    if (modelDevice->getShortAddress() == ShortAddressStaleDevice) {
      modelDevice->setShortAddress(vdc.lastShortAddress++);
    }
    modelDevice->setDSMeter(dsm);
    if (modelDevice->getName().empty()) { // dss is data owner
      modelDevice->setName(spec.name);
    }
    modelDevice->addToGroup(spec.primaryGroup);
    modelDevice->setActiveGroup(spec.primaryGroup);
    modelDevice->setIsPresent(spec.active);
    modelDevice->setIsConnected(spec.active);
    modelDevice->setButtonInputMode(ButtonInputMode::DEACTIVATED);

    // Always valid. Read is atomic in backendVdc
    modelDevice->setIsValid(true);
    dsm->addDevice(modelDevice->reference());

    // Binary inputs.
    //
    // For simplicity, all binary input arrays need to be in the same order
    // and ordered by dSIndex
    //
    // Binary input settings and descriptions are applied only once for simplicy.
    // Device->setBinaryInputs has side effects when called again
    auto binaryInputsSize = spec.binaryInputSettings.size();
    if (modelDevice->getBinaryInputCount() != binaryInputsSize) {
      std::vector<DeviceBinaryInputSpec_t> deviceBinaryInputs;
      for (uint i = 0; i < binaryInputsSize; i++) {
        deviceBinaryInputs.push_back(DeviceBinaryInputSpec_t{});
        auto& x = deviceBinaryInputs.back();
        x.TargetGroup = spec.binaryInputSettings[i].second.group;
        x.InputType = static_cast<BinaryInputType>(spec.binaryInputDescriptions[i].second.sensorFunction);
        x.InputID = static_cast<BinaryInputId>(15); // Avoids magic numbers in property tree.
        x.Independent = true;
      }
      modelDevice->setBinaryInputs(deviceBinaryInputs);
      DS_REQUIRE(binaryInputsSize == spec.binaryInputDescriptions.size());
      for (size_t i = 0; i < binaryInputsSize; i++) {
        const auto& description = spec.binaryInputDescriptions.at(i);
        DS_REQUIRE(spec.binaryInputSettings.at(i).first == description.first,
                   "setting and description ids don't match 1:1");
        DS_REQUIRE(description.second.dSIndex == int(i), "descriptions must be in dSIndex order");
        modelDevice->setBinaryInputsInfo(description.second.dSIndex, description.first, description.second.name);
      }
    }
    DS_REQUIRE(binaryInputsSize == spec.binaryInputStates.size());
    for (size_t i = 0; i < binaryInputsSize; i++) {
      const auto& state = spec.binaryInputStates.at(i);
      DS_REQUIRE(spec.binaryInputSettings.at(i).first == state.first, "setting and states ids don't match 1:1");
      const auto& stateValue = static_cast<BinaryInputStateValue>(state.second.value);
      modelDevice->handleBinaryInputEvent(i, stateValue);
    }

    if (vdcDataQueryNeeded && m_apartment.getModelMaintenance()) {
      TaskProcessor& tp = m_apartment.getModelMaintenance()->getTaskProcessor();
      tp.addEvent(std::make_shared<VdcDataQuery>(modelDevice, *m_apartment.getPropertySystem()));
    }

    if (!wasReady) {
      m_apartment.pushEvent(
          std::make_shared<Event>(EventName::DeviceEvent, modelDevice->reference(), Properties{{"action", "ready"}}));
    }
    if (isAdded) {
      m_apartment.pushEvent(
          std::make_shared<Event>(EventName::DeviceEvent, modelDevice->reference(), Properties{{"action", "added"}}));
    }

    // Hekatron device presence enables mute fire support within dss.
    if (vdcId == HEKATRON) {
      const auto& origin = CallOrigin::json; // CallOrigin::system would be ignored by system_states code
      m_apartment.getNonScriptState(StateName::FireMuteEnabled)->setState(origin, State_Active);
    }
    DS_DEBUG_LEAVE(vdcId, id, specIn);
  }

  void deleteVdcDevice(const BackendVdcId& vdcId, const BackendVdcDeviceId& id) {
    DS_DEBUG_ENTER(vdcId, id);
    DS_NOTICE("Delete vdc device", vdcId, id);
    auto& vdc = this->vdc(vdcId);
    auto& devices = vdc.devices;
    const auto& deviceIt = devices.find(id);
    if (deviceIt != devices.end()) {
      const auto& device = deviceIt->second;
      m_deviceDsuidToVdc.erase(device.spec.dSUID);
      if (const auto& modelDevice = m_apartment.tryGetDeviceByDSID(str2dsuid(device.spec.dSUID))) {
        modelDevice->setIsPresent(false);
      }
      devices.erase(deviceIt);
    } else {
      DS_WARNING("Vdc device to delete not found.", vdcId, id);
    }
    DS_DEBUG_LEAVE(vdcId, id);
  }

  void reregister(BackendVdcsReregisterHooks&& hooks = {}) override {
    DS_DEBUG_ENTER("");
    m_reregisterRetryer.asyncPerform({.attempt =
                                          [this] {
                                            updateVdcs();
                                            for (const auto& vdcEntry : m_vdcs) {
                                              updateVdcDevices(vdcEntry.first);
                                            }
                                          },
                                      .succeeded = std::move(hooks.succeeded),
                                      .attemptFailed = std::move(hooks.attemptFailed)});
  }

  void updateVdcs() {
    DS_CONTEXT("updateVdcs");
    DS_DEBUG_ENTER("");
    auto vdcs = [&] {
      if (hasBackend()) {
        return ds::fromJson<std::vector<std::pair<BackendVdcId, VdcSpec>>>(
            http(HttpRequest{.url = "/v1/cloudvdc/vdcs"}));
      } else {
        DS_NOTICE("Backend not available.");
        return std::vector<std::pair<BackendVdcId, VdcSpec>>{};
      }
    }();

    std::set<BackendVdcId> idsToDelete;
    boost::copy(m_vdcs | boost::adaptors::map_keys, std::inserter(idsToDelete, idsToDelete.end()));
    for (auto& vdc : vdcs) {
      const auto id = vdc.first;
      try {
        idsToDelete.erase(id);
        putVdc(id, std::move(vdc.second));
      } catch (const std::exception& e) {
        DS_ERROR("Failed to put vdc", id, e);
      }
    }
    for (auto& id : idsToDelete) {
      deleteVdc(id);
    }
  }

  void updateVdcDevices(const BackendVdcId& vdcId) {
    DS_CONTEXT("updateVdcDevices", vdcId);
    DS_DEBUG_ENTER(vdcId);
    auto& vdc = this->vdc(vdcId);
    auto devices = ds::fromJson<std::vector<VdsdSpec_t>>(
        http(HttpRequest{.url = ds::str("/v1/cloudvdc/vdcs/", vdcId, "/devices")}));

    std::set<BackendVdcDeviceId> idsToDelete;
    boost::copy(vdc.devices | boost::adaptors::map_keys, std::inserter(idsToDelete, idsToDelete.end()));
    for (auto& device : devices) {
      const auto id = BackendVdcDeviceId{device.dSUID};
      try {
        idsToDelete.erase(id);
        putVdcDevice(vdcId, id, std::move(device));
      } catch (const std::exception& e) {
        DS_ERROR("Failed to put vdc device", vdcId, id, e);
      }
    }
    for (auto& id : idsToDelete) {
      deleteVdcDevice(vdcId, id);
    }
  }

  void reregisterDevices(const dsuid_t& dsuid) override {
    DS_DEBUG_ENTER(dsuid);
    updateVdcDevices(vdcIdForDsuid(dsuid));
  }

  StructureQueryBusInterface& structureQueryBusInterface() override { return *this; }
  StructureModifyingBusInterface& structureModifyingBusInterface() override { return *this; }

  bool isBackendVdcDevice(const dsuid_t& device) const override {
    return m_apartment.dsMeterByDeviceDsuid(device)->getBusMemberType() == BusMember_backendVdc;
  }

  BackendVdcId vdcIdForDeviceDsuid(const dsuid_t& dsuid) const override {
    const auto& it = m_deviceDsuidToVdc.find(ds::str(dsuid));
    DS_REQUIRE(it != m_deviceDsuidToVdc.end(), "Unknown vdc device.", dsuid);
    return it->second;
  }

  BackendVdcId vdcIdForDsuid(const dsuid_t& dsuid) const override {
    DS_DEBUG_ENTER(dsuid, m_dsuidToVdc);
    const auto& it = m_dsuidToVdc.find(ds::str(dsuid));
    DS_REQUIRE(it != m_dsuidToVdc.end(), "Unknown vdc.", dsuid);
    return it->second;
  }

  std::string vdcDsuidForId(const BackendVdcId& vdcId) const override {
    const auto& it = m_vdcs.find(vdcId);
    if (it != m_vdcs.end()) {
      return it->second.spec.dSUID;
    }
    DS_FAIL_REQUIRE("Unknown vdc", vdcId);
  }

  ds::JsonFragment http(HttpRequest&& x, const std::optional<std::locale>& locale = {},
                        ds::LogSeverity logSeverity = ds::LogSeverity::NOTICE) const override {
    DS_DEBUG_ENTER(x);
    // PC build does not have remote connectivity available.
    // Ds-conf options `dssBackendVdcsDssId` and `dssBackendVdcsOspToken`
    // allow to mock values from remote connectivity add-on
    const auto& dssId = this->dssId();
    const auto& ospToken = this->ospToken();
    x.url = ds::str(baseUrl(), x.url);
    x.headers.insert({"x-ds-dss-id", dssId});
    x.headers.insert({"x-ds-osp-token", ospToken});
    if (locale) {
      x.headers.insert({"Accept-Language", ds::locale::httpLanguage(*locale)});
    }

    ds::Elapsed elapsed;
    auto response = m_httpClient.request(x);
    DS_LOG(logSeverity, "Http request", x.url, response, elapsed);
    DS_REQUIRE(response.code / 100 == 2, "Http request failed.", x.url, response);
    return ds::fromJson<ds::JsonFragment>(response.data);
  }

  bool hasBackend() const override { return !ospToken().empty(); }
  bool hasVdcs() const override { return !m_vdcs.empty(); }

private:
  Apartment& m_apartment;
  SimpleHttpClient& m_httpClient;
  PropertyNodePtr m_enabledNode;
  ds::asio::Retryer m_reregisterRetryer{*this, "reregisterRetryer"};

  std::map<BackendVdcId, Vdc> m_vdcs;
  std::map<std::string, BackendVdcId> m_deviceDsuidToVdc;
  std::map<std::string, BackendVdcId> m_dsuidToVdc;

  PropertySystem& propertySystem() { return *m_apartment.getPropertySystem(); }
  const PropertySystem& propertySystem() const { return *m_apartment.getPropertySystem(); }
  std::string baseUrl() const { return propertySystem().getStringValue(pp_websvc_private_url); }

  std::string name() const override { return dsLogChannel.name(); }

  std::string dssId() const {
    return conf().readOr<std::string>("dssId", propertySystem().getStringValue(pp_sysinfo_dsid));
  }

  std::string ospToken() const {
    return conf().readOr<std::string>("ospToken", propertySystem().getStringValue(pp_websvc_rc_osptoken));
  }

  //
  // StructureQueryBusInterface
  //
  std::vector<DSMeterSpec_t> getBusMembers() override { DS_FAIL_REQUIRE("Not implemented"); }
  DSMeterSpec_t getDSMeterSpec(const dsuid_t& dsm) override { DS_FAIL_REQUIRE("Not implemented"); }
  std::vector<int> getZones(const dsuid_t& dsm) override { DS_FAIL_REQUIRE("Not implemented"); }
  std::vector<DeviceSpec> getDevicesInZone(const dsuid_t& dsm, const int zone, bool complete = true) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  std::vector<DeviceSpec> getInactiveDevicesInZone(const dsuid_t& dsm, const int zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  std::vector<GroupSpec_t> getGroups(const dsuid_t& dsm, const int zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  std::vector<ClusterSpec_t> getClusters(const dsuid_t& dsm) override { DS_FAIL_REQUIRE("Not implemented"); }
  std::vector<CircuitPowerStateSpec_t> getPowerStates(const dsuid_t& dsm) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  MeteringConfigurationSpec_t getMeteringConfiguration(const dsuid_t& _dsMeterID) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  std::vector<std::pair<int, int>> getLastCalledScenes(const dsuid_t& dsm, const int zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  std::bitset<7> getZoneStates(const dsuid_t& dsm, const int zone) override { DS_FAIL_REQUIRE("Not implemented"); }
  bool getEnergyBorder(const dsuid_t& dsm, int& lower, int& upper) override { DS_FAIL_REQUIRE("Not implemented"); }
  DeviceSpec deviceGetSpec(devid_t id, dsuid_t dsm) override { DS_FAIL_REQUIRE("Not implemented"); }
  std::string getSceneName(dsuid_t dsm, std::shared_ptr<Group> group, const uint8_t scene) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  DSMeterHash_t getDSMeterHash(const dsuid_t& dsm) override { DS_FAIL_REQUIRE("Not implemented"); }
  void getDSMeterState(const dsuid_t& dsm, uint8_t* state) override { DS_FAIL_REQUIRE("Not implemented"); }
  ZoneHeatingConfigSpec_t getZoneHeatingConfig(const dsuid_t& dsm, const uint16_t zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  ZoneHeatingInternalsSpec_t getZoneHeatingInternals(const dsuid_t& dsm, const uint16_t zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  ZoneHeatingStateSpec_t getZoneHeatingState(const dsuid_t& dsm, const uint16_t zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  ZoneHeatingOperationModeSpec_t getZoneHeatingOperationModes(const dsuid_t& dsm, const uint16_t zone) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  dsuid_t getZoneSensor(const dsuid_t& dsm, const uint16_t zone, SensorType sensor) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  void getZoneSensorValue(const dsuid_t& dsm, const uint16_t zone, SensorType sensor, uint16_t* value,
                          uint32_t* age) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  int getDevicesCountInZone(const dsuid_t& dsm, const int zone, bool onlyActive) override {
    DS_FAIL_REQUIRE("Not implemented");
  }
  void protobufMessageRequest(const dsuid_t dsm, const uint16_t requestSize, const uint8_t* request,
                              uint16_t* responseSize, uint8_t* response) override {
    DS_FAIL_REQUIRE("Not implemented");
  }

  //
  // StructureModifyingBusInterface
  //
  void setZoneID(const dsuid_t&, const devid_t, const int) override {
    // zones are not propagated
  }
  void createZone(const dsuid_t&, const int) override {
    // zones are not propagated
  }
  void removeZone(const dsuid_t&, const int) override {
    // zones are not propagated
  }
  void addToGroup(const dsuid_t&, const int, const int) override { DS_FAIL_REQUIRE("Failed."); }
  void removeFromGroup(const dsuid_t&, const int, const int) override { DS_FAIL_REQUIRE("Failed."); }
  void removeDeviceFromDSMeter(const dsuid_t&, const int) override { DS_FAIL_REQUIRE("Failed."); }
  void removeDeviceFromDSMeters(const dsuid_t& dsuid) override {
    http(HttpRequest{
        .url = ds::str("/v1/cloudvdc/vdcs/", vdcIdForDeviceDsuid(dsuid), "/devices/", dsuid),
        .type = HttpMethod::DELETE,
    });
  }
  void sceneSetName(uint16_t, uint8_t, uint8_t, const std::string&) override {}
  void deviceSetName(dsuid_t, devid_t, const std::string&) override {} // dss is data owner
  void meterSetName(dsuid_t, const std::string&) override {}           // dss is data owner
  void createGroup(uint16_t, uint8_t, ApplicationType, uint32_t, const std::string&) override {}
  void removeGroup(uint16_t, uint8_t) override {}
  void groupSetName(uint16_t, uint8_t, const std::string&) override {}
  void createCluster(uint8_t, ApplicationType, uint32_t, const std::string&) override {}
  void removeCluster(uint8_t) override {}
  void groupSetApplication(uint16_t, uint8_t, ApplicationType, uint32_t) override {}
  void clusterSetName(uint8_t, const std::string&) override {}
  void clusterSetApplication(uint8_t, ApplicationType, uint32_t) override {}
  void clusterSetProperties(uint8_t, uint16_t _location, uint16_t _floor, uint16_t _protectionClass) override {}
  void clusterSetLockedScenes(uint8_t, const std::vector<int>&) override {}
  void clusterSetConfigurationLock(uint8_t, bool) override {}
  void clusterSetOperationLock(uint8_t, bool) override {}
  void setButtonSetsLocalPriority(const dsuid_t&, const devid_t, bool) override { DS_FAIL_REQUIRE("Failed."); }
  void setButtonCallsPresent(const dsuid_t&, const devid_t, bool) override { DS_FAIL_REQUIRE("Failed."); }
  void setZoneHeatingConfig(const dsuid_t&, const uint16_t, const ZoneHeatingConfigSpec_t) override {
    DS_FAIL_REQUIRE("Failed.");
  }
  void setZoneHeatingOperationModes(const dsuid_t&, const uint16_t, const ZoneHeatingOperationModeSpec_t) override {
    DS_FAIL_REQUIRE("Failed.");
  }
  void setZoneSensor(const uint16_t, SensorType, const dsuid_t&) override {}
  void resetZoneSensor(const uint16_t, SensorType) override {}
  void setCircuitPowerStateConfig(const dsuid_t&, const int, const int, const int) override {
    DS_FAIL_REQUIRE("Failed.");
  }
  void setMeteringConfiguration(const dsuid_t&, const MeteringConfigurationSpec_t&) override {}
  void setProperty(const dsuid_t&, const ::google::protobuf::RepeatedPtrField<::vdcapi::PropertyElement>&) override {
    DS_FAIL_REQUIRE("Failed.");
  }
  vdcapi::Message getProperty(const dsuid_t&,
                              const ::google::protobuf::RepeatedPtrField<::vdcapi::PropertyElement>&) override {
    DS_FAIL_REQUIRE("Failed.");
  }

  void setFireMuteEnabled(dsuid_t dsm, bool) override {}
  void removeVdc(const dsuid_t& vdsmID, const dsuid_t& vdcID) override {}
  void clusterSetStopTimeout(uint8_t cluster, uint8_t timeoutSeconds) override {}
  void zoneSetStopTimeout(uint16_t zone, uint8_t group, uint8_t timeoutSeconds) override {}
};

BackendVdcs::~BackendVdcs(){};

std::unique_ptr<BackendVdcs> makeBackendVdcs(ds::asio::ResourceBase& parent, Apartment& apartment,
                                             EventInterpreter& eventInterpreter, SimpleHttpClient& httpClient) {
  return std::make_unique<BackendVdcsImpl>(parent, apartment, eventInterpreter, httpClient);
}

} // namespace dss
