/*
    Copyright (c) 2009,2011 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <bitset>
#include <memory>
#include <string>
#include <vector>

#include "businterface-fwd.h"
#include "datetools.h"
#include "ds485/specification.h"
#include "ds485types.h"
#include "dssfwd.h"
#include "exception.h"
#include "model/behavior.h"
#include "model/deviceinterface.h"

namespace google {
namespace protobuf {
template <typename T>
class RepeatedPtrField;
}
} // namespace google
namespace vdcapi {
class PropertyElement;
class Message;
} // namespace vdcapi

namespace dss {

class AddressableModelItem;
class BusInterface;

class DeviceBusInterface {
public:
  //------------------------------------------------ Device manipulation
  virtual uint8_t getDeviceConfig(const Device& _device, uint8_t _configClass, uint8_t _configIndex) = 0;

  virtual uint16_t getDeviceConfigWord(const Device& _device, uint8_t _configClass, uint8_t _configIndex) = 0;

  virtual void setDeviceConfig(const Device& _device, uint8_t _configClass, uint8_t _configIndex, uint8_t _value) = 0;

  /** Set the active group for the button */
  virtual void setDeviceButtonActiveGroup(const Device& _device, uint8_t _groupID) = 0;

  /** Enable or disable programming mode */
  virtual void setDeviceProgMode(const Device& _device, uint8_t modeId) = 0;

  /** Increase Increase value of selected output channel. */
  virtual void increaseDeviceOutputChannelValue(const Device& _device, uint8_t _channel) = 0;

  /** Decrease value of selected output channel. */
  virtual void decreaseDeviceOutputChannelValue(const Device& _device, uint8_t _channel) = 0;

  /** Stop dimming value of selected output channel. */
  virtual void stopDeviceOutputChannelValue(const Device& _device, uint8_t _channel) = 0;

  virtual uint16_t getDeviceOutputChannelValue(const Device& _device, uint8_t _channel) = 0;

  virtual void setDeviceOutputChannelValue(const Device& _device, uint8_t _channel, uint8_t _size, uint16_t _value,
                                           bool _applyNow = true) = 0;

  virtual uint16_t getDeviceOutputChannelSceneValue(const Device& _device, uint8_t _channel, uint8_t _scene) = 0;

  virtual void setDeviceOutputChannelSceneValue(const Device& _device, uint8_t _channel, uint8_t _size, uint8_t _scene,
                                                uint16_t _value) = 0;

  virtual uint16_t getDeviceOutputChannelSceneConfig(const Device& _device, uint8_t _scene) = 0;

  virtual void setDeviceOutputChannelSceneConfig(const Device& _device, uint8_t _scene, uint16_t _value) = 0;
  virtual void setDeviceOutputChannelDontCareFlags(const Device& _device, uint8_t _scene, uint16_t _value) = 0;
  virtual uint16_t getDeviceOutputChannelDontCareFlags(const Device& _device, uint8_t _scene) = 0;

  /** Tests transmission quality to a device, where the first returned
    value is the DownstreamQuality and the second value the UpstreamQuality */
  virtual std::pair<uint8_t, uint16_t> getTransmissionQuality(const Device& _device) = 0;

  /** Set generic device output value */
  virtual void setValue(const Device& _device, uint8_t _value) = 0;

  /** Queries sensor value from a device -> returns the dsmapi normalized value */
  virtual uint16_t getSensorValue(const Device& _device, const int _sensorIndex) = 0;

  /** Queries extended sensor value from a device */
  virtual DeviceSensorValue getSensorValueEx(const Device& _device, const int _sensorIndex) = 0;

  /** add device to a user group */
  virtual void addGroup(const Device& _device, const int _groupId) = 0;

  /** remove device from a user group */
  virtual void removeGroup(const Device& _device, const int _groupId) = 0;

  /** Tells the dSM to lock the device if \a _lock is true. */
  virtual void lockOrUnlockDevice(const Device& _device, const bool _lock) = 0;

  /** Invokes generic request on vdc device and waits for reply. */
  virtual void genericRequest(const Device& _device, const std::string& methodName,
                              const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& params,
                              const std::locale&) = 0;

  virtual void setProperty(const Device& _device,
                           const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& properties) = 0;
  virtual vdcapi::Message getProperty(
      const Device& _device, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& query) = 0;

  virtual void reregister(const Device& _device) = 0;

  virtual ~DeviceBusInterface(){}; // please the compiler (virtual dtor)
};                                 // DeviceBusInterface

class StructureQueryBusInterface {
public:
  /** Returns an std::vector containing the dsMeter-spec of all dsMeters present. */
  virtual std::vector<DSMeterSpec> getBusMembers() = 0;

  /** Returns the dsMeter-spec for a dsMeter */
  virtual DSMeterSpec getDSMeterSpec(const dsuid_t& _dsMeterID) = 0;

  /** Returns a std::vector conatining the zone-ids of the specified dsMeter */
  virtual std::vector<int> getZones(const dsuid_t& _dsMeterID) = 0;
  /** Returns the bus-ids of the devices present in the given zone of the specified dsMeter */
  virtual std::vector<DeviceSpec> getDevicesInZone(const dsuid_t& _dsMeterID, const int _zoneID,
                                                   bool complete = true) = 0;
  virtual std::vector<DeviceSpec> getInactiveDevicesInZone(const dsuid_t& _dsMeterID, const int _zoneID) = 0;

  /** Returns the a std::vector containing the group-ids of the given zone on the specified dsMeter */
  virtual std::vector<GroupSpec> getGroups(const dsuid_t& _dsMeterID, const int _zoneID) = 0;
  virtual std::vector<ClusterSpec> getClusters(const dsuid_t& _dsMeterID) = 0;
  virtual std::vector<CircuitPowerStateSpec> getPowerStates(const dsuid_t& _dsMeterID) = 0;
  virtual MeteringConfigurationSpec getMeteringConfiguration(const dsuid_t& _dsMeterID) = 0;

  virtual std::vector<std::pair<int, int> > getLastCalledScenes(const dsuid_t& _dsMeterID, const int _zoneID) = 0;
  virtual std::bitset<7> getZoneStates(const dsuid_t& _dsMeterID, const int _zoneID) = 0;
  virtual bool getEnergyBorder(const dsuid_t& _dsMeterID, int& _lower, int& _upper) = 0;

  /** Returns the function, product and revision id of the device. */
  virtual DeviceSpec deviceGetSpec(devid_t _id, dsuid_t _dsMeterID) = 0;

  virtual ~StructureQueryBusInterface(){}; // please the compiler (virtual dtor)
  virtual std::string getSceneName(dsuid_t _dsMeterID, std::shared_ptr<Group> _group, const uint8_t _sceneNumber) = 0;

  /** returns the hash over the dSMeter's datamodel */
  virtual DSMeterHash_t getDSMeterHash(const dsuid_t& _dsMeterID) = 0;
  virtual void getDSMeterState(const dsuid_t& _dsMeterID, uint8_t* state) = 0;

  /** return heating controller data model */
  virtual ZoneHeatingConfigSpec getZoneHeatingConfig(const dsuid_t& _dsMeterID, const uint16_t _ZoneID) = 0;
  virtual ZoneHeatingInternalsSpec getZoneHeatingInternals(const dsuid_t& _dsMeterID, const uint16_t _ZoneID) = 0;
  virtual ZoneHeatingStateSpec getZoneHeatingState(const dsuid_t& _dsMeterID, const uint16_t _ZoneID) = 0;
  virtual ZoneHeatingOperationModeSpec getZoneHeatingOperationModes(const dsuid_t& _dsMeterID,
                                                                      const uint16_t _ZoneID) = 0;
  virtual dsuid_t getZoneSensor(const dsuid_t& _meterDSUID, const uint16_t _zoneID, SensorType _sensorType) = 0;
  virtual void getZoneSensorValue(const dsuid_t& _meterDSUID, const uint16_t _zoneID, SensorType _sensorType,
                                  uint16_t* SensorValue, uint32_t* SensorAge) = 0;
  virtual int getDevicesCountInZone(const dsuid_t& _dsMeterID, const int _zoneID, bool _onlyActive = false) = 0;
  /** vdsm property queries */
  virtual void protobufMessageRequest(const dsuid_t _dSMdSUID, const uint16_t _request_size, const uint8_t* _request,
                                      uint16_t* _response_size, uint8_t* _response) = 0;
  // default implementation is a wrapper above protobufMessageRequest
  virtual vdcapi::Message getVdcProperty(const dsuid_t& vdcId, const dsuid_t& targetId,
                                         const std::vector<std::string>& keys);

  struct SensorDesc {
    std::string sensorName;
    uint8_t index;
    int sensorUsage;
    SensorType sensorType;
    double min, max, resolution;
    double updateInterval;
    SensorDesc()
        : index(255),
          sensorUsage(0),
          sensorType(SensorType::UnknownType),
          min(0),
          max(0),
          resolution(0),
          updateInterval(0) {}
  };

  struct BinaryInputDesc {
    std::string inputName;
    uint8_t dsIndex;
    BinaryInputType sensorFunction;
    double updateInterval;
    BinaryInputDesc() : dsIndex(255), sensorFunction(BinaryInputType::AppMode), updateInterval(0) {}
  };

  struct VdcInputDesc {
    std::map<std::string, SensorDesc> m_sensorInputs;
    std::map<std::string, BinaryInputDesc> m_binaryInputs;
  };
  struct VdcInputDesc getVdcInputDesc(dsuid_t _vdsm, dsuid_t _device);

}; // StructureQueryBusInterface

class StructureModifyingBusInterface {
public:
  virtual std::string name() const = 0;
  /** Adds the given device to the specified zone. */
  virtual void setZoneID(const dsuid_t& _dsMeterID, const devid_t _deviceID, const int _zoneID) = 0;

  /** Creates a new Zone on the given dsMeter */
  virtual void createZone(const dsuid_t& _dsMeterID, const int _zoneID) = 0;

  /** Removes the zone \a _zoneID on the dsMeter \a _dsMeterID */
  virtual void removeZone(const dsuid_t& _dsMeterID, const int _zoneID) = 0;

  /** Adds a device to a given group */
  virtual void addToGroup(const dsuid_t& _dsMeterID, const int _groupID, const int _deviceID) = 0;
  /** Removes a device from a given group */
  virtual void removeFromGroup(const dsuid_t& _dsMeterID, const int _groupID, const int _deviceID) = 0;

  /** Removes the device from the dSM data model **/
  virtual void removeDeviceFromDSMeter(const dsuid_t& _dsMeterID, const int _deviceID) = 0;
  virtual void removeDeviceFromDSMeters(const dsuid_t& _DeviceDSID) = 0;

  /** Sets the name of a scene */
  virtual void sceneSetName(uint16_t _zoneID, uint8_t _groupID, uint8_t _sceneNumber, const std::string& _name) = 0;
  virtual void deviceSetName(dsuid_t _meterDSID, devid_t _deviceID, const std::string& _name) = 0;
  virtual void meterSetName(dsuid_t _meterDSID, const std::string& _name) = 0;

  /** Create and manage user groups */
  virtual void createGroup(uint16_t _zoneID, uint8_t _groupID, ApplicationType applicationType,
                           uint32_t applicationConfig, const std::string& _name) = 0;
  virtual void removeGroup(uint16_t _zoneID, uint8_t _groupID) = 0;
  virtual void groupSetApplication(uint16_t _zoneID, uint8_t _groupID, ApplicationType applicationType,
                                   uint32_t applicationConfig) = 0;
  virtual void groupSetName(uint16_t _zoneID, uint8_t _groupID, const std::string& _name) = 0;

  virtual void createCluster(uint8_t _groupID, ApplicationType applicationType, uint32_t applicationConfig,
                             const std::string& _name) = 0;
  virtual void removeCluster(uint8_t _clusterID) = 0;
  virtual void clusterSetName(uint8_t _clusterID, const std::string& _name) = 0;
  virtual void clusterSetApplication(uint8_t _clusterID, ApplicationType applicationType,
                                     uint32_t applicationConfig) = 0;
  virtual void clusterSetProperties(uint8_t _clusterID, uint16_t _location, uint16_t _floor,
                                    uint16_t _protectionClass) = 0;
  virtual void clusterSetLockedScenes(uint8_t _clusterID, const std::vector<int>& lockedScenes) = 0;
  virtual void clusterSetConfigurationLock(uint8_t _clusterID, bool _lock) = 0;
  virtual void clusterSetOperationLock(uint8_t _clusterID, bool _lock) = 0;

  virtual void setButtonSetsLocalPriority(const dsuid_t& _dsMeterID, const devid_t _deviceID, bool _setsPriority) = 0;
  virtual void setButtonCallsPresent(const dsuid_t& _dsMeterID, const devid_t _deviceID, bool _callsPresent) = 0;

  /** Create and manage heating controller */
  virtual void setZoneHeatingConfig(const dsuid_t& _dsMeterID, const uint16_t _ZoneID,
                                    const ZoneHeatingConfigSpec _spec) = 0;
  void disableZoneHeatingConfig(const dsuid_t& dsMeterId, uint16_t zoneId) {
    setZoneHeatingConfig(dsMeterId, zoneId, ZoneHeatingConfigSpec{});
  }
  virtual void setZoneHeatingOperationModes(const dsuid_t& _dsMeterID, const uint16_t _ZoneID,
                                            const ZoneHeatingOperationModeSpec _spec) = 0;
  virtual void setZoneSensor(const uint16_t _zoneID, SensorType _sensorType, const dsuid_t& _sensorDSUID) = 0;
  virtual void resetZoneSensor(const uint16_t _zoneID, SensorType _sensorType) = 0;

  virtual void setCircuitPowerStateConfig(const dsuid_t& _dsMeterID, const int _index, const int _setThreshold,
                                          const int _resetThreshold) = 0;
  virtual void setMeteringConfiguration(const dsuid_t& _dsMeterID, const MeteringConfigurationSpec& _spec) = 0;

  virtual void setProperty(const dsuid_t& _meter,
                           const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& properties) = 0;
  virtual vdcapi::Message getProperty(
      const dsuid_t& _meter, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& query) = 0;

  virtual void setFireMuteEnabled(dsuid_t dsm, bool) = 0;

  virtual void removeVdc(const dsuid_t& _VdsmID, const dsuid_t& _VdcID) = 0;

  virtual void clusterSetStopTimeout(uint8_t cluster, uint8_t timeoutSeconds) = 0;

  virtual void zoneSetStopTimeout(uint16_t zone, uint8_t group, uint8_t timeoutSeconds) = 0;

  virtual ~StructureModifyingBusInterface(){}; // please the compiler (virtual dtor)
};
std::ostream& operator<<(std::ostream& ostream, const StructureModifyingBusInterface& x);

struct Caller {
  CallOrigin origin{};
  SceneAccessCategory category{};
  const std::string token{};
};

class ActionRequestInterface {
public:
  virtual ~ActionRequestInterface(){}; // please the compiler (virtual dtor)
  virtual void callScene(AddressableModelItem* pTarget, const uint16_t scene, Caller&& caller, const bool _force) = 0;
  virtual void callSceneMin(AddressableModelItem* pTarget, const uint16_t _scene, Caller&& caller) = 0;
  virtual void saveScene(AddressableModelItem* pTarget, const uint16_t scene, Caller&& caller) = 0;
  virtual void undoScene(AddressableModelItem* pTarget, const uint16_t scene, Caller&& caller) = 0;
  virtual void undoSceneLast(AddressableModelItem* pTarget, Caller&& caller) = 0;
  virtual void blink(AddressableModelItem* pTarget, Caller&& caller) = 0;
  virtual void setValue(AddressableModelItem* pTarget, const uint8_t _value, Caller&& caller) = 0;
  virtual void increaseOutputChannelValue(AddressableModelItem* pTarget, ChannelType _channel, Caller&& caller) = 0;
  virtual void decreaseOutputChannelValue(AddressableModelItem* pTarget, ChannelType _channel, Caller&& caller) = 0;
  virtual void stopOutputChannelValue(AddressableModelItem* pTarget, ChannelType _channel, Caller&& caller) = 0;
  virtual void pushSensor(AddressableModelItem* pTarget, dsuid_t _sourceID, SensorType _sensorType,
                          double _sensorValueFloat, Caller&& caller) = 0;
  virtual void muteFire() = 0;
  virtual void setState(Group* group, const uint8_t zoneState, const uint8_t areaState1, const uint8_t areaState2,
                        const uint8_t areaState3, const uint8_t areaState4) = 0;
  virtual void setState(Cluster* cluster, const uint8_t state) = 0;
  virtual void stopStopTimer(Group* group, const uint8_t zoneOrArea) = 0;
  virtual void stopStopTimer(Cluster* cluster) = 0;

  //< @ret true if locked
  virtual bool isOperationLock(const dsuid_t& _dSM, int _clusterId) = 0;
}; // ActionRequestInterface

class MeteringBusInterface {
public:
  /** Sends a message to all devices to report their meter-data */
  virtual void requestMeterData() = 0;

  /** Returns the current power-consumption in mW */
  virtual unsigned long getPowerConsumption(const dsuid_t& _dsMeterID) = 0;

  /** Returns the meter value in Wh */
  virtual unsigned long getEnergyMeterValue(const dsuid_t& _dsMeterID) = 0;

  virtual ~MeteringBusInterface(){}; // please the compiler (virtual dtor)
};                                   // MeteringBusInterface

class BusEventSink {
public:
  virtual void onGroupCallScene(BusInterface* _source, const dsuid_t& _dsMeterID, const int _zoneID, const int _groupID,
                                const int _originDeviceId, const int _sceneID, Caller&& caller, const bool _force) = 0;
  virtual void onGroupCallSceneMin(BusInterface* _source, const dsuid_t& _dsMeterID, const int _zoneID,
                                   const int _groupID, const int _originDeviceId, const int _sceneID,
                                   Caller&& caller) = 0;
  virtual void onGroupSaveScene(BusInterface* _source, const dsuid_t& _dsMeterID, const int _zoneID, const int _groupID,
                                const int _originDeviceId, const int _sceneID, Caller&& caller) = 0;

  virtual void onGroupUndoScene(BusInterface* _source, const dsuid_t& _dsMeterID, const int _zoneID, const int _groupID,
                                const int _originDeviceId, const int _sceneID, const bool _explicit,
                                Caller&& caller) = 0;
  virtual void onGroupBlink(BusInterface* _source, const dsuid_t& _dsMeterID, const int _zoneID, const int _groupID,
                            const int _originDeviceId, Caller&& caller) = 0;
  virtual void onDeviceCallScene(BusInterface* _source, const dsuid_t& _dsMeterID, const int _deviceID,
                                 const int _originDeviceId, const int _sceneID, Caller&& caller, const bool _force) = 0;
  virtual void onDeviceSaveScene(BusInterface* _source, const dsuid_t& _dsMeterID, const int _deviceID,
                                 const int _originDeviceId, const int _sceneID, Caller&& caller) = 0;
  virtual void onDeviceBlink(BusInterface* _source, const dsuid_t& _dsMeterID, const int _deviceID,
                             const int _originDeviceId, Caller&& caller) = 0;
  virtual void onDeviceUndoScene(BusInterface* _source, const dsuid_t& _dsMeterID, const int _deviceID,
                                 const int _originDeviceId, const int _sceneID, const bool _explicit,
                                 Caller&& caller) = 0;
  virtual void onMeteringEvent(BusInterface* _source, const dsuid_t& _dsMeterID, const unsigned int _powerW,
                               const unsigned int _energyWs, const unsigned int _powerProducedW,
                               const unsigned int _energyProducedWs) = 0;
  virtual void onZoneSensorValue(BusInterface* _source, const dsuid_t& _dsMeterID, const dsuid_t& _sourceDevice,
                                 const int& _zoneID, const int& _groupID, SensorType _sensorType,
                                 const int& _sensorValue, const int& _precision, Caller&& caller) = 0;
  virtual void onDeviceSetValue(BusInterface* _source, const dsuid_t& _dsMeterID, const int _deviceID,
                                const uint8_t _value, Caller&& caller) = 0;
  virtual void onZoneGroupSetValue(BusInterface* _source, const int _zoneID, const int _groupID, const uint8_t _value,
                                   Caller&& caller) = 0;

  virtual ~BusEventSink(){};
};

/** Interface to be implemented by any bus interface provider */
class BusInterface {
public:
  virtual ~BusInterface(){};

  virtual DeviceBusInterface* getDeviceBusInterface() = 0;
  virtual StructureQueryBusInterface* getStructureQueryBusInterface() = 0;
  virtual MeteringBusInterface* getMeteringBusInterface() = 0;
  virtual StructureModifyingBusInterface* getStructureModifyingBusInterface() = 0;
  virtual ActionRequestInterface* getActionRequestInterface() = 0;
  virtual void setBusEventSink(BusEventSink* _eventSink) = 0;
  virtual const std::string getConnectionURI() = 0;
  virtual void add(StructureModifyingBusInterface&){};
};

class BusApiError : public DSSException {
public:
  BusApiError(const std::string& _what, int _error = INT_MIN) : DSSException(_what), error(_error) {}
  int error;
}; // BusApiError

} // namespace dss

DS_REFLECT_DECLARE_STRUCT((dss), Caller, (origin)(category)(token));
