/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include "cache/apartment-proxy.h"

#include <ds/asio/thread-pool.h>
#include <ds/conf.h>
#include <ds/system.h>
#include <event.h>

#include <boost/range/algorithm/find.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <cmath>
#include <future>
#include <map>
#include <string>
#include <thread>
#include <tuple>

#include "businterface.h"
#include "cache/cache-updater.h"
#include "ds-model/model.h"
#include "ds-scenarios.h"
#include "ds/log.h"
#include "dss.h"
#include "event.h"
#include "hal/hardware-abstraction.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/modelconst.h"
#include "model/scenehelper.h"
#include "model/set.h"
#include "model/zone.h"
#include "propertysystem.h"
#include "propertysystem_common_paths.h"

using ds::asio::ThreadPool;
using dss::Cluster;
using dss::hal::DeviceChannelValues;
using dss::hal::DeviceState;
using namespace std::chrono_literals;

DS_JSON_DEFINE_STRUCT((dss), SceneFlags);
DS_JSON_DEFINE_STRUCT((dss), SceneChannel);
DS_JSON_DEFINE_STRUCT((dss), SceneEntry);
DS_JSON_DEFINE_STRUCT((dss), DeviceSceneCache);
DS_JSON_DEFINE_STRUCT((dss), DeviceCache);
DS_JSON_DEFINE_STRUCT((dss), ScenarioCache);
DS_JSON_DEFINE_STRUCT((dss), CacheScenarioState);

namespace dss {

static ds::LogChannel dsLogChannel("Cache");

static constexpr ds::Clock::duration ReadoutTimePerParameter = 5s;

class ApartmentProxyEventPlugin : public EventInterpreterPlugin {
private:
  ApartmentProxy* m_proxy;

public:
  ApartmentProxyEventPlugin(ApartmentProxy* proxy, EventInterpreter* pInterpreter)
      : EventInterpreterPlugin("ApartmentProxy", pInterpreter), m_proxy(proxy) {}

  void subscribe() {
    getEventInterpreter().subscribe(
        std::make_shared<EventSubscription>(*this, EventName::ApartmentProxyDeviceTimeout));
    getEventInterpreter().subscribe(
        std::make_shared<EventSubscription>(*this, EventName::ApartmentProxyDeviceDimmingTimeout));
    getEventInterpreter().subscribe(
        std::make_shared<EventSubscription>(*this, EventName::StopDsmTimer));
  }

  void handleEvent(Event& event, const EventSubscription& subscription) {
    if (!DSS::hasInstance()) {
      return;
    }

    if (event.getName() == EventName::StopDsmTimer) {
      m_proxy->stopDsmTimer();
    }

    auto deviceRef = event.getRaisedAtDevice();
    if (!deviceRef) {
      return;
    }
    auto device = deviceRef->getDevice();
    if (!device) {
      return;
    }
    if (event.getName() == EventName::ApartmentProxyDeviceTimeout) {
      m_proxy->completeMoving(device->getDSID());
    } else if (event.getName() == EventName::ApartmentProxyDeviceDimmingTimeout) {
      m_proxy->stopDimming(device->getDSID());
    }
  }
};

uint8_t toDSMState(dsModel::LightsStatusStatus status) {
  switch (status) {
    case dsModel::LightsStatusStatus::none: return -1; // dsm state machine ignores value
    case dsModel::LightsStatusStatus::off: return 0;
    case dsModel::LightsStatusStatus::on: [[fallthrough]];
    case dsModel::LightsStatusStatus::dimming: return 1;
  }
  return -1; // [gcc] error: control reaches end of non-void function [-Werror=return-type]
}

uint8_t toDSMState(dsModel::ShadesStatusStatus status) {
  switch (status) {
    case dsModel::ShadesStatusStatus::none: return -1;
    case dsModel::ShadesStatusStatus::closed: return 0;
    case dsModel::ShadesStatusStatus::open: [[fallthrough]];
    case dsModel::ShadesStatusStatus::moving: [[fallthrough]];
    case dsModel::ShadesStatusStatus::mixed: return 1;
  }
  return -1; // [gcc] error: control reaches end of non-void function [-Werror=return-type]
}

uint8_t toDSMState(dsModel::AudioStatusStatus status) {
  switch (status) {
    case dsModel::AudioStatusStatus::none: return -1;
    case dsModel::AudioStatusStatus::off: [[fallthrough]];
    case dsModel::AudioStatusStatus::pause: [[fallthrough]];
    case dsModel::AudioStatusStatus::powerOff: [[fallthrough]];
    case dsModel::AudioStatusStatus::standby: return 0;
    case dsModel::AudioStatusStatus::playing: [[fallthrough]];
    case dsModel::AudioStatusStatus::stop: return 1;
  }
  return -1; // [gcc] error: control reaches end of non-void function [-Werror=return-type]
}

class ApartmentProxyImpl : public ApartmentProxy {
  Apartment* m_apartment;

  PropertySystem& m_propertySystem = *m_apartment->getPropertySystem();
  PropertyNodePtr m_statesNode{m_propertySystem.createProperty("/usr/states/")};
  PropertyNodePtr m_apartmentVentilationNode{
      m_propertySystem.createProperty("/scripts/system-addon-ventilation-controller/apartmentVentilation")};
  PropertyNodePtr m_zoneVentilationNode{
      m_propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-ventilation")};
  PropertyNodePtr m_zoneRecirculationNode{
      m_propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-recirculation")};

public:
  ApartmentProxyImpl(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db,
                     EventInterpreter* pInterpreter, const ds::Conf& conf)
      : m_apartment(apt),
        m_hal(hal),
        m_scenarioCache(db),
        m_deviceCache(hal->getCache()),
        m_pool(1),
        m_cacheReady(true),
        m_emptyClickPreventionActiveLight(true),
        m_emptyClickPreventionActiveNonLight(true) {
    m_emptyClickPreventionActiveLight = conf.readOr<bool>("emptyClickPreventionLight", true);
    m_emptyClickPreventionActiveNonLight = conf.readOr<bool>("emptyClickPreventionNonLight", false);
    if (pInterpreter) {
      EventInterpreterPlugin* plugin = new ApartmentProxyEventPlugin(this, pInterpreter);
      pInterpreter->addPlugin(plugin);
    }
    if (auto m_pPropSystem = m_apartment->getPropertySystem()) {
      auto cacheReadyProperty = m_pPropSystem->createProperty(pp_device_cache_ready);
      cacheReadyProperty->linkToProxy(PropertyProxyReference<bool>(m_cacheReady, false));
    }
  }

  void stopDsmTimer() override {
    if (m_emptyClickPreventionActiveNonLight && m_cacheReady &&
      m_apartment->getBusInterface() &&
      m_apartment->getBusInterface()->getActionRequestInterface()) {

      if (m_stopTimerZonesList.empty()) {
        return;
      }

      time_t now = DateTime().secondsSinceEpoch();
      time_t next = 0;

      {
        boost::recursive_mutex::scoped_lock lock(m_stopTimerListMutex);
        std::vector<std::tuple<int, int, int, time_t> >::iterator it;

        for (it = m_stopTimerZonesList.begin(); it != m_stopTimerZonesList.end();) {
          if (std::get<3>(*it) <= now) {
            int zoneId = std::get<0>(*it);
            int groupId = std::get<1>(*it);
            int sceneId = std::get<2>(*it);

            it = m_stopTimerZonesList.erase(it);
            try {
              if ((zoneId == ZONE_ID_BROADCAST) && isCluster(groupId)) {
                std::shared_ptr<Cluster> cluster = m_apartment->getCluster(groupId);
                m_apartment->getBusInterface()->getActionRequestInterface()->stopStopTimer(cluster.get());
              } else { // zone scene call, need to check areas
                std::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
                uint8_t zoneOrArea = 0; // zone
                if (SceneHelper::isAreaScene(sceneId)) {
                  zoneOrArea = sceneId % 5; // get area number 1..4
                }

                m_apartment->getBusInterface()->getActionRequestInterface()->stopStopTimer(zone->getGroup(GroupIDGray).get(), zoneOrArea);
              }
            } catch(ItemNotFoundException& e) {
              DS_WARNING("Not stopping timer on zone ", zoneId, ": zone not found");
            }
          } else {
            if ((next == 0) || (std::get<3>(*it) < next)) {
              next = std::get<3>(*it);
            }
            it++;
          }
        }

        if (m_stopTimerZonesList.empty()) {
            return;
        }
      }
      std::shared_ptr<Event> event = std::make_shared<Event>(EventName::StopDsmTimer);
      time_t newNow = DateTime().secondsSinceEpoch();
      time_t diffSeconds = next - newNow;
      if (diffSeconds <= 0) {
        diffSeconds = 1; // don't overspam the queue, leave at least 1s delay
      }
      event->setProperty("time", "+" + intToString(diffSeconds));
      event->setProperty(EventProperty::Unique, "Yes");
      m_apartment->pushEvent(event);
    }
  }

  bool emptyClickEnabled(int color) const {
    return (color == GroupIDYellow) ? m_emptyClickPreventionActiveLight : m_emptyClickPreventionActiveNonLight;
  }

  void updateDSMState(Group& group, int zoneState, int area1State, int area2State, int area3State,
                      int area4State) const override {
    if (!emptyClickEnabled(group.getColor())) {
      DS_NOTICE("updateDSMState skipping", emptyClickEnabled(group.getColor()));
      return;
    }

    if (!m_cacheReady) {
      DS_NOTICE("updateDSMState skipping", m_cacheReady);
      return;
    }

    auto actionRequestInterface = m_apartment->getBusInterface()->getActionRequestInterface();
    actionRequestInterface->setState(&group, zoneState, area1State, area2State, area3State, area4State);
  }

  void updateDSMState(Cluster& cluster, int state) const override {
    if (!emptyClickEnabled(cluster.getColor())) {
      DS_NOTICE("updateDSMState skipping", emptyClickEnabled(cluster.getColor()));
      return;
    }

    if (!m_cacheReady) {
      DS_NOTICE("updateDsmClusterState skip", m_cacheReady);
      return;
    }

    auto actionRequestInterface = m_apartment->getBusInterface()->getActionRequestInterface();
    actionRequestInterface->setState(&cluster, state);
  }

  void syncWithDsm(int zoneId, int groupId) const override {
    DS_DEBUG_ENTER(zoneId, groupId);

    auto zone = m_apartment->getZone(zoneId);
    auto devices = zone->getDevices().getByPresence(true).getByGroup(groupId);

    if (!devices.empty()) {
      if (groupId == GroupIDYellow) {
        auto result = getLightApplicationStatus(devices);
        updateDSMState(*zone->getGroup(groupId), toDSMState(result.nonLocalPriority), toDSMState(result.area1),
                       toDSMState(result.area2), toDSMState(result.area3), toDSMState(result.area4));
      } else if (groupId == GroupIDGray) {
        auto result = getShadeApplicationStatus(devices);
        updateDSMState(*zone->getGroup(groupId), toDSMState(result.nonLocalPriority), toDSMState(result.area1),
                       toDSMState(result.area2), toDSMState(result.area3), toDSMState(result.area4));
      } else if (groupId == GroupIDCyan) {
        auto result = getAudioApplicationStatus(devices, {});
        updateDSMState(*zone->getGroup(groupId), toDSMState(result.nonLocalPriority), -1, -1, -1, -1);
      }
    }
    DS_DEBUG_LEAVE()
  }

  void syncWithDsm(std::shared_ptr<Cluster> cluster) const override {
    DS_DEBUG_ENTER(cluster->getID());
    dsModel::GroupStatus gs{};
    gs.id = intToString(cluster->getID());

    Set devices = cluster->getDevices().getByPresence(true);
    if (devices.empty()) {
      return;
    }

    if (cluster->getColor() == GroupIDYellow) {
      dsModel::LightsStatusStatus zoneStateDsm{};
      std::tie(gs.application, zoneStateDsm) = getLightGroupApplicationStatus(devices);
      if (zoneStateDsm != dsModel::LightsStatusStatus::none) {
        updateDSMState(*cluster, toDSMState(zoneStateDsm));
      }
    } else if (cluster->getColor() == GroupIDGray) {
      dsModel::ShadesStatusStatus zoneStateDsm{};
      std::tie(gs.application, zoneStateDsm) = getShadeGroupApplicationStatus(devices);
      if (zoneStateDsm != dsModel::ShadesStatusStatus::none) {
        updateDSMState(*cluster, toDSMState(zoneStateDsm));
      }
    } else if (cluster->getColor() == GroupIDCyan) {
      dsModel::AudioStatusStatus zoneStateDsm{};
      std::tie(gs.application, zoneStateDsm) = getAudioGroupApplicationStatus(devices);
      if (zoneStateDsm != dsModel::AudioStatusStatus::none) {
        updateDSMState(*cluster, toDSMState(zoneStateDsm));
      }
    }
    DS_DEBUG_LEAVE()
  }

  void completeMoving(const Dsuid& device) const override {
    auto now = std::chrono::system_clock::now();
    m_pool.post([this, device, now]() {
      DS_DEBUG_ENTER(device, now);
      try {
        if (m_hal->completeMoving(device, now)) {
          updateScenarios(device, now);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("completeMoving", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void saveScene(const Dsuid& device, int sceneId) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId]() {
      DS_DEBUG_ENTER(device, sceneId);
      try {
        m_hal->saveScene(device, sceneId, std::chrono::system_clock::now());
        m_cacheReady = false;
        m_apartment->pushEvent(std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent));
      } catch (const std::exception& e) {
        DS_NOTICE("saveScene", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void callScene(const Dsuid& device, int sceneId, bool forced, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId, forced, time]() {
      DS_DEBUG_ENTER(device, sceneId, forced, time);
      try {
        if (m_hal->callScene(device, sceneId, forced, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("callScene", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  std::vector<Dsuid> iterateDevices(int zoneId, int groupId, std::function<bool(Dsuid)> action) const {
    std::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
    Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);
    std::vector<Dsuid> changedDevices{};
    for (int i = 0; i < s.length(); i++) {
      try {
        if (action(s.get(i).getDSID())) {
          changedDevices.push_back(s.get(i).getDSID());
        }
      } catch (const std::exception& e) {
        DS_NOTICE(e.what()); // TODO action type, e.g. callScene not printed
      }
    }
    return changedDevices;
  }

  void callScene(int zoneId, int groupId, int sceneId, bool forced, ds::DateTime time) override {
    m_pool.post([this, zoneId, groupId, sceneId, forced, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, forced, time);

      auto changedDevices = iterateDevices(
          zoneId, groupId, [&](Dsuid dsuid) -> bool { return m_hal->callScene(dsuid, sceneId, forced, time); });

      updateScenarios(changedDevices, zoneId, groupId, time);

      if (((groupId == GroupIDGray) || (groupId == GroupIDBroadcast)) &&
          !changedDevices.empty()) {
        processDsmStopTimers(changedDevices, sceneId, zoneId, groupId);
      }

      std::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);
      if (!s.empty() && SceneHelper::rememberScene(sceneId) && groupId != GroupIDControlTemperature) {
        // mark current scenario active
        auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
        scenarios.scenarios[encodeId(ApplicationZoneScenarioId{.zone = zoneId, .group = groupId, .scene = sceneId})] = {
            .sceneId = sceneId,
            .zoneId = zoneId,
            .groupId = groupId,
            .status = dsModel::ScenarioStatusStatus::active,
            .lastChanged = time};
        m_scenarioCache.update(DSUID_NULL, scenarios);
      }
      dumpScenarioCache(zoneId, groupId);

      if (static_cast<ApplicationType>(groupId) == ApplicationType::audio) {
        DS_DEBUG(groupId);
        auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
        std::vector<int> audioScenariosInZone = m_hal->getPlayerScenarios(zoneId);
        for (int scenario : audioScenariosInZone) {
          if (scenario != sceneId) {
            scenarios.scenarios[encodeId(ApplicationZoneScenarioId{.zone = zoneId, .group = groupId, .scene = scenario})] = {
                .sceneId = sceneId,
                .zoneId = zoneId,
                .groupId = groupId,
                .status = dsModel::ScenarioStatusStatus::inactive,
                .lastChanged = time};
          }
        }
        m_scenarioCache.update(DSUID_NULL, scenarios);
      }
      DS_DEBUG_LEAVE();
    });
  }

  void callSceneMin(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, zoneId, groupId, sceneId, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, time);

      auto changedDevices = iterateDevices(
          zoneId, groupId, [&](Dsuid dsuid) -> bool { return m_hal->callSceneMin(dsuid, sceneId, time); });

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void callSceneMin(const Dsuid& device, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId, time]() {
      DS_DEBUG_ENTER(device, sceneId, time);
      try {
        if (m_hal->callSceneMin(device, sceneId, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("callSceneMin", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void undoScene(int zoneId, int groupId, int sceneId, ds::DateTime time) override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, zoneId, groupId, sceneId, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, time);

      auto changedDevices =
          iterateDevices(zoneId, groupId, [&](Dsuid dsuid) -> bool { return m_hal->undoScene(dsuid, sceneId, time); });

      updateScenarios(changedDevices, zoneId, groupId, time);

      if (((groupId == GroupIDGray) || (groupId == GroupIDBroadcast)) &&
          !changedDevices.empty()) {
        processDsmStopTimers(changedDevices, sceneId, zoneId, groupId);
      }

      DS_DEBUG_LEAVE();
    });
  }

  void undoScene(const Dsuid& device, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId, time]() {
      DS_DEBUG_ENTER(device, sceneId, time);
      try {
        if (m_hal->undoScene(device, sceneId, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("undoScene", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setOutputValue(int zoneId, int groupId, uint8_t value, ds::DateTime time) const override {
    m_pool.post([this, zoneId, groupId, value, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, time);

      auto changedDevices = iterateDevices(
          zoneId, groupId, [&](Dsuid dsuid) -> bool { return m_hal->setOutputValue(dsuid, value, time); });

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void setOutputValue(const Dsuid& device, uint8_t value, ds::DateTime time) const override {
    m_pool.post([this, device, value, time]() {
      DS_DEBUG_ENTER(device, time);
      try {
        if (m_hal->setOutputValue(device, value, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("setOutputValue", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setOutputValue(const Dsuid& device, ChannelType channel, double value, ds::DateTime time) const override {
    m_pool.post([this, device, channel, value, time]() {
      DS_DEBUG_ENTER(device, channel, time);
      try {
        if (m_hal->setOutputValue(device, channel, value, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("setOutputValue", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setSceneLocalPrio(int zoneId, int groupId, int sceneId) const override {
    m_pool.post([this, zoneId, groupId, sceneId]() {
      DS_DEBUG_ENTER(zoneId, groupId);

      auto changedDevices = iterateDevices(
          zoneId, groupId, [&](Dsuid dsuid) -> bool { return m_hal->setSceneLocalPrio(dsuid, sceneId); });

      if (!changedDevices.empty()) {
        syncWithDsm(zoneId, groupId);
        sendChangeEvent();
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setOperationLock(const Dsuid& device, bool lock) const override {
    m_pool.post([this, device, lock]() {
      DS_DEBUG_ENTER(device);
      try {
        m_hal->setOperationLock(device, lock);
      } catch (const std::exception& e) {
        DS_NOTICE("setOperationLock", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void stopDimming(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {
    m_pool.post([this, zoneId, groupId, sceneId, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, time);

      auto changedDevices = iterateDevices(
          zoneId, groupId, [&](Dsuid dsuid) -> bool { return m_hal->stopDimming(dsuid, sceneId, time); });

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void stopDimming(const Dsuid& device) const override {
    auto now = std::chrono::system_clock::now();
    m_pool.post([this, device, now]() {
      DS_DEBUG_ENTER(device);

      try {
        if (m_hal->stopDimming(device, 0, now)) {
          updateScenarios(device, now);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("stopDimming", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  dsModel::ScenarioStatus zoneScenarioStatus(const dsModel::ScenarioId& scenarioId) const override {
    DS_DEBUG_ENTER(scenarioId);
    try {
      auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
      for (const auto& scenario : scenarios.scenarios) {
        if (scenarioId == dsModel::ScenarioId{scenario.first}) {
          return {.id = dsModel::ScenarioId{scenario.first},
                  .status = scenario.second.status,
                  .completed = {},
                  .scheduledAt = scenario.second.lastChanged,
                  .terminatesAt = scenario.second.lastChanged};
        }
      }
      DS_INFO("Zone scenario could not be found in cache", scenarioId);
    } catch (std::exception& e) {
      DS_ERROR("Error looking up zone scenario in cache", e);
    }
    DS_DEBUG_LEAVE();
    return {.id = scenarioId};
  }

  dsModel::DSDeviceStatus dsDeviceStatus(const dsModel::DSDeviceId& id) const override {
    DS_DEBUG_ENTER(id);
    auto state = m_hal->getState(Dsuid{id});

    auto dsOutputs = std::vector<dsModel::OutputStatus>{};
    auto cache = m_hal->getCache().tryLookup(Dsuid{id});
    if (!(cache.property[DeviceProperty::outputMode] &&
          *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED)) {
      for (const auto& output : state.outputs) {
        dsOutputs.push_back(dsModel::OutputStatus{
            .id = dsModel::OutputId{ds::str(output.first)},
            .value = (state.status == dsModel::OutputStatusStatus::moving ||
                      state.status == dsModel::OutputStatusStatus::dimming)
                         ? std::optional<double>{}
                         : output.second.value,
            .initialValue = (state.status == dsModel::OutputStatusStatus::moving ||
                             state.status == dsModel::OutputStatusStatus::dimming)
                                ? output.second.value
                                : std::optional<double>{},
            .status = state.status,
            .targetValue = output.second.targetValue,
            .startedAt = (state.status == dsModel::OutputStatusStatus::moving ||
                          state.status == dsModel::OutputStatusStatus::dimming)
                             ? state.lastChanged
                             : ds::DateTime{},
            .terminatesAt = (state.terminatesAt) ? *state.terminatesAt : ds::DateTime{},
            .level = output.second.level,
        });
      }
    }

    DS_DEBUG_LEAVE();
    return dsModel::DSDeviceStatus{.id = id,
                                   .type = "dsDeviceStatus",
                                   .lastChanged = {},
                                   .attributes = {.functionBlocks = {{
                                                      .id = dsModel::FunctionBlockId{id},
                                                      .outputs = std::move(dsOutputs),
                                                  }},
                                                  .submodules = {{
                                                    .id = dsModel::SubmoduleId{id},
                                                    .operationsLocked = m_apartment->device(id)->operationsLock(),
                                                  }}}};
  }

  dsModel::LightsStatus lightsStatus(const dsModel::ZoneId& id) const override {
    DS_DEBUG_ENTER(id);
    auto result = dsModel::LightsStatus{};
    auto zone = m_apartment->zone(id);

    auto devices = zone->getDevices().getByPresence(true).getByGroup(GroupIDYellow);
    if (!devices.empty()) {
      result = getLightApplicationStatus(devices);
    }
    DS_DEBUG_LEAVE();
    return result;
  }
  dsModel::ShadesStatus shadesStatus(const dsModel::ZoneId& id) const override {
    DS_DEBUG_ENTER(id);
    auto result = dsModel::ShadesStatus{};
    auto zone = m_apartment->zone(id);

    auto devices = zone->getDevices().getByPresence(true).getByGroup(GroupIDGray);
    if (!devices.empty()) {
      result = getShadeApplicationStatus(devices);
    }
    return result;
  }

  dsModel::AudioStatus audioStatus(const dsModel::ZoneId& id) const override {
    DS_DEBUG_ENTER(id);
    auto result = dsModel::AudioStatus{};
    auto zone = m_apartment->zone(id);
    auto zoneId = strToInt(id);

    dsModel::ZoneApplicationsStatus zoneApplicationsCache = m_hal->getZoneState(zoneId);

    result = zoneApplicationsCache.audio;

    auto devices = zone->getDevices().getByPresence(true).getByGroup(GroupIDCyan);
    if (!devices.empty()) {
      DS_DEBUG("Audio devices:", devices.size());
      result = getAudioApplicationStatus(devices, result);
    }
    DS_DEBUG(result);
    return result;
  }

  dsModel::Ventilation ventilationUnitLevels() const override {
    DS_DEBUG_ENTER();
    const auto& zone = *m_apartment->zone(dsModel::ZoneId{"0"});

    if (!boost::find<boost::return_found_end>(m_apartment->apartmentApplications(),
                                              dsModel::ApartmentApplicationType::ventilationUnit)) {
      DS_DEBUG("Apartment ventilation application does not exist.");
      return {};
    }

    // check for "present" and "invisible" devices
    auto ventilationUnitDevices = zone.getDevices().getByGroup(GroupIDGlobalAppDsVentilation);
    Set devices;
    for (int i = 0; i < ventilationUnitDevices.length(); i++) {
      const DeviceReference& d = ventilationUnitDevices.get(i);
      if (d.getDevice()->isVisible() && d.getDevice()->isPresent()) {
        devices.addDevice(d);
      }
    }
    DS_DEBUG(devices.size());
    if (devices.empty()) {
      return {};
    }

    std::vector<dsModel::VentilationLevel> availableLevels{};
    for (const auto& level : zone.getGroup(GroupIDGlobalAppDsVentilation)->getAvailableScenes()) {
      if (level == 0) {
        availableLevels.push_back(dsModel::VentilationLevel::off);
      } else if (level == 5) {
        availableLevels.push_back(dsModel::VentilationLevel::level1);
      } else if (level == 17) {
        availableLevels.push_back(dsModel::VentilationLevel::level2);
      } else if (level == 18) {
        availableLevels.push_back(dsModel::VentilationLevel::level3);
      } else if (level == 19) {
        availableLevels.push_back(dsModel::VentilationLevel::level4);
      } else if (level == 6) {
        availableLevels.push_back(dsModel::VentilationLevel::boost);
      }
    }

    DS_DEBUG(availableLevels);
    return dsModel::Ventilation{.availableLevels = availableLevels};
  };

  dsModel::Ventilation ventilationLevelsImpl(const dsModel::ZoneId& zoneId,
                                             const dsModel::ZoneApplicationType application) const {
    DS_DEBUG_ENTER();
    DS_CONTEXT(zoneId);
    const auto& zone = *m_apartment->zone(zoneId);
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.applications(), application), "Application does not exist.",
               application, zoneId);

    int groupNumber{};
    if (application == dsModel::ZoneApplicationType::ventilation) {
      groupNumber = GroupIDVentilation;
    } else if (application == dsModel::ZoneApplicationType::recirculation) {
      groupNumber = GroupIDRecirculation;
    }

    // check for "present" and "invisible" devices
    auto ventilationDevices = zone.getDevices().getByGroup(groupNumber);
    Set devices;
    for (int i = 0; i < ventilationDevices.length(); i++) {
      const DeviceReference& d = ventilationDevices.get(i);
      if (d.getDevice()->isVisible() && d.getDevice()->isPresent()) {
        devices.addDevice(d);
      }
    }
    DS_DEBUG(devices.size());
    if (devices.empty()) {
      return {};
    }

    PropertyNodePtr node;
    int levels{};
    bool offAvailable{};
    if (application == dsModel::ZoneApplicationType::ventilation) {
      node = m_zoneVentilationNode->getPropertyByName(ds::str(zoneId));
      if (node) {
        levels = m_zoneVentilationNode->tryChildValue<int>(ds::str(zoneId, "/ventilationLevel")).value_or(1);
        offAvailable =
            m_zoneVentilationNode->tryChildValue<bool>(ds::str(zoneId, "/ventilationManual")).value_or(false);
      } else {
        DS_DEBUG("Empty return");
        return {};
      }
    } else if (application == dsModel::ZoneApplicationType::recirculation) {
      node = m_zoneRecirculationNode->getPropertyByName(ds::str(zoneId));
      if (node) {
        levels = m_zoneRecirculationNode->tryChildValue<int>(ds::str(zoneId, "/ventilationLevel")).value_or(1);
        offAvailable =
            m_zoneRecirculationNode->tryChildValue<bool>(ds::str(zoneId, "/ventilationManual")).value_or(false);
      } else {
        DS_DEBUG("Empty return");
        return {};
      }
    }

    std::vector<dsModel::VentilationLevel> availableLevels{};
    if (offAvailable) {
      availableLevels.push_back(dsModel::VentilationLevel::off);
    }
    availableLevels.push_back(dsModel::VentilationLevel::level1);
    if (levels > 1) {
      availableLevels.push_back(dsModel::VentilationLevel::level2);
    }
    if (levels > 2) {
      availableLevels.push_back(dsModel::VentilationLevel::level3);
    }
    if (levels > 3) {
      availableLevels.push_back(dsModel::VentilationLevel::level4);
    }
    availableLevels.push_back(dsModel::VentilationLevel::boost);

    DS_DEBUG(availableLevels);
    return dsModel::Ventilation{.availableLevels = availableLevels};
  };

  [[nodiscard]] dsModel::Ventilation ventilationLevels(const dsModel::ZoneId& zoneId) const override {
    return ventilationLevelsImpl(zoneId, dsModel::ZoneApplicationType::ventilation);
  };

  [[nodiscard]] dsModel::Ventilation recirculationLevels(const dsModel::ZoneId& zoneId) const override {
    return ventilationLevelsImpl(zoneId, dsModel::ZoneApplicationType::recirculation);
  }

  dsModel::VentilationLevel ventilationLastScene(const dss::Group& group) const {
    const auto& lastScene = group.getLastCalledScene();
    if (lastScene == 0) {
      return dsModel::VentilationLevel::off;
    } else if (lastScene == 5) {
      return dsModel::VentilationLevel::level1;
    } else if (lastScene == 17) {
      return dsModel::VentilationLevel::level2;
    } else if (lastScene == 18) {
      return dsModel::VentilationLevel::level3;
    } else if (lastScene == 19) {
      return dsModel::VentilationLevel::level4;
    } else if (lastScene == 6) {
      return dsModel::VentilationLevel::boost;
    }
    DS_FAIL_REQUIRE("Invalid last called scene in ventilation application", lastScene);
  }

  bool propertyValueToBool(std::optional<int> value) const {
    if (value == 1) {
      return true;
    } else {
      return false;
    }
  }

  dsModel::VentilationStatus ventilationUnitStatus() const override {
    DS_DEBUG_ENTER();
    const auto& zone = *m_apartment->zone(dsModel::ZoneId{"0"});

    try {
      if (!boost::find<boost::return_found_end>(m_apartment->apartmentApplications(),
                                                dsModel::ApartmentApplicationType::ventilationUnit)) {
        DS_DEBUG("Apartment ventilation application does not exist.");
        return {};
      }

      const auto& group = *zone.getGroup(64);
      dsModel::VentilationLevel level = ventilationLastScene(group);

      if (!m_statesNode->getPropertyByName("zone.0.group.64.status.malfunction") ||
          !m_statesNode->getPropertyByName("zone.0.group.64.status.service")) {
        return dsModel::VentilationStatus({});
      }

      bool malfunction{};
      bool serviceRequired{};

      if (auto&& malfunctionNode = m_statesNode->getPropertyByName(ds::str("zone.0.group.64.status.malfunction"))) {
        std::optional<int> value = malfunctionNode->tryChildValue<int>("value");
        malfunction = propertyValueToBool(value);
      }
      if (auto&& serviceNode = m_statesNode->getPropertyByName(ds::str("zone.0.group.64.status.service"))) {
        std::optional<int> value = serviceNode->tryChildValue<int>("value");
        serviceRequired = propertyValueToBool(value);
      }
      DS_DEBUG(level, malfunction, serviceRequired);

      return dsModel::VentilationStatus{
          .level = level,
          .malfunction = malfunction,
          .serviceRequired = serviceRequired,
      };
    } catch (const std::exception& e) {
      DS_ERROR("Failed to get apartment ventilation status", e);
      return {};
    }
  }

  dsModel::VentilationStatus ventilationStatusImpl(const dsModel::ZoneId& zoneId,
                                                   const dsModel::ZoneApplicationType application) const {
    DS_CONTEXT(zoneId);
    const auto& zone = *m_apartment->zone(zoneId);
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.applications(), application), "Application does not exist.",
               application, zoneId);

    try {
      int groupNumber{};

      PropertyNodePtr node;
      if (application == dsModel::ZoneApplicationType::ventilation) {
        node = m_zoneVentilationNode->getPropertyByName(ds::str(zoneId));
        groupNumber = GroupIDVentilation;
      } else if (application == dsModel::ZoneApplicationType::recirculation) {
        node = m_zoneRecirculationNode->getPropertyByName(ds::str(zoneId));
        groupNumber = GroupIDRecirculation;
      }
      auto devices = zone.getDevices().getByPresence(true).getByGroup(groupNumber);
      DS_DEBUG(devices.size());
      if (devices.empty()) {
        return {};
      }

      const auto& group = *zone.getGroup(groupNumber);
      dsModel::VentilationLevel level = ventilationLastScene(group);

      if (!m_statesNode->getPropertyByName(ds::str("zone.", zoneId, ".group.", groupNumber, ".status.malfunction")) ||
          !m_statesNode->getPropertyByName(ds::str("zone.", zoneId, ".group.", groupNumber, ".status.service"))) {
        return dsModel::VentilationStatus({});
      }

      bool malfunction{};
      bool serviceRequired{};

      if (auto&& malfunctionNode = m_statesNode->getPropertyByName(
              ds::str("zone.", zoneId, ".group.", groupNumber, ".status.malfunction"))) {
        std::optional<int> value = malfunctionNode->tryChildValue<int>("value");
        malfunction = propertyValueToBool(value);
      }
      if (auto&& serviceNode =
              m_statesNode->getPropertyByName(ds::str("zone.", zoneId, ".group.", groupNumber, ".status.service"))) {
        std::optional<int> value = serviceNode->tryChildValue<int>("value");
        serviceRequired = propertyValueToBool(value);
      }
      DS_DEBUG(level, malfunction, serviceRequired);

      return dsModel::VentilationStatus{
          .level = level,
          .malfunction = malfunction,
          .serviceRequired = serviceRequired,
      };
    } catch (const std::exception& e) {
      DS_ERROR("Failed to get ventilation status", e);
      return {};
    }
  }

  [[nodiscard]] dsModel::VentilationStatus ventilationStatus(const dsModel::ZoneId& zoneId) const override {
    return ventilationStatusImpl(zoneId, dsModel::ZoneApplicationType::ventilation);
  }

  [[nodiscard]] dsModel::VentilationStatus recirculationStatus(const dsModel::ZoneId& zoneId) const override {
    return ventilationStatusImpl(zoneId, dsModel::ZoneApplicationType::recirculation);
  }

  std::vector<dsModel::GroupStatus> groupsStatus() const override {
    DS_DEBUG_ENTER();
    std::vector<std::shared_ptr<Cluster>> clusters = m_apartment->getClusters();
    auto result = std::vector<dsModel::GroupStatus>{};

    for (auto& cluster : clusters) {
      if ((cluster->getColor() != GroupIDYellow) && (cluster->getColor() != GroupIDGray)) {
        continue;
      }

      Set devices = cluster->getDevices().getByPresence(true);
      if (devices.empty()) {
        continue;
      }

      dsModel::GroupStatus gs{};
      gs.id = intToString(cluster->getID());

      if (cluster->getColor() == GroupIDYellow) {
        dsModel::LightsStatusStatus zoneStateDsm{};
        std::tie(gs.application, zoneStateDsm) = getLightGroupApplicationStatus(devices);
        result.push_back(gs);
      } else if (cluster->getColor() == GroupIDGray) {
        dsModel::ShadesStatusStatus zoneStateDsm{};
        std::tie(gs.application, zoneStateDsm) = getShadeGroupApplicationStatus(devices);
        result.push_back(gs);
      } else if (cluster->getColor() == GroupIDCyan) {
        dsModel::AudioStatusStatus zoneStateDsm{};
        std::tie(gs.application, zoneStateDsm) = getAudioGroupApplicationStatus(devices);
        result.push_back(gs);
      }
    }
    DS_DEBUG_LEAVE();
    return result;
  }

  void updateSceneConfig(const DeviceSceneChannelId& deviceSceneChannelId, double value) const override {
    m_pool.post([this, &dsuid = deviceSceneChannelId.scene.dsuid, path = toPath(deviceSceneChannelId), value]() {
      m_hal->write(dsuid, path, value);
    });
  }
  void updateDeviceSceneChannelProperty(const DeviceSceneChannelPropertyId& deviceSceneChannelPropertyId,
                                        bool value) const override {
    m_pool.post([this, &dsuid = deviceSceneChannelPropertyId.channel.scene.dsuid,
                 path = toPath(deviceSceneChannelPropertyId), value]() { m_hal->write(dsuid, path, value); });
  }

  void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const override {
    m_pool.post([this, dsuid, bank, offset, value]() { m_hal->updateDeviceConfig(dsuid, bank, offset, value); });
  }

  std::vector<double> getDeviceThresholds(const Dsuid& dsuid) const override {
    std::vector<double> levelThresholds;
    try {
      if (m_hal->deviceCacheable(dsuid)) {
        levelThresholds = m_hal->getDeviceThresholds(dsuid);
      }
    } catch (const ds::Exception& e) {
      DS_NOTICE("Could not scan DSUID: ", dsuid, e.what());
    }
    return levelThresholds;
  }

  void completeProcessing() const override {
    auto task = std::packaged_task<void(void)>([] {});
    m_pool.post([&] { task(); });
    task.get_future().get();
  }

  std::vector<std::string> scenarioDeviceList(Group& group, int scene) const override {
    DS_DEBUG_ENTER(group, scene);
    auto devices = group.getDevices();
    auto result = std::vector<std::string>{};
    for (int i = 0; i < devices.length(); i++) {
      auto dsid = devices.get(i).getDSID();
      auto cache = m_deviceCache.tryLookup(dsid);
      if (cache.scene.sceneTable[scene].flags.dontCare.value_or(true) == false) {
        result.push_back(ds::str(dsid));
      }
    }
    std::sort(result.begin(), result.end());
    DS_DEBUG_LEAVE();
    return result;
  }

  ParameterCacheStatus getParameterCacheStatus() const override {
    DS_DEBUG_ENTER();
    ParameterCacheStatus result{};

    auto devices = m_apartment->getDevices().getByPresence(true);
    for (size_t i = 0; i < devices.size(); ++i) {
      try {
        if (m_hal->deviceCacheable(devices[i].getDSID())) {
          result.devices.push_back(m_hal->cacheStatus(devices[i].getDSID()));
        } else {
          DS_DEBUG("Skip non-cacheable", devices[i].getDSID());
        }
      } catch (const ds::Exception& e) {
        DS_DEBUG("Skip", devices[i].getDSID(), e.what());
        continue;
      }
    }
    auto total = 0;
    auto completed = 0;
    result.cache.total = 0;
    result.cache.completed = 0;
    for (auto& d : result.devices) {
      (*result.cache.total)++;
      if (d.completed == d.total) {
        (*result.cache.completed)++;
      }
      total += d.total;
      completed += d.completed;
    }
    result.cache.completion = (total > 0) ? (static_cast<double>(completed) / total) * 100 : 100;
    result.cache.remainingTime = (total - completed) * ReadoutTimePerParameter;

    if (m_apartment->getPropertySystem() &&
        m_apartment->getPropertySystem()->getBoolValue(pp_device_cache_update_status)) {
      result.cache.status = dsModel::ParameterCacheStatus::updating;
    } else {
      if (result.cache.completion < 100) {
        result.cache.status = dsModel::ParameterCacheStatus::incomplete;
      } else {
        result.cache.status = dsModel::ParameterCacheStatus::complete;
      }
    }
    m_cacheReady = (result.cache.status == dsModel::ParameterCacheStatus::complete);

    DS_DEBUG_LEAVE();
    return result;
  }

  void updateParameterCacheStatus() const override {
    getParameterCacheStatus();
  }

  bool getEmptyClickPreventionActive() const override {
    return m_emptyClickPreventionActiveLight;
  }
  void setEmptyClickPreventionActive(bool active) override {
    m_emptyClickPreventionActiveLight = active;
    ds::system().conf().writeJson("emptyClickPreventionLight", ds::toJson(active));
  }

private:
  void updateScenarios(const Dsuid& changedDevice, ds::DateTime now) const {
    auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
    auto device = m_apartment->getDeviceByDSID(changedDevice);
    auto cache = m_deviceCache.tryLookup(changedDevice);

    for (auto& scenario : scenarios.scenarios) {
      if (scenario.second.status == dsModel::ScenarioStatusStatus::active &&
          (scenario.second.zoneId == 0 || (scenario.second.zoneId == device->getZoneID())) &&
          (scenario.second.groupId == 0 || (device->isInGroup(scenario.second.groupId)))) {
        DS_DEBUG("Checking scenario", scenario);
        if (cache.scene.sceneTable[scenario.second.sceneId].flags.dontCare.value_or(true) == false) {
          scenario.second.status = dsModel::ScenarioStatusStatus::inactive;
          scenario.second.lastChanged = now;
        }
      }
    }
    m_scenarioCache.update(DSUID_NULL, scenarios);

    int groupId = GroupIDBlack;
    if (device->isInGroup(GroupIDYellow)) {
      groupId = GroupIDYellow;
    } else if (device->isInGroup(GroupIDGray)) {
      groupId = GroupIDGray;
    } else if (device->isInGroup(GroupIDCyan)) {
      groupId = GroupIDCyan;
    }
    auto zone = m_apartment->getZone(device->getZoneID());
    syncWithDsm(device->getZoneID(), groupId);

    auto clusters = device->getClusters();
    for (auto cluster : clusters) {
      syncWithDsm(cluster);
    }
    sendChangeEvent();
  }

  void updateScenarios(const std::vector<Dsuid>& changedDevices, int zoneId, int groupId, ds::DateTime now) const {
    DS_DEBUG_ENTER();

    if (changedDevices.empty()) {
      DS_DEBUG("No devices changed");
      return;
    }

    auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
    for (auto& scenario : scenarios.scenarios) {
      if (scenario.second.status == dsModel::ScenarioStatusStatus::active &&
          (scenario.second.zoneId == 0 || zoneId == 0 || (scenario.second.zoneId == zoneId)) &&
          (scenario.second.groupId == 0 || groupId == 0 || (scenario.second.groupId == groupId))) {
        DS_DEBUG("Checking scenario", scenario);
        for (auto& device : changedDevices) {
          auto cache = m_deviceCache.tryLookup(device);
          if (cache.scene.sceneTable[scenario.second.sceneId].flags.dontCare.value_or(true) == false) {
            scenario.second.status = dsModel::ScenarioStatusStatus::inactive;
            scenario.second.lastChanged = now;
            break; // scenario is inactive, skip remaining devices
          }
        }
      }
    }
    m_scenarioCache.update(DSUID_NULL, scenarios);

    if (zoneId != 0) {
      syncWithDsm(zoneId, groupId);
    } else {
      for (auto& zone : m_apartment->getZones()) {
        if (zone->getID() != 0) {
          auto group = zone->getGroup(groupId);
          if (group != nullptr) {
            if (!group->getDevices().isEmpty()) {
              syncWithDsm(zone->getID(), groupId);
            }
          }
        }
      }
    }

    //get all clusters from devices
    std::map<int, std::shared_ptr<Cluster>> clustersProcessed;
    for (const auto& dsuid : changedDevices) {
      auto device = m_apartment->getDeviceByDSID(dsuid);
      for (auto cluster : device->getClusters()) {
        if (clustersProcessed.find(cluster->getID()) != clustersProcessed.end()) {
          syncWithDsm(cluster);
          clustersProcessed[cluster->getID()] = cluster;
        }
      }
    }
    sendChangeEvent();
    DS_DEBUG_LEAVE();
  }

  void dumpScenarioCache(int zoneId, int groupId) const {
    DS_DEBUG_ENTER();
    auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
    for (auto& scenario : scenarios.scenarios) {
      if ((scenario.second.zoneId == zoneId) && (scenario.second.groupId == groupId)) {
        DS_NOTICE(scenario);
      }
    }
  }

  bool sceneNeedsStopping(int sceneId) const {
    // info obtained from Christoph, no extra stopping for these scenes:
    // 10-12, 15, 42-49, 52-55
    switch (sceneId) {
      case SceneDimArea:
      case SceneDec:
      case SceneInc:
      case SceneStop:
      case SceneDecA1:
      case SceneIncA1:
      case SceneDecA2:
      case SceneIncA2:
      case SceneDecA3:
      case SceneIncA3:
      case SceneDecA4:
      case SceneIncA4:
      case SceneStopA1:
      case SceneStopA2:
      case SceneStopA3:
      case SceneStopA4:
        return false;
      default:
        return true;
    }
  }

  void processDsmStopTimers(const std::vector<Dsuid>& changedDevices, int sceneId, int zoneId, int groupId) {
    if (m_emptyClickPreventionActiveNonLight && m_cacheReady &&
      m_apartment->getBusInterface() &&
      m_apartment->getBusInterface()->getActionRequestInterface()) {
      time_t shadeMovementCompletion = 0;
      time_t soonestEventForApartment = 0;
      {
        // check if an event is already queued for this zone and remove it
        // from our list. The new scene call will override an already queued
        // timer for this zone/group scene call.
        boost::recursive_mutex::scoped_lock lock(m_stopTimerListMutex);
        std::vector<std::tuple<int, int, int, time_t> >::iterator it;

        for (it = m_stopTimerZonesList.begin(); it != m_stopTimerZonesList.end();) {
          if ((std::get<0>(*it) == zoneId) && (std::get<1>(*it) == groupId)) {
            it = m_stopTimerZonesList.erase(it);
          } else {
            // remember next soonest event for the whole apartment
            if ((soonestEventForApartment == 0) ||
                (std::get<3>(*it) < soonestEventForApartment)) {
              soonestEventForApartment = std::get<3>(*it);
            }
            it++;
          }
        }
      }
      // if the scene is a "stop" scene, there won't be any target values
      // cached, then we only need to remove it from the list
      if (sceneNeedsStopping(sceneId)) {
        for (const auto& dsuid : changedDevices) {
          if (groupId == GroupIDBroadcast) { // ignore non grey devices
            try {
              std::shared_ptr<Device> dev = m_apartment->getDeviceByDSID(dsuid);
              if (dev->getDeviceClass() != DeviceClasses_t::DEVICE_CLASS_GR) {
                continue;
              }
            } catch(ItemNotFoundException& e) {
              continue;
            }
          }

          auto sc = m_hal->getStateCache().tryLookup(dsuid);
          // if there is at least one device we can't calculate, we leave it to
          // the default stop timer logic
          if (!sc.terminatesAt.has_value()) {
            return;
          }

          // value of terminatesAt is stored in milliseconds)
          time_t termination = std::chrono::system_clock::to_time_t(sc.terminatesAt.value());
          if (termination > shadeMovementCompletion) {
            shadeMovementCompletion = termination;
          }
        }
      }

      if (!sceneNeedsStopping(sceneId) || !shadeMovementCompletion) {
          return;
      }

      {
        boost::recursive_mutex::scoped_lock lock(m_stopTimerListMutex);
        m_stopTimerZonesList.push_back(std::make_tuple(zoneId, groupId, sceneId, shadeMovementCompletion));
      }

      if ((soonestEventForApartment == 0) ||
          (shadeMovementCompletion < soonestEventForApartment)) {
        soonestEventForApartment = shadeMovementCompletion;
      }

      std::shared_ptr<Event> event = std::make_shared<Event>(EventName::StopDsmTimer);
      time_t diffSeconds = soonestEventForApartment -
                                DateTime().secondsSinceEpoch();
      if (diffSeconds <= 0) {
        diffSeconds = 1;
      }
      event->setProperty("time", "+" + intToString(diffSeconds));
      event->setProperty(EventProperty::Unique, "Yes");
      m_apartment->pushEvent(event);
    }
  }

  dsModel::LightsStatusStatus getLightStateFor(dsModel::LightsStatusStatus roomState,
                                               dsModel::OutputStatusStatus deviceState,
                                               DeviceChannelValues& channel) const {
    if (roomState == dsModel::LightsStatusStatus::dimming) {
      // dimming is highest state, no further transition possible
      return roomState;
    }
    if ((deviceState == dsModel::OutputStatusStatus::dimming) || (deviceState == dsModel::OutputStatusStatus::moving)) {
      return dsModel::LightsStatusStatus::dimming;
    }
    if (roomState == dsModel::LightsStatusStatus::on) {
      // next highest state after dimming, no further transition possible
      return roomState;
    }
    if (deviceState == dsModel::OutputStatusStatus::ok) {
      if ((!channel.level && channel.value > 0) || (channel.level > 0)) {
        return dsModel::LightsStatusStatus::on;
      } else {
        return dsModel::LightsStatusStatus::off;
      }
    }
    return roomState;
  }

  dsModel::LightsStatus getLightApplicationStatus(Set devices) const {
    DS_DEBUG_ENTER(devices.size());
    dsModel::LightsStatus aps{.status = {}, .area1 = {}, .area2 = {}, .area3 = {}, .area4 = {}};

    /* If at least one device is "dimming", then the whole zone is
     * dimming, otherwise, if the device returns "ok", then check
     * its brightness channel output. If output of at least one
     * device is >0 then the zone is "on", else it is "off".  */

    auto lookupState = [](auto area, auto& status) -> dsModel::LightsStatusStatus& {
      switch (area) {
        case SceneA11: return status.area1;
        case SceneA21: return status.area2;
        case SceneA31: return status.area3;
        case SceneA41: return status.area4;
      }
      DS_FAIL_REQUIRE("invalid area", area);
    };

    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::exception& ex) {
        DS_WARNING("Ignoring state for device", devRef.getDSID(), ex.what());
        continue;
      } catch (...) {
        /* I was not able to figure out which exception was thrown, perhaps
         * something custom originating from the testing framework? */
        DS_WARNING("Ignoring state for device", devRef.getDSID(), " unknown exception.");
        continue;
      }

      for (auto area : m_hal->areaMember(devRef.getDSID())) {
        auto& areaState = lookupState(area, aps);
        areaState = getLightStateFor(areaState, state.status, state.outputs[ChannelType::brightness]);
      }

      // union set: public API, smartphone, is any device is on?
      aps.status = getLightStateFor(aps.status, state.status, state.outputs[ChannelType::brightness]);
      if (!state.localPriority) {
        // difference set: for empty clicks, DSM needs to know what the next click in the room should do.
        aps.nonLocalPriority =
            getLightStateFor(aps.nonLocalPriority, state.status, state.outputs[ChannelType::brightness]);
      }
    }
    
    DS_DEBUG_LEAVE();
    return aps;
  }

  dsModel::ShadesStatusStatus getShadeStateFor(dsModel::ShadesStatusStatus roomStatus,
                                               dsModel::OutputStatusStatus deviceStatus, double value) const {
    if (roomStatus == dsModel::ShadesStatusStatus::moving) {
      return dsModel::ShadesStatusStatus::moving;
    }
    if (deviceStatus == dsModel::OutputStatusStatus::moving || deviceStatus == dsModel::OutputStatusStatus::dimming) {
      return dsModel::ShadesStatusStatus::moving;
    }
    if (roomStatus == dsModel::ShadesStatusStatus::mixed) {
      return dsModel::ShadesStatusStatus::mixed;
    }
    if (deviceStatus == dsModel::OutputStatusStatus::ok) {
      value = round(value);
      if (value == 0) {
        if (roomStatus == dsModel::ShadesStatusStatus::open) {
          return dsModel::ShadesStatusStatus::mixed;
        } else {
          return dsModel::ShadesStatusStatus::closed;
        }
      } else if (value == 100) {
        if (roomStatus == dsModel::ShadesStatusStatus::closed) {
          return dsModel::ShadesStatusStatus::mixed;
        } else {
          return dsModel::ShadesStatusStatus::open;
        }
      } else {
        return dsModel::ShadesStatusStatus::mixed;
      }
    }

    return roomStatus;
  }

  dsModel::ShadesStatus getShadeApplicationStatus(Set devices) const {
    dsModel::ShadesStatus aps{.status = {}, .area1 = {}, .area2 = {}, .area3 = {}, .area4 = {}};

    /* When all are closed, the state is closed, when all are open, the
     * the state is open, if some are open and some closed, then the
     * state is mixed, but if at least one is moving, then the state is
     * moving. */

    auto lookupState = [](auto area, auto& status) -> dsModel::ShadesStatusStatus& {
      switch (area) {
        case SceneA11: return status.area1;
        case SceneA21: return status.area2;
        case SceneA31: return status.area3;
        case SceneA41: return status.area4;
      }
      DS_FAIL_REQUIRE("invalid area", area);
    };

    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::runtime_error& ex) {
        DS_WARNING("Ignoring state for device", devRef.getDSID(), ex.what());
        continue;
      }

      for (auto area : m_hal->areaMember(devRef.getDSID())) {
        auto& areaState = lookupState(area, aps);
        areaState = getShadeStateFor(areaState, state.status, state.outputs[ChannelType::shadePositionOutside].value);
      }

      // union set of device states: high level view, e.g. public API, smartphone, is any device is on in the room?
      aps.status = getShadeStateFor(aps.status, state.status, state.outputs[ChannelType::shadePositionOutside].value);

      // difference set of device states: needed for empty clicks, DSM needs
      // to know what the next click in the room should do.  for this it
      // excludes devices that are in a local priority areas hence don't
      // react to callScene in a room
      if (!state.localPriority) {
        // difference set: for empty clicks, DSM needs to know what the next click in the room should do.
        // this excludes devices that are in a local priority areas
        aps.nonLocalPriority = getShadeStateFor(aps.nonLocalPriority, state.status,
                                                state.outputs[ChannelType::shadePositionOutside].value);
      }
    }

    return aps;
  }

  dsModel::AudioStatusStatus getAudioStateFor(dsModel::AudioStatusStatus roomStatus,
                                              dsModel::OutputStatusStatus deviceStatus,
                                              DeviceChannelValues& channel) const {
    if (roomStatus == dsModel::AudioStatusStatus::playing) {
      return dsModel::AudioStatusStatus::playing;
    }
    if (deviceStatus == dsModel::OutputStatusStatus::standby) {
      if (roomStatus == dsModel::AudioStatusStatus::powerOff || roomStatus == dsModel::AudioStatusStatus::off || roomStatus == dsModel::AudioStatusStatus::none) {
        roomStatus = dsModel::AudioStatusStatus::standby;
      }
    }
    else if (deviceStatus == dsModel::OutputStatusStatus::ok) {
      if (channel.level > 0) {
        roomStatus = dsModel::AudioStatusStatus::playing;
      } else if (roomStatus != dsModel::AudioStatusStatus::standby && roomStatus != dsModel::AudioStatusStatus::pause) {
        roomStatus = dsModel::AudioStatusStatus::off;
      }
    }
    return roomStatus;
  }

  dsModel::AudioStatus getAudioApplicationStatus(Set devices, dsModel::AudioStatus existingZoneState) const {
    DS_DEBUG(devices.size());
    dsModel::AudioStatus aps = existingZoneState;

    dsModel::AudioStatusStatus zoneState = existingZoneState.status;

    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::runtime_error& ex) {
        DS_WARNING("Ignoring state for device ", devRef.getDSID(), ex.what());
        continue;
      }

      zoneState = getAudioStateFor(zoneState, state.status, state.outputs[ChannelType::powerState]);
      if (!state.localPriority) {
        // difference set: for empty clicks, DSM needs to know what the next click in the room should do.
        aps.nonLocalPriority =
            getAudioStateFor(aps.nonLocalPriority, state.status, state.outputs[ChannelType::powerState]);
      }
    }

    aps.status = zoneState;

    return aps;
  }

  std::pair<dsModel::ZoneApplicationsStatus, dsModel::ShadesStatusStatus> getShadeGroupApplicationStatus(
      Set devices) const {
    if (devices.empty()) {
      return {{}, {}};
    }

    auto shadesStatus = getShadeApplicationStatus(devices);
    return {dsModel::ZoneApplicationsStatus{.shades = {.status = shadesStatus.status}}, shadesStatus.nonLocalPriority};
  }

  std::pair<dsModel::ZoneApplicationsStatus, dsModel::LightsStatusStatus> getLightGroupApplicationStatus(
      Set devices) const {
    if (devices.empty()) {
      return {{}, {}};
    }

    auto lightStatus = getLightApplicationStatus(devices);
    return {{.lights = {.status = lightStatus.status}}, lightStatus.nonLocalPriority};
  }

  std::pair<dsModel::ZoneApplicationsStatus, dsModel::AudioStatusStatus> getAudioGroupApplicationStatus(
      Set devices) const {
    if (devices.empty()) {
      return {{}, {}};
    }

    auto audioStatus = getAudioApplicationStatus(devices, {});
    return {dsModel::ZoneApplicationsStatus{.audio = {.status = audioStatus.status}}, audioStatus.nonLocalPriority};
  }

  void sendChangeEvent() const {
    m_apartment->pushEvent(std::make_shared<Event>(EventName::ApartmentProxyStateChanged));
  }

private:
  HardwareAbstraction* m_hal;
  mutable CacheDb<ScenarioCache> m_scenarioCache;
  mutable DsmZoneStateCache m_dsmCache;
  CacheDb<DeviceCache>& m_deviceCache;
  mutable ThreadPool m_pool;
  mutable bool m_cacheReady;
  bool m_emptyClickPreventionActiveLight;
  bool m_emptyClickPreventionActiveNonLight;
  // zone id, group id, scene id, timestamp
  std::vector<std::tuple<int, int, int, time_t> > m_stopTimerZonesList;
  boost::recursive_mutex m_stopTimerListMutex;
}; // namespace dss

std::unique_ptr<ApartmentProxy> makeApartmentProxy(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db,
                                                   EventInterpreter* pInterpreter, const ds::Conf& conf) {
  return std::make_unique<ApartmentProxyImpl>(apt, hal, db, pInterpreter, conf);
}

class NullApartmentProxy : public ApartmentProxy {
public:
  void callScene(const Dsuid& device, int sceneId, bool forced, ds::DateTime time) const override {}
  void callScene(int zoneId, int groupId, int sceneId, bool forced, ds::DateTime time) override {}
  void saveScene(const Dsuid& device, int sceneId) const override {}
  void callSceneMin(const Dsuid& device, int sceneId, ds::DateTime time) const override {}
  void callSceneMin(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {}
  void undoScene(const Dsuid& device, int sceneId, ds::DateTime time) const override {}
  void undoScene(int zoneId, int groupId, int sceneId, ds::DateTime time) override {}
  void setOutputValue(const Dsuid& device, uint8_t value, ds::DateTime time) const override {}
  void setOutputValue(int zoneId, int groupId, uint8_t value, ds::DateTime time) const override {}
  void setOutputValue(const Dsuid& device, ChannelType channel, double value, ds::DateTime time) const override {}
  void setSceneLocalPrio(int zoneId, int groupId, int sceneId) const override {}
  void setOperationLock(const Dsuid& device, bool lock) const override {}
  void stopDimming(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {}
  void stopDimming(const Dsuid& device) const override {}
  void completeMoving(const Dsuid& device) const override {}
  void stopDsmTimer() override {}
  void updateDSMState(Group& group, int zoneState, int area1State, int area2State, int area3State,
                      int area4State) const override {}
  void updateDSMState(Cluster& group, int state) const override {}
  void syncWithDsm(int zoneId, int groupId) const override {}
  void syncWithDsm(std::shared_ptr<Cluster> cluster) const override {}

  dsModel::ScenarioStatus zoneScenarioStatus(const dsModel::ScenarioId&) const override { return {}; }
  dsModel::DSDeviceStatus dsDeviceStatus(const dsModel::DSDeviceId& id) const override { return {}; }
  dsModel::LightsStatus lightsStatus(const dsModel::ZoneId&) const override { return {}; }
  dsModel::ShadesStatus shadesStatus(const dsModel::ZoneId&) const override { return {}; }
  dsModel::AudioStatus audioStatus(const dsModel::ZoneId&) const override { return {}; }
  dsModel::VentilationStatus ventilationUnitStatus() const override { return {}; }
  dsModel::VentilationStatus ventilationStatus(const dsModel::ZoneId& zoneId) const override { return {}; }
  dsModel::VentilationStatus recirculationStatus(const dsModel::ZoneId& zoneId) const override { return {}; }
  std::vector<dsModel::GroupStatus> groupsStatus() const override { return {}; }
  std::vector<std::string> scenarioDeviceList(Group& group, int scene) const override { return {}; }

  void updateSceneConfig(const DeviceSceneChannelId& deviceSceneChannelId, double value) const override {}
  void updateDeviceSceneChannelProperty(const DeviceSceneChannelPropertyId& deviceSceneChannelPropertyId,
                                        bool value) const override {}

  void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const override {}
  std::vector<double> getDeviceThresholds(const Dsuid& dsuid) const override { return {}; }

  dsModel::Ventilation ventilationUnitLevels() const override { return {}; }
  dsModel::Ventilation ventilationLevels(const dsModel::ZoneId& zoneId) const override { return {}; }
  dsModel::Ventilation recirculationLevels(const dsModel::ZoneId& zoneId) const override { return {}; }

  void completeProcessing() const override {}
  ParameterCacheStatus getParameterCacheStatus() const override { return {}; }

  void updateParameterCacheStatus() const override {}
  bool getEmptyClickPreventionActive() const override { return {}; }
  void setEmptyClickPreventionActive(bool active) override { };
};

std::unique_ptr<ApartmentProxy> makeNullApartmentProxy() {
  return std::make_unique<NullApartmentProxy>();
}

} /* namespace dss */
