/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "cache/device-cache.h"
#include <trompeloeil.hpp>
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"

static ds::LogChannel dsLogChannel("deviceCache", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][deviceCache]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

TEST_CASE("serializeToJSON", TAGS) {
  DeviceCache cache = {
      ds::DateTime(std::chrono::milliseconds(1505731033777)),
      {120,
       {{5,
         {
             ds::DateTime(std::chrono::milliseconds(1505731033777)),
             {false, true, false, true, LedConfig::ledCon1, DimTime::noDimTime},
             std::string("Preset1"),
             {{ChannelType::brightness, {141.65, false, {}}}, {ChannelType::shadePositionIndoor, {52.5, true, {}}}},
         }},
        {6,
         {ds::DateTime(std::chrono::milliseconds(150573103378)),
          {false, true, false, true, LedConfig::ledCon2, DimTime::dimTime1},
          std::string("Preset2"),
          {{ChannelType::brightness, {100, false, {}}}}}}}},
      {{DeviceProperty::outputMode, 1}}};

  REQUIRE(cache.scene.sceneTable[5].action == std::string("Preset1"));

  REQUIRE(!ds::toJson(cache).empty());
}

TEST_CASE("serializeFromJSON", TAGS) {
  std::string jsonString = R"(
  {
    "lastChanged": "2017-09-18T10:37:13.777Z",
    "scene": {
      "changeCounter": 120,
      "sceneTable": {
        "5": {
          "lastChanged": "2017-09-18T10:37:13.777Z",
          "flags": {
            "dontCare": false,
            "ignoreLocalPrio": true,
            "specialHandling": false,
            "blink": true,
            "ledConfig": "ledCon1",
            "dimTime": "noDimTime"
          },
          "action": "Preset1",
          "channels": {
            "brightness": {
              "value": 141.65,
              "dontCare": false
            },
            "shadePositionIndoor": {
              "value": 52.5,
              "dontCare": true
            }
          }
        },
        "6": {
          "lastChanged": "1974-10-09T17:51:43.378Z",
          "flags": {
            "dontCare": false,
            "ignoreLocalPrio": true,
            "specialHandling": false,
            "blink": true,
            "ledConfig": "ledCon2",
            "dimTime": "dimTime1"
          },
          "action": "Preset2",
          "channels": {
            "brightness": {
              "value": 100,
              "dontCare": false
            }
          }
        }
      }
    },
    "property": {
      "outputMode": 1
    }
  })";

  DeviceCache cache = ds::fromJson<DeviceCache>(jsonString);

  REQUIRE(*cache.scene.sceneTable[5].channels[ChannelType::brightness].value == 141.65);

  REQUIRE(!ds::toJson(cache).empty());
}

TEST_CASE("comparisonTest", TAGS) {
  DeviceCache cacheFromInitializer = {
      ds::DateTime(std::chrono::milliseconds(1505731033777)),
      {120,
       {{5,
         {
             ds::DateTime(std::chrono::milliseconds(1505731033777)),
             {false, true, false, true, LedConfig::ledCon1, DimTime::noDimTime},
             std::string("Preset1"),
             {{ChannelType::brightness, {141.65, false, {}}}, {ChannelType::shadePositionIndoor, {52.5, true, {}}}},
         }},
        {6,
         {ds::DateTime(std::chrono::milliseconds(150573103378)),
          {false, true, false, true, LedConfig::ledCon2, DimTime::dimTime1},
          std::string("Preset2"),
          {{ChannelType::brightness, {100, false, {}}}}}}}},
      {{DeviceProperty::outputMode, 1}}};

  std::string jsonString = R"(
  {
    "lastChanged": "2017-09-18T10:37:13.777Z",
    "scene": {
      "changeCounter": 120,
      "sceneTable": {
        "5": {
          "lastChanged": "2017-09-18T10:37:13.777Z",
          "flags": {
            "dontCare": false,
            "ignoreLocalPrio": true,
            "specialHandling": false,
            "blink": true,
            "ledConfig": "ledCon1",
            "dimTime": "noDimTime"
          },
          "action": "Preset1",
          "channels": {
            "brightness": {
              "value": 141.65,
              "dontCare": false
            },
            "shadePositionIndoor": {
              "value": 52.5,
              "dontCare": true
            }
          }
        },
        "6": {
          "lastChanged": "1974-10-09T17:51:43.378Z",
          "flags": {
            "dontCare": false,
            "ignoreLocalPrio": true,
            "specialHandling": false,
            "blink": true,
            "ledConfig": "ledCon2",
            "dimTime": "dimTime1"
          },
          "action": "Preset2",
          "channels": {
            "brightness": {
              "value": 100,
              "dontCare": false
            }
          }
        }
      }
    },
    "property": {
      "outputMode": 1
    }
  })";

  DeviceCache cacheFromString = ds::fromJson<DeviceCache>(jsonString);

  REQUIRE(cacheFromInitializer == cacheFromString);
}

TEST_CASE("enumsToStringTest", TAGS) {
  std::stringstream testString;
  ChannelType channel = ChannelType::airLouverAuto;

  testString << channel;
  REQUIRE(testString.str() == "airLouverAuto");

  testString.str(std::string());

  LedConfig config = LedConfig::ledCon6;
  testString << config;
  REQUIRE(testString.str() == "ledCon6");

  testString.str(std::string());

  DimTime dimTime = DimTime::notSupported;
  testString << dimTime;
  REQUIRE(testString.str() == "notSupported");

  testString.str(std::string());

  DeviceProperty property = DeviceProperty::dimStepPause;
  testString << property;
  REQUIRE(testString.str() == "dimStepPause");
}

} // namespace
