/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <ds-api/translatable.h>
#include <ds-model/devices-fwd.h>
#include <ds-model/model-fwd.h>
#include <ds-model/scenarios-fwd.h>
#include <ds/enum.h>
#include <ds/system-fwd.h>
#include <gsl.h>

#include <locale>
#include <string>

#include "businterface-fwd.h"
#include "vdc-db.h"

DS_REFLECTED_ENUM((dss), DeviceInfoPrefer,
                  /// Read events, properties, etc from db unless vdc sets `dynamicDefinitions` property
                  (none)
                  /// Read events, properties, etc from db.
                  /// Old behavior, we intent to remove it. But we need to sync db and vdcs.
                  (db)
                  /// Read events, properties, etc from device.
                  /// Mew behavior, we intent to switch to it. But we need to sync db and vdcs.
                  (device));

namespace dss {

struct DeviceInfoAction {
  std::string id;
  dsApi::Translatable name;
  std::string title;
  std::string action;
  std::vector<std::pair<std::string, std::string>> params;
  std::string tags;
};

class Device;
class JSONWriter;

class DeviceInfo {
public:
  using Prefer = DeviceInfoPrefer;

  /// device - nullptr means no device is available
  DeviceInfo(Prefer prefer, StructureQueryBusInterface &query, VdcDb &db, const Device *device, std::string gtin,
             const std::locale &locale, JSONWriter &json, const ds::System &system);

  void addSpec();
  void addStateDescriptions();
  void addEventDescriptions();
  void addPropertyDescriptions();
  void addSensorDescriptions();
  void addActionDescriptions();
  std::vector<dsModel::DeviceScenario> scenarios();
  void addStandardActions();
  void addCustomActions();
  void addDynamicDeviceActions();
  std::vector<dsModel::State> deviceStates();
  std::vector<dsModel::StateStatus> deviceStateStatuses();
  void addOperational();
  void addCategories();
  void addOutputDescriptions();
  std::string getTranslatedStateName(const std::string& id);
  std::string getTranslatedStateValue(const std::string& stateId, const std::string& optionId);
  std::string getTranslatedOutputChannelEnumValue(const std::string& outputId, const std::string& enumValue);
  std::string getTranslatedSensorName(const std::string& id);


  struct Filter {
    bool spec : 1;
    bool stateDesc : 1;
    bool eventDesc : 1;
    bool propertyDesc : 1;
    bool sensorDesc : 1;
    bool actionDesc : 1;
    bool standardActions : 1;
    bool customActions : 1;
    bool dynamicDeviceActions : 1;
    bool operational : 1;
    bool categories : 1;
    bool outputDesc : 1;

    std::vector<bool> asBoolVector() const {
      return std::vector<bool>{spec, stateDesc, eventDesc, propertyDesc, sensorDesc, actionDesc, standardActions,
          customActions, dynamicDeviceActions, operational, categories, outputDesc};
    }
    bool operator==(const DeviceInfo::Filter &x) const;
  };

  static Filter parseFilter(const std::string &filterParam);
  void addByFilter(Filter filter);

  const Device *device() const { return m_device; }
  const std::string &gtin() const { return m_gtin; }
  const std::locale &locale() const { return m_locale; }

private:
  Prefer m_prefer;
  StructureQueryBusInterface &m_query;
  VdcDb &m_db;
  const Device *m_device;
  std::string m_gtin;
  const std::locale &m_locale;
  std::string m_vdcTrDomain;
  JSONWriter &m_json;
  const ds::System &m_system;
  const bool m_hasDynamicDefinitions;

  VdcDb::Specs m_dbSpecs;
  VdcDb::States m_dbStates;
  VdcDb::Events m_dbEvents;
  VdcDb::Properties m_dbProperties;
  VdcDb::Sensors m_dbSensors;
  VdcDb::Actions m_dbActions;
  VdcDb::StandardActions m_dbStandardActions;
  VdcDb::Categories m_dbCategories;
  VdcDb::Outputs m_dbOutputs;

  /// lookup item in db, return default instance if not present.
  template <class Items>
  typename Items::value_type dbItemOrDefault(const Items &items, const std::string &name);
  VdcDb::SensorDesc dbSensorOrDefault(unsigned index, const std::string &name);

  void addParameterDescription(const VdcDb::PropertyDesc &prop, gsl::czstring optionsTrConTranslatable);
  void addOperationalIntern();
  std::string vdcTr(gsl::czstring conTranslatable, gsl::czstring id) const;
  void getStateDescriptions(std::vector<VdcDb::StateDesc>& states);
  void getSensorDescriptions(std::vector<VdcDb::SensorDesc>& sensors);
  std::vector<DeviceInfoAction> deviceActions();
  std::vector<DeviceInfoAction> actions(const std::string &propertyType);
};

std::ostream &operator<<(std::ostream &, const DeviceInfo &);
std::ostream &operator<<(std::ostream &, const DeviceInfo::Filter &);

} // namespace dss

DS_REFLECT_STRUCT((dss), DeviceInfoAction, (id)(name)(title)(action)(params)(tags));
