#include "installation.h"

#include <ds-api/rest.h>
#include <ds-model/installation.h>
#include <ds/asio/resource.h>
#include <ds/vector.h>
#include <model/dsmeter.h>
#include <model/floor.h>

#include <boost/range/adaptor/transformed.hpp>

#include "propertysystem.h"
#include "protobufjson.h"
#include "util.h"
#include "vdc-connection.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsApartmentInstallation");

namespace dsApartment {

class InstallationImpl : public Installation, ds::asio::Resource<> {
public:
  InstallationImpl(ds::asio::ResourceBase& parent, dss::Apartment& apartment, dsApi::AsyncCallee& bus)
      : ds::asio::Resource<>{parent, "installation"}, m_apartment{apartment}, m_bus{bus} {}

private:
  dss::Apartment& m_apartment;
  dsApi::AsyncCallee& m_bus;
  dss::PropertySystem& m_propertySystem = *m_apartment.getPropertySystem();
  dss::PropertyNodePtr m_configNode{m_propertySystem.getProperty("/config")};
  dss::PropertyNodePtr m_systemNode{m_propertySystem.createProperty("/system")};

  dsModel::Floor floor(const dsModel::FloorId& id) const override {
    DS_INFO_ENTER(id);
    const auto floorPtr = m_apartment.floor(id);
    const auto& floor = *floorPtr;
    try {
      return dsModel::Floor{
          .id = id,
          .type = "floor",
          .lastChanged = {},
          .attributes = {
              .name = dss::unescapeHTML(floor.getName()), .orderId = floor.getOrder(), .zones = floor.zones()}};
    } catch (const std::exception& e) {
      DS_ERROR("Floor failed.", e);
      return {.id = id};
    }
  }

  dsModel::Floors floors() const override {
    return dsModel::Floors{.floors =
                               ds::makeVector(m_apartment.floorIds() |
                                              boost::adaptors::transformed([&](const auto& x) { return floor(x); }))};
  }

  void setFloorName(const dsModel::FloorId& id, const std::string& value) override {
    DS_NOTICE("Set floor name.", id, value);
    DS_REQUIRE(!value.empty(), "New name cannot be empty.");
    m_apartment.floor(id)->setName(dss::escapeHTML(value));
  }

  dsModel::Installation installation() const override {
    DS_INFO_ENTER("");
    const auto installationId = dsModel::InstallationId{m_apartment.id()};
    try {
      return dsModel::Installation{
          .id = installationId, // we use apartment id for installation id for now
          .type = "installation",
          .lastChanged = {},
          .attributes = {
              .location = {.longitude = [&]() -> std::optional<double> {
                             if (auto x = m_configNode->tryChildValue<std::string>("geodata/longitude")) {
                               return stod(*x);
                             }
                             return {};
                           }(),
                           .latitude = [&]() -> std::optional<double> {
                             if (auto x = m_configNode->tryChildValue<std::string>("geodata/latitude")) {
                               return stod(*x);
                             }
                             return {};
                           }()},
              .countryCode = m_configNode->tryChildValue<std::string>("geodata/country").value_or(std::string{}),
              .city = m_configNode->tryChildValue<std::string>("geodata/location").value_or(std::string{}),
              .timeZone = system().timeZone()}};
    } catch (const std::exception& e) {
      DS_ERROR("Installation failed.", e);
      return {.id = installationId};
    }
  }

  [[nodiscard]] dsModel::DSServer dsServer() const override {
    DS_INFO_ENTER("");
    const auto serverId =
        dsModel::DSServerId{m_systemNode->tryChildValue<std::string>("dSUID").value_or(std::string{})};
    try {
      return dsModel::DSServer{
          .id = serverId,
          .type = "dsServer",
          .lastChanged = {},
          .attributes = {
              .name = dss::unescapeHTML(m_apartment.getName()),
              .model = m_systemNode->tryChildValue<std::string>("version/model").value_or(std::string{}),
              .gtin = m_systemNode->tryChildValue<std::string>("version/gtin").value_or(std::string{}),
              .dsid = m_systemNode->tryChildValue<std::string>("dSID").value_or(std::string{}),
              .version = m_systemNode->tryChildValue<std::string>("version/distroVersion").value_or(std::string{}),
              .environment = [&]() {
                switch (system().release()) {
                  case ds::Release::devel: return dsModel::Environment::development;
                  case ds::Release::testing: return dsModel::Environment::alpha;
                  case ds::Release::fieldtest: return dsModel::Environment::beta;
                  case ds::Release::production: return dsModel::Environment::production;
                }
                return dsModel::Environment::none;
              }()}};
    } catch (const std::exception& e) {
      DS_ERROR("Server failed.", e);
      return {.id = serverId};
    }
  }

  dsModel::Controller controller(const dsModel::ControllerId& id) const override {
    DS_INFO_ENTER(id);
    const auto controllerPtr = m_apartment.controller(id);
    const auto& controller = *controllerPtr;
    try {
      return dsModel::Controller{.id = id,
                                 .type = "controller",
                                 .lastChanged = {},
                                 .attributes = {.name = dss::unescapeHTML(controller.getName())}};
    } catch (const std::exception& e) {
      DS_ERROR("Controller failed.", e);
      return {.id = id};
    }
  }

  dsModel::AuthorizeControllerResult authorizeController(const dsModel::ControllerId& id,
                                                         dsModel::AuthorizeControllerArgs&& args) override {
    DS_NOTICE("Authorize controller", id, args);
    const auto controllerPtr = m_apartment.controller(id);
    args.controller = id;
    return m_bus.blockingInvokeMethod<dsModel::AuthorizeControllerResult>(context(), "/backend/oauth/v2/authorizeApps",
                                                                          std::move(args));
  }

  void storeControllerAccessToken(const dsModel::ControllerId& id,
                                  dsModel::StoreControllerAccessTokenArgs&& args) override {
    const auto controllerPtr = m_apartment.controller(id);
    const auto& controller = *controllerPtr;
    DS_NOTICE("Store controller access token.", id, args);
    dss::VdcHelper::callAuthenticate(controller.getDSID(), args.authData, args.authScope,
                                     dss::ProtobufToJSon::jsonToElement(args.params));
  }

  dsModel::Controllers controllers() const override {
    return dsModel::Controllers{
        .controllers = ds::makeVector(m_apartment.controllerIds() |
                                      boost::adaptors::transformed([&](const auto& x) { return controller(x); }))};
  }
};

std::unique_ptr<Installation> makeInstallation(ds::asio::ResourceBase& parent, dss::Apartment& apartment,
                                               dsApi::AsyncCallee& bus) {
  return std::make_unique<InstallationImpl>(parent, apartment, bus);
}

} // namespace dsApartment
