/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <functional>

#include "rest.h"
#include "service-fwd.h"

namespace dsApi {

/// Route describing one resource
struct Route {
  /// Each route pattern segment is delimited by '/' character.
  /// When a segment is enclosed in braces ('{' and '}'), the segment is an argument.
  ///
  /// Example route patterns:
  /// - "/apartments/{apartmentId}
  /// - "/apartments/{apartmentId}/status
  /// - "/apartments/{apartmentId}/zones
  /// - "/apartments/{apartmentId}/zones/{zoneId}
  /// - "/apartments/{apartmentId}/zones/{zoneId}/status
  /// - "/apartments/{apartmentId}/{*path}
  std::string pattern{};

  /// @param args arguments specified by pattern parsed from url path
  /// @param HttpArgs http payload
  std::function<HttpResult(HttpRequest&&)> invoke{};
};

/// make wildcard route forwarding all subpaths to service
Route makeServiceRoute(std::string&& pattern, dsApi::Service& service);
std::string routeIdImpl(const RouteHttpArgs&);
template <class Id>
Id routeId(const RouteHttpArgs& x) {
  return Id{routeIdImpl(x)};
}
template <class Id>
Id routeId(const HttpRequest& x) {
  return Id{routeIdImpl(x.routeArgs)};
}

/// make route from function `void ()`
template <class F, class = decltype(std::declval<F>()())>
Route makeMethodRoute(std::string&& pattern, F&& f) {
  return Route{.pattern = std::move(pattern), .invoke = [f = std::forward<F>(f)](HttpRequest&& request) {
                 auto method = request.httpArgs.method;
                 if (method != HttpMethod::POST) {
                   return methodNotAllowedResult(method, {HttpMethod::POST});
                 }
                 if constexpr (std::is_void_v<std::invoke_result_t<F>>) {
                   f();
                   return noContentResult();
                 } else {
                   return dsApi::okDocumentResult(f());
                 }
               }};
}

/// make route from function `void (Args &&args)` or `Result (Args &&args)` with automatic conversion
/// from http request content to Args and Result to response content
template <class Args, class F, class = decltype(std::declval<F>()(std::declval<Args>()))>
Route makeMethodRoute(std::string&& pattern, F&& f) {
  return Route{.pattern = std::move(pattern), .invoke = [f = std::forward<F>(f)](HttpRequest&& request) {
                 auto method = request.httpArgs.method;
                 if (method != HttpMethod::POST) {
                   return methodNotAllowedResult(method, {HttpMethod::POST});
                 }
                 if constexpr (std::is_void_v<std::invoke_result_t<F, Args&&>>) {
                   f(ds::fromJson<Args>(ds::toJsonFragment(request.httpArgs.content)));
                   return noContentResult();
                 } else {
                   return dsApi::okDocumentResult(
                       f(ds::fromJson<Args>(ds::toJsonFragment(request.httpArgs.content))));
                 }
               }};
}

/// make route with {id} parameter from function `void (const Id&)`
template <class Id, class F, class = decltype(std::declval<F>()(std::declval<Id>()))>
Route makeMethodByIdRoute(std::string&& pattern, F&& f) {
  return Route{.pattern = std::move(pattern), .invoke = [f = std::forward<F>(f)](HttpRequest&& request) {
                 auto method = request.httpArgs.method;
                 if (method != HttpMethod::POST) {
                   return methodNotAllowedResult(method, {HttpMethod::POST});
                 }
                 f(routeId<Id>(request));
                 return noContentResult();
               }};
}

/// Make route with {id} parameter from function `void (const Id&, Args &&)`
/// or `Result (const Id&, Args &&)` with automatic conversion
/// from http request content to Args and Result to response content
template <class Id, class Args, class F, class = decltype(std::declval<F>()(std::declval<Id>(), std::declval<Args>()))>
Route makeMethodByIdRoute(std::string&& pattern, F&& f) {
  return Route{.pattern = std::move(pattern), .invoke = [f = std::forward<F>(f)](HttpRequest&& request) {
                 auto method = request.httpArgs.method;
                 if (method != HttpMethod::POST) {
                   return methodNotAllowedResult(method, {HttpMethod::POST});
                 }
                 if constexpr (std::is_void_v<std::invoke_result_t<F, const Id&, Args&&>>) {
                   f(routeId<Id>(request), ds::fromJson<Args>(ds::toJsonFragment(request.httpArgs.content)));
                   return noContentResult();
                 } else {
                   return dsApi::okResult(
                       f(routeId<Id>(request), ds::fromJson<Args>(ds::toJsonFragment(request.httpArgs.content))));
                 }
               }};
}

/// Route and invoke http request
HttpResult invoke(const std::vector<Route>& routes, HttpRequest&&);

/// Service routing requests according to available routes
struct Service {
  Service() = default;
  Service(std::vector<Route>&& routes) : routes{std::move(routes)} {}
  virtual ~Service() = default; // be usable as base class in std::unique_ptr

  std::vector<Route> routes{};

  // const because it does not modify Service.
  // `const Service&` is for invocation and reflection purposes.
  // `Service&` allows to also edit routes.
  HttpResult invoke(HttpRequest&&) const;
};

GetQuery makeGetQuery(const HttpRequest&);

/// make GET function returning document
template <class F, class = decltype(std::declval<F>()())>
auto makeGet(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET, [f = std::forward<F>(f)](auto&&) { return dsApi::okDocumentResult(f()); }};
}

/// make GET function with arguments returning document
template <class Args, class F, class = decltype(std::declval<F>()(std::declval<Args>()))>
auto makeGet(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET, [f = std::forward<F>(f)](auto&& request) {
        return dsApi::okDocumentResult(f(ds::fromJson<Args>(ds::toJsonFragment(request.routeArgs))));
      }};
}

/// make GET by id function returning document
template <class Id, class F, class = decltype(std::declval<F>()(std::declval<Id>()))>
auto makeGetById(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET,
      [f = std::forward<F>(f)](const auto& request) { return dsApi::okDocumentResult(f(routeId<Id>(request))); }};
}

/// make GET function with query returning document
template <class F, class = decltype(std::declval<F>()(std::declval<GetQuery>()))>
auto makeGetWithQuery(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET,
      [f = std::forward<F>(f)](const auto& request) { return dsApi::okDocumentResult(f(makeGetQuery(request))); }};
}

/// make GET by id with query function returning document with query
template <class Id, class F, class = decltype(std::declval<F>()(std::declval<Id>(), std::declval<GetQuery>()))>
auto makeGetByIdWithQuery(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET, [f = std::forward<F>(f)](const auto& request) {
        return dsApi::okDocumentResult(f(routeId<Id>(request), makeGetQuery(request)));
      }};
}

/// make DELETE with {id} parameter from function(id) deleting document
template <class Id, class F, class = decltype(std::declval<F>()(std::declval<Id>()))>
auto makeDeleteById(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::DELETE, [f = std::forward<F>(f)](HttpRequest&& request) {
        f(routeId<Id>(request));
        return noContentResult();
      }};
}

/// make POST function creating document and returning id of created resource as std::string
template <class Data, class F>
auto makeCreate(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::POST, [f = std::forward<F>(f)](auto&& context) {
        auto data = dsApi::data<Data>(context.httpArgs);
        auto id = static_cast<std::string>(f(std::move(data)));
        return HttpResult{.status = 201, .content = {}, .headers = {{"Location", {std::move(id)}}}};
      }};
}

/// make POST function invoking method `void ()`
template <class Args, class F, class = decltype(std::declval<F>()(std::declval<Args>()))>
auto makeMethod(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::POST, [f = std::forward<F>(f)](auto&& context) {
        // TODO(someday): this could be more effective without conversion to std::map<string, json>
        auto args = ds::fromJson<std::map<std::string, ds::JsonFragment>>(context.httpArgs.content);
        for (const auto& routeArg : context.routeArgs) {
          args[routeArg.first] = ds::toJsonFragment(routeArg.second);
        }
        f(ds::fromJson<Args>(ds::toJsonFragment(args)));
        return noContentResult();
      }};
}

/// make PUT function creating document.
template <class Args, class Data, class F>
auto makePut(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::PUT, [f = std::forward<F>(f)](auto&& context) {
        f(ds::fromJson<Args>(ds::toJsonFragment(context.routeArgs)), dsApi::data<Data>(context.httpArgs));
        return noContentResult();
      }};
}

/// Route describing one path operation.
///
/// Functions `makeReplacePatch`, etc
/// are intended to make instances of PatchRoute in application code.
struct PatchRoute {
  std::string pattern{};

  std::map<PatchOperationType, std::function<void(PatchRequest&&)>> operations{};
};

/// make replace route from function `void (Args &&args)`
/// with automatic conversion from route and http content to Args
template <class Args, class Value, class F,
          class = decltype(std::declval<F>()(std::declval<Args>(), std::declval<Value>()))>
auto makeReplacePatch(F&& f) {
  return std::pair<PatchOperationType, std::function<void(PatchRequest &&)>>{
      dsApi::PatchOperationType::replace, [f = std::forward<F>(f)](auto&& context) {
        f(ds::fromJson<Args>(ds::toJsonFragment(context.routeArgs)), ds::fromJson<Value>(context.operation.value));
      }};
}

/// make replace route with {id} from function `void (Args &&args)`
template <class Id, class Value, class F,
          class = decltype(std::declval<F>()(std::declval<Id>(), std::declval<Value>()))>
auto makeReplaceByIdPatch(F&& f) {
  return std::pair<PatchOperationType, std::function<void(PatchRequest &&)>>{
      dsApi::PatchOperationType::replace, [f = std::forward<F>(f)](auto&& patchRequest) {
        f(routeId<Id>(patchRequest.routeArgs), ds::fromJson<Value>(patchRequest.operation.value));
      }};
}

struct DocumentMethods {
  std::map<HttpMethod, std::function<HttpResult(HttpRequest&&)>> methods{};

  /// default PATCH method implementation
  /// interates over all patch operations and
  /// invokes matching patch route
  std::vector<PatchRoute> patchRoutes{};

  std::vector<HttpMethod> allowedMethods() const;
};

Route makeDocumentRoute(std::string&& pattern, DocumentMethods&& document);

/// Makes test service suitable for dsApi integration tests
std::unique_ptr<Service> makeApartmentTestService();

} // namespace dsApi
