/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <avahi-client/client.h>
#include <avahi-common/error.h>

#include <boost/asio/io_service.hpp>
#include <boost/noncopyable.hpp>
#include <chrono>
#include <list>
#include <memory>

namespace dsAvahi {


/// DS_FAIL_REQUIRE with avahi error as context
#define DS_AVAHI_FAIL(_error, ...)                           \
  do {                                                       \
    const char *_errorStr = avahi_strerror(_error);          \
    DS_CONTEXT("Avahi error:", _errorStr, "(", _error, ")"); \
    DS_FAIL_REQUIRE("", ##__VA_ARGS__);                      \
  } while (0)

/// DS_FAIL_REQUIRE with avahi error in client as context
#define DS_AVAHI_FAIL_CLIENT_ERROR(client, ...) DS_AVAHI_FAIL(avahi_client_errno((client).ptr()), ##__VA_ARGS__)

/// Client connection to avahi deamon.
/// Maintains dbus connection to avahi deamon and notifies other avahi classes when
/// client is in running state.
class Client : boost::noncopyable {
public:
  Client(boost::asio::io_service &);
  ~Client();

  ::AvahiClient *ptr() const { return m_ptr; }

  void reconnectOnError(const std::exception &e);

  /// Connection to deamon is fully established and deamon is ready.
  /// In other words, avahi client is in AVAHI_CLIENT_S_RUNNING state.
  bool isRunning() const { return m_isRunning; }

  struct Hook : boost::noncopyable {
    virtual ~Hook();

    /// Called from addHook or when connection to deamon is up.
    /// Default implementation does nothing.
    virtual void onRunning();

    /// Function `avahi_client_free` internally
    /// automatically frees all associated
    /// browser, resolve and entry group objects.
    /// So they must be released from owning pointers
    /// on this callback.
    ///
    /// Do not call any callbacks from freed(),
    /// we need to notify ALL hooks
    /// because calling any Hook destructor
    /// leads to double free.
    virtual void freed() = 0;
  };
  void addHook(Hook &);
  void removeHook(Hook &);

  /// Type prefix added to all service types.
  /// Tests set it to random value so that test runs are independent
  /// of each other and production instances.
  const std::string &typePrefix() const { return m_typePrefix; }
  void setRandomTypePrefix();

  boost::asio::io_service &ioService();

  void setOnReconnectOnError(std::function<void(const std::exception &)> &&x) { m_onReconnectOnError = std::move(x); }

private:
  class Poll;
  const std::unique_ptr<Poll> m_poll;

  struct Deleter {
    void operator()(::AvahiClient *x) { avahi_client_free(x); }
  };
  std::unique_ptr< ::AvahiClient, Deleter> m_uniquePtr;
  // Avahi calls callback from within `avahi_client_new` invocation,
  // i.e. before the owning `m_uniquePtr` is set.
  // So we use this non-owning pointer populated in callback.
  ::AvahiClient *m_ptr;

  bool m_isRunning;
  std::list<Hook *> m_hooks;
  std::string m_typePrefix;
  std::function<void(const std::exception &)> m_onReconnectOnError;

  void setRunning();
  void reconnect();

  static void sCallback(AvahiClient *, AvahiClientState state, void *data);
  void callback(AvahiClient *ptr, AvahiClientState state);
};

} // namespace dsAvahi
