/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "connector.h"

#include <ds-api/event-mock.h>
#include <ds-api/rest.h>
#include <ds-net/message-stream-mock.h>
#include <ds/asio/ds-signal-mock.h>
#include <ds/asio/promise.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/log.h>

#include <boost/algorithm/string/predicate.hpp>

#include "connector-mock.h"

using namespace trompeloeil;
using namespace std::literals;

namespace {

static const char* TAGS = "[dsBackendConnector][dsBackend]";

struct HttpResultFulfilled {
public:
  MAKE_MOCK1(call, void(std::optional<dsApi::HttpResult>));
};

SCENARIO("dsBackendConnector", TAGS) {
  auto clock = ds::ClockMocker{};
  auto loop = ds::asio::TestLoopResource{};
  dsNet::MessageStreamClientMock streamClient{};
  const auto reconnectSchedule = std::chrono::seconds{10};
  const auto keepAlivePeriod = std::chrono::seconds{120};
  const auto deadPeriod = keepAlivePeriod * 2;
  dsBackend::ConnectorHookMock busHook;
  ALLOW_CALL(busHook, authenticate())
      .RETURN(dsBackend::Authenticate{.client = "dss", .secret = "dssSecret", .apartmentId = "dssApartmentId"});

  auto eventHub = dsApi::EventHubMock{};
  auto connector = dsBackend::makeConnector(loop, "connector", streamClient, busHook, eventHub);
  auto isConnectedChanged = dsBackend::BusIsConnectedChangedMock{loop, connector->isConnectedChanged()};

  const auto error = boost::system::errc::make_error_code(boost::system::errc::operation_canceled);
  WHEN("just created") {
    THEN("is disconnected") { REQUIRE(!connector->isConnected()); }
    THEN("discards published events") { eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}}); }
    THEN("invoke fails") {
      auto httpResultFulfilled = HttpResultFulfilled{};
      REQUIRE_CALL(httpResultFulfilled, call(_))
          .LR_SIDE_EFFECT(CHECK(_1.value().status == 500); CHECK(boost::contains(
              ds::fromJson<dsApi::Document>(_1.value().content).error.detail, "Not connected."));
                          loop.stop());
      connector->asyncInvoke({.path = "/backend/test", .httpArgs = {.method = dsApi::HttpMethod::GET}},
                             ds::makePromise<dsApi::HttpResult>(loop.io, [&](std::optional<dsApi::HttpResult>&& x) {
                               httpResultFulfilled.call(std::move(x));
                             }));
      loop.run();
      SUCCEED();
    }
  }

  connector->setActive(true);
  THEN("after reconnectSchedule, connects") {
    clock.advance(reconnectSchedule);

    auto streamPtr = std::make_shared<dsNet::MessageStreamMock>(loop, "stream");
    auto& stream = *streamPtr;
    stream.allowClose = NAMED_ALLOW_CALL(stream, close());
    auto streamHooks = dsNet::MessageStreamHooks{};

    // when old stream is deleted, we connect 2nd one
    auto stream2Ptr = std::make_shared<dsNet::MessageStreamMock>(loop, "stream2");
    auto& stream2 = *stream2Ptr;
    stream2.allowClose = NAMED_ALLOW_CALL(stream2, close());

    REQUIRE_CALL(streamClient, asyncConnect()).LR_RETURN(std::move(streamPtr));
    REQUIRE_CALL(stream, asyncRun(_)).LR_SIDE_EFFECT(loop.stop(); streamHooks = std::move(_1));

    loop.run();

    WHEN("connect fails") {
      {
        stream.allowClose.reset();
        REQUIRE_CALL(stream, close());
        streamHooks.closed(boost::system::errc::make_error_code(boost::system::errc::operation_not_supported));
      }

      THEN("after reconnectSchedule, connects") {
        clock.advance(reconnectSchedule);
        REQUIRE_CALL(streamClient, asyncConnect()).LR_RETURN(std::move(stream2Ptr));
        REQUIRE_CALL(stream2, asyncRun(_)).LR_SIDE_EFFECT(loop.stop(); streamHooks = std::move(_1));
        loop.run();
      }
    }
    WHEN("connect succeeds, authenticate is written") {
      // unhandled reads stops io
      auto pendingReadHandler = dsNet::MessageStream::ReadHandler{};
      ALLOW_CALL(stream, asyncRead(_)).LR_SIDE_EFFECT(loop.stop(); pendingReadHandler = std::move(_1));
      REQUIRE_CALL(
          stream,
          asyncWrite(R"({"authenticate":{"client":"dss","secret":"dssSecret","apartmentId":"dssApartmentId"}})", _))
          .LR_SIDE_EFFECT(_2({}));

      streamHooks.opened();

      auto reconnectsAfter = [&](ds::Clock::duration duration) {
        REQUIRE(!connector->isConnected());
        clock.advance(duration);
        REQUIRE_CALL(streamClient, asyncConnect()).LR_RETURN(std::move(stream2Ptr));
        REQUIRE_CALL(stream2, asyncRun(_)).LR_SIDE_EFFECT(loop.stop(); streamHooks = std::move(_1));
        loop.run();
      };
      auto reconnects = [&]() { return reconnectsAfter(reconnectSchedule); };
      auto doesNotReconnectFor = [&](ds::Clock::duration duration) {
        REQUIRE(!connector->isConnected());
        clock.advance(duration);
        loop.postPostStopRun();
      };

      THEN("Write buffer size is sane") {
        /// Apartment structure and status may be quite big.
        const auto config = connector->maybeSession()->bufferedWriter().config();
        CHECK(config.countCapacity == 256);
        CHECK(config.sizeCapacity == 4 * 1024 * 1024);
      }

      WHEN("read fails") {
        stream.allowClose.reset();
        REQUIRE_CALL(stream, close());
        pendingReadHandler(error, {});
        reconnects();
      }

      WHEN("authenticated times out") {
        clock.advance(deadPeriod);
        stream.allowClose.reset();
        REQUIRE_CALL(stream, close()).LR_SIDE_EFFECT(loop.stop());
        loop.run();
        reconnects();
      }

      THEN("receives close message -> reconnects") {
        stream.allowClose.reset();
        REQUIRE_CALL(stream, close());
        pendingReadHandler({}, R"({"close":{"reason":"Test disconnect", "blockedFor":600}})");
        INFO(connector->lastDisconnectReason());
        REQUIRE(boost::contains(connector->lastDisconnectReason(), "Test disconnect"));
        doesNotReconnectFor(5min);
        reconnectsAfter(30min);
      }

      WHEN("receives unknown message") {
        stream.allowClose.reset();
        REQUIRE_CALL(stream, close());
        pendingReadHandler({}, R"({"unknown":{}})");
        reconnects();
      }

      WHEN("authenticated") {
        {
          REQUIRE_CALL(isConnectedChanged, call(true)).LR_SIDE_EFFECT(loop.stop());
          pendingReadHandler({}, R"({"authenticated":{}})");
          loop.run();
        }

        THEN("is connected") { REQUIRE(connector->isConnected()); }

        WHEN("read fails") {
          {
            stream.allowClose.reset();
            REQUIRE_CALL(stream, close());
            REQUIRE_CALL(isConnectedChanged, call(false)).LR_SIDE_EFFECT(loop.stop());
            pendingReadHandler(error, {});
            loop.run();
          }
          reconnects();
        }

        WHEN("ignores unknown messages") {
          pendingReadHandler({}, R"({"unknown":{}})");
          REQUIRE(connector->isConnected());
        }

        THEN("responds to pings") {
          REQUIRE_CALL(stream, asyncWrite(R"({"pong":{"id":"4"}})", _)).LR_SIDE_EFFECT(_2({}));
          pendingReadHandler({}, R"({"ping":{"id":"4"}})");
        }

        THEN("sends ping within keepAlivePeriod") {
          clock.advance(keepAlivePeriod);
          REQUIRE_CALL(stream, asyncWrite(R"({"ping":{"id":"1"}})", _)).LR_SIDE_EFFECT(loop.stop(); _2({}));
          loop.run();

          WHEN("pong is not received within keepAlivePeriod") {
            {
              stream.allowClose.reset();
              REQUIRE_CALL(stream, close());
              REQUIRE_CALL(isConnectedChanged, call(false)).LR_SIDE_EFFECT(loop.stop());
              clock.advance(keepAlivePeriod);
              loop.run();
            }
            reconnects();
          }

          WHEN("after keepAlivePeriod / 2") {
            clock.advance(keepAlivePeriod / 2);
            loop.postStopRun();

            THEN("is connected") { REQUIRE(connector->isConnected()); }

            WHEN("wrong pong is received") {
              pendingReadHandler({}, R"({"pong":{"id":"7"}})");
              THEN("is connected") { REQUIRE(connector->isConnected()); }
            }

            WHEN("pong is received") {
              pendingReadHandler({}, ds::str(R"({"pong":{"id":"1"}})"));
              THEN("sends another ping within keepAlivePeriod") {
                clock.advance(keepAlivePeriod * 2);
                REQUIRE_CALL(stream, asyncWrite(R"({"ping":{"id":"2"}})", _)).LR_SIDE_EFFECT(loop.stop(); _2({}));
                loop.run();
                REQUIRE(connector->isConnected());
              }
            }
          }
        }

        WHEN("request is read") {
          std::unique_ptr<ds::Promise<dsApi::HttpResult>> promise;
          REQUIRE_CALL(busHook, asyncInvoke(_, _))
              .LR_SIDE_EFFECT(
                  CHECK(_1 == dsApi::HttpRequest{.path = "/a", .httpArgs = {.method = dsApi::HttpMethod::GET}});
                  promise = std::move(_2););
          pendingReadHandler({}, R"({"request":{"id":"id-1","path":"/a","args":{"method":"GET"}}})");

          THEN("reply is written") {
            REQUIRE_CALL(stream, asyncWrite(R"({"response":{"id":"id-1","result":{"status":200}}})", _))
                .LR_SIDE_EFFECT(_2({}); loop.stop());
            promise->value = {.status = 200};
            promise.reset();
            loop.run();
          }
        }

        THEN("non-subscribed event is discarded") {
          eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}});
        }

        WHEN("subscribed") {
          REQUIRE_CALL(eventHub, subscribe(_, _))
              .LR_SIDE_EFFECT(
                  CHECK(_1 == dsApi::Subscription{.type = dsApi::EventType{"a"}, .expiresAt = ds::Clock::now() + 60s});
                  CHECK(&_2 == &connector->eventSink()); loop.stop());
          pendingReadHandler({}, R"({"subscribe":{"type":"a", "timeout":60}})");

          THEN("writes subscribed event") {
            REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"a"}})", _)).LR_SIDE_EFFECT(_2({}); loop.stop(););
            eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}});
            loop.run();
          }
          WHEN("write fails") {
            {
              stream.allowClose.reset();
              REQUIRE_CALL(stream, close());
              REQUIRE_CALL(stream, asyncWrite(_, _)).LR_SIDE_EFFECT(_2(error));
              REQUIRE_CALL(isConnectedChanged, call(false)).LR_SIDE_EFFECT(loop.stop());
              eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}});
              loop.run();
            }
            reconnects();
          }

          THEN("writes are queued when stream write is blocking") {
            dsNet::MessageStream::WriteHandler pendingWriteHandler{};
            REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"a","path":"1"}})", _))
                .LR_SIDE_EFFECT(loop.postStop(); pendingWriteHandler = std::move(_2));
            eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}, .path = "1"});
            eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}, .path = "2"});
            eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}, .path = "3"});
            loop.run();

            THEN("queued writes are written when write is unblocked") {
              REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"a","path":"2"}})", _)).LR_SIDE_EFFECT(_2({}););
              REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"a","path":"3"}})", _)).LR_SIDE_EFFECT(_2({}););
              pendingWriteHandler({});
            }

            THEN("queued writes are discarded on disconnect and not written to new connection") {
              {
                stream.allowClose.reset();
                REQUIRE_CALL(stream, close());
                REQUIRE_CALL(isConnectedChanged, call(false));
                pendingWriteHandler(error);
                reconnects();
              }

              ALLOW_CALL(stream2, asyncRead(_)).LR_SIDE_EFFECT(loop.stop(); pendingReadHandler = std::move(_1));
              {
                REQUIRE_CALL(
                    stream2,
                    asyncWrite(
                        R"({"authenticate":{"client":"dss","secret":"dssSecret","apartmentId":"dssApartmentId"}})", _))
                    .LR_SIDE_EFFECT(_2({}));
                streamHooks.opened();
              }
              {
                REQUIRE_CALL(isConnectedChanged, call(true)).LR_SIDE_EFFECT(loop.stop());
                pendingReadHandler({}, R"({"authenticated":{}})");
                loop.run();
              }
              {
                REQUIRE_CALL(stream2, asyncWrite(R"({"event":{"type":"a","path":"9"}})", _))
                    .LR_SIDE_EFFECT(_2({}); loop.stop());
                eventHub.m_event.emit(dsApi::Event{.type = dsApi::EventType{"a"}, .path = "9"});
                loop.run();
              }
            }
          }
        }

        THEN("callee invoke with /backend prefix is sent as request") {
          auto httpResultFulfilled = HttpResultFulfilled{};
          auto message = dsBackend::Message{};
          REQUIRE_CALL(stream, asyncWrite(_, _))
              .LR_SIDE_EFFECT(message = ds::fromJson<dsBackend::Message>(std::string{_1}); _2({}); loop.stop(););
          connector->asyncInvoke({.path = "/backend/test", .httpArgs = {.method = dsApi::HttpMethod::GET}},
                                 ds::makePromise<dsApi::HttpResult>(loop.io, [&](std::optional<dsApi::HttpResult>&& x) {
                                   httpResultFulfilled.call(std::move(x));
                                 }));
          loop.run();
          REQUIRE(message.request);
          const auto& request = *message.request;
          CHECK(request.id != dsBackend::RequestId{});
          CHECK(request.path == "/test");
          CHECK(request.args == dsApi::HttpArgs{.method = dsApi::HttpMethod::GET});

          THEN("callee is replied on response") {
            REQUIRE_CALL(httpResultFulfilled, call(dsApi::HttpResult{.status = 200})).LR_SIDE_EFFECT(loop.stop());
            pendingReadHandler(
                {}, ds::str(R"({"response":{"id":")", message.request->id, R"(", "result":{"status":200}}})"));
            loop.run();
            SUCCEED();
          }

          THEN("call is replied with cancel and disconnect on timeout") {
            clock.advance(90s);
            ALLOW_CALL(stream, asyncWrite(R"({"ping":{"id":"1"}})", _))
                .LR_SIDE_EFFECT(_2({}))
                .LR_SIDE_EFFECT(loop.stop());
            loop.run();
            clock.advance(90s);
            REQUIRE_CALL(isConnectedChanged, call(false)).LR_SIDE_EFFECT(loop.stop());
            REQUIRE_CALL(httpResultFulfilled, call(std::nullopt)).LR_SIDE_EFFECT(loop.stop());
            loop.run();
            loop.run();
            REQUIRE(!connector->isConnected());
            SUCCEED();
          }

          THEN("call is replied with cancel on disconnect") {
            ALLOW_CALL(isConnectedChanged, call(false));
            REQUIRE_CALL(httpResultFulfilled, call(std::nullopt)).LR_SIDE_EFFECT(loop.stop(););
            pendingReadHandler(error, {});
            loop.run();
            SUCCEED();
          }
        }

        THEN("callee invoke without /backend prefix fails") {
          auto httpResultFulfilled = HttpResultFulfilled{};
          REQUIRE_CALL(httpResultFulfilled, call(_))
              .LR_SIDE_EFFECT(CHECK(_1.value().status == 500);
                              CHECK(boost::contains(ds::fromJson<dsApi::Document>(_1.value().content).error.detail,
                                                    "Paths in backend must start with backendPrefix"));
                              loop.stop());
          connector->asyncInvoke({.path = "/test", .httpArgs = {.method = dsApi::HttpMethod::GET}},
                                 ds::makePromise<dsApi::HttpResult>(loop.io, [&](std::optional<dsApi::HttpResult>&& x) {
                                   httpResultFulfilled.call(std::move(x));
                                 }));
          loop.run();
          SUCCEED();
        }

        THEN("receives close message -> reconnect") {
          stream.allowClose.reset();
          REQUIRE_CALL(stream, close());
          REQUIRE_CALL(isConnectedChanged, call(false)).LR_SIDE_EFFECT(loop.stop());
          pendingReadHandler({}, R"({"close":{"reason":"Test disconnect", "blockedFor":600}})");
          INFO(connector->lastDisconnectReason());
          REQUIRE(boost::contains(connector->lastDisconnectReason(), "Test disconnect"));
          loop.run();
          doesNotReconnectFor(5min);
          reconnectsAfter(30min);
        }

        WHEN("setActive(false)") {
          connector->setActive(false);
          THEN("is !connected") {
            REQUIRE_CALL(isConnectedChanged, call(false)).LR_SIDE_EFFECT(loop.stop());
            loop.run();
            REQUIRE(!connector->isConnected());
          }
        }
      }
    }
  }
}

} // namespace
