/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "signals-source.h"

#include <ds-apartment/ds-devices.h>
#include <ds-backend/connector.h>
#include <ds-model/scenarios.h>
#include <ds/asio/ds-signal.h>
#include <ds/asio/promise.h>
#include <ds/asio/resource.h>
#include <event/event_create.h>
#include <event/event_fields.h>
#include <model/group.h>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsBackendSignalsSource");

namespace dsBackend {

class SignalsSourceImpl : public SignalsSource,
                          public ds::asio::Resource<>,
                          public dss::RegisteredEventInterpreterPlugin {
public:
  SignalsSourceImpl(ds::asio::ResourceBase& parent, dss::EventInterpreter& eventInterpreter)
      : ds::asio::Resource<>(parent, "signalsSource"),
        dss::RegisteredEventInterpreterPlugin("signalsSource", eventInterpreter) {
    eventInterpreter.subscribe(*this, dss::EventName::CallScene);
  }

private:
  ds::Signal<dsModel::Signal> m_next;

  ds::Slots<dsModel::Signal>& next() override { return m_next; }

  void handleEvent(dss::Event& event, const dss::EventSubscription&) override {
    DS_INFO_ENTER(event);
    handleEventImpl(event);
  }

  void handleEventImpl(const dss::Event& event) const {
    DS_DEBUG_ENTER(event);
    // Running from foreign thread(s) -> make sure to not modify this. signal emit is thread safe
    [& next = m_next, &event, &dsSystem = context().system] {
      const auto& name = event.getName();
      const auto& location = event.getRaiseLocation();
      if (name == dss::EventName::CallScene && location == dss::erlGroup) {
        auto sceneId = [&]() {
          switch (ds::requireInt(event.getPropertyByName(dss::ef_sceneID))) {
            case dss::SceneAbsent: return dsModel::ApartmentActionType::absent;
            case dss::ScenePresent: return dsModel::ApartmentActionType::present;
          }
          return dsModel::ApartmentActionType::none;
        }();
        const auto& group = *event.getRaisedAtGroup();
        if (group.getZoneID() == 0 && group.getID() == 0 && sceneId != dsModel::ApartmentActionType::none) {
          next.emit(
              dsModel::Signal{.id = dsModel::SignalId{dsSystem.randomUuidString()},
                              .type = "signal",
                              .lastChanged = static_cast<ds::DateTime>(event.getTimestamp()),
                              .attributes = {.target = dsModel::SignalTargetApartment{.type = "apartment"},
                                             .action = dsModel::ApartmentAction{.type = "apartment", .id = sceneId}}});
        }
      }
    }();
  }
};

std::unique_ptr<SignalsSource> makeSignalsSource(ds::asio::ResourceBase& parent,
                                                 dss::EventInterpreter& eventInterpreter) {
  return std::make_unique<SignalsSourceImpl>(parent, eventInterpreter);
}

} // namespace dsBackend
