/*
    Copyright (c) 2016 digitalSTROM.org," " Zurich," " Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/
#include "timed-measurements-source.h"

#include <ds-model/model.h>
#include <ds/asio/ds-signal-mock.h>
#include <ds/asio/test.h>
#include <ds/clock-mock.h>
#include <metering/metering-mock.h>
#include <model/apartment-mock.h>
#include <model/dsmeter.h>
#include <model/zone.h>

#include "event.h"

using namespace std::literals;
using namespace trompeloeil;
static ds::LogChannel dsLogChannel("dsBackendTimedMeasurementsSourceTest");
static const char* TAGS = "[dss][dsBackendTimedMeasurementsSource]";

namespace {

DS_SLOT_MOCK1(MeasurementSlotMock, dsModel::Measurement);

TEST_CASE("dsBackendTimedMeasurementsSource", TAGS) {
  auto clock = ds::ClockMocker{};
  ds::asio::TestLoopResource loop;
  dss::ApartmentMock apartment{};

  dss::metering::MeteringMock metering{};
  ALLOW_CALL(apartment, getMetering()).LR_RETURN(metering);

  dss::EventInterpreter eventInterpreter{nullptr};
  dss::EventQueue queue{&eventInterpreter};
  eventInterpreter.initialize();
  eventInterpreter.setEventQueue(&queue);
  auto source = dsBackend::makeTimedMeasurementsSource(loop, eventInterpreter, apartment);
  auto nextMeasurement = MeasurementSlotMock{loop, source->next()};

  const auto lastChanged = loop.system.now();

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});
  controller1->setCapability_HasMetering(true);
  controller1->setIsPresent(true);
  controller1->setIsConnected(true);

  SECTION("measurements are not emitted before ModelReady event is issued") {
    clock.advance(60min);
    loop.postStopRun();
  }

  SECTION("measurements are emitted after ModelReady event") {
    queue.pushEvent(std::make_shared<dss::Event>(dss::EventName::ModelReady));
    clock.advance(15min); // No measurements emitted before ModelReady event
    eventInterpreter.executePendingEvent();
    clock.advance(5s);

    /// Controller energy
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("d"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energyDelta))
        .RETURN(2)
        .LR_SIDE_EFFECT(CHECK(_1[0] == dss::Dsuid{controllerId1}));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"d"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = dsModel::MeasurementType::energy,
                                    .origin =
                                        dsModel::MeasurementOriginController{
                                            .type = "controller",
                                            .id = dsModel::ControllerId{"1111111111111111111111111111111100"}},
                                    .value = 2}}))
        .LR_SIDE_EFFECT(loop.stop());

    /// Controller energy counter
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("c"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energy))
        .RETURN(5)
        .LR_SIDE_EFFECT(CHECK(_1[0] == dss::Dsuid{controllerId1}));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"c"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = dsModel::MeasurementType::energyCounter,
                                    .origin =
                                        dsModel::MeasurementOriginController{
                                            .type = "controller",
                                            .id = dsModel::ControllerId{"1111111111111111111111111111111100"}},
                                    .value = 5}}));

    /// Apartment energy
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("b"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energyDelta)).RETURN(10);
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"b"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = dsModel::MeasurementType::energy,
                                                     .origin = dsModel::MeasurementOriginApartment{.type = "apartment"},
                                                     .value = 10}}));

    /// Apartment energy counter
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("a"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energy)).RETURN(20);
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"a"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = dsModel::MeasurementType::energyCounter,
                                                     .origin = dsModel::MeasurementOriginApartment{.type = "apartment"},
                                                     .value = 20}}));
    loop.run();
  }

  SECTION("controller set to producer") {
    controller1->setCapability_HasMeteringProducer(true);
    controller1->setCapability_HasMeteringProducerEnabled(true);

    queue.pushEvent(std::make_shared<dss::Event>(dss::EventName::ModelReady));
    clock.advance(15min); // No measurements emitted before ModelReady event
    eventInterpreter.executePendingEvent();
    clock.advance(5s);

    // NOTE: order is reversed, the last calls in timedUploadLoop, are checked first

    /// Controller energy
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("d-p"));
    REQUIRE_CALL(metering, get15minProduced(_, dss::metering::SeriesTypes::energyDelta))
        .RETURN(2)
        .LR_SIDE_EFFECT(CHECK(_1[0] == dss::Dsuid{controllerId1}));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"d-p"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = dsModel::MeasurementType::energyProduced,
                                    .origin =
                                        dsModel::MeasurementOriginController{
                                            .type = "controller",
                                            .id = dsModel::ControllerId{"1111111111111111111111111111111100"}},
                                    .value = 2}}))
        .LR_SIDE_EFFECT(loop.stop()); // terminate test

    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("d-c"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energyDelta))
        .RETURN(0)
        .LR_SIDE_EFFECT(CHECK(_1[0] == dss::Dsuid{controllerId1}));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"d-c"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = dsModel::MeasurementType::energy,
                                    .origin =
                                        dsModel::MeasurementOriginController{
                                            .type = "controller",
                                            .id = dsModel::ControllerId{"1111111111111111111111111111111100"}},
                                    .value = 0}}));

    /// Controller energy counter
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("c-p"));
    REQUIRE_CALL(metering, get15minProduced(_, dss::metering::SeriesTypes::energy))
        .RETURN(5)
        .LR_SIDE_EFFECT(CHECK(_1[0] == dss::Dsuid{controllerId1}));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"c-p"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = dsModel::MeasurementType::energyProducedCounter,
                                    .origin =
                                        dsModel::MeasurementOriginController{
                                            .type = "controller",
                                            .id = dsModel::ControllerId{"1111111111111111111111111111111100"}},
                                    .value = 5}}));

    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("c-c"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energy))
        .RETURN(0)
        .LR_SIDE_EFFECT(CHECK(_1[0] == dss::Dsuid{controllerId1}));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"c-c"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = dsModel::MeasurementType::energyCounter,
                                    .origin =
                                        dsModel::MeasurementOriginController{
                                            .type = "controller",
                                            .id = dsModel::ControllerId{"1111111111111111111111111111111100"}},
                                    .value = 0}}));

    /// Apartment energy
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("b-p"));
    REQUIRE_CALL(metering, get15minProduced(_, dss::metering::SeriesTypes::energyDelta)).RETURN(10);
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"b-p"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = dsModel::MeasurementType::energyProduced,
                                                     .origin = dsModel::MeasurementOriginApartment{.type = "apartment"},
                                                     .value = 10}}));

    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("b-c"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energyDelta)).RETURN(0);
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"b-c"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = dsModel::MeasurementType::energy,
                                                     .origin = dsModel::MeasurementOriginApartment{.type = "apartment"},
                                                     .value = 0}}));

    /// Apartment energy counter
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("a-p"));
    REQUIRE_CALL(metering, get15minProduced(_, dss::metering::SeriesTypes::energy)).RETURN(20);
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"a-p"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = dsModel::MeasurementType::energyProducedCounter,
                                                     .origin = dsModel::MeasurementOriginApartment{.type = "apartment"},
                                                     .value = 20}}));

    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("a-c"));
    REQUIRE_CALL(metering, get15minConsumed(_, dss::metering::SeriesTypes::energy)).RETURN(0);
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"a-c"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = dsModel::MeasurementType::energyCounter,
                                                     .origin = dsModel::MeasurementOriginApartment{.type = "apartment"},
                                                     .value = 0}}));

    loop.run();
  }
}

} // namespace
