/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "ds-meterings.h"

#include <ds-api/event.h>
#include <ds-model/apartment-api.h>
#include <ds-model/meterings.h>
#include <ds-model/model.h> // extern meteringValues
#include <ds/asio/resource.h>
#include <ds/asio/timer.h>
#include <ds/log.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <boost/range/algorithm/find.hpp>
#include <boost/range/algorithm/sort.hpp>

#include "metering/metering.h"
#include "model/apartment.h"
#include "model/dsmeter.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsMeterings");

namespace dss {

using metering::EnergyUnit;
using metering::Metering;
using metering::SeriesTypes;
using metering::TimeDuration;
using metering::ValueRange;

DecodedMeteringId decodeId(const dsModel::MeteringId& meteringId) {
  auto result = DecodedMeteringId{};
  std::stringstream ss(meteringId);

  std::string originType;
  std::getline(ss, originType, '-');
  if (originType != "dsm" && originType != "apartment" && originType != "openMeteringSystem") {
    DS_FAIL_REQUIRE("Unsupported measuring device.", originType);
  }

  if (originType == "dsm") {
    std::string meterId;
    std::getline(ss, meterId, '-');
    result.origin = dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{meterId}};
  } else if (originType == dsModel::OpenMeteringSystemOrigin{}.type) {
    result.origin = dsModel::OpenMeteringSystemOrigin{};
  } else if (originType == dsModel::ApartmentMeteringOrigin{}.type) {
    result.origin = dsModel::ApartmentMeteringOrigin{};
  }

  std::string meteringType;
  std::getline(ss, meteringType);

  // power -> powerMetering
  if (auto tmp = ds::tryToEnum<dsModel::MeteringType>(ds::str(meteringType, "Metering"))) {
    result.meteringType = *tmp;
  } else {
    DS_FAIL_REQUIRE("Unsupported measuring type.", meteringType);
  }
  return result;
}

dsModel::MeteringId encodeId(const dsModel::MeteringOrigin& meteringOrigin, dsModel::MeteringType meteringType) {
  std::string meteringId{};
  std::visit(::ds::Visitor{[&](const std::monostate& origin) { DS_FAIL_REQUIRE("monostate"); },
                           [&](const dsModel::ApartmentMeteringOrigin& origin) {
                             meteringId = ds::str(origin.type, "-", meteringType);
                           },
                           [&](const dsModel::ControllerMeteringOrigin& origin) {
                             meteringId = ds::str("dsm-", origin.id, "-", meteringType);
                           },
                           [&](const dsModel::OpenMeteringSystemOrigin& origin) {
                             meteringId = ds::str(origin.type, "-", meteringType);
                           }},
             meteringOrigin);
  size_t meteringPos = meteringId.rfind("Metering"); // inconsistency in openapi schema
  DS_REQUIRE(meteringPos != std::string::npos, "invalid metering type", meteringType);
  return dsModel::MeteringId{meteringId.substr(0, meteringPos)};
};

namespace {

auto toUnit = [](const dsModel::MeteringType& type) {
  switch (type) {
    case dsModel::MeteringType::powerMetering: [[fallthrough]];
    case dsModel::MeteringType::powerProducedMetering: return "W"s;
    case dsModel::MeteringType::energyMetering: [[fallthrough]];
    case dsModel::MeteringType::energyProducedMetering:
    case dsModel::MeteringType::energyConsumedTariff1Metering:
    case dsModel::MeteringType::energyConsumedTariff2Metering:
    case dsModel::MeteringType::energyProducedTariff1Metering:
    case dsModel::MeteringType::energyProducedTariff2Metering: return "Wh"s; // TODO rrd returns Ws by default
    case dsModel::MeteringType::none:
    default: DS_FAIL_REQUIRE("bad metering type");
  }
};

class DsMetering final : public dsModel::MeteringsApi, public ds::asio::Resource<> {
  Apartment& m_apartment;
  Metering& m_meterings;
  dsApi::EventHub& m_eventHub;
  ds::asio::Timer m_pushMeteringValuesTimer{io()};

public:
  DsMetering(ds::asio::ResourceBase& parent, Apartment& apartment, dsApi::EventHub& eventHub)
      : ds::asio::Resource<>{parent, "meterings"},
        m_apartment{apartment},
        m_meterings{m_apartment.getMetering()},
        m_eventHub{eventHub} {
    pushMeteringValuesAsyncLoop();
  }

  [[nodiscard]] dsModel::Metering metering(const dsModel::MeteringArgs& args) const override {
    auto& meteringId = args.meteringId;
    DS_CONTEXT(meteringId);
    try {
      std::string technicalName{};

      auto toString = [](const dsModel::MeteringType& type) {
        switch (type) {
          case dsModel::MeteringType::powerMetering: return "power"s;
          case dsModel::MeteringType::energyMetering: return "energy"s;
          case dsModel::MeteringType::powerProducedMetering: return "power produced"s;
          case dsModel::MeteringType::energyProducedMetering: return "energy produced"s;
          case dsModel::MeteringType::energyConsumedTariff1Metering: return "energy consumed tariff 1"s;
          case dsModel::MeteringType::energyConsumedTariff2Metering: return "energy consumed tariff 2"s;
          case dsModel::MeteringType::energyProducedTariff1Metering: return "energy produced tariff 1"s;
          case dsModel::MeteringType::energyProducedTariff2Metering: return "energy produced tariff 2"s;
          case dsModel::MeteringType::none:
          default: DS_FAIL_REQUIRE("bad metering type");
        }
      };

      auto meteringDecoded = decodeId(args.meteringId);

      if (std::holds_alternative<dsModel::OpenMeteringSystemOrigin>(meteringDecoded.origin)) {
        auto omsOrigin = m_meterings.openMeteringSystemOrigin();
        DS_REQUIRE(omsOrigin, "OpenMeteringSystem not present");
        meteringDecoded.origin = *omsOrigin;
      }

      std::visit(::ds::Visitor{[&](const std::monostate& origin) { DS_FAIL_REQUIRE("monostate"); },
                               [&](dsModel::ApartmentMeteringOrigin& origin) {
                                 technicalName = ds::str("apartment ", toString(meteringDecoded.meteringType));
                               },
                               [&](dsModel::ControllerMeteringOrigin& origin) {
                                 auto meter = m_apartment.getDSMeterByDSID(Dsuid(origin.id));
                                 DS_REQUIRE(meter->isPresent(), "Metering device not present.", origin.id); // xx
                                 technicalName =
                                     ds::str(meter->getHardwareName(), " ", toString(meteringDecoded.meteringType));
                               },
                               [&](dsModel::OpenMeteringSystemOrigin& origin) {
                                 technicalName = ds::str("openMeteringSystem ", toString(meteringDecoded.meteringType));
                               }},
                 meteringDecoded.origin);

      return dsModel::Metering{.id = args.meteringId,
                               .type = meteringDecoded.meteringType,
                               .lastChanged = {},
                               .attributes = {.unit = toUnit(meteringDecoded.meteringType),
                                              .technicalName = technicalName,
                                              .origin = meteringDecoded.origin}};
    } catch (const std::exception& e) {
      DS_ERROR("Metering failed.", e);
      return dsModel::Metering{.id = meteringId};
    }
  }

  [[nodiscard]] dsModel::Meterings meterings() const override {
    DS_NOTICE("Get meterings.");

    auto result = dsModel::Meterings{};
    auto meteringChannels = m_meterings.meterings();
    std::transform(std::cbegin(meteringChannels), std::cend(meteringChannels), std::back_inserter(result.meterings),
                   [this](const auto& elt) {
                     return metering(dsModel::MeteringArgs{.meteringId = encodeId(elt.first, elt.second)});
                   });

    boost::sort(result.meterings, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  [[nodiscard]] dsModel::MeteringValue meteringValue(const dsModel::MeteringArgs& args) const override {
    const auto& meteringId = args.meteringId;
    DS_CONTEXT(meteringId);
    try {
      double value{};
      ds::DateTime lastChanged{};

      auto meteringDecoded = decodeId(meteringId);
      std::visit(::ds::Visitor{[&](const std::monostate& origin) { DS_FAIL_REQUIRE("monostate"); },
                               [&](dsModel::ControllerMeteringOrigin& origin) {
                                 auto meter = m_apartment.getDSMeterByDSID(Dsuid(origin.id));
                                 DS_REQUIRE(meter->isPresent(), "Metering device not present.", origin.id);

                                 if (auto lastReading = m_meterings.getLastReading(meteringDecoded.origin,
                                                                                   meteringDecoded.meteringType)) {
                                   value = lastReading->value;
                                   lastChanged = lastReading->lastChanged;
                                 }
                               },

                               [&](dsModel::ApartmentMeteringOrigin& origin) {
                                 if (auto lastReading = m_meterings.getLastReading(meteringDecoded.origin,
                                                                                   meteringDecoded.meteringType)) {
                                   value = lastReading->value;
                                   lastChanged = lastReading->lastChanged;
                                 }
                               },
                               [&](dsModel::OpenMeteringSystemOrigin origin) {
                                 // we announced we deliver Wh in the meterings() call, while getLastReading defaults do
                                 // Ws. Don't touch the others calls for now, since existing UI code might break
                                 // -> TODO it later
                                 if (auto lastReading = m_meterings.getLastReading(meteringDecoded.origin,
                                                                                   meteringDecoded.meteringType,
                                                                                   metering::EnergyUnit::Wh)) {
                                   value = lastReading->value;
                                   lastChanged = lastReading->lastChanged;
                                 }
                               }},
                 meteringDecoded.origin);

      return dsModel::MeteringValue{
          .id = meteringId, .type = "meteringValue", .lastChanged = lastChanged, .attributes = {.value = value}};
    } catch (const std::exception& e) {
      DS_ERROR("Metering value failed.", e);
      return dsModel::MeteringValue{.id = meteringId};
    }
  }

  [[nodiscard]] dsModel::MeteringValues meteringValues() const override {
    DS_INFO_ENTER("Get metering values.");
    return meteringValuesImpl();
  }

  [[nodiscard]] dsModel::MeteringValues meteringValuesImpl() const {
    DS_DEBUG_ENTER("Get metering values.");
    auto result = dsModel::MeteringValues{};

    auto meteringChannels = m_meterings.meterings();
    std::transform(std::cbegin(meteringChannels), std::cend(meteringChannels), std::back_inserter(result.values),
                   [this](const auto& elt) {
                     return meteringValue(dsModel::MeteringArgs{.meteringId = encodeId(elt.first, elt.second)});
                   });

    boost::sort(result.values, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  [[nodiscard]] dsModel::SeriesCompatValue seriesCompatValue([
      [maybe_unused]] const dsModel::MeteringSeriesArgs& args) const override {
    DS_CONTEXT(args.meteringId);
    auto meteringId = decodeId(args.meteringId);
    ValueRange valuesConsumption;
    ValueRange valuesProduction;

    metering::ValueRange values{};
    std::visit(::ds::Visitor{
                   [&](const std::monostate& origin) { DS_FAIL_REQUIRE("monostate"); },
                   [](dsModel::ApartmentMeteringOrigin& origin) { DS_NOTICE(origin); },
                   [&](dsModel::ControllerMeteringOrigin& origin) {
                     DS_NOTICE(origin);
                     values = m_meterings.getSeriesValue(meteringId.origin, meteringId.meteringType,
                                                         std::chrono::seconds{args.resolution}, 0, EnergyUnit::Wh);
                   },
                   [&](dsModel::OpenMeteringSystemOrigin& origin) {
                     auto mapToProduction = [](auto meteringType) {
                       switch (meteringType) {
                         case dsModel::MeteringType::powerMetering: return dsModel::MeteringType::powerProducedMetering;
                         case dsModel::MeteringType::energyMetering:
                           return dsModel::MeteringType::energyProducedMetering;
                         case dsModel::MeteringType::energyConsumedTariff1Metering:
                           return dsModel::MeteringType::energyProducedTariff1Metering;
                         case dsModel::MeteringType::energyConsumedTariff2Metering:
                           return dsModel::MeteringType::energyProducedTariff2Metering;

                         case dsModel::MeteringType::powerProducedMetering:
                         case dsModel::MeteringType::energyProducedMetering:
                         case dsModel::MeteringType::energyProducedTariff1Metering:
                         case dsModel::MeteringType::energyProducedTariff2Metering:
                         case dsModel::MeteringType::none:
                         default: DS_FAIL_REQUIRE("query consumption, production is implicit then");
                       }
                     };

                     DS_NOTICE(meteringId);

                     valuesConsumption =
                         m_meterings.getSeriesValue(meteringId.origin, meteringId.meteringType,
                                                    std::chrono::seconds{args.resolution}, 0, EnergyUnit::Wh);
                     valuesProduction =
                         m_meterings.getSeriesValue(meteringId.origin, mapToProduction(meteringId.meteringType),
                                                    std::chrono::seconds{args.resolution}, 0, EnergyUnit::Wh);
                   },
               },
               meteringId.origin);

    dsModel::SeriesCompatValue result{
        .meterID = args.meteringId, .unit = toUnit(meteringId.meteringType), .resolution = ds::str(args.resolution)};

    if (meteringId.meteringType == dsModel::MeteringType::energyMetering) {
      result.type = "energyDelta";
    }

    if (!valuesConsumption.values.empty()) {
      for (size_t i = 0; i < valuesConsumption.values.size(); i++) {
        if (i < valuesProduction.values.size()) {
          dss::DateTime stamp(valuesConsumption.values[i].timeStamp);
          dsModel::SeriesRealCompatValueValue compatValue{.value = {static_cast<double>(stamp.secondsSinceEpoch()),
                                                                    valuesConsumption.values[i].value,
                                                                    valuesProduction.values[i].value}};
          result.values.push_back(compatValue);
        }
      }
    }

    return result;
  }

  [[nodiscard]] dsModel::SeriesCompatValues seriesCompatValues() const override { return {}; }

  void pushMeteringValuesAsyncLoop() {
    // TODO(someday): avoid polling when there is no subscription
    m_pushMeteringValuesTimer.randomlyExpiresFromNow(1s, [this] {
      DS_CONTEXT("pushMeteringValues");
      DS_DEFER(pushMeteringValuesAsyncLoop());
      try {
        m_eventHub.emitEvent(dsApi::Event{
            .type = dsModel::meteringValues, .path = {}, .data = ds::toJsonFragment(meteringValuesImpl())});
      } catch (std::exception& e) {
        DS_ERROR("Failed to push event", e);
      }
    });
  }
};

} // namespace

std::unique_ptr<dsModel::MeteringsApi> makeDsMeterings(ds::asio::ResourceBase& parent, Apartment& apartment,
                                                       dsApi::EventHub& eventHub) {
  return std::make_unique<DsMetering>(parent, apartment, eventHub);
}

} // namespace dss
