/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/asio/resource.h>

DS_REFLECTED_ENUM((dsNet), SessionStateType, (init)(opening)(open)(closed));

namespace dsNet {

/// Session is implemented as shared_ptr by inheriting
/// enable_shared_from_this to  to be able
/// to prolong session livetime
/// until all callbacks are finished.
/// This is common idiom in asio world.
///
/// Therefore session livetime cannot be maintained
/// by unique_ptr.
/// Session can be only closed
/// and session can be closed by many reasons
/// like session not needed any more, communication error, etc.
/// Once the session is closed it cannot be opened and will be destroyed.
///
/// Closed session stops all active asynchronous operations
/// and will be destroyed when all references die.
///
/// `ClosePtr` passes "ownership" of session
/// to higher level code instead of `std::unique_ptr`.
/// Session will be closed when ClosePtr goes out of scope.
///
/// ISession inherits IResourceBase because
/// * every session is resource
/// * `id()` is used to identify the session among sessions
///
/// See also `Session`, `Sessions`.
struct ISession : virtual ds::asio::IResourceBase {
  /// Closes all active asynchronous operations.
  /// This leads to asynchronous drop of all internal ownership references.
  virtual void close() = 0;

  virtual std::shared_ptr<ISession> sharedSession() = 0;
  std::weak_ptr<ISession> weakSession() { return sharedSession(); }
};

/// ClosePtr is like unique pointer,
/// but it holds `shared_ptr` referce to session and
/// closes the session (`Session::close()`) before
/// releasing the session `shared_ptr` reference.
///
/// Technically more ClosePtr instances can link to the same session.
/// Session is closed on first call to close()
template <class T>
class ClosePtr final : private std::shared_ptr<T> {
  using Ptr = std::shared_ptr<T>;
  static_assert(std::is_base_of_v<ISession, T>);
  const Ptr& ptr() const { return static_cast<const Ptr&>(*this); }
  Ptr& ptr() { return static_cast<Ptr&>(*this); }

public:
  ClosePtr() = default;
  ClosePtr(Ptr&& x) : Ptr{std::move(x)} {}
  template <class Y>
  ClosePtr(std::shared_ptr<Y>&& x) : Ptr{std::move(x)} {}
  ClosePtr(const ClosePtr<T>&) = delete;
  ClosePtr& operator=(const ClosePtr<T>&) = delete;
  ClosePtr(ClosePtr<T>&& x) = default;
  ClosePtr& operator=(ClosePtr<T>&& x) = default;

  ~ClosePtr() {
    if (ptr()) {
      (*this)->close();
    }
  }

  using Ptr::operator*;
  using Ptr::operator->;
  using Ptr::swap;
  using Ptr::operator bool;

  void reset() {
    if (ptr()) {
      ptr()->close();
    }
    ptr().reset();
  }
};

// Defined in sessions-base.h, but appears a lot in interfaces.
class Sessions;

/// `SessionState` mixin provides `SessionStateType state()`
/// with methods supporting state transitions.
///
/// Session is created in `init` state.
///
/// Session method `asyncRun(&&hooks)` enters `opening` state
/// and starts initial handshake.
/// Hooks are passed to asyncRun from higher level code.
/// Hooks cannot be passed to session contructor
/// to support dependency injection of session into higher level code.
///
/// After succesful handshake,  session enters `open` state.
/// Session is fully operational.
///
/// Session enters `closed` on any error or
/// on close from higher level code (see `ClosePtr`).
struct SessionState {
  virtual ~SessionState() = default;
  virtual SessionStateType state() const = 0;
  bool isClosed() const { return state() == SessionStateType::closed; }
};

} // namespace dsNet
