/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "tcp-acceptor.h"

#include <ds/log.h>

static ds::LogChannel dsLogChannel("dsNetTcpAcceptor");

using namespace std::literals;
using boost::system::error_code;
using tcp = boost::asio::ip::tcp;

namespace dsNet {

tcp::endpoint Endpoint::toAsio() const {
  return tcp::endpoint{boost::asio::ip::make_address(address), port};
}

Endpoint makeEndpoint(const tcp::endpoint& x) {
  return {.address = x.address().to_string(), .port = x.port()};
}

namespace {

struct TcpAcceptorImpl final : public TcpAcceptor, public ds::asio::Resource<TcpAcceptorConfig> {
  struct Shared final : public std::enable_shared_from_this<Shared> {
    TcpAcceptorImpl& m_parent;
    tcp::acceptor m_acceptor{m_parent.io()};
    tcp::socket m_socket{m_parent.io()};
    bool m_running = false;

    Shared(TcpAcceptorImpl& parent) : m_parent{parent} {
      DS_DEBUG_ENTER("");
      const auto endpoint = m_parent.config().endpoint.toAsio();
      m_acceptor.open(endpoint.protocol());
      m_acceptor.set_option(tcp::socket::reuse_address(true));
      m_acceptor.bind(endpoint);
      m_acceptor.listen();
      // We cannot start the loop from constructor because
      // `shared_from_this()` cannot be used from constructor.
    }
    ~Shared() { DS_DEBUG_ENTER(""); }

    void asyncRun() {
      DS_REQUIRE(!m_running);
      m_running = true;
      // "Loop" forever accepting new connections.
      asynAcceptLoop();
    }

    void asynAcceptLoop() {
      DS_INFO_ENTER("");
      m_acceptor.async_accept(m_socket, [weakSelf = weak_from_this()](error_code ec) {
        if (auto self = weakSelf.lock()) {
          // m_arent is our only owner. So it is alive.
          DS_INFO_ENTER(ec);
          if (!ec) {
            self->m_parent.m_hooks.accept(std::move(self->m_socket));
          } else {
            if (ec != boost::system::errc::operation_canceled) {
              DS_ERROR("Accept failed", ec);
            }
          }
          self->asynAcceptLoop();
        }
      });
    }
  };
  TcpAcceptorHooks m_hooks;
  const std::shared_ptr<Shared> m_shared = std::make_shared<Shared>(*this);

  TcpAcceptorImpl(ds::asio::ResourceBase& parent, TcpAcceptorConfig&& config)
      : ds::asio::Resource<TcpAcceptorConfig>{parent, "tcpAcceptor", std::move(config)} {
    DS_DEBUG_ENTER("");
  }

  void asyncRun(TcpAcceptorHooks&& hooks) override {
    m_hooks = std::move(hooks);
    m_shared->asyncRun();
  }
  Endpoint localEndpoint() const override { return makeEndpoint(localAsioEndpoint()); }
  tcp::endpoint localAsioEndpoint() const override { return m_shared->m_acceptor.local_endpoint(); }
};
} // namespace

std::unique_ptr<TcpAcceptor> makeTcpAcceptor(ds::asio::ResourceBase& parent, TcpAcceptorConfig&& config) {
  return std::make_unique<TcpAcceptorImpl>(parent, std::move(config));
}

} // namespace dsNet

DS_JSON_DEFINE_STRUCT((dsNet), Endpoint);
DS_JSON_DEFINE_STRUCT((dsNet), TcpAcceptorConfig);
