#pragma once

#include <ds/asio/move-only.h>

#include <boost/asio/io_service.hpp>
#include <functional>
#include <memory>
#include <optional>

namespace ds {

// Promise of value fullfilled by setting `value` and destroying the `Promise` object.
//
// Similar to std::promise, but std::promise (as of c++17)
// is usable only for thread synchronization.
// This promise is intended for asynchronous notifications.
// The same story for boost::promise.
template <class T>
class Promise {
public:
  /// @args fulfilled Callback called when promise is fulfilled (i.e. destroyed)
  Promise(boost::asio::io_service& io, std::function<void(std::optional<T>&&)>&& fulfilled);
  ~Promise();

  // Promissed value. Nullopt means that promise was cancelled.
  std::optional<T> value{};

private:
  boost::asio::io_service& m_io;
  std::function<void(std::optional<T>&&)> m_fulfilled{};
};

template <class T>
std::unique_ptr<Promise<T>> makePromise(boost::asio::io_service& io,
                                        std::function<void(std::optional<T>&&)>&& fulfilled) {
  return std::make_unique<Promise<T>>(io, std::move(fulfilled));
}

// implementation details
template <class T>
Promise<T>::Promise(boost::asio::io_service& io, std::function<void(std::optional<T>&&)>&& fulfilled)
    : m_io{io}, m_fulfilled{std::move(fulfilled)} {};

template <class T>
Promise<T>::~Promise() {
  m_io.post(ds::asio::MoveOnly{
      [fulfilled = std::move(m_fulfilled), value = std::move(value)]() mutable { fulfilled(std::move(value)); }});
}

} // namespace ds
