/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "resource.h"

#include <ds/asio/loop.h>
#include <ds/asio/test.h>

#include <trompeloeil.hpp>

namespace {

struct Config {
  std::string s;
};
} // namespace

DS_REFLECT_ANONYMOUS_STRUCT(Config, (s));

struct IsSynchronizedMock {
public:
  MAKE_MOCK1(call, void(bool));
};

namespace {

static const char* TAGS = "[dsAsioResource][dsAsio][ds]";
using namespace std::chrono_literals;
using namespace trompeloeil;

TEST_CASE("dsAsioInvoke", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  auto loop2 = ds::asio::Loop{loop.system.conf(), "testLoop2", {}};

  SECTION("is executed in other thread, passes returned value") {
    REQUIRE(ds::asio::invoke(loop2.io(), [&] { return std::this_thread::get_id(); }) == loop2.context().threadId);
  }

  SECTION("passes exceptions") {
    REQUIRE_THROWS_FIND(ds::asio::invoke(loop2.io(), [&] { DS_FAIL_REQUIRE("Test failure"); }), "Test failure");
  }

  SECTION("Current log context is passed to invoked") {
    DS_CONTEXT("aba");
    REQUIRE_THAT(ds::asio::invoke(loop2.io(), [&] { return ds::str(ds::logContext()); }), Catch::Contains("aba"));
  }
}

TEMPLATE_TEST_CASE("dsAsioResource", TAGS, ds::asio::EmptyResourceConfig, Config) {
  auto loop = ds::asio::TestLoopResource{};

  WHEN("Created from context") {
    auto resource = ds::asio::Resource<TestType>(loop, "resource");
    THEN("context, id, path, io are set") {
      CHECK(&resource.context() == &loop.context);
      CHECK(resource.path() == "/resource");
      CHECK(resource.id() == "resource");
      CHECK(&resource.io() == &loop.io);
    }

    WHEN("child resource") {
      auto childResource = ds::asio::Resource<>(resource, "child");
      THEN("context, id, io are set") {
        CHECK(&childResource.context() == &loop.context);
        CHECK(childResource.path() == "/resource/child");
        CHECK(childResource.id() == "child");
        CHECK(&childResource.io() == &loop.io);
      }
      THEN("resource is serialized as path") { CHECK(ds::str(childResource) == childResource.path()); }
    }
  }
  THEN("Resource can be destroyed in io descructor") {
    loop.post([x = std::make_shared<ds::asio::Resource<>>(loop, "x")] {});
    // intentionally no loop.run()
  }
}

TEST_CASE("dsAsioResource with config without conf entry", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  WHEN("Created from context") {
    auto defaultConfig = Config{.s = "default"};
    auto resource = ds::asio::Resource<Config>(loop, "resource", Config{defaultConfig});
    THEN("config() matches defaultConfig") { CHECK(resource.config() == defaultConfig); }
    THEN("jsonConfig() matches defaultConfig") { CHECK(resource.jsonConfig() == ds::toJsonFragment(defaultConfig)); }

    WHEN("child resource") {
      auto defaultChildConfig = Config{.s = "childDefault"};
      auto childResource = ds::asio::Resource<Config>(resource, "child", Config{defaultChildConfig});
      THEN("config matches defaultConfig") { CHECK(childResource.config() == defaultChildConfig); }
      THEN("jsonConfig() matches defaultConfig") {
        CHECK(childResource.jsonConfig() == ds::toJsonFragment(defaultChildConfig));
      }
    }
  }
}

TEST_CASE("dsAsioResource with config with conf entry", TAGS) {
  auto loop = ds::asio::TestLoopResource{};

  WHEN("Created from context") {
    const auto config = Config{.s = "X"};
    const auto defaultConfig = Config{.s = "default"};
    REQUIRE_CALL(loop.system.conf(), readJsonIfExists(eq("/resource"))).RETURN(ds::toJsonFragment(config));
    auto resource = ds::asio::Resource<Config>(loop, "resource", Config{defaultConfig});
    THEN("config matches conf") { CHECK(resource.config() == config); }
    THEN("jsonConfig() matches defaultConfig") { CHECK(resource.jsonConfig() == ds::toJsonFragment(config)); }

    WHEN("child resource") {
      const auto childConfig = Config{.s = "childX"};
      const auto defaultChildConfig = Config{.s = "childDefault"};
      REQUIRE_CALL(loop.system.conf(), readJsonIfExists(eq("/resource/child"))).RETURN(ds::toJsonFragment(childConfig));

      auto childResource = ds::asio::Resource<Config>(resource, "child", Config{defaultChildConfig});
      THEN("config matches conf") { CHECK(childResource.config() == childConfig); }
      THEN("jsonConfig() matches defaultConfig") {
        CHECK(childResource.jsonConfig() == ds::toJsonFragment(childConfig));
      }
    }
  }
  THEN("Cannot be created with id with special characters") {
    CHECK_THROWS_FIND(ds::asio::Resource<>(loop, "/"), "Special characters in id are not allowed");
  }
}

TEST_CASE("dsAsioResource children", TAGS) {
  auto loop = ds::asio::TestLoopResource{};

  WHEN("resource1, resource1/1, resource1/2, resource2") {
    auto resource1 = ds::asio::Resource<>(loop, "resource1");
    auto resource11 = ds::asio::Resource<>(resource1, "1");
    auto resource12 = ds::asio::Resource<>(resource1, "2");
    auto resource2 = ds::asio::Resource<>(loop, "resource2");

    THEN("childIds()") { CHECK(resource1.childIds() == std::vector<std::string>{"1", "2"}); }
  }
}

TEST_CASE("dsAsioResource synchronization", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  auto loop2 = ds::asio::Loop{loop.system.conf(), "testLoop2", {}};
  auto resource =
      ds::asio::invoke(loop2.io(), [&] { return std::make_unique<ds::asio::Resource<>>(loop2.context(), "resource"); });
  auto isSynchronized = IsSynchronizedMock{};
  WHEN("outside of io") {
    THEN("!isSynchronized") {
      CHECK(!resource->isSynchronized());
      CHECK(resource->isBlockingAllowed());
    }
  }
  WHEN("in invoke(io)") {
    THEN("isSynchronized") {
      REQUIRE_CALL(isSynchronized, call(true));
      ds::asio::invoke(loop2.io(), [&] {
        isSynchronized.call(resource->isSynchronized());
        CHECK(!resource->isBlockingAllowed());
      });
    }
  }
  SECTION("isSynchronized in resource.invoke()") {
    SECTION("returns void") {
      REQUIRE_CALL(isSynchronized, call(true));
      resource->invoke([&] { isSynchronized.call(resource->isSynchronized()); });
    }
    SECTION("returns value") {
      REQUIRE_CALL(isSynchronized, call(true));
      const auto x = resource->invoke([&] {
        isSynchronized.call(resource->isSynchronized());
        return 7;
      });
      CHECK(x == 7);
    }
  }
  // resource lives on loop2, it must be deleted in loop2
  ds::asio::invoke(loop2.io(), [&] { resource.reset(); });
}

TEST_CASE("dsAsioResourceNormalizeId", TAGS) {
  CHECK(ds::asio::ResourceBase::isIdNormalized("azAZ_-09"));
  CHECK(!ds::asio::ResourceBase::isIdNormalized("azAZ_!-09"));
  CHECK(!ds::asio::ResourceBase::isIdNormalized("azAZ_/-09"));
  CHECK(ds::asio::ResourceBase::normalizeId("azAZ!@#$%^&*()/\\_-09") == "azAZ------------_-09");
}

} // namespace
