/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "state-poller.h"
#include <ds/asio/test.h>
#include <ds/clock-mock.h>
#include <trompeloeil.hpp>

namespace {

static const char* TAGS = "[dsAsioStatePoller][dsAsio][ds]";

using namespace trompeloeil;
using namespace std::literals;

struct Poll {
public:
  MAKE_MOCK0(call, ds::JsonFragment());
};

struct Fire {
public:
  MAKE_MOCK1(call, void(ds::JsonFragment&&));
};

struct PolledAndNoFire {
public:
  MAKE_MOCK0(call, void());
};

SCENARIO("dsAsioStatePoller", TAGS) {
  auto clock = ds::ClockMocker{};
  auto loop = ds::asio::TestLoop{};

  auto poll = Poll{};
  auto fire = Fire{};
  auto polledAndNoFire = PolledAndNoFire{};
  const auto moderatePeriod = 1000ms;
  const auto loopPeriod = 60000ms;
  auto poller = ds::asio::StatePoller{loop.io,
                                      "test",
                                      {.moderate = moderatePeriod, .loop = loopPeriod},
                                      {
                                          .poll = [&] { return poll.call(); },
                                          .fire = [&](auto&& x) { fire.call(std::move(x)); },
                                          .polledAndNoFire = [&] { return polledAndNoFire.call(); },
                                      }};

  auto pollAfterLoopPeriod = [&] {
    clock.advance(loopPeriod * 3 / 4);
    loop.postPostStopRun();
    REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(9));
    REQUIRE_CALL(fire, call(_)).LR_SIDE_EFFECT(CHECK(ds::toJson(_1) == "9"); loop.stop());
    clock.advance(loopPeriod * 2 / 4);
    loop.run();
  };

  THEN("can be created and deleted") {}

  WHEN("set active") {
    poller.setActive(true);

    THEN("can be created and deleted") {}

    THEN("polls and fires immediately (any state is new)") {
      {
        REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(5));
        REQUIRE_CALL(fire, call(_)).LR_SIDE_EFFECT(CHECK(ds::toJson(_1) == "5"); loop.stop());
        loop.run();
      }
      WHEN("polled") {
        poller.poll();
        loop.postStopRun();

        THEN("no poll until moderate period") {
          clock.advance(moderatePeriod * 3 / 4);
          loop.postPostStopRun();
        }

        THEN("poll the same value after moderate period does no fire") {
          {
            REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(5));
            REQUIRE_CALL(polledAndNoFire, call()).LR_SIDE_EFFECT(loop.stop());
            clock.advance(moderatePeriod * 5 / 4);
            loop.run();
          }

          THEN("no poll until loop period") {
            clock.advance(loopPeriod * 3 / 4);
            loop.postPostStopRun();
            loop.postPostStopRun();
          }

          THEN("poll the same value after loop period does no fire") {
            {
              REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(5));
              REQUIRE_CALL(polledAndNoFire, call()).LR_SIDE_EFFECT(loop.stop());
              clock.advance(loopPeriod * 5 / 4);
              loop.run();
            }

            THEN("poll after moderation period on poll") {
              {
                poller.poll();
                clock.advance(moderatePeriod * 3 / 4);
                loop.postPostStopRun();
                loop.postStopRun();
                clock.advance(moderatePeriod * 2 / 4);
                REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(6));
                REQUIRE_CALL(fire, call(_)).LR_SIDE_EFFECT(CHECK(ds::toJson(_1) == "6"); loop.stop());
                loop.run();
              }

              THEN("poll after loop period") { pollAfterLoopPeriod(); }
            }

            THEN("poll after loop period") { pollAfterLoopPeriod(); }
          }

          THEN("poll different value after loop period fires") {
            {
              REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(6));
              REQUIRE_CALL(fire, call(_)).LR_SIDE_EFFECT(CHECK(ds::toJson(_1) == "6"); loop.stop());
              clock.advance(loopPeriod * 5 / 4);
              loop.run();
            }

            THEN("poll after loop period") { pollAfterLoopPeriod(); }
          }
        }

        THEN("poll different value after moderate period fires") {
          {
            REQUIRE_CALL(poll, call()).LR_RETURN(ds::toJsonFragment(6));
            REQUIRE_CALL(fire, call(_)).LR_SIDE_EFFECT(CHECK(ds::toJson(_1) == "6"); loop.stop());
            clock.advance(moderatePeriod * 5 / 4);
            loop.run();
          }

          THEN("poll after loop period") { pollAfterLoopPeriod(); }
        }
      }
    }
  }
}

} // namespace
