/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "reflect.h"
#include <ds/catch/catch.h>

static const char* TAGS = "[dsReflect][ds]";

namespace dsReflectTest {

enum class Transport { CAR, PLANE };
inline std::ostream& operator<<(std::ostream& ostream, const Transport& x) {
  return ostream << static_cast<int>(x);
}

struct X {
  int i;
  char c;
  unsigned char uc;
  std::string s;
};
} // namespace dsReflectTest

using dsReflectTest::Transport;
using dsReflectTest::X;

namespace {
struct Y {
  X x;
  std::vector<X> xs;
  Transport transport;
};
} // namespace

DS_REFLECT_STRUCT((dsReflectTest), X, (i)(c)(uc)(s));
DS_REFLECT_ANONYMOUS_STRUCT(Y, (x)(xs)(transport));

namespace {
TEST_CASE("ds::reflect_t<T> carries type of reflected type", TAGS) {
  static_assert(std::is_same<ds::reflect_t<X>, X>::value, "ds::reflect_t does not work for reflected type");
}

TEST_CASE("dsIsEmptyInstance", TAGS) {
  CHECK(ds::isEmptyInstance(static_cast<const char*>(nullptr)));
  CHECK(ds::isEmptyInstance((const char*)""));
  CHECK(!ds::isEmptyInstance("a"));

  CHECK(ds::isEmptyInstance(std::string()));
  CHECK(ds::isEmptyInstance(std::string("")));
  CHECK(!ds::isEmptyInstance(std::string("a")));

  CHECK(ds::isEmptyInstance((char)0));
  CHECK(ds::isEmptyInstance((unsigned char)0));
  CHECK(ds::isEmptyInstance((signed char)0));
  CHECK(ds::isEmptyInstance((unsigned short)0));
  CHECK(ds::isEmptyInstance((signed short)0));
  CHECK(ds::isEmptyInstance(0));
  CHECK(ds::isEmptyInstance(0U));
  CHECK(ds::isEmptyInstance(0LL));
  CHECK(ds::isEmptyInstance(0ULL));
  CHECK(!ds::isEmptyInstance((char)1));
  CHECK(!ds::isEmptyInstance((unsigned char)1));
  CHECK(!ds::isEmptyInstance((signed char)1));
  CHECK(!ds::isEmptyInstance((unsigned short)1));
  CHECK(!ds::isEmptyInstance((signed short)1));
  CHECK(!ds::isEmptyInstance(1));
  CHECK(!ds::isEmptyInstance(1U));
  CHECK(!ds::isEmptyInstance(1LL));
  CHECK(!ds::isEmptyInstance(1ULL));

  CHECK(ds::isEmptyInstance(0.0));
  CHECK(ds::isEmptyInstance(0.0f));
  CHECK(!ds::isEmptyInstance(1.0));
  CHECK(!ds::isEmptyInstance(1.0f));

  CHECK(ds::isEmptyInstance(Transport::CAR));
  CHECK(!ds::isEmptyInstance(Transport::PLANE));

  CHECK(ds::isEmptyInstance(std::vector<int>()));
  CHECK(!ds::isEmptyInstance(std::vector<int>(1, 1)));

  CHECK(ds::isEmptyInstance(X({})));
  CHECK(ds::isEmptyInstance(Y({})));
  {
    X x = {};
    x.i = 1;
    CHECK(!ds::isEmptyInstance(x));
  }
  {
    Y y = {};
    y.x.i = 1;
    CHECK(!ds::isEmptyInstance(y));
  }
}

TEST_CASE("dsReflectedEqual", TAGS) {
  CHECK(ds::reflectedEqual(X{}, X{}));
  CHECK(ds::reflectedEqual(X{.i = 1}, X{.i = 1}));
  CHECK(!ds::reflectedEqual(X{.i = 0}, X{.i = 1}));
  CHECK(ds::reflectedEqual(X{.i = 0, .c = 1}, X{.i = 0, .c = 1}));
  CHECK(!ds::reflectedEqual(X{.i = 0, .c = 0}, X{.i = 0, .c = 1}));
}

TEST_CASE("operator ==, !=", TAGS) {
  CHECK((X{} == X{}));
  CHECK((X{.i = 1} == X{.i = 1}));
  CHECK((X{.i = 0} != X{.i = 1}));
  CHECK((X{.i = 0, .c = 1} == X{.i = 0, .c = 1}));
  CHECK((X{.i = 0, .c = 0} != X{.i = 0, .c = 1}));
}

} // namespace
