/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "stack-trace.h"
#include <cxxabi.h>
#include <execinfo.h>
#include <algorithm>
#include <fstream>
#include <string_view>

namespace ds {

StackTrace StackTrace::capture(int ignoreCount) {
  auto result = StackTrace{};
  ignoreCount += 2; // this stack frame
  int size = ::backtrace(result.buffer.data(), result.buffer.size());
  result.first = std::min(ignoreCount, size);
  result.afterLast = size;
  return result;
}

template <class Write>
static void writeSymbolsToFd(Write&& write, const StackTrace& stackTrace) {
  // Based on https://panthema.net/2008/0901-stacktrace-demangled/

  const auto& values = stackTrace.span();
  if (values.empty()) {
    return write("  <empty bactrace, possibly corrupt>\n");
  }

  char** symbols = backtrace_symbols(values.data(), values.size());
  DS_DEFER(free(symbols));

  // allocate string which will be filled with the demangled function name
  size_t demangledNameSize = 8096;
  char* demangledName = static_cast<char*>(malloc(demangledNameSize));
  DS_DEFER(free(demangledName));

  write("\n");
  for (size_t i = 0; i < values.size(); i++) {
    auto&& symbol = symbols[i];
    char *beginName = 0, *beginOffset = 0, *endOffset = 0;

    // find parentheses and +address offset surrounding the mangled name:
    // ./module(function+0x15c) [0x8048a6d]
    for (char* p = symbol; *p; ++p) {
      if (*p == '(')
        beginName = p;
      else if (*p == '+')
        beginOffset = p;
      else if (*p == ')' && beginOffset) {
        endOffset = p;
        break;
      }
    }

    if (beginName && beginOffset && endOffset && beginName < beginOffset) {
      *beginName++ = '\0';
      *beginOffset++ = '\0';
      *endOffset = '\0';
      // mangled name is now in [beginName, beginOffset) and caller
      // offset in [beginOffset, endOffset). now apply __cxa_demangle():

      int status;
      write("  ");
      write(symbol);
      write(" : ");
      if (auto x = abi::__cxa_demangle(beginName, demangledName, &demangledNameSize, &status)) {
        write(x);
      } else { // demangling failed
        write(beginName);
      }
      write(" +");
      write(beginOffset);
      write("\n");
    } else {
      // couldn't parse the line? print the whole line.
      write("  ");
      write(symbol);
      write("\n");
    }
  }
}

std::ostream& operator<<(std::ostream& stream, const StackTrace& stackTrace) {
  writeSymbolsToFd([&](const char* x) { stream << x; }, stackTrace);
  return stream;
}

void writeSymbolsToFd(const StackTrace& stackTrace, int fd) {
  writeSymbolsToFd(
      [&](std::string_view x) {
        auto size = x.size();
        auto data = x.data();
        while (size) {
          auto written = ::write(fd, data, size);
          if (written < 0) {
            break;
          }
          size -= written;
          data += written;
        }
      },
      stackTrace);
}

} // namespace ds
