/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "ds485-device.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ds/locale/ds-locale.h>
#include <ds/log.h>
#include <bitset>

#include "cache/apartment-proxy.h"
#include "ds485/dsbusinterface.h"
#include "dss.h"
#include "vdc-messages.pb.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/modelconst.h"
#include "model/modelevent.h"
#include "model/modelmaintenance.h"
#include "model/dsmeter.h"
#include "model/set.h"

static ds::LogChannel dsLogChannel("deviceBusInterface");

namespace dss {

//================================================== DSDeviceBusInterface

// default timeout for synchronous calls to dsm api
static const int kDSM_API_TIMEOUT = 8;

uint8_t DSDeviceBusInterface::getDeviceConfig(const Device& device, uint8_t class_, uint8_t index) {
  DS_CONTEXT(device, class_, index);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint8_t retVal;

  int ret = DeviceConfig_get_sync_8(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, kDSM_API_TIMEOUT, &retVal);
  DSBusInterface::checkResultCode(ret);

  DS_NOTICE("Config.", retVal);

  /* The set config call that originates from us triggers a ds485 callback,
   * but the get call  does not, so we intercept it here directly */
  if (DSS::hasInstance()) {
    DSS::getInstance()->getApartment().proxy().updateDeviceConfig(device.getDSID(), class_, index, retVal);
  }
  return retVal;
} // getDeviceConfig

uint16_t DSDeviceBusInterface::getDeviceConfigWord(const Device& device, uint8_t class_, uint8_t index) {
  DS_CONTEXT(device, class_, index);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t retVal;

  int ret = DeviceConfig_get_sync_16(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, kDSM_API_TIMEOUT, &retVal);
  DSBusInterface::checkResultCode(ret);

  DS_NOTICE("Config.", retVal);
  return retVal;
} // getDeviceConfigWord

void DSDeviceBusInterface::setDeviceConfig(const Device& device, uint8_t class_, uint8_t index, uint8_t value) {
  DS_CONTEXT(device, class_, index, value);

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set config.");
  int ret;
  if (device.getDsmApiVersion() >= 0x301) {
    ret = DeviceConfig_set_sync(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, value, 30);
  } else {
    ret = DeviceConfig_set(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, value);
  }
  DSBusInterface::checkResultCode(ret);
} // setDeviceConfig

void DSDeviceBusInterface::setDeviceButtonActiveGroup(const Device& device, uint8_t group) {
  DS_CONTEXT(device, group);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set button active group.");
  int ret = DeviceProperties_set_button_active_group(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group);
  DSBusInterface::checkResultCode(ret);
} // setDeviceButtonActiveGroup

void DSDeviceBusInterface::setDeviceProgMode(const Device& device, uint8_t mode) {
  DS_CONTEXT(device, mode);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set prog mode.");
  int ret;
  if (mode) {
    ret = DeviceActionRequest_action_programming_mode_on(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress());
  } else {
    ret = DeviceActionRequest_action_programming_mode_off(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress());
  }
  DSBusInterface::checkResultCode(ret);
} // setDeviceProgMode

void DSDeviceBusInterface::setValue(const Device& device, uint8_t value) {
  DS_CONTEXT(device, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set value.", value);
  int ret =
      DeviceActionRequest_action_set_outval(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), value);
  DSBusInterface::checkResultCode(ret);

  if (m_pModelMaintenance) {
    auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etDeviceSetOutputValue, device.getDSMeterDSID());
    pEvent->addParameter(device.getShortAddress());
    pEvent->addParameter(value);
    m_pModelMaintenance->addModelEvent(pEvent);
  }
} // setValue

uint16_t DSDeviceBusInterface::getSensorValue(const Device& device, const int sensorIndex) {
  DS_CONTEXT(device, sensorIndex);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  uint16_t retVal;

  int ret = DeviceSensor_get_value_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), sensorIndex, kDSM_API_TIMEOUT, &retVal);
  DSBusInterface::checkResultCode(ret);

  auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etDeviceSensorValue, device.getDSMeterDSID());
  pEvent->addParameter(device.getShortAddress());
  pEvent->addParameter(sensorIndex);
  pEvent->addParameter(retVal);

  if (m_pModelMaintenance) {
    m_pModelMaintenance->addModelEvent(pEvent);
  }

  DS_NOTICE("Sensor value.", retVal);
  return retVal;
} // getSensorValue

DeviceSensorValue_t DSDeviceBusInterface::getSensorValueEx(const Device& device, const int sensorIndex) {
  DS_CONTEXT(device, sensorIndex);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  uint8_t contextMsg[64];
  DeviceSensorValue_t retVal;

  int ret = DeviceSensor_get_value_extended_sync(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(),
      sensorIndex, kDSM_API_TIMEOUT, &retVal.value, &retVal.valueAge, &retVal.contextId, contextMsg);
  DSBusInterface::checkResultCode(ret);

  retVal.contextMsg = std::string((const char*)(contextMsg));
  retVal.timestamp = DateTime().addMilliSeconds((int)retVal.valueAge * (-1));

  auto pEvent = std::make_shared<ModelEventWithSensorEx>();
  pEvent->m_deviceDSID = device.getDSMeterDSID();
  pEvent->addParameter(device.getShortAddress());
  pEvent->addParameter(sensorIndex);
  pEvent->addParameter(retVal.valueAge);
  pEvent->addParameter(retVal.contextId);
  pEvent->setSingleStringParameter(retVal.contextMsg);
  pEvent->m_sensorValue = retVal.value;

  if (m_pModelMaintenance) {
    m_pModelMaintenance->addModelEvent(pEvent);
  }

  DS_NOTICE("Sensor value.", retVal);
  return retVal;
} // getSensorValueEx

void DSDeviceBusInterface::addGroup(const Device& device, const int group) {
  DS_CONTEXT(device, group);

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }
  DS_NOTICE("Add group.");
  int ret = 0;
  if (device.getDsmApiVersion() >= 0x301) {
    ret = DeviceGroupMembershipModify_add_sync(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group, 30);
  } else {
    ret = DeviceGroupMembershipModify_add(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group);
  }
  if (ret == ERROR_WRONG_PARAMETER) {
    DS_WARNING("addGroup: dsm-api returned wrong parameter");
  } else {
    DSBusInterface::checkResultCode(ret);
  }
} // addGroup

void DSDeviceBusInterface::removeGroup(const Device& device, const int group) {
  DS_CONTEXT(device, group);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Remove group.");
  int ret =
      DeviceGroupMembershipModify_remove(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group);
  if (ret == ERROR_WRONG_PARAMETER) {
    DS_WARNING("addGroup: dsm-api returned wrong parameter");
  } else {
    DSBusInterface::checkResultCode(ret);
  }
} // removeGroup

void DSDeviceBusInterface::lockOrUnlockDevice(const Device& device, const bool locked) {
  DS_CONTEXT(device, locked);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set locked.");
  int ret = DeviceProperties_set_locked_flag(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), locked);
  DSBusInterface::checkResultCode(ret);
} // lockOrUnlockDevice

std::pair<uint8_t, uint16_t> DSDeviceBusInterface::getTransmissionQuality(const Device& device) {
  DS_CONTEXT(device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint8_t downstream;
  uint16_t upstream;

  int ret = TestTransmissionQuality_get_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), kDSM_API_TIMEOUT, &downstream, &upstream);
  DSBusInterface::checkResultCode(ret);

  const auto result = std::make_pair(downstream, upstream);
  DS_NOTICE("Transmission quality.", result);
  return result;
}

void DSDeviceBusInterface::increaseDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Increase output channel value.");
  int ret =
      DeviceActionRequest_action_opc_inc(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel);
  DSBusInterface::checkResultCode(ret);
} // increaseDeviceOutputChannelValue

void DSDeviceBusInterface::decreaseDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Decrease output channel value.");
  int ret =
      DeviceActionRequest_action_opc_dec(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel);
  DSBusInterface::checkResultCode(ret);
} // decreaseDeviceOutputChannelValue

void DSDeviceBusInterface::stopDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Stop output channel value.");
  int ret =
      DeviceActionRequest_action_opc_stop(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel);
  DSBusInterface::checkResultCode(ret);
} // stopDeviceOutputChannelValue

uint16_t DSDeviceBusInterface::getDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
    ;
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_current_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, 2 * kDSM_API_TIMEOUT, &out);
  DSBusInterface::checkResultCode(ret);
  DS_NOTICE("Output channel value.", out);
  return out;
} // getDeviceOutputChannelValue

void DSDeviceBusInterface::setDeviceOutputChannelValue(
    const Device& device, uint8_t channel, uint8_t size, uint16_t value, bool applyNow) {
  DS_CONTEXT(device, channel, size, value, applyNow);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  int ret;

  DS_NOTICE("Set output channel value.");
  if (applyNow) {
    ret = DeviceOPCConfig_set_current_and_apply(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, size, value);
  } else {
    ret = DeviceOPCConfig_set_current(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, size, value);
  }
  DSBusInterface::checkResultCode(ret);
} // setDeviceOutputChannelValue

uint16_t DSDeviceBusInterface::getDeviceOutputChannelSceneValue(const Device& device, uint8_t channel, uint8_t scene) {
  DS_CONTEXT(device, channel, scene);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_scene_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, scene, 2 * kDSM_API_TIMEOUT, &out);
  DSBusInterface::checkResultCode(ret);
  DS_NOTICE("Output channel scene value.", out);
  return out;
} // getDeviceOutputChannelSceneValue

void DSDeviceBusInterface::setDeviceOutputChannelSceneValue(
    const Device& device, uint8_t channel, uint8_t size, uint8_t scene, uint16_t value) {
  DS_CONTEXT(device, channel, size, scene, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set output channel scene value.");
  int ret = DeviceOPCConfig_set_scene(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, size, scene, value);
  DSBusInterface::checkResultCode(ret);
} // setDeviceOutputChannelSceneValue

uint16_t DSDeviceBusInterface::getDeviceOutputChannelSceneConfig(const Device& device, uint8_t scene) {
  DS_CONTEXT(device, scene);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_scene_config_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, 2 * kDSM_API_TIMEOUT, &out);
  DSBusInterface::checkResultCode(ret);
  DS_NOTICE("Output channel scene config.", out);
  return out;
} // getDeviceOutputChannelSceneConfig

void DSDeviceBusInterface::setDeviceOutputChannelSceneConfig(const Device& device, uint8_t scene, uint16_t value) {
  DS_CONTEXT(device, scene, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set output channel scene config.");
  int ret =
      DeviceOPCConfig_set_scene_config(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, value);
  DSBusInterface::checkResultCode(ret);
} // setDeviceOutputChannelSceneConfig

void DSDeviceBusInterface::setDeviceOutputChannelDontCareFlags(const Device& device, uint8_t scene, uint16_t value) {
  DS_CONTEXT(device, scene, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set output channel don't care.");
  int ret =
      DeviceOPCConfig_set_dc_flags(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, value);
  DSBusInterface::checkResultCode(ret);
}

uint16_t DSDeviceBusInterface::getDeviceOutputChannelDontCareFlags(const Device& device, uint8_t scene) {
  DS_CONTEXT(device, scene);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_dc_flags_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, 2 * kDSM_API_TIMEOUT, &out);
  DS_NOTICE("Output channel don't care.", ret);
  DSBusInterface::checkResultCode(ret);
  return out;
}

void DSDeviceBusInterface::genericRequest(const Device& device, const std::string& methodName,
    const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& params, const std::locale& locale) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  vdcapi::Message message;
  message.set_type(vdcapi::VDSM_REQUEST_GENERIC_REQUEST);
  auto genericRequest = message.mutable_vdsm_request_generic_request();
  genericRequest->set_dsuid(dsuid2str(device.getDSID()));
  genericRequest->set_methodname(methodName);
  *genericRequest->mutable_params() = params;
  uint8_t arrayOut[REQUEST_LEN];
  if (!message.SerializeToArray(arrayOut, sizeof(arrayOut))) {
    throw std::runtime_error("SerializeToArray failed");
  }
  uint8_t arrayIn[RESPONSE_LEN];
  uint16_t arrayInSize;
  int ret = UserProtobufMessageRequest(
      m_DSMApiHandle, device.getDSMeterDSID(), message.ByteSizeLong(), arrayOut, &arrayInSize, arrayIn);
  DSBusInterface::checkResultCode(ret);
  if (!message.ParseFromArray(arrayIn, arrayInSize)) {
    throw std::runtime_error("ParseFromArray failed");
  }
  if (message.type() != vdcapi::GENERIC_RESPONSE) {
    throw std::runtime_error("Invalid vdc response");
  }
  const vdcapi::GenericResponse& response = message.generic_response();
  if (response.code() != vdcapi::ERR_OK) {
    auto type = [&] {
      switch (response.errortype()) {
        case vdcapi::ERROR_TYPE_FAILED:
          return ds::Exception::Type::FAILED;
        case vdcapi::ERROR_TYPE_OVERLOADED:
          return ds::Exception::Type::OVERLOADED;
        case vdcapi::ERROR_TYPE_DISCONNECTED:
          return ds::Exception::Type::DISCONNECTED;
        case vdcapi::ERROR_TYPE_UNIMPLEMENTED:
          return ds::Exception::Type::UNIMPLEMENTED;
      }
      DS_ERROR("Unrecognized errorType", static_cast<int>(response.errortype()));
      return ds::Exception::Type::FAILED;
    }();
    throw ds::Exception(type, response.description(),
        dsDtr(locale, device.vdcTrDomain(), "error", response.usermessagetobetranslated()));
  }
}

void DSDeviceBusInterface::setProperty(
    const Device& device, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& properties) {
  return setVdcProperty(device.getDSID(), device.getDSMeterDSID(), properties);
}

vdcapi::Message DSDeviceBusInterface::getProperty(
    const Device& device, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& query) {
  return getVdcProperty(device.getDSID(), device.getDSMeterDSID(), query);
}

void DSDeviceBusInterface::reregister(const Device& device) {
  DS_CONTEXT(device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Reregister.");
  int ret;
  ret = DeviceConfig_trigger_readout(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getDSID());
  DSBusInterface::checkResultCode(ret);
} // reregister


} // namespace dss
