/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hal-device.h"

#include "model/apartment.h"

namespace dss::hal {

static ds::LogChannel dsLogChannel("halAudioDevice");

dsModel::AudioStatus HalAudioDevice::getPlayerStatus(PropertyNodePtr player) const {
  dsModel::AudioStatus deviceAudioState;
  PropertyNodePtr deviceState = player->getPropertyByName("state");
  if (deviceState != nullptr) {
    PropertyNodePtr state = deviceState->getPropertyByName("state");
    if (state != nullptr) {
      if (state->getStringValue() == "playing") {
        deviceAudioState.status = dsModel::AudioStatusStatus::playing;
      } else if (state->getStringValue() == "stop") {
        deviceAudioState.status = dsModel::AudioStatusStatus::stop;
      } else if (state->getStringValue() == "pause") {
        deviceAudioState.status = dsModel::AudioStatusStatus::pause;
      }
    }
    PropertyNodePtr muted = deviceState->getPropertyByName("muted");
    if (muted != nullptr) {
      deviceAudioState.muted = muted->getBoolValue();
    }
    PropertyNodePtr volume = deviceState->getPropertyByName("volume");
    if (volume != nullptr) {
      deviceAudioState.volume = volume->getIntegerValue();
    }
  }

  return deviceAudioState;
}

void HalAudioDevice::mergeStatus(dsModel::AudioStatus& zoneStatus, dsModel::AudioStatus deviceStatus) const {
  if (zoneStatus.status == dsModel::AudioStatusStatus::none) {
    zoneStatus.status = deviceStatus.status;
  } else if (zoneStatus.status == dsModel::AudioStatusStatus::playing || deviceStatus.status == dsModel::AudioStatusStatus::playing) {
    zoneStatus.status = dsModel::AudioStatusStatus::playing;
  } else if (zoneStatus.status == dsModel::AudioStatusStatus::stop) {
    if (deviceStatus.status == dsModel::AudioStatusStatus::pause) {
      zoneStatus.status = deviceStatus.status;
    }
  }

  if (deviceStatus.volume) {
    if (!zoneStatus.volume || *zoneStatus.volume < *deviceStatus.volume) {
      zoneStatus.volume = deviceStatus.volume;
    }
  }

  if (deviceStatus.muted) {
    if (!zoneStatus.muted) {
      zoneStatus.muted = deviceStatus.muted;
    } else if (*zoneStatus.muted && !*deviceStatus.muted) {
      zoneStatus.muted = deviceStatus.muted;
    }
  }
}

dsModel::AudioStatus HalAudioDevice::getZoneState(int zoneId) const {
  dsModel::AudioStatus zoneStatus;
  DS_DEBUG_ENTER(zoneId);
  PropertyNodePtr players = m_apartment->getPropertySystem()->getProperty("/scripts/sonos-controller/players");
  if (players != nullptr) {
    int childCount = players->getChildCount();
    DS_DEBUG(childCount);
    for (int i = 0; i < childCount; i++) {
      PropertyNodePtr player = players->getChild(i);
      DS_DEBUG(player->getPropertyByName("id")->getStringValue());
      PropertyNodePtr zonePtr = player->getPropertyByName("zoneID");
      if (zonePtr != nullptr) {
        int propZoneId = -1;
        try {
          propZoneId = std::stoi(zonePtr->getStringValue());
        } catch (const PropertyTypeMismatch& mismatch) {
          propZoneId = zonePtr->getIntegerValue();
        }
        DS_DEBUG("Player in Zone",propZoneId);
        if (propZoneId == zoneId) {
          PropertyNodePtr onlinePtr = player->getPropertyByName("online");
          if (onlinePtr != nullptr) {
            DS_DEBUG(onlinePtr->getBoolValue());
            if (onlinePtr->getBoolValue()) {
              dsModel::AudioStatus deviceStatus = getPlayerStatus(player);
              DS_DEBUG(deviceStatus);
              mergeStatus(zoneStatus, deviceStatus);
              DS_DEBUG("Merged", zoneStatus);
            }
          }
        }
      }
    }
  }
  DS_DEBUG(zoneStatus);
  return zoneStatus;
}

std::vector<int> HalAudioDevice::playerZoneScenarios(int zoneId) const {
  std::vector<int> scenarios;

  DS_DEBUG_ENTER(zoneId);
  PropertyNodePtr entries = m_apartment->getPropertySystem()->getProperty("/scripts/sonos-controller/entries");
  if (entries != nullptr) {
    int childCount = entries->getChildCount();
    DS_DEBUG_ENTER("Scene entries: ", childCount);
    for (int i = 0; i < childCount; i++) {
      PropertyNodePtr entry = entries->getChild(i);
      PropertyNodePtr conditions = entry->getProperty("conditions/enabled");
      if (conditions != nullptr) {
        bool enabled = conditions->getBoolValue();
        DS_DEBUG_ENTER(enabled);
        if (!enabled) {
          continue;
        }
      }
      PropertyNodePtr triggers = entry->getPropertyByName("triggers");
      if (triggers != nullptr) {
        int triggerCount = triggers->getChildCount();
        DS_DEBUG_ENTER("Triegger entries: ", triggerCount);
        if (triggerCount > 0) {
          DS_DEBUG("Fetch scene property");
          PropertyNodePtr scenePtr = triggers->getChild(0)->getPropertyByName("scene");
          if (scenePtr != nullptr) {
            DS_DEBUG(scenePtr->getIntegerValue());
            scenarios.push_back(scenePtr->getIntegerValue());
          }
        }
      }
    }
  }

  return scenarios;
}

} // namespace dss::hal
