/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once
#include <boost/range/algorithm_ext/push_back.hpp>
#include <boost/range/irange.hpp>
#include <vector>

#include "cache/cache-db.h"
#include "cache/device-cache.h"
#include "ds-model/model-fwd.h"
#include "ds/date-time.h"
#include "dssfwd.h"
#include "hal/hardware-abstraction.h"
#include "model/modelconst.h"
#include "propertysystem.h"

namespace dss::hal {

class HalDevice {
public:
  HalDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
      : m_sceneCacheDb(sceneCacheDb), m_StateCache(stateCache){};
  virtual ~HalDevice() = default;
  virtual void write(const Device& device, const std::string& parameter, double value) const;
  virtual void write(const Device& device, const std::string& parameter, bool value) const;
  virtual void write(const Device& device, const std::string& parameter, const std::string& value) const;

  virtual double readDouble(const Device& device, const std::string& parameter) const;
  virtual bool readBool(const Device& device, const std::string& parameter) const;
  virtual std::string readString(const Device& device, const std::string& parameter) const;
  virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const;
  void populateCache(const Device& device) const;

  virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const {
    DS_FAIL_REQUIRE("Not supported on this device type.");
  };
  virtual bool undoScene(const Device& device, int sceneId, ds::DateTime now) const;
  virtual void saveScene(const Device& device, int sceneId, ds::DateTime now) const;
  virtual bool setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
    DS_FAIL_REQUIRE("Not supported on this device type.");
  };
  virtual bool setOutputValue(const Device& device, ChannelType channel, double value, ds::DateTime now) const;
  virtual bool setSceneLocalPrio(const Device& device, int sceneId) const;
  virtual bool callSceneMin(const Device& device, int sceneId, ds::DateTime now) const;
  virtual void setOperationLock(const Device& device, bool lock) const;
  virtual bool stopDimming(const Device& device, int sceneId, ds::DateTime now) const;
  virtual bool completeMoving(const Device& device, ds::DateTime now) const;

  virtual DeviceState getState(const Device& device) const;
  virtual void publishToPropertyTree(const Device& device) const;
  virtual DeviceCacheState cacheStatus(const Device& device) const;
  virtual std::vector<double> getDeviceThresholds(const Device& device) const;

protected:
  virtual ChannelType getDefaultChannel() const = 0;
  void initializeStatus(const Device& device) const;
  bool updateCurrentValues(const Device& device, const ds::DateTime& now, DeviceInternalState& state) const;
  bool isDimmingUp(int sceneId, bool forceLocalPrio, DeviceCache& cache) const;
  bool isDimmingDown(int sceneId, bool forceLocalPrio, DeviceCache& cache) const;
  bool isStop(int sceneId, bool forceLocalPrio, DeviceCache& cache) const;
  virtual void updateSwitchStatus(const Device& device, DeviceInternalState& state) const;
  virtual const std::vector<int>& getApplicableScenes() const { return m_applicableScenes; }
  virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const {
    return std::vector<ChannelType>{ChannelType::none};
  }
  virtual const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const { return {}; }
  virtual int getSceneSize() const { return 1; }
  void populateCache(const Device& device, DeviceCache& cache) const;
  virtual std::map<DeviceProperty, std::tuple<int, int, int>> getDeviceParameters(const Device& device) const;
  virtual std::tuple<int, int, int> getDeviceParameter(const Device& device, DeviceProperty property) const;
  double convertThresholdToPercentage(const double& value) const;

  CacheDb<DeviceCache>& m_sceneCacheDb;
  CacheDb<DeviceInternalState>& m_StateCache;
  std::vector<int> m_applicableScenes;
  mutable std::map<DeviceProperty, std::tuple<int, int, int>> m_deviceParameters;
};

class HalLightDevice : public HalDevice {
public:
  using HalDevice::setOutputValue;
  using HalDevice::write;
  HalLightDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
      : HalDevice(sceneCacheDb, stateCache) {
    boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneDimArea));
    boost::push_back(m_applicableScenes, boost::irange(SceneMin, SceneStop));
    boost::push_back(m_applicableScenes, boost::irange(Scene2, SceneDecA1));
    boost::push_back(m_applicableScenes, boost::irange(SceneLocalOff, SceneStopA1));
    boost::push_back(m_applicableScenes, boost::irange(SceneAutoStandBy, SceneApartmentUndefined));

    m_deviceParameters = {{DeviceProperty::outputMode, {3, 0x00, 1}}, {DeviceProperty::switchedThreshold, {3, 0x43, 1}},
                          {DeviceProperty::dimTime0Up, {3, 0x06, 1}}, {DeviceProperty::dimTime0Down, {3, 0x07, 1}},
                          {DeviceProperty::dimTime1Up, {3, 0x08, 1}}, {DeviceProperty::dimTime1Down, {3, 0x09, 1}},
                          {DeviceProperty::dimTime2Up, {3, 0x0A, 1}}, {DeviceProperty::dimTime2Down, {3, 0x0B, 1}},
                          {DeviceProperty::dimStepUp, {3, 0x0C, 1}},  {DeviceProperty::dimStepDown, {3, 0x0D, 1}},
                          {DeviceProperty::minDim, {3, 0x04, 1}}};
  }
  virtual void write(const Device& device, const std::string& parameter, double value) const override;
  virtual void write(const Device& device, const std::string& parameter, const std::string& value) const override;

  virtual double readDouble(const Device& device, const std::string& parameter) const override;
  virtual std::string readString(const Device& device, const std::string& parameter) const override;

  virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const override;

  virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const override;
  virtual bool setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const override;
  virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;

protected:
  std::map<DeviceProperty, std::tuple<int, int, int>> getDeviceParameters(const Device& device) const override;
  std::tuple<int, int, int> getDeviceParameter(const Device& device, DeviceProperty property) const override;
  const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;

private:
  void invalidateReleatedChannels(SceneEntry& sceneCache, const ChannelType channel) const;
  virtual ChannelType getDefaultChannel() const override { return ChannelType::brightness; }
  bool updateOutputValue(const Device& device, uint8_t value, ds::DateTime now) const;
  bool handleDimming(DeviceInternalState& state, dss::DeviceCache& cache, const ds::DateTime& now, const int dimmScene,
                     bool up) const;
};

class HalShadeDevice : public HalDevice {
public:
  using HalDevice::setOutputValue;
  using HalDevice::write;
  HalShadeDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
      : HalDevice(sceneCacheDb, stateCache) {
    boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneDimArea));
    boost::push_back(m_applicableScenes, boost::irange(SceneMin, SceneStop));
    boost::push_back(m_applicableScenes, boost::irange(Scene2, SceneDecA1));
    boost::push_back(m_applicableScenes, boost::irange(SceneLocalOff, SceneStopA1));
    boost::push_back(m_applicableScenes, boost::irange(SceneSunProtection, SceneZoneUndefined));
    boost::push_back(m_applicableScenes, boost::irange(SceneAutoStandBy, SceneApartmentUndefined));

    m_deviceParameters = {{DeviceProperty::dimStepNum, {3, 0x3A, 1}},
                          {DeviceProperty::dimStepLen, {3, 0x3B, 1}},
                          {DeviceProperty::dimStepPause, {3, 0x3C, 1}},
                          {DeviceProperty::posTimeUp, {3, 0x30, 2}},
                          {DeviceProperty::posTimeDown, {3, 0x32, 2}}};
  }
  virtual void write(const Device& device, const std::string& parameter, double value) const override;

  virtual double readDouble(const Device& device, const std::string& parameter) const override;
  double readDoubleRAM(const Device& device, int sceneId, ChannelType channelType) const;

  virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const override;

  virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const override;
  virtual bool setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const override;
  virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;
  virtual bool setOutputValue(const Device& device, ChannelType channel, double value, ds::DateTime now) const override;
  DeviceCacheState cacheStatus(const Device& device) const override;

private:
  virtual const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;
  void addPositionOutsideRegister(const Device& device, int scene, int value, bool firstRegister) const;
  virtual ChannelType getDefaultChannel() const override { return ChannelType::shadePositionOutside; }

  bool handleDimming(int sceneId, DeviceCache& cache, const ds::DateTime& now, DeviceInternalState& state,
                     DeviceInternalOutputState& outputState, bool up) const;
  double calculateRuntime(DeviceCache& cache, DeviceInternalState& state) const;
  int getSceneSize() const override { return 2; }
  DeviceCacheState cacheStatusFw387(const Device& device) const;
};

class HalJokerDevice : public HalLightDevice {
public:
  using HalDevice::setOutputValue;
  using HalDevice::write;
  HalJokerDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
      : HalLightDevice(sceneCacheDb, stateCache) {
    boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneBurglary));

    m_deviceParameters = {{DeviceProperty::outputMode, {3, 0x00, 1}}, {DeviceProperty::switchedThreshold, {3, 0x3C, 1}},
                          {DeviceProperty::dimTime0Up, {3, 0x06, 1}}, {DeviceProperty::dimTime0Down, {3, 0x07, 1}},
                          {DeviceProperty::dimTime1Up, {3, 0x08, 1}}, {DeviceProperty::dimTime1Down, {3, 0x09, 1}},
                          {DeviceProperty::dimTime2Up, {3, 0x0A, 1}}, {DeviceProperty::dimTime2Down, {3, 0x0B, 1}},
                          {DeviceProperty::dimStepUp, {3, 0x0C, 1}},  {DeviceProperty::dimStepDown, {3, 0x0D, 1}},
                          {DeviceProperty::minDim, {3, 0x04, 1}}};
  }
  virtual void write(const Device& device, const std::string& parameter, double value) const override;
  virtual double readDouble(const Device& device, const std::string& parameter) const override;
  virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const override;
  virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;
  virtual DeviceState getState(const Device& device) const override;
  virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const override;
  virtual std::vector<double> getDeviceThresholds(const Device& device) const override;

protected:
  std::map<DeviceProperty, std::tuple<int, int, int>> getDeviceParameters(const Device& device) const override;
  std::tuple<int, int, int> getDeviceParameter(const Device& device, DeviceProperty property) const override;
  void updateSwitchStatus(const Device& device, DeviceInternalState& state) const override;
  const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;

private:
  void updateSceneConfig(const Device& device, int scene, dss::ChannelType channel, int value,
                         bool firstRegister) const;
  virtual ChannelType getDefaultChannel() const override { return ChannelType::powerState; }

  bool isIgnoredInPowersave(int sceneId) const;
  bool isWipeStandbyScene(int sceneId) const;
  bool updateOutputValue(const Device& device, uint8_t value, ds::DateTime now) const;
};

class HalAudioDevice : public HalDevice {
public:
  HalAudioDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache, Apartment* apt)
      : HalDevice(sceneCacheDb, stateCache), m_apartment(apt) {}
  dsModel::AudioStatus getZoneState(int zoneId) const;
  std::vector<int> playerZoneScenarios(int zoneId) const;

private:
  virtual ChannelType getDefaultChannel() const override { return ChannelType::audioVolume; }
  dsModel::AudioStatus getPlayerStatus(PropertyNodePtr player) const;
  void mergeStatus(dsModel::AudioStatus& zoneStatus, dsModel::AudioStatus deviceStatus) const;

  Apartment* m_apartment;
};

class HalClimateDevice : public HalLightDevice {
public:
  using HalDevice::setOutputValue;
  using HalDevice::write;
  HalClimateDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
      : HalLightDevice(sceneCacheDb, stateCache) {
    boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneBurglary));

    m_deviceParameters = {{DeviceProperty::outputMode, {3, 0x00, 1}},
                          {DeviceProperty::switchedThreshold, {3, 0x43, 1}}};
  }
  virtual void write(const Device& device, const std::string& parameter, double value) const override;
  virtual double readDouble(const Device& device, const std::string& parameter) const override;
  virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;
  virtual DeviceState getState(const Device& device) const override;

protected:
  const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;

private:
  void updateSceneConfig(const Device& device, int scene, dss::ChannelType channel, int value,
                         bool firstRegister) const;
  virtual ChannelType getDefaultChannel() const override { return ChannelType::waterFlowRate; }
};

} // namespace dss::hal

