/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelconst.h"
#include "model/modelmaintenance-mock.h"
#include "propertysystem-mock.h"

static ds::LogChannel dsLogChannel("hardwareAbstraction", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][hardwareAbstraction]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid lightId{"0000000000000000000000000000000001"};
static const Dsuid dummyId{"0000000000000000000000000000000003"};

TEST_CASE("legacy apiRoute conversion", TAGS) {
  SECTION("DeviceProperty") {
    CHECK("/property/dimStepLen" == toPath({lightId, DeviceProperty::dimStepLen}));
    CHECK("/property/dimStepNum" == toPath({lightId, DeviceProperty::dimStepNum}));
    CHECK("/property/dimTime0Up" == toPath({lightId, DeviceProperty::dimTime0Up}));
    CHECK("/property/outputMode" == toPath({lightId, DeviceProperty::outputMode}));
    CHECK("/property/posTimeUp" == toPath({lightId, DeviceProperty::posTimeUp}));
    CHECK("/property/switchedThreshold" == toPath({lightId, DeviceProperty::switchedThreshold}));
  }

  SECTION("DeviceScene") {
    CHECK("/scene/5" == toPath(DeviceSceneId{lightId, Scene1}));
  }

  SECTION("DeviceSceneProperty") {
    CHECK("/scene/5/blink" == toPath({{lightId, Scene1}, SceneProperty::blink}));
    CHECK("/scene/5/dimTime" == toPath({{lightId, Scene1}, SceneProperty::dimTime}));
    CHECK("/scene/5/dontCare" == toPath({{lightId, Scene1}, SceneProperty::dontCare}));
    CHECK("/scene/5/ignoreLocalPrio" == toPath({{lightId, Scene1}, SceneProperty::ignoreLocalPrio}));
    CHECK("/scene/5/ledConfig" == toPath({{lightId, Scene1}, SceneProperty::ledConfig}));
    CHECK("/scene/5/specialHandling" == toPath({{lightId, Scene1}, SceneProperty::specialHandling}));
  }

  SECTION("DeviceSceneChannel") {
    CHECK("/scene/5/brightness" == toPath({{lightId, Scene1}, ChannelType::brightness}));
    CHECK("/scene/5/powerState" == toPath({{lightId, Scene1}, ChannelType::powerState}));
    CHECK("/scene/5/shadeOpeningAngleOutside" == toPath({{lightId, Scene1}, ChannelType::shadeOpeningAngleOutside}));
    CHECK("/scene/5/shadePositionOutside" == toPath({{lightId, Scene1}, ChannelType::shadePositionOutside}));
    CHECK("/scene/5/waterFlowRate" == toPath({{lightId, Scene1}, ChannelType::waterFlowRate}));
  }

  SECTION("DeviceSceneChannelProperty") {
    CHECK("/scene/5/brightness/dontCare" ==
          toPath({{{.dsuid = lightId, .sceneId = 5}, ChannelType::brightness}, ChannelProperty::dontCare}));
  }
}

SCENARIO("Scene table access through HAL", TAGS) {
  PropertySystemMock propertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance(propertySystem, eventSink);

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  auto light = createDevice(apt, {.hardwareType = HardwareType::GE_TKM210});
  auto shade = createDevice(apt, {.hardwareType = HardwareType::GR_KL200});
  auto jalusie = createDevice(apt, {.hardwareType = HardwareType::GR_KL220});
  auto joker = createDevice(apt, {.hardwareType = HardwareType::SW_KL200});
  auto lightNonCacheable = createDevice(apt, {.hardwareType = HardwareType::GE_TKM220});
  auto BLKM300 = createDevice(apt, {.hardwareType = HardwareType::BL_KM300});
  auto umr200TerminalBlocks = createSWUMR200(apt, {.hardwareType = HardwareType::SW_UMR200});
  auto& jokerUMR = umr200TerminalBlocks[2]; // 3rd/4th terminal blocks have outputs
  auto& jokerUMR2 = umr200TerminalBlocks[3]; // 4th terminal block is secondary output

  auto dummyDevice = apt.allocateDevice(dummyId);

  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  WHEN("A light device hat no outputs") {
    THEN("It is not cacheable") { REQUIRE(hal.deviceCacheable(lightNonCacheable->getDSID()) == false); }
  }

  WHEN("a not implemented device") {
    THEN("hal access fails") {
      REQUIRE_THROWS(hal.write(dummyId, "/scene/5/brightness", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "/scene/5/dontCare", true));
      REQUIRE_THROWS(hal.write(dummyId, "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "/scene/5/brightness"), 1) == 62.0);
      REQUIRE_THROWS(hal.readBool({.scene = {.dsuid = dummyId, .sceneId = 5}, .property = SceneProperty::dontCare}) ==
                     false);
      REQUIRE_THROWS(hal.readString(dummyId, "/scene/5/dimTime") == "dimTime1");
    }
  }

  WHEN("not implemented paths") {
    THEN("hal access fails") {
      REQUIRE_THROWS(hal.write(dummyId, "/xxx", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "/", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "//", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "/xxx", true));
      REQUIRE_THROWS(hal.write(dummyId, "", true));
      REQUIRE_THROWS(hal.write(dummyId, "/", true));
      REQUIRE_THROWS(hal.write(dummyId, "//", true));
      REQUIRE_THROWS(hal.write(dummyId, "/xxx", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(dummyId, "", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(dummyId, "/", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(dummyId, "//", std::string("ajhakjdha")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "/xxx"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, ""), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "/"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "//"), 1) == 62.0);
      REQUIRE_THROWS(hal.readString(dummyId, "/xxx") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(dummyId, "") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(dummyId, "/") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(dummyId, "//") == "ajhasdjka");

      REQUIRE_THROWS(hal.write(light->getDSID(), "/xxx", 50.0));
      REQUIRE_THROWS(hal.write(light->getDSID(), "", 50.0));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/", 50.0));
      REQUIRE_THROWS(hal.write(light->getDSID(), "//", 50.0));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/xxx", true));
      REQUIRE_THROWS(hal.write(light->getDSID(), "", true));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/", true));
      REQUIRE_THROWS(hal.write(light->getDSID(), "//", true));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/xxx", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(light->getDSID(), "", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(light->getDSID(), "//", std::string("ajhakjdha")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(light->getDSID(), "/xxx"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(light->getDSID(), ""), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(light->getDSID(), "/"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(light->getDSID(), "//"), 1) == 62.0);
      REQUIRE_THROWS(hal.readString(light->getDSID(), "/xxx") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(light->getDSID(), "") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(light->getDSID(), "/") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(light->getDSID(), "//") == "ajhasdjka");

      REQUIRE_THROWS(hal.write(shade->getDSID(), "/xxx", 50.0));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "", 50.0));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/", 50.0));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "//", 50.0));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/xxx", true));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "", true));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/", true));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "//", true));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/xxx", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "//", std::string("ajhakjdha")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(shade->getDSID(), "/xxx"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(shade->getDSID(), ""), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(shade->getDSID(), "/"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(shade->getDSID(), "//"), 1) == 62.0);
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "/xxx") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "/") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "//") == "ajhasdjka");
    }
  }

  WHEN("writing a value") {
    THEN("lastChanged is updated") {
      ALLOW_CALL(busInterface.device, setDeviceConfig(_, _, _, _));
      ALLOW_CALL(busInterface.device, getDeviceConfig(_, _, _)).RETURN(0xAF);

      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/brightness", 50.0));
      auto devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/dontCare", true));
      devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/ledConfig", std::string("ledCon1")));
      devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside", 50.0));
      devCache = hal.getCache().tryLookup(shade->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside", 34.0));
      devCache = hal.getCache().tryLookup(jalusie->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(joker->getDSID(), "/scene/5/powerState", 2.0));
      devCache = hal.getCache().tryLookup(joker->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(BLKM300->getDSID(), "/scene/5/waterFlowRate", 25.0));
      devCache = hal.getCache().tryLookup(BLKM300->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);
    }
  }

  WHEN("channel dontCare of terminal blocks") {
    THEN("writing succeeds, but writes 'false'") {
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/brightness/dontCare", true));
      auto devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::brightness].dontCare == false);
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside/dontCare", true));
      devCache = hal.getCache().tryLookup(shade->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].dontCare == false);
      REQUIRE_NOTHROW(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside/dontCare", true));
      devCache = hal.getCache().tryLookup(jalusie->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::shadeOpeningAngleOutside].dontCare == false);
      REQUIRE_NOTHROW(hal.write(joker->getDSID(), "/scene/5/powerState/dontCare", true));
      devCache = hal.getCache().tryLookup(joker->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::powerState].dontCare == false);
      REQUIRE_NOTHROW(hal.write(BLKM300->getDSID(), "/scene/5/waterFlowRate/dontCare", true));
      devCache = hal.getCache().tryLookup(BLKM300->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::waterFlowRate].dontCare == false);
    }
    THEN("reading succeeds, but reads 'false'") {
      REQUIRE_NOTHROW(hal.readBool({{{.dsuid = light->getDSID(), .sceneId = 5}, ChannelType::brightness},
                                    ChannelProperty::dontCare}) == false);
      REQUIRE_NOTHROW(hal.readBool({{{.dsuid = shade->getDSID(), .sceneId = 5}, ChannelType::shadePositionOutside},
                                    ChannelProperty::dontCare}) == false);
      REQUIRE_NOTHROW(hal.readBool({{{.dsuid = jalusie->getDSID(), .sceneId = 5}, ChannelType::shadeOpeningAngleOutside},
                                    ChannelProperty::dontCare}) == false);
      REQUIRE_NOTHROW(hal.readBool({{{.dsuid = joker->getDSID(), .sceneId = 5}, ChannelType::powerState},
                                    ChannelProperty::dontCare}) == false);
      REQUIRE_NOTHROW(hal.readBool({{{.dsuid = BLKM300->getDSID(), .sceneId = 5}, ChannelType::waterFlowRate},
                                    ChannelProperty::dontCare}) == false);
    }
  }

  WHEN("Writing a brightness value") {
    THEN("it writes to correct bank/offset") {
      // "/scene/5/brightness" -> 128, 5
      // "/scene/6/dontcare"   -> 128, 134 (=128+6)
      // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 128));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/brightness", 50.0));

      auto devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::brightness].value == 50.0);
    }
  }
  WHEN("Writing extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/brightness", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 255));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/brightness", 100));
    }
  }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/brightness", 100.1)); }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/brightness", -1)); }
  WHEN("Writing to an unavailable channel fails") {
    hal.getCache().remove(light->getDSID());
    REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/unavailable", true));
    REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/unavailable", std::string("jasdfhasdjklfh")));
    hal.getCache().remove(shade->getDSID());
    REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/unavailable", true));
    REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/unavailable", std::string("jasdfhasdjklfh")));
    REQUIRE_THROWS(hal.write(joker->getDSID(), "/scene/5/brightness", 50.0));
    REQUIRE_THROWS(hal.write(BLKM300->getDSID(), "/scene/5/brightness", 50.0));
  }
  WHEN("Writing to an unavailable property fails") {
    hal.getCache().remove(light->getDSID());
    REQUIRE_THROWS(hal.write(light->getDSID(), "/property/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(light->getDSID(), "/property/unavailable", true));
    REQUIRE_THROWS(hal.write(light->getDSID(), "/property/unavailable", std::string("jasdfhasdjklfh")));
    hal.getCache().remove(shade->getDSID());
    REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/unavailable", true));
    REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/unavailable", std::string("jasdfhasdjklfh")));
  }

  WHEN("Writing flags") {
    THEN("that are not cached yet requires read-modify-write bool") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("that are not cached yet requires read-modify-write string") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/dimTime", std::string("dimTime1")));

      auto devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }
  WHEN("Writing flags with cache to a light device") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(light->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("flags need to be boolean") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/dontCare", 25.0)); }
    THEN("Write ignoreLocalPrio") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x52));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/ignoreLocalPrio", true));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.ignoreLocalPrio == true);
    }
    THEN("Write specialHandling") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x54));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/specialHandling", true));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.specialHandling == true);
    }
    THEN("Write blink") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x58));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/blink", true));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.blink == true);
    }
    THEN("Write blink as number fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/blink", 3.5)); }
    THEN("Write ledConfig") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x60));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/ledConfig", std::string("ledCon1")));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon1);
    }
    THEN("Write ledConfig as bool fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/ledConfig", true)); }
    THEN("Write ledConfig out of range") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/ledConfig", std::string("ledCon3")));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/ledConfig", std::string("1")));
    }
    THEN("Write dimTime") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
    THEN("Write dimTime as bool fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/dimTime", true)); }
    THEN("Write dimTime out of range") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/dimTime", std::string("dimTime3")));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/5/dimTime", std::string("1")));
    }
  }
  WHEN("Writing flags with cache to shade device") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool({{.dsuid = shade->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(shade->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("Write dimTime") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }

  WHEN("Writing flags with cache to joker device") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool({{.dsuid = joker->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(joker->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(joker->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("Write dimTime") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(joker->getDSID(), "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }

  WHEN("Writing flags with cache to climate device") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool({{.dsuid = BLKM300->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(BLKM300->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(BLKM300->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
  }

  WHEN("reading brightness channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(158);
      REQUIRE(roundDigits(hal.readDouble(light->getDSID(), "/scene/5/brightness"), 1) == 62.0);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].channels[ChannelType::brightness].value, 1) ==
              62.0);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0);
      REQUIRE(hal.readDouble(light->getDSID(), "/scene/5/brightness") == 0);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].channels[ChannelType::brightness].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(255);
      REQUIRE(hal.readDouble(light->getDSID(), "/scene/5/brightness") == 100);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].channels[ChannelType::brightness].value == 100);
    }
  }
  WHEN("reading powerState channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(158);
      REQUIRE(roundDigits(hal.readDouble(joker->getDSID(), "/scene/5/powerState"), 1) == 1.2);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[5].channels[ChannelType::powerState].value, 1) ==
              1.2);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0);
      REQUIRE(hal.readDouble(joker->getDSID(), "/scene/5/powerState") == 0);
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[5].channels[ChannelType::powerState].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(255);
      REQUIRE(hal.readDouble(joker->getDSID(), "/scene/5/powerState") == 2.0);
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[5].channels[ChannelType::powerState].value == 2.0);
    }
  }
  WHEN("reading waterFlowRate channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(158);
      REQUIRE(roundDigits(hal.readDouble(BLKM300->getDSID(), "/scene/5/waterFlowRate"), 1) == 62);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(BLKM300->getDSID()).scene.sceneTable[5].channels[ChannelType::waterFlowRate].value, 1) ==
              62);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0);
      REQUIRE(hal.readDouble(BLKM300->getDSID(), "/scene/5/waterFlowRate") == 0);
      REQUIRE(*hal.getCache().tryLookup(BLKM300->getDSID()).scene.sceneTable[5].channels[ChannelType::waterFlowRate].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(255);
      REQUIRE(hal.readDouble(BLKM300->getDSID(), "/scene/5/waterFlowRate") == 100.0);
      REQUIRE(*hal.getCache().tryLookup(BLKM300->getDSID()).scene.sceneTable[5].channels[ChannelType::waterFlowRate].value == 100.0);
    }
  }
  WHEN("reading from an unavailable channel fails") {
    REQUIRE_THROWS(hal.readDouble(light->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(light->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(shade->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(shade->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(joker->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(joker->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(joker->getDSID(), "/scene/5/brightness"));
    REQUIRE_THROWS(hal.readDouble(BLKM300->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(BLKM300->getDSID(), "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(BLKM300->getDSID(), "/scene/5/brightness"));
  }
  WHEN("reading from an unavailable property fails") {
    REQUIRE_THROWS(hal.readDouble(light->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(light->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readDouble(shade->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(shade->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readDouble(joker->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(joker->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readDouble(BLKM300->getDSID(), "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(BLKM300->getDSID(), "/property/unavailable"));
  }
  WHEN("reading from an unavailable scene fails") {
    REQUIRE_THROWS(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 128}, SceneProperty::dontCare}));
    REQUIRE_THROWS(hal.readDouble(light->getDSID(), "/scene/128/brightness"));
  }

  WHEN("Reading flags") {
    THEN("can be read directly") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.dontCare == false);
    }
    THEN("flags need to be boolean") { REQUIRE_THROWS(hal.readDouble(light->getDSID(), "/scene/5/dontCare") == 25.0); }
    THEN("Read ignoreLocalPrio") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x52);
      REQUIRE(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::ignoreLocalPrio}) == true);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.ignoreLocalPrio == true);
    }
    THEN("Read specialHandling") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x56);
      REQUIRE(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::specialHandling}) == true);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.specialHandling == true);
    }
    THEN("Read blink") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x5E);
      REQUIRE(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::blink}) == true);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.blink == true);
    }
    THEN("Read blink as number fails") { REQUIRE_THROWS(hal.readDouble(light->getDSID(), "/scene/5/blink") == 3.5); }
    THEN("Read ledConfig") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x6E);
      REQUIRE(hal.readString(light->getDSID(), "/scene/5/ledConfig") == "ledCon1");
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon1);
    }
    THEN("Read ledConfig as bool fails") {
      REQUIRE_THROWS(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::ledConfig}) == true);
    }
    THEN("Read dimTime") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0xAE);
      REQUIRE(hal.readString(light->getDSID(), "/scene/5/dimTime") == "dimTime1");
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
    THEN("Read dimTime as bool fails") {
      REQUIRE_THROWS(hal.readBool({{.dsuid = light->getDSID(), .sceneId = 5}, SceneProperty::dimTime}) == true);
    }
  }
  WHEN("Reading flags from shade device") {
    THEN("can be read directly") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE(hal.readBool({{.dsuid = shade->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[5].flags.dontCare == false);
    }
    THEN("Read dimTime") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0xAE);
      REQUIRE(hal.readString(shade->getDSID(), "/scene/5/dimTime") == "dimTime1");
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }

  WHEN("Writing a shadePositionOutside value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x02);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0x94));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0xF2));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside", 58.2));

      REQUIRE(
          *hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value ==
          58.2);

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0xF8);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 4, 0x26));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0xF2));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/4/shadePositionOutside", 14.9));

      REQUIRE(
          *hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[4].channels[ChannelType::shadePositionOutside].value ==
          14.9);
    }
  }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside", 100.1)); }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside", -1)); }
  WHEN("Writing extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0x05));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside", 0));

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 4, 0));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0x40));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/4/shadePositionOutside", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0xFF));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0xF5));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/5/shadePositionOutside", 100));

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 4, 0xFF));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0x4F));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/scene/4/shadePositionOutside", 100));
    }
  }

  WHEN("reading shadePositionOutside channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5))
          .RETURN(0xE4)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 128, 5, 0xE4));
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2))
          .RETURN(0x08)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 129, 2, 0x08));
      REQUIRE(roundDigits(hal.readDouble(shade->getDSID(), "/scene/5/shadePositionOutside"), 1) == 89.1);
      REQUIRE(
          roundDigits(
              *hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value,
              1) == 89.1);
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 4))
          .RETURN(0x0E)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 128, 4, 0x0E));
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2))
          .RETURN(0x5D)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 129, 2, 0x5D));
      REQUIRE(roundDigits(hal.readDouble(shade->getDSID(), "/scene/4/shadePositionOutside"), 1) == 5.8);
      REQUIRE(
          roundDigits(
              *hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[4].channels[ChannelType::shadePositionOutside].value,
              1) == 5.8);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5))
          .RETURN(0)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 128, 5, 0));
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2))
          .RETURN(0x03)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 129, 2, 0x03));
      REQUIRE(hal.readDouble(shade->getDSID(), "/scene/5/shadePositionOutside") == 0);
      REQUIRE(
          *hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value ==
          0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 4))
          .RETURN(255)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 128, 4, 255));
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2))
          .RETURN(0xEF)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(shade->getDSID(), 129, 2, 0xEF));
      REQUIRE(hal.readDouble(shade->getDSID(), "/scene/4/shadePositionOutside") == 100);
      REQUIRE(
          *hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[4].channels[ChannelType::shadePositionOutside].value ==
          100);
    }
  }

  WHEN("Writing a shadeOpeningAngleOutside value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 130, 5, 0x79));
      REQUIRE_NOTHROW(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside", 47.5));

      REQUIRE(*hal.getCache()
                   .tryLookup(jalusie->getDSID())
                   .scene.sceneTable[5]
                   .channels[ChannelType::shadeOpeningAngleOutside]
                   .value == 47.5);
    }
    THEN("a unsupporting device fails") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/shadeOpeningAngleOutside", 47.5));
    }
  }
  WHEN("Writing a illegal value fails") {
    REQUIRE_THROWS(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside", 100.1));
  }
  WHEN("Writing a illegal value fails") {
    REQUIRE_THROWS(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside", -1));
  }
  WHEN("Writing extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 130, 5, 0x00));
      REQUIRE_NOTHROW(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 130, 5, 0xFF));
      REQUIRE_NOTHROW(hal.write(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside", 100));
    }
  }

  WHEN("reading shadeOpeningAngleOutside channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 130, 5))
          .RETURN(0x8F)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(jalusie->getDSID(), 130, 5, 0X8F));
      REQUIRE(roundDigits(hal.readDouble(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside"), 1) == 56.1);
      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusie->getDSID())
                               .scene.sceneTable[5]
                               .channels[ChannelType::shadeOpeningAngleOutside]
                               .value,
                          1) == 56.1);
    }
    THEN("a unsupporting device fails") {
      REQUIRE_THROWS(hal.readDouble(shade->getDSID(), "/scene/5/shadeOpeningAngleOutside") == 56.2);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 130, 5))
          .RETURN(0)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(jalusie->getDSID(), 130, 5, 0));
      REQUIRE(hal.readDouble(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside") == 0);
      REQUIRE(*hal.getCache()
                   .tryLookup(jalusie->getDSID())
                   .scene.sceneTable[5]
                   .channels[ChannelType::shadeOpeningAngleOutside]
                   .value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 130, 5))
          .RETURN(255)
          .LR_SIDE_EFFECT(hal.updateDeviceConfig(jalusie->getDSID(), 130, 5, 255));
      REQUIRE(hal.readDouble(jalusie->getDSID(), "/scene/5/shadeOpeningAngleOutside") == 100);
      REQUIRE(*hal.getCache()
                   .tryLookup(jalusie->getDSID())
                   .scene.sceneTable[5]
                   .channels[ChannelType::shadeOpeningAngleOutside]
                   .value == 100);
    }
  }
  WHEN("Writing angle to a device that does not support angles") {
    THEN("fails") { REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/5/shadeOpeningAngleOutside", 15.6)); }
  }

  WHEN("Updating cache via setDeviceConfig on bank 128 trapped by HAL") {
    THEN("should work for light device scene value") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(light->getDSID(), 128, 4, 10));
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[4].channels[ChannelType::brightness].value, 1) ==
              3.9);
    }
    THEN("should work for light device scene configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(light->getDSID(), 128, 128 + 4, 1));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[4].flags.dontCare == true);
    }
  }

  WHEN("Updating cache via device config trap on bank 128 for light devices") {
    THEN("should work for light device scene settings configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(light->getDSID(), 128, 4, 10));
      REQUIRE(hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[4].channels[ChannelType::brightness].value);

      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(light->getDSID()).scene.sceneTable[4].channels[ChannelType::brightness].value, 1) ==
              3.9);
    }
  }

  WHEN("Updating cache via setDeviceConfig on bank 128 trapped by HAL for joker device") {
    THEN("should work for joker device scene value") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(joker->getDSID(), 128, 4, 10));
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[4].channels[ChannelType::powerState].value, 1) ==
                0.1);
    }
    THEN("should work for joker device scene configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(joker->getDSID(), 128, 128 + 4, 1));
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[4].flags.dontCare == true);
    }
  }

  WHEN("Updating cache via device config trap on bank 128 for joker devices") {
    THEN("should work for light device scene settings configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(joker->getDSID(), 128, 4, 10));
      REQUIRE(hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[4].channels[ChannelType::powerState].value);

      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(joker->getDSID()).scene.sceneTable[4].channels[ChannelType::powerState].value, 1) ==
              0.1);
    }
  }


  WHEN("Updating cache via device config trap on bank 128/129 for shade devices") {
    THEN("should work only if the whole 16bit are known") {
      /* check "even" scene */
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), CfgClassScene, 56, 145));
      REQUIRE(
          !hal.getCache().tryLookup(jalusie->getDSID()).scene.sceneTable[56].channels[ChannelType::shadePositionOutside].value);

      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), CfgClassSceneExtention, 28, 14));
      REQUIRE(
          hal.getCache().tryLookup(jalusie->getDSID()).scene.sceneTable[56].channels[ChannelType::shadePositionOutside].value);
      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusie->getDSID())
                               .scene.sceneTable[56]
                               .channels[ChannelType::shadePositionOutside]
                               .value,
                          1) == 57.0);

      /* check "odd" scene, 2nd register already known */
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), CfgClassScene, 57, 145));
      REQUIRE(
          hal.getCache().tryLookup(jalusie->getDSID()).scene.sceneTable[57].channels[ChannelType::shadePositionOutside].value);
      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusie->getDSID())
                               .scene.sceneTable[57]
                               .channels[ChannelType::shadePositionOutside]
                               .value,
                          1) == 56.6);
    }
    THEN("should work for shade device scene configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(shade->getDSID(), 128, 128 + 4, 1));
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).scene.sceneTable[4].flags.dontCare == true);
    }
  }

  WHEN("Updating cache via device config trap on bank 130 for shade devices") {
    THEN("should work") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), CfgClassSceneAngle, 4, 128));
      REQUIRE(hal.getCache()
                  .tryLookup(jalusie->getDSID())
                  .scene.sceneTable[4]
                  .channels[ChannelType::shadeOpeningAngleOutside]
                  .value);

      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusie->getDSID())
                               .scene.sceneTable[4]
                               .channels[ChannelType::shadeOpeningAngleOutside]
                               .value,
                          1) == 50.2);

      std::optional<ds::DateTime> before = hal.getCache().tryLookup(jalusie->getDSID()).lastChanged;
      REQUIRE(before);
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), CfgClassSceneExtention, 64, 255));
      std::optional<ds::DateTime> after = hal.getCache().tryLookup(jalusie->getDSID()).lastChanged;
      REQUIRE(after);
      REQUIRE(*before == *after);
    }
  }

  WHEN("Updating cache via device config trap on unsupported bank") {
    REQUIRE_NOTHROW(hal.updateDeviceConfig(light->getDSID(), 128, 4, 128));
    REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), CfgClassSceneAngle, 4, 128));

    THEN("on light device should do nothing") {
      std::optional<ds::DateTime> before = hal.getCache().tryLookup(light->getDSID()).lastChanged;
      REQUIRE(before);
      REQUIRE_NOTHROW(hal.updateDeviceConfig(light->getDSID(), 127, 64, 255));
      std::optional<ds::DateTime> after = hal.getCache().tryLookup(light->getDSID()).lastChanged;
      REQUIRE(after);
      REQUIRE(*before == *after);
    }
    THEN("on shade device should do nothing") {
      std::optional<ds::DateTime> before = hal.getCache().tryLookup(jalusie->getDSID()).lastChanged;
      REQUIRE(before);
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusie->getDSID(), 127, 64, 255));
      std::optional<ds::DateTime> after = hal.getCache().tryLookup(jalusie->getDSID()).lastChanged;
      REQUIRE(after);
      REQUIRE(*before == *after);
    }
  }

  WHEN("Updating cache via device config update properties for light") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::outputMode, 3, 0x00},
                                        {DeviceProperty::switchedThreshold, 3, 0x43},
                                        {DeviceProperty::dimTime0Up, 3, 0x06},
                                        {DeviceProperty::dimTime0Down, 3, 0x07},
                                        {DeviceProperty::dimTime1Up, 3, 0x08},
                                        {DeviceProperty::dimTime1Down, 3, 0x09},
                                        {DeviceProperty::dimTime2Up, 3, 0x0A},
                                        {DeviceProperty::dimTime2Down, 3, 0x0B},
                                        {DeviceProperty::dimStepUp, 3, 0x0C},
                                        {DeviceProperty::dimStepDown, 3, 0x0D}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(light->getDSID(), p.bank, p.offset, 128));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(light->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[p.prop] == 128);
    }
  }

  WHEN("Updating cache via device config update simple properties for shade") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::dimStepNum, 3, 0x3A},
                                        {DeviceProperty::dimStepLen, 3, 0x3B},
                                        {DeviceProperty::dimStepPause, 3, 0x3C}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(shade->getDSID(), p.bank, p.offset, 128));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(shade->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[p.prop] == 128);
    }
  }

  WHEN("Updating cache via device config update properties for joker") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::outputMode, 3, 0x00},
                                        {DeviceProperty::switchedThreshold, 3, 0x3C},
                                        {DeviceProperty::dimTime0Up, 3, 0x06},
                                        {DeviceProperty::dimTime0Down, 3, 0x07},
                                        {DeviceProperty::dimTime1Up, 3, 0x08},
                                        {DeviceProperty::dimTime1Down, 3, 0x09},
                                        {DeviceProperty::dimTime2Up, 3, 0x0A},
                                        {DeviceProperty::dimTime2Down, 3, 0x0B},
                                        {DeviceProperty::dimStepUp, 3, 0x0C},
                                        {DeviceProperty::dimStepDown, 3, 0x0D}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(joker->getDSID(), p.bank, p.offset, 128));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(joker->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).property[p.prop] == 128);
    }
  }

  WHEN("Updating cache via device config update properties for BL-KM300") {
      struct Property {
        DeviceProperty prop;
        int bank;
        int offset;
      };
      auto p = GENERATE(values<Property>({{DeviceProperty::switchedThreshold, 3, 0x43},
                                          {DeviceProperty::outputMode, 3, 0x0}}));

      REQUIRE_NOTHROW(hal.updateDeviceConfig(BLKM300->getDSID(), p.bank, p.offset, 152));

      THEN("Cache should be updated") {
        REQUIRE(hal.getCache().tryLookup(BLKM300->getDSID()).property[p.prop] == 152);
      }
    }


  WHEN("Updating cache via device config update properties for joker UMR primary output") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::outputMode, 3, 0x00},
                                        {DeviceProperty::switchedThreshold, 3, 0x30},
                                        {DeviceProperty::switchedThresholdTwoStep1, 3, 0x33},
                                        {DeviceProperty::switchedThreshold3Step0, 3, 0x34},
                                        {DeviceProperty::dimTime0Up, 3, 0x06},
                                        {DeviceProperty::dimTime0Down, 3, 0x07},
                                        {DeviceProperty::dimTime1Up, 3, 0x08},
                                        {DeviceProperty::dimTime1Down, 3, 0x09},
                                        {DeviceProperty::dimTime2Up, 3, 0x0A},
                                        {DeviceProperty::dimTime2Down, 3, 0x0B},
                                        {DeviceProperty::dimStepUp, 3, 0x0C},
                                        {DeviceProperty::dimStepDown, 3, 0x0D}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerUMR->getDSID(), p.bank, p.offset, 152));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(jokerUMR->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(jokerUMR->getDSID()).property[p.prop] == 152);
    }
  }

  WHEN("Updating cache via device config update properties for joker UMR secondary output") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::switchedThresholdTwoStep1, 3, 0x33},
                                        {DeviceProperty::switchedThreshold3Step0, 3, 0x34}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerUMR2->getDSID(), p.bank, p.offset, 152));

    THEN("Cache should not be updated") {
      REQUIRE(hal.getCache().tryLookup(jokerUMR2->getDSID()).property[p.prop] == std::nullopt);
    }
  }

  WHEN("Updating cache via device config update 2-byte properties for shade") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset1;
      int offset2;
    };
    auto p = GENERATE(
        values<Property>({{DeviceProperty::posTimeUp, 3, 0x30, 0x31}, {DeviceProperty::posTimeDown, 3, 0x32, 0x33}}));

    THEN("Cache should update") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(shade->getDSID(), p.bank, p.offset1, 128));
      REQUIRE(hal.getCache().tryLookup(shade->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[p.prop] == 128);

      REQUIRE_NOTHROW(hal.updateDeviceConfig(shade->getDSID(), p.bank, p.offset2, 128));
      REQUIRE(hal.getCache().tryLookup(shade->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[p.prop] == ((128 << 8) | 128));
    }

    THEN("Cache should update") {
      {
        auto cache = hal.getCache().tryLookup(shade->getDSID());
        cache.property[p.prop] = 0xfeed;
        hal.getCache().update(shade->getDSID(), cache);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(shade->getDSID(), p.bank, p.offset1, 0xad));
      REQUIRE(hal.getCache().tryLookup(shade->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[p.prop] == 0xfead);

      REQUIRE_NOTHROW(hal.updateDeviceConfig(shade->getDSID(), p.bank, p.offset2, 0x55));
      REQUIRE(hal.getCache().tryLookup(shade->getDSID()).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[p.prop] == ((0x55 << 8) | 0xad));
    }
  }

  WHEN("Writing outputMode") {
    THEN("Write outputMode") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0, 16));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/property/outputMode", std::string("KM_Switch")));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::outputMode] ==
              static_cast<int>(OutputMode::KM_Switch));
    }
    THEN("Write outputMode as bool fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/outputMode", true)); }
    THEN("Write outputMode out of range") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/outputMode", std::string("KM_Switch45")));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/outputMode", std::string("16")));
    }
  }

  WHEN("Writing dimTime") {
    THEN("Write dimTime exact") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 6, 0x26));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/property/dimTime0Up", 0.2875));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::dimTime0Up] == 0x26);
    }
    THEN("Write dimTime approx") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 6, 0xE3));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/property/dimTime0Up", 1000));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::dimTime0Up] == 0xE3);
    }
    THEN("Write outpdimTimeutMode as bool fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/property/dimTime0Up", true)); }
    THEN("Write dimTime as string fails") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/dimTime0Up", std::string("ajfhadl")));
    }
    THEN("Write dimTime out of range") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/outputMode", 0.052));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/outputMode", 3276.9));
    }
  }

  WHEN("Writing known property but wrong for this device") {
    THEN("fails") { REQUIRE_THROWS(hal.write(light->getDSID(), "/property/dimTime0Up", std::string("ajfhadl"))); }
  }

  WHEN("Writing switchedThreshold") {
    THEN("Write switchedThreshold") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x43, 190));
      REQUIRE_NOTHROW(hal.write(light->getDSID(), "/property/switchedThreshold", 74.5));
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Write switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/switchedThreshold", true));
    }
    THEN("Write switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/scene/switchedThreshold", std::string("KM_Switch45")));
    }
    THEN("Write switchedThreshold out of range") {
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/switchedThreshold", -.2));
      REQUIRE_THROWS(hal.write(light->getDSID(), "/property/switchedThreshold", 100.1));
    }
  }

  WHEN("Writing switchedThreshold for joker") {
    THEN("Write switchedThreshold") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x3C, 190));
      REQUIRE_NOTHROW(hal.write(joker->getDSID(), "/property/switchedThreshold", 74.5));
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Write switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.write(joker->getDSID(), "/scene/switchedThreshold", true));
    }
    THEN("Write switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.write(joker->getDSID(), "/scene/switchedThreshold", std::string("KM_Switch45")));
    }
    THEN("Write switchedThreshold out of range") {
      REQUIRE_THROWS(hal.write(joker->getDSID(), "/property/switchedThreshold", -.2));
      REQUIRE_THROWS(hal.write(joker->getDSID(), "/property/switchedThreshold", 100.1));
    }
  }

  WHEN("Writing switchedThreshold for climate") {
    THEN("Write switchedThreshold") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x43, 190));
      REQUIRE_NOTHROW(hal.write(BLKM300->getDSID(), "/property/switchedThreshold", 74.5));
      REQUIRE(*hal.getCache().tryLookup(BLKM300->getDSID()).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Write switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.write(BLKM300->getDSID(), "/scene/switchedThreshold", true));
    }
    THEN("Write switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.write(BLKM300->getDSID(), "/scene/switchedThreshold", std::string("KM_Switch45")));
    }
    THEN("Write switchedThreshold out of range") {
      REQUIRE_THROWS(hal.write(BLKM300->getDSID(), "/property/switchedThreshold", -.2));
      REQUIRE_THROWS(hal.write(BLKM300->getDSID(), "/property/switchedThreshold", 100.1));
    }
  }

  WHEN("Reading outputMode") {
    THEN("Read outputMode") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0)).RETURN(19);
      REQUIRE(hal.readString(light->getDSID(), "/property/outputMode") == "LeadingEdgeDimmer");
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::outputMode] ==
              static_cast<int>(OutputMode::LeadingEdgeDimmer));
    }
    THEN("Read outputMode as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = light->getDSID(), .property = DeviceProperty::outputMode}) == true);
    }
    THEN("Read outputMode as double fails") {
      REQUIRE_THROWS(hal.readDouble(light->getDSID(), "/property/outputMode") == 45.54);
    }
  }

  WHEN("Reading dimTime") {
    THEN("Read dimTime exact") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 6)).RETURN(0x26);
      REQUIRE_NOTHROW(hal.readDouble(light->getDSID(), "/property/dimTime0Up") == 0.2875);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::dimTime0Up] == 0x26);
    }
    THEN("Read outpdimTimeutMode as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = light->getDSID(), .property = DeviceProperty::dimTime0Up}) == true);
    }
    THEN("Read dimTime as string fails") {
      REQUIRE_THROWS(hal.readString(light->getDSID(), "/property/dimTime0Up") == std::string("ajfhadl"));
    }
  }

  WHEN("Reading switchedThreshold") {
    THEN("Read switchedThreshold") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0x43)).RETURN(190);
      REQUIRE_NOTHROW(hal.readDouble(light->getDSID(), "/property/switchedThreshold") == 74.5);
      REQUIRE(*hal.getCache().tryLookup(light->getDSID()).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Read switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = light->getDSID(), .property = DeviceProperty::switchedThreshold}) == true);
    }
    THEN("Read switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.readString(light->getDSID(), "/scene/switchedThreshold") == std::string("KM_Switch45"));
    }
  }
  WHEN("Reading switchedThreshold for joker devices") {
    THEN("Read switchedThreshold") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0x3C)).RETURN(190);
      REQUIRE_NOTHROW(hal.readDouble(joker->getDSID(), "/property/switchedThreshold") == 1.5);
      REQUIRE(*hal.getCache().tryLookup(joker->getDSID()).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Read switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = joker->getDSID(), .property = DeviceProperty::switchedThreshold}) == true);
    }
    THEN("Read switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.readString(joker->getDSID(), "/scene/switchedThreshold") == std::string("KM_Switch45"));
    }
  }
  WHEN("Reading switchedThreshold for climate devices") {
    THEN("Read switchedThreshold") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0x43)).RETURN(190);
      REQUIRE_NOTHROW(hal.readDouble(BLKM300->getDSID(), "/property/switchedThreshold") == 190);
      REQUIRE(*hal.getCache().tryLookup(BLKM300->getDSID()).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Read switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = BLKM300->getDSID(), .property = DeviceProperty::switchedThreshold}) == true);
    }
    THEN("Read switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.readString(BLKM300->getDSID(), "/scene/switchedThreshold") == std::string("KM_Switch45"));
    }
  }

  WHEN("Writing dimStepNum") {
    THEN("Write dimStepNum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x3A, 245));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/property/dimStepNum", 245));
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[DeviceProperty::dimStepNum] == 245);
    }
    THEN("Write dimStepNum as bool fails") { REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepNum", true)); }
    THEN("Write dimStepNum as string fails") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepNum", std::string("KM_Switch45")));
    }
    THEN("Write dimStepNum out of range") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepNum", -1));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepNum", 256));
    }
  }

  WHEN("Reading dimStepNum") {
    THEN("Read dimStepNum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfigWord(_, 3, 0x3A)).RETURN(0x6730);
      REQUIRE_NOTHROW(hal.readDouble(shade->getDSID(), "/property/dimStepNum") == 48);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[DeviceProperty::dimStepNum] == 48);
    }
    THEN("Read dimStepNum as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = shade->getDSID(), .property = DeviceProperty::dimStepNum}) == true);
    }
    THEN("Read dimStepNum as string fails") {
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "/property/dimStepNum") == std::string("KM_Switch45"));
    }
  }

  WHEN("Writing dimStepLen") {
    THEN("Write dimStepLen") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x3B, 178));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/property/dimStepLen", 1.78));
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[DeviceProperty::dimStepLen] == 178);
    }
    THEN("Write dimStepLen as bool fails") { REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepLen", true)); }
    THEN("Write dimStepLen as string fails") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepLen", std::string("KM_Switch45")));
    }
    THEN("Write dimStepLen out of range") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepLen", -1));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/dimStepLen", 256));
    }
  }

  WHEN("Reading dimStepLen") {
    THEN("Read dimStepLen") {
      REQUIRE_CALL(busInterface.device, getDeviceConfigWord(_, 3, 0x3A)).RETURN(0x6730);
      REQUIRE_NOTHROW(hal.readDouble(shade->getDSID(), "/property/dimStepLen") == 1.03);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[DeviceProperty::dimStepLen] == 103);
    }
    THEN("Read dimStepLen as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = shade->getDSID(), .property = DeviceProperty::dimStepLen}) == true);
    }
    THEN("Read dimStepLen as string fails") {
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "/property/dimStepLen") == std::string("KM_Switch45"));
    }
  }

  WHEN("Writing posTimeUp") {
    THEN("Write posTimeUp") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x30, 45));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x31, 201));
      REQUIRE_NOTHROW(hal.write(shade->getDSID(), "/property/posTimeUp", 515.01));
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[DeviceProperty::posTimeUp] == 51501);
    }
    THEN("Write posTimeUp as bool fails") { REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/posTimeUp", true)); }
    THEN("Write posTimeUp as string fails") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/scene/posTimeUp", std::string("KM_Switch45")));
    }
    THEN("Write posTimeUp out of range") {
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/posTimeUp", -1));
      REQUIRE_THROWS(hal.write(shade->getDSID(), "/property/posTimeUp", 65536));
    }
  }

  WHEN("Reading posTimeUp") {
    THEN("Read posTimeUp") {
      REQUIRE_CALL(busInterface.device, getDeviceConfigWord(_, 3, 0x30)).RETURN(0x4D4A);
      REQUIRE_NOTHROW(hal.readDouble(shade->getDSID(), "/property/posTimeUp") == 197.86);
      REQUIRE(*hal.getCache().tryLookup(shade->getDSID()).property[DeviceProperty::posTimeUp] == 19786);
    }
    THEN("Read posTimeUp as bool fails") {
      REQUIRE_THROWS(hal.readBool({.dsuid = shade->getDSID(), .property = DeviceProperty::posTimeUp}) == true);
    }
    THEN("Read posTimeUp as string fails") {
      REQUIRE_THROWS(hal.readString(shade->getDSID(), "/scene/posTimeUp") == std::string("KM_Switch45"));
    }
  }

  WHEN("Querying area membership") {
    THEN("the answer should be correct") {
      DeviceCache cache = hal.getCache().tryLookup(light->getDSID());
      /* device is only in area 1 */
      cache.scene.sceneTable[SceneA11].flags.dontCare = false;
      cache.scene.sceneTable[SceneA21].flags.dontCare = true;
      cache.scene.sceneTable[SceneA31].flags.dontCare = true;
      cache.scene.sceneTable[SceneA41].flags.dontCare = true;
      hal.getCache().update(light->getDSID(), cache);
      std::vector<int> areas = hal.areaMember(light->getDSID());
      REQUIRE(areas.size() == 1);
      REQUIRE(areas.at(0) == SceneA11);
    }
  }

  WHEN("Querying cache status") {
    auto status = hal.cacheStatus(light->getDSID());
    THEN("cache is empty") {
      REQUIRE(status.completed == 0);
      REQUIRE(status.completion == 0);
      REQUIRE(status.total == 149);
    }
  }

  WHEN("Quering cache status of partially populated device") {
    {
      DeviceCache cache = hal.getCache().tryLookup(light->getDSID());
      /* device is only in area 1 */
      cache.scene.sceneTable[SceneA11].flags.dontCare = false;
      cache.scene.sceneTable[SceneA11].channels[ChannelType::brightness].value = 50;
      cache.scene.sceneTable[SceneA21].flags.dontCare = true;
      cache.scene.sceneTable[SceneA21].channels[ChannelType::brightness].value = 50;
      cache.scene.sceneTable[SceneA31].flags.dontCare = true;
      cache.scene.sceneTable[SceneA31].channels[ChannelType::brightness].value = 50;
      cache.scene.sceneTable[SceneA41].flags.dontCare = true;
      cache.scene.sceneTable[SceneA41].channels[ChannelType::brightness].value = 50;
      hal.getCache().update(light->getDSID(), cache);
    }
    auto status = hal.cacheStatus(light->getDSID());
    THEN("cache is partially filled") {
      REQUIRE(status.completed == 8);
      REQUIRE(status.completion == (8.0 / 149.0) * 100);
      REQUIRE(status.total == 149);
    }
  }
}

SCENARIO("Performance testing for cache status", "[dss][hardwareAbstraction][!hide]") {
  PropertySystemMock propertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance(propertySystem, eventSink);

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

 
  auto light = createDevice(apt, {.hardwareType = HardwareType::GE_TKM210});

//  ds::sqlite3::Database d("./cache.db");
  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  WHEN("Querying cache status speed-test") {
    auto start = std::chrono::steady_clock::now();
    for (int i = 0; i < 1000; ++i) {
      auto status = hal.cacheStatus(light->getDSID());
      //      DS_DEBUG(status.completion);
    }
    auto end = std::chrono::steady_clock::now();
    REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count() == 10);
  }
}

SCENARIO("Sonos room test", TAGS) {
  PropertySystemMock propSystem;
  ALLOW_CALL(propSystem, getProperty("/system/dSUID")).RETURN(nullptr);
  auto pApartment = makeApartment(nullptr, &propSystem);
  auto& apt = *pApartment;
  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  WHEN("There is no sonos device in the zone") {
    REQUIRE_CALL(propSystem, getProperty("/scripts/sonos-controller/players")).RETURN(nullptr);
    dsModel::ZoneApplicationsStatus zoneState = hal.getZoneState(1);

    THEN("no state is reflected") {
      REQUIRE(zoneState.audio.status == dsModel::AudioStatusStatus::none);
    }
  }

  WHEN("There is one sonos player in the room and and status queried") {
    std::shared_ptr<PropertyNodeMock> nodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> playerNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> zoneIdNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> onlineNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateStateNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateMutedNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateVolumeNodePtr(new PropertyNodeMock());
    REQUIRE_CALL(propSystem, getProperty("/scripts/sonos-controller/players")).RETURN(nodePtr);
    REQUIRE_CALL(*(nodePtr.get()), getChildCount()).RETURN(1);
    REQUIRE_CALL(*(nodePtr.get()), getChild(0)).RETURN(playerNodePtr);
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("online")).RETURN(onlineNodePtr);
    REQUIRE_CALL(*(onlineNodePtr.get()), getBoolValue()).RETURN(true);
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("zoneID")).RETURN(zoneIdNodePtr);
    REQUIRE_CALL(*(zoneIdNodePtr.get()), getStringValue()).RETURN("10");
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("state")).RETURN(stateNodePtr);
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("state")).RETURN(stateStateNodePtr);
    REQUIRE_CALL(*(stateStateNodePtr.get()), getStringValue()).RETURN("playing");
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("muted")).RETURN(stateMutedNodePtr);
    REQUIRE_CALL(*(stateMutedNodePtr.get()), getBoolValue()).RETURN("true");
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("volume")).RETURN(stateVolumeNodePtr);
    REQUIRE_CALL(*(stateVolumeNodePtr.get()), getIntegerValue()).RETURN(33);

    dsModel::ZoneApplicationsStatus zoneState = hal.getZoneState(10);
    THEN("Room state must be device state") {
      REQUIRE(zoneState.audio.status == dsModel::AudioStatusStatus::playing);
      REQUIRE(zoneState.audio.muted == true);
      REQUIRE(zoneState.audio.volume == 33);
    }
  }

  WHEN("There are 2 sonoses in the room and the status is queried") {
    std::shared_ptr<PropertyNodeMock> nodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> playerNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> zoneIdNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> onlineNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateStateNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateMutedNodePtr(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateVolumeNodePtr(new PropertyNodeMock());
    REQUIRE_CALL(propSystem, getProperty("/scripts/sonos-controller/players")).RETURN(nodePtr);
    REQUIRE_CALL(*(nodePtr.get()), getChildCount()).RETURN(2);
    REQUIRE_CALL(*(nodePtr.get()), getChild(0)).RETURN(playerNodePtr);
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("online")).RETURN(onlineNodePtr);
    REQUIRE_CALL(*(onlineNodePtr.get()), getBoolValue()).RETURN(true);
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("zoneID")).RETURN(zoneIdNodePtr);
    REQUIRE_CALL(*(zoneIdNodePtr.get()), getStringValue()).RETURN("10");
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("state")).RETURN(stateNodePtr);
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("state")).RETURN(stateStateNodePtr);
    REQUIRE_CALL(*(stateStateNodePtr.get()), getStringValue()).RETURN("playing");
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("muted")).RETURN(stateMutedNodePtr);
    REQUIRE_CALL(*(stateMutedNodePtr.get()), getBoolValue()).RETURN("false");
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("volume")).RETURN(stateVolumeNodePtr);
    REQUIRE_CALL(*(stateVolumeNodePtr.get()), getIntegerValue()).RETURN(33);

    /*std::shared_ptr<PropertyNodeMock> playerNodePtr1(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> zoneIdNodePtr1(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> onlineNodePtr1(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateNodePtr1(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateStateNodePtr1(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateMutedNodePtr1(new PropertyNodeMock());
    std::shared_ptr<PropertyNodeMock> stateVolumeNodePtr1(new PropertyNodeMock());*/
    REQUIRE_CALL(*(nodePtr.get()), getChild(1)).RETURN(playerNodePtr);
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("online")).RETURN(onlineNodePtr);
    REQUIRE_CALL(*(onlineNodePtr.get()), getBoolValue()).RETURN(true);
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("zoneID")).RETURN(zoneIdNodePtr);
    REQUIRE_CALL(*(zoneIdNodePtr.get()), getStringValue()).RETURN("10");
    REQUIRE_CALL(*(playerNodePtr.get()), getPropertyByName("state")).RETURN(stateNodePtr);
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("state")).RETURN(nullptr);
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("muted")).RETURN(stateMutedNodePtr);
    REQUIRE_CALL(*(stateMutedNodePtr.get()), getBoolValue()).RETURN("true");
    REQUIRE_CALL(*(stateNodePtr.get()), getPropertyByName("volume")).RETURN(stateVolumeNodePtr);
    REQUIRE_CALL(*(stateVolumeNodePtr.get()), getIntegerValue()).RETURN(45);

    dsModel::ZoneApplicationsStatus zoneState = hal.getZoneState(10);

    THEN("Room state should reflect a combination of the 2 devices") {
      REQUIRE(zoneState.audio.status == dsModel::AudioStatusStatus::playing);
      REQUIRE(zoneState.audio.muted == true);
      REQUIRE(zoneState.audio.volume == 45);
    }
  }

}

} // namespace
