/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <iostream>
#include <ostream>
#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "cache/device-cache.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"
#include "propertysystem-mock.h"
#include "util.h"

static ds::LogChannel dsLogChannel("halLightStateMachine", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][halLightStateMachine]";

using namespace dss;
using namespace trompeloeil;
using Catch::Generators::values;
using trompeloeil::_;

namespace {

struct Area {
  int area;
  int inc;
  int dec;
};


SCENARIO("Standard Light device", TAGS) {
  PropertySystemMock propertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance(propertySystem, eventSink);

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  auto dev1 = createDevice(apt, {
                                    .hardwareType = HardwareType::GE_KM200,
                                });
  dev1->setOutputChannels({});

  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  auto& lightState = hal.getStateCache();
  lightState.empty();
  ds::DateTime now = std::chrono::system_clock::now();

  GIVEN("the cache is empty") {
    WHEN("a scene is called") {
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == false);
      THEN("no action is taken") {}
    }
  }

  GIVEN("the cache is populated") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(0.2);
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.property[DeviceProperty::minDim] = 10;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::brightness].value = 25.0;
      cache.scene.sceneTable[Scene4].channels[ChannelType::brightness].value = 0;

      cache.scene.sceneTable[SceneAutoOff].flags.dontCare = false;
      cache.scene.sceneTable[SceneAutoOff].flags.dimTime = DimTime::dimTime1;
      cache.scene.sceneTable[SceneAutoOff].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneAutoOff].channels[ChannelType::brightness].value = 0.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }

    WHEN("No state is known") {
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(!state.channels[ChannelType::brightness].target);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
        REQUIRE(!state.channels[ChannelType::brightness].target);
      }
      THEN("DEC do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
        REQUIRE(!state.channels[ChannelType::brightness].target);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 25.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 38.73);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 38.73);
        REQUIRE(!ret.terminatesAt);
        REQUIRE(hal.completeMoving(dev1->getDSID(), now + std::chrono::milliseconds(5000)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 38.73);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 38.73);
        REQUIRE(!ret.terminatesAt);
      }
      THEN("INC does not adjust the current value if the last called scene is unknown") {
        {
          auto state = lightState.tryLookup(dev1->getDSID());
          state.lastSceneCalled = Scene3;
          lightState.update(dev1->getDSID(), state);
        }
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 25);
      }
      THEN("INC does not adjust the current value if the last called scene is 0") {
        {
          auto state = lightState.tryLookup(dev1->getDSID());
          state.lastSceneCalled = Scene4;
          lightState.update(dev1->getDSID(), state);
        }
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 25);
      }
      THEN("INC will not exceed max value") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(7 + 3);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(1 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(3 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(4 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(5 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(6 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(7 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(8 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(9 * 250)) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), SceneInc, now + std::chrono::milliseconds(9 * 250 + 10)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 100);
      }
      THEN("DEC adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        REQUIRE(hal.completeMoving(dev1->getDSID(), now + std::chrono::milliseconds(5000)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 15.20);
      }
      THEN("DEC will not reach 0") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(4 + 1);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), SceneDec, now + std::chrono::milliseconds(10)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 15.20);

        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(1 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(3 * 250)) == true);
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(4 * 250)) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), SceneDec, now + std::chrono::milliseconds(4 * 250 + 10)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) > 0);
      }
      THEN("callSceneMin does not change the output") {
        REQUIRE(hal.callSceneMin(dev1->getDSID(), Scene1, now) == false);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 25.0);
      }
      THEN("after callScene(AUTO_OFF) state is moving") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneAutoOff, false, now) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::moving);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 25.0);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 0.0);
        REQUIRE(hal.completeMoving(dev1->getDSID(), now + std::chrono::milliseconds(60900)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 0.0);
      }
      THEN("after callScene(AUTO_OFF) with same vale state is not moving") {
        {
          auto cache = hal.getCache().tryLookup(dev1->getDSID());
          cache.scene.sceneTable[SceneAutoOff].channels[ChannelType::brightness].value = 25.0;
          hal.getCache().update(dev1->getDSID(), cache);
        }
        REQUIRE(hal.callScene(dev1->getDSID(), SceneAutoOff, false, now) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 25.0);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 25.0);
      }
    }
    WHEN("Current device state is zero") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 0.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("INC does not change the value") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 0);
        REQUIRE(state.dimmStatus == DeviceDimmingStatus::none);
        REQUIRE(!state.terminatesAt);
      }
      THEN("DEC not change the value") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 0);
        REQUIRE(state.dimmStatus == DeviceDimmingStatus::none);
        REQUIRE(!state.terminatesAt);
      }
      THEN("INC changes the value after callSceneMin") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(2);
        REQUIRE(hal.callSceneMin(dev1->getDSID(), Scene1, now) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 13.73);

        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now + std::chrono::milliseconds(1 * 250)) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), SceneInc, now + std::chrono::milliseconds(1 * 250 + 10)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 27.45);
      }
    }
    WHEN("Current device state is max") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 100.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("INC does not change the value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), SceneInc, now + std::chrono::milliseconds(10)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 100);
      }
      THEN("DEC changes the value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), SceneDec, now + std::chrono::milliseconds(10)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 90.20);
      }
    }
    WHEN("Area dimming is set") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 25.0;
        state.lastDimmCommand = SceneDecA4;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("A call scene resets the lastDimmCommand") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.lastDimmCommand);
      }
    }
  }

  GIVEN("the cache is populated for undo scene") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(0.2);
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;

      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::brightness].value = 25.0;

      cache.scene.sceneTable[Scene3].flags.dontCare = false;
      cache.scene.sceneTable[Scene3].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene3].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene3].channels[ChannelType::brightness].value = 75.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }

    WHEN("No state is known") {
      THEN("after undoScene does not make the state defined") {
        REQUIRE(hal.undoScene(dev1->getDSID(), -1, now) == false);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
      }
      THEN("after undoSceneNr does not make the state defined") {
        REQUIRE(hal.undoScene(dev1->getDSID(), Scene1, now) == false);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 100.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("after undoScene previous scene is called") {
        REQUIRE(hal.undoScene(dev1->getDSID(), -1, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 75);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with matching sceneId previous scene is called") {
        REQUIRE(hal.undoScene(dev1->getDSID(), Scene2, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 75);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with not matching sceneId previous scene is not called") {
        REQUIRE(hal.undoScene(dev1->getDSID(), Scene44, now) == false);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 100);
        REQUIRE(*state.lastSceneCalled == Scene2);
        REQUIRE(*state.beforeLastSceneCalled == Scene3);
      }
    }
  }

  GIVEN("Area dimming") {
    auto area = GENERATE(values<Area>({{SceneA11, SceneIncA1, SceneDecA1},
                                       {SceneA21, SceneIncA2, SceneDecA2},
                                       {SceneA31, SceneIncA3, SceneDecA3},
                                       {SceneA41, SceneIncA4, SceneDecA4}}));
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.property[DeviceProperty::minDim] = 10;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;

      cache.scene.sceneTable[area.area].flags.dontCare = false;
      cache.scene.sceneTable[area.area].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[area.area].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[area.area].channels[ChannelType::brightness].value = 50.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }

    WHEN("No state is known") {
      THEN("INC " << area.inc << " do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), area.inc, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
      }
      THEN("DEC " << area.dec << " do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), area.dec, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
      }
      THEN("CONT do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDimArea, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 25.0;
        state.lastSceneCalled = area.area;
        state.localPriority = true;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("INC " << area.inc << " adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(3);
        REQUIRE(hal.callScene(dev1->getDSID(), area.inc, false, now) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 38.73);

        REQUIRE(hal.callScene(dev1->getDSID(), SceneDimArea, false, now + std::chrono::milliseconds(1 * 250)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 52.45);

        REQUIRE(hal.callScene(dev1->getDSID(), SceneDimArea, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), area.inc, now + std::chrono::milliseconds(2 * 250 + 10)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 66.18);
      }
      THEN("DEC " << area.dec << " adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(3);
        REQUIRE(hal.callScene(dev1->getDSID(), area.dec, false, now) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 15.20);

        REQUIRE(hal.callScene(dev1->getDSID(), SceneDimArea, false, now + std::chrono::milliseconds(1 * 250)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].targetValue, 2) == 5.39);

        REQUIRE(hal.callScene(dev1->getDSID(), SceneDimArea, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.stopDimming(dev1->getDSID(), area.dec, now + std::chrono::milliseconds(2 * 250 + 10)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 3.92);
      }
    }
  }

  GIVEN("Device is not locally prioritized") {
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 75.0;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("A localOn scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[SceneLocalOn].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOn].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOn].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOn].flags.specialHandling = true;
        cache.scene.sceneTable[SceneLocalOn].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneLocalOn, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("A localOn scene is called, but not configured to be prioritized") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[SceneLocalOn].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOn].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOn].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOn].flags.specialHandling = false;
        cache.scene.sceneTable[SceneLocalOn].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("It does not becomes prioritized") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneLocalOn, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("Device is locally prioritized") {
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 75.0;
      state.localPriority = true;
      state.lastSceneCalled = Scene1;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("A normal scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[Scene1].flags.dontCare = false;
        cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("Without force, it does not change") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == false);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 75.0);
        REQUIRE(state.localPriority == true);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, true, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        REQUIRE(state.localPriority == false);
      }
    }
    WHEN("A dimming scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("Without force, it does not change") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == false);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 75.0);
        REQUIRE(state.localPriority == true);
      }
      THEN("With force, it does change") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, true, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current > 50);
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("A scene with ignoreLocalProp is called") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[Scene1].flags.dontCare = false;
        cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = true;
        cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("Without force, it does change") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        REQUIRE(state.localPriority == false);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, true, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        REQUIRE(state.localPriority == false);
      }
    }
    WHEN("A local off scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[SceneLocalOff].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOff].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOff].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOff].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneLocalOff, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A scene with a long transition time") {
    DeviceProperty down;
    DeviceProperty up;
    DimTime dim;
    std::tie(down, up, dim) = GENERATE(values<std::tuple<DeviceProperty, DeviceProperty, DimTime>>(
        {{DeviceProperty::dimTime0Down, DeviceProperty::dimTime0Up, DimTime::dimTime0},
         {DeviceProperty::dimTime1Down, DeviceProperty::dimTime1Up, DimTime::dimTime1},
         {DeviceProperty::dimTime2Down, DeviceProperty::dimTime2Up, DimTime::dimTime2}}));
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[down] = convertDimTimeToDevice(10);
      cache.property[up] = convertDimTimeToDevice(10);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = dim;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 100.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 50.0;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After the scene is called") {
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
      THEN("termination time is initially set to lastChanged") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(state.channels[ChannelType::brightness].target);
        REQUIRE(*state.channels[ChannelType::brightness].target == 100);
        REQUIRE(state.lastChanged > state.lastChanged.min());
        REQUIRE(
            std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() ==
            10000);
      }
      THEN("current value changes with subsequent calls to getStatus") {
        auto state = lightState.tryLookup(dev1->getDSID());
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::moving);
        REQUIRE(ret.outputs[ChannelType::brightness].targetValue == 100.0);
        REQUIRE(ret.outputs[ChannelType::brightness].value == 50.0);

        REQUIRE(hal.completeMoving(dev1->getDSID(), now + std::chrono::milliseconds(10100)) == true);
        ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(ret.outputs[ChannelType::brightness].targetValue == 100.0);
        REQUIRE(ret.outputs[ChannelType::brightness].value == 100.0);
      }
    }
  }

  GIVEN("Stop handling") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(10);
      cache.property[DeviceProperty::dimTime0Up] = convertDimTimeToDevice(10);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::dimTime0;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 100.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 0.0;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After a scene is called") {
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
      auto state = lightState.tryLookup(dev1->getDSID());
      auto origCallTime = state.lastChanged;
      THEN("the stop scene halts the value before it reaches the target") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneStop, false, origCallTime + std::chrono::milliseconds(5000)) ==
                true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(!state.channels[ChannelType::brightness].target);
        REQUIRE(
            roundDigits(*state.channels[ChannelType::brightness].current, 2) * 100 ==
            roundDigits(std::chrono::duration_cast<std::chrono::milliseconds>(state.lastChanged - origCallTime).count(),
                        2));
      }
    }
  }

  GIVEN("A scene with dontCare") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.scene.sceneTable[Scene1].flags.dontCare = true;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 100.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 50.0;
      state.lastSceneCalled = 13;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After the scene is called") {
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == false);
      THEN("state has not changed") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(!state.channels[ChannelType::brightness].target);
        REQUIRE(*state.lastSceneCalled == 13);
      }
    }
  }

  GIVEN("A setOutputValue call") {
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 50.0;
      state.channels[ChannelType::brightness].target = 55.0;
      state.lastSceneCalled = 13;
      state.localPriority = true;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After setOutputValue is called") {
      REQUIRE(hal.setOutputValue(dev1->getDSID(), 245, now) == true);
      THEN("state is changed") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 96.08);
        REQUIRE(!state.channels[ChannelType::brightness].target);
        REQUIRE(!state.lastSceneCalled);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A setLocalPrioCall") {
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 50.0;
      state.channels[ChannelType::brightness].target = 55.0;
      state.lastSceneCalled = Scene42;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After setLocalPrioCall is called with matching sceneId") {
      REQUIRE_NOTHROW(hal.setSceneLocalPrio(dev1->getDSID(), Scene42));
      THEN("state is changed") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("After setLocalPrioCall is called with matching sceneId") {
      REQUIRE_NOTHROW(hal.setSceneLocalPrio(dev1->getDSID(), Scene43));
      THEN("state is changed") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("The device is set to switched mode") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KM_Switch);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 100.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::brightness].value = 25.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    WHEN("after callScene with value above the threshold") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 25.0;
        state.channels[ChannelType::brightness].level = 0;
        state.lastSceneCalled = 13;
        lightState.update(dev1->getDSID(), state);
      }
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
      THEN("the output is on (level1)") {
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.outputs[ChannelType::brightness].value == 100);
        REQUIRE(ret.outputs[ChannelType::brightness].level == 1);
      }
    }
    WHEN("after callScene with value below the threshold") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 75.0;
        state.channels[ChannelType::brightness].level = 1;
        state.lastSceneCalled = 13;
        lightState.update(dev1->getDSID(), state);
      }
      REQUIRE(hal.callScene(dev1->getDSID(), Scene2, false, now) == true);
      THEN("the output is off (level0)") {
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.outputs[ChannelType::brightness].value == 25);
        REQUIRE(ret.outputs[ChannelType::brightness].level == 0);
      }
    }
    WHEN("device is reconfigured to dimmed mode") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 75.0;
        state.channels[ChannelType::brightness].level = 1;
        state.lastSceneCalled = 13;
        lightState.update(dev1->getDSID(), state);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 3, 0, 22));
      THEN("the output level is reset") {
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(!ret.outputs[ChannelType::brightness].level);
      }
    }
    WHEN("device switch threshold is reconfigured") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 75.0;
        state.channels[ChannelType::brightness].level = 1;
        state.lastSceneCalled = 13;
        lightState.update(dev1->getDSID(), state);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 3, 0x43, 205));
      THEN("the output level is reset") {
        DS_DEBUG("Just here to make the test work");
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.outputs[ChannelType::brightness].level == 0);
      }
    }
    THEN("DEC across the threshold will change level") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 100.0;
        state.channels[ChannelType::brightness].level = 1;
        state.lastSceneCalled = Scene1;
        lightState.update(dev1->getDSID(), state);
      }
      REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(15);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
      auto ret = hal.getState(dev1->getDSID());
      REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
      REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 95.69);
      REQUIRE(ret.outputs[ChannelType::brightness].level == 1);

      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(1 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(2 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(3 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(4 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(5 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(6 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(7 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(8 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(9 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(10 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(11 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(12 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(13 * 250)) == true);
      REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now + std::chrono::milliseconds(14 * 250)) == true);
      REQUIRE(hal.stopDimming(dev1->getDSID(), SceneDec, now + std::chrono::milliseconds(14 * 250 + 10)) == true);
      ret = hal.getState(dev1->getDSID());
      REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
      REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 35.29);
      REQUIRE(ret.outputs[ChannelType::brightness].level == 0);
    }
  }

  GIVEN("The device is set to dimmed mode") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::TrailingEdgeDimmerCurve);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 100.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::brightness].value = 25.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 75.0;
      state.lastSceneCalled = 13;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("device is reconfigured to switched mode") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 3, 0, 16));
      THEN("the output level is correct") {
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.outputs[ChannelType::brightness].level == 1);
      }
    }
  }

  GIVEN("a switched light device with a scene call configured at threshold") {
    {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 3, 0, 16));
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 128 / 255.0 * 100;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      hal.getCache().update(dev1->getDSID(), cache);
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 0.0;
      state.channels[ChannelType::brightness].level = 0;
      state.lastSceneCalled = SceneOff;
      lightState.update(dev1->getDSID(), state);
    }
    THEN("calling the scene will change the switced level") {
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
      auto ret = hal.getState(dev1->getDSID());
      REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 50.2);
      REQUIRE(ret.outputs[ChannelType::brightness].level == 1);
    }
  }

  GIVEN("A setOutputValue call must change switched level") {
    {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 3, 0, 16));
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::switchedThreshold] = 128;
      hal.getCache().update(dev1->getDSID(), cache);
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 0.0;
      state.channels[ChannelType::brightness].level = 0;
      state.lastSceneCalled = SceneOff;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After setOutputValue is called") {
      REQUIRE(hal.setOutputValue(dev1->getDSID(), 200, now) == true);
      THEN("state is changed") {
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 78.43);
        REQUIRE(ret.outputs[ChannelType::brightness].level == 1);
      }
    }
  }
}

SCENARIO("Multi-color Light device", TAGS) {
  PropertySystemMock propertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance(propertySystem, eventSink);

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  auto dev1 = createDevice(apt, {
                                    .hardwareType = HardwareType::GE_KM200,
                                });
  dev1->setOutputChannels({ChannelType::brightness, ChannelType::hue, ChannelType::saturation});

  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  auto& lightState = hal.getStateCache();
  lightState.empty();
  ds::DateTime now = std::chrono::system_clock::now();

  GIVEN("the cache is populated") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(0.2);
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;
      cache.scene.sceneTable[Scene1].channels[ChannelType::hue].value = 250.5;
      cache.scene.sceneTable[Scene1].channels[ChannelType::saturation].value = 75.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }

    WHEN("No state is known") {
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(state.channels[ChannelType::hue].current);
        REQUIRE(*state.channels[ChannelType::hue].current == 250.5);
        REQUIRE(state.channels[ChannelType::saturation].current);
        REQUIRE(*state.channels[ChannelType::saturation].current == 75);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
        REQUIRE(!state.channels[ChannelType::hue].current);
        REQUIRE(!state.channels[ChannelType::saturation].current);
      }
      THEN("DEC do not make the state defined") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
        REQUIRE(!state.channels[ChannelType::hue].current);
        REQUIRE(!state.channels[ChannelType::saturation].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = lightState.tryLookup(dev1->getDSID());
        state.channels[ChannelType::brightness].current = 25.0;
        state.channels[ChannelType::hue].current = 25.0;
        state.channels[ChannelType::saturation].current = 25.0;
        state.lastSceneCalled = Scene1;
        lightState.update(dev1->getDSID(), state);
      }
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(state.channels[ChannelType::hue].current);
        REQUIRE(*state.channels[ChannelType::hue].current == 250.5);
        REQUIRE(state.channels[ChannelType::saturation].current);
        REQUIRE(*state.channels[ChannelType::saturation].current == 75);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneInc, false, now) == true);
        REQUIRE(hal.completeMoving(dev1->getDSID(), now + std::chrono::milliseconds(5000)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 38.73);
        REQUIRE(roundDigits(ret.outputs[ChannelType::hue].value, 2) == 25.0);
        REQUIRE(roundDigits(ret.outputs[ChannelType::saturation].value, 2) == 25);
      }
      THEN("DEC adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dev1->getDSID(), SceneDec, false, now) == true);
        REQUIRE(hal.completeMoving(dev1->getDSID(), now + std::chrono::milliseconds(5000)) == true);
        auto ret = hal.getState(dev1->getDSID());
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::brightness].value, 2) == 15.20);
        REQUIRE(roundDigits(ret.outputs[ChannelType::hue].value, 2) == 25.0);
        REQUIRE(roundDigits(ret.outputs[ChannelType::saturation].value, 2) == 25);
      }
    }
  }

  GIVEN("Device is not locally prioritized") {
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 75.0;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("A localOn scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dev1->getDSID());
        cache.scene.sceneTable[SceneLocalOn].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOn].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOn].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOn].flags.specialHandling = true;
        cache.scene.sceneTable[SceneLocalOn].channels[ChannelType::brightness].value = 50.0;
        hal.getCache().update(dev1->getDSID(), cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(dev1->getDSID(), SceneLocalOn, false, now) == true);
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50.0);
        DS_DEBUG(state.localPriority);
        REQUIRE(state.localPriority == true);
      }
    }
  }

  GIVEN("A scene with dontCare on individual channels") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(0.01);
      cache.property[DeviceProperty::dimTime0Up] = convertDimTimeToDevice(0.01);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].value = 50.0;
      cache.scene.sceneTable[Scene1].channels[ChannelType::brightness].dontCare = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::hue].value = 250.5;
      cache.scene.sceneTable[Scene1].channels[ChannelType::hue].dontCare = true;
      cache.scene.sceneTable[Scene1].channels[ChannelType::saturation].value = 75.0;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 25.0;
      state.channels[ChannelType::hue].current = 150.0;
      state.channels[ChannelType::saturation].current = 175.0;
      state.localPriority = false;
      lightState.update(dev1->getDSID(), state);
    }
    WHEN("After the scene is called") {
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == true);
      THEN("only the active channels are applied") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(state.channels[ChannelType::brightness].current);
        REQUIRE(*state.channels[ChannelType::brightness].current == 50);
        REQUIRE(state.channels[ChannelType::hue].current);
        REQUIRE(*state.channels[ChannelType::hue].current == 150);
        REQUIRE(state.channels[ChannelType::saturation].current);
        REQUIRE(*state.channels[ChannelType::saturation].current == 75);
      }
    }
  }

  GIVEN("A device has a disabled output") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::outputMode] = OUTPUT_MODE_DEACTIVATED;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    WHEN("After the scene is called") {
      REQUIRE(hal.callScene(dev1->getDSID(), Scene1, false, now) == false);
      THEN("state is not changed") {
        auto state = lightState.tryLookup(dev1->getDSID());
        REQUIRE(!state.channels[ChannelType::brightness].current);
        REQUIRE(!state.lastDimmCommand);
        REQUIRE(!state.lastSceneCalled);
      }
    }
  }

  GIVEN("a light device is controlled by writing to the runtime bank") {
    THEN("writing the output value will change it") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 64, 0x00, 0x34));
      auto state = lightState.tryLookup(dev1->getDSID());
      REQUIRE(state.channels[ChannelType::brightness].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 20.39);
      REQUIRE(!state.terminatesAt);
    }
  }

  GIVEN("a light device is switched and controlled by writing to the runtime bank") {
    {
      auto cache = hal.getCache().tryLookup(dev1->getDSID());
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KL_Switch);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      hal.getCache().update(dev1->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(dev1->getDSID());
      state.channels[ChannelType::brightness].current = 0;
      state.channels[ChannelType::brightness].level = 0;
      state.lastSceneCalled = 0;
      lightState.update(dev1->getDSID(), state);
    }
    THEN("writing the output value will change it and the switched state") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dev1->getDSID(), 64, 0x00, 0xac));
      auto state = lightState.tryLookup(dev1->getDSID());
      REQUIRE(state.channels[ChannelType::brightness].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::brightness].current, 2) == 67.45);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dev1->getDSID());
      REQUIRE(ret.outputs[ChannelType::brightness].level == 1);
    }
  }
}

} // namespace
