/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "system_triggers.h"

#include <event/event_create.h>
#include <metering/metering-mock.h>
#include <metering/power-states-fwd.h>
#include <model/apartment-mock.h>
#include <model/state.h>

#include <trompeloeil.hpp>

#include "ds/catch/catch.h"

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

static ds::LogChannel dsLogChannel("handlerSystemTriggersTest");

namespace {

static const char* TAGS = "[dss][handler][systemTriggers]";

SCENARIO("TriggerBuilder", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto triggerBuilder = makeTriggerBuilder(apartment);

  dss::metering::MeteringMock metering{};
  ALLOW_CALL(apartment, getMetering()).LR_RETURN(metering);

  SECTION("invalid trigger path") {
    auto& x = GENERATE("/scripts/system-addon-user-defined-states/apartment-power-states-extra-suffix"s,
                       "/scripts/system-addon-user-defined-states/open-meteirng-sysem"s);
    triggerBuilder->create(ds::str(x, "/12345678/setTrigger"));
  }

  SECTION("electricityPower triggers") {
    auto x = GENERATE(table<std::string, MeteringSystem>({
        {"/scripts/system-addon-user-defined-states/apartment-power-states", MeteringSystem::apartmentSummation},
        {"/scripts/system-addon-user-defined-states/open-metering-system", MeteringSystem::openMeteringSystem},
    }));

    const std::string& udsPath = std::get<std::string>(x);
    auto& triggerConfigs = *propertySystem.createProperty(std::get<std::string>(x));

    WHEN("configured correctly") {
      auto& configTrigger0 = *triggerConfigs.createProperty("12345678");
      configTrigger0.setChildValue("category", ds::str(Category::production));
      configTrigger0.setChildValue("activeValue", 7);
      configTrigger0.setChildValue("inactiveValue", 2);

      auto& state = *apartment.allocateState(StateType_Circuit, "apartment-power-production-on7-off2"s, {});

      WHEN("setTrigger is set up") {
        auto& setTriggerNode = *configTrigger0.createProperty("setTrigger");
        auto& trigger0 = *setTriggerNode.createProperty("triggers"s)->createProperty("0"s);

        REQUIRE_CALL(metering, createState(std::get<MeteringSystem>(x), _)).LR_RETURN(state.shared_from_this());
        triggerBuilder->create(ds::str(udsPath, "/12345678/setTrigger"));

        THEN("state name is filled in") {
          CHECK(trigger0.getPropertyByName("name")->getStringValue() == state.getName());

          WHEN("all fields in trigger are set up") {
            state.setState(CallOrigin::system, State_Active);
            SystemTrigger trigger(*createStateChangeEvent(state.shared_from_this(), {}, CallOrigin::system));

            THEN("checkTriggerNode fires upon inactive event") {
              CHECK(trigger.checkTriggerNode(trigger0.shared_from_this()));
            }
            THEN("checkTrigger fires upon inactive event") {
              CHECK(trigger.checkTrigger(setTriggerNode.shared_from_this()));
            }
          }
        }

        WHEN("remove trigger") {
          REQUIRE_CALL(metering, removeState(std::get<MeteringSystem>(x), _));
          triggerBuilder->remove(ds::str(udsPath, "/12345678/setTrigger"));
        }
      }

      THEN("without setup, state name is missing") {
        auto& setTriggerNode = *configTrigger0.createProperty("setTrigger");
        auto& trigger0 = *setTriggerNode.createProperty("triggers"s)->createProperty("0"s);
        CHECK(trigger0.getPropertyByName("name") == nullptr);
      }

      WHEN("resetTrigger is set up") {
        auto& resetTriggerNode = *configTrigger0.createProperty("resetTrigger");
        auto& trigger0 = *resetTriggerNode.createProperty("triggers"s)->createProperty("0"s);

        REQUIRE_CALL(metering, createState(std::get<MeteringSystem>(x), _)).LR_RETURN(state.shared_from_this());
        triggerBuilder->create(ds::str(udsPath, "/12345678/resetTrigger"));

        THEN("state name is filled in") {
          CHECK(trigger0.getPropertyByName("name")->getStringValue() == state.getName());

          WHEN("all fields in trigger are set up") {
            state.setState(CallOrigin::system, State_Inactive);
            SystemTrigger trigger(*createStateChangeEvent(state.shared_from_this(), {}, CallOrigin::system));

            THEN("checkTriggerNode fires upon inactive event") {
              CHECK(trigger.checkTriggerNode(trigger0.shared_from_this()));
            }
            THEN("checkTrigger fires upon inactive event") {
              CHECK(trigger.checkTrigger(resetTriggerNode.shared_from_this()));
            }
          }
        }
      }
    }

    WHEN("configuration values are incorrect") {
      WHEN("integer values are given as strings") {
        auto& configTrigger1 = *triggerConfigs.createProperty("id1");
        configTrigger1.setChildValue("category", ds::str(Category::production));
        configTrigger1.setChildValue("activeValue", "8"s);
        configTrigger1.setChildValue("inactiveValue", "1"s);

        THEN("then setup fails") {
          configTrigger1.createProperty("setTrigger");
          CHECK_THROWS(triggerBuilder->create(ds::str(udsPath, "/id1/setTrigger")));
        }
      }

      WHEN("invalid category is given") {
        auto& configTrigger1 = *triggerConfigs.createProperty("id1");
        configTrigger1.setChildValue("category", "froggle"s);
        configTrigger1.setChildValue("activeValue", 8);
        configTrigger1.setChildValue("inactiveValue", 1);

        THEN("that setup fails") {
          configTrigger1.createProperty("setTrigger");
          CHECK_THROWS(triggerBuilder->create(ds::str(udsPath, "/id1/setTrigger")));
        }
      }
    }
  }
}

} // namespace
