/*
    Copyright (c) 2012, 2014 digitalSTROM.org, Zurich, Switzerland

    Authors: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>
             Andreas Fenkart <andreas.fenkart@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#pragma once

#include <curl/curl.h>
#include <ds/reflect.h>
#include <memory>
#include <unordered_map>

#include "logger.h"

namespace dss {

enum class HttpMethod {
  GET,
  POST,
  DELETE,
};
inline std::ostream &operator<<(std::ostream &os, HttpMethod x) { return os << static_cast<int>(x); }

struct HttpRequest {
  std::string url;
  HttpMethod type;
  std::unordered_map<std::string, std::string> headers;
  std::string postdata;
};
bool operator==(const HttpRequest &a, const HttpRequest &b);
std::ostream &operator<<(std::ostream &os, const HttpRequest &x);

struct HttpResponse {
  int code;
  std::string data;
};

struct HttpStatistics {
  double totalTime;
  double sumUp, sumDown;
  double speedUp, speedDown;
};

/// Http client.
///
/// The methods are not thread safe.
/// Curl handle is stored as member variable and
/// parallel calls reuse the same curl handle.
/// But one curl handle
/// [must not be used from more threads](https://curl.haxx.se/libcurl/c/threadsafe.html).
///
class HttpClient {
  __DECL_LOG_CHANNEL__
public:
  HttpClient(bool reuseHandle = false);
  ~HttpClient();

  long get(const std::string &url, std::string *result, bool insecure = false);

  long get(const std::string &url, const std::unordered_map<std::string, std::string> &headers, std::string *result,
      bool insecure = false);

  long post(const std::string &_url, const std::string &_postdata, std::string *_result);

  /* TODO make postdata const-by-reference */
  long post(const std::string &_url, const std::unordered_map<std::string, std::string> &headers,
      const std::string &_postdata, std::string *_result);

  long request(const HttpRequest &req, std::string *result);
  HttpResponse request(const HttpRequest &);

  long downloadFile(const std::string &_url, const std::string &_filename);

  /* TODO evtl. URI wrapper easy appending/parsing of query options */

  HttpStatistics getStats();

private:
  long internalRequest(const std::string &_url, HttpMethod _type,
      const std::unordered_map<std::string, std::string> &headers, const std::string &_postdata, std::string *_result,
      bool _insecure);

  static size_t writeCallbackMute(void *contents, size_t size, size_t nmemb, void *userp);
  bool m_reuse_handle;
  struct CurlHandleDeleter {
    void operator()(CURL *);
  };
  std::unique_ptr<CURL, CurlHandleDeleter> m_curl_handle;
};

// Contrary to HttpClient methods, these methods are thread safe
class SimpleHttpClient {
public:
  virtual ~SimpleHttpClient(){};
  virtual HttpResponse request(const HttpRequest &) = 0;
};
std::unique_ptr<SimpleHttpClient> makeSimpleHttpClient();

} // namespace dss

DS_REFLECT_STRUCT((dss), HttpRequest, (url)(type)(headers)(postdata));
DS_REFLECT_STRUCT((dss), HttpResponse, (code)(data));
