/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "archives-sanity.h"

#include <ds/catch/catch.h>
#include <ds/system-mock.h>

#include "archives-connection-mock.h"
#include "archives-storage-mock.h"

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

static ds::LogChannel dsLogChannel("dssMeteringArchivesSanityTest");

namespace {

static const char* TAGS = "[dSS][metering][archivesStorage]";

SCENARIO("archivesSanity", TAGS) {
  auto archivesConnection = ArchivesConnectionMock();
  auto archivesStorage = ArchivesStorageMock();
  auto system = ds::SystemMock();

  auto archivesSanity = makeArchivesSanity(archivesConnection, archivesStorage, system);

  using system_clock = std::chrono::system_clock;

  SECTION("files no longer receiving sensor values") {
    auto now = system_clock::now();
    GIVEN("file updates a little behind, but still within limits") {
      REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(now - 10s);
      CHECK(archivesSanity->withoutUpdates("/storage/archive.rrd") == false);
    }
    GIVEN("outdated file") {
      // MAX_INTERPOLATION_INTERVAL
      REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(now - 7 * 24h - 1s);
      CHECK(archivesSanity->withoutUpdates("/storage/archive.rrd") == true);
    }
    GIVEN("lastTimestamp fails") {
      THEN("file is kept") {
        // rely on file format checks
        REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(std::nullopt);
        CHECK(archivesSanity->withoutUpdates("/storage/archive.rrd"s) == false);
      }
    }
    GIVEN("lastTimestamp is a little in the future") {
      THEN("file is still good") {
        REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(now + 24h);
        CHECK(archivesSanity->withoutUpdates("/storage/archive.rrd"s) == false);
      }
    }
    GIVEN("lastTimestamp is far in the future") {
      THEN("file is still good") {
        REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(now + 7 * 24h + 1s);
        CHECK(archivesSanity->withoutUpdates("/storage/archive.rrd"s) == true);
      }
    }
  }

  SECTION("remove dead files") {
    auto now = system_clock::now();
    auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
    using vector_t = decltype(std::declval<ArchivesStorage>().searchArchives());
    REQUIRE_CALL(archivesStorage, searchArchives())
        .RETURN(vector_t{std::pair{ArchiveId{controllerId1, Category::consumption},
                                   "/storage/1111111111111111111111111111111100.rrd"s}});
    GIVEN("old file") {
      REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(now - 7 * 24h - 1s);
      REQUIRE_CALL(system, remove(_));
      CHECK(archivesSanity->removeDeadArchives() == true);
    }
    GIVEN("valid file") {
      REQUIRE_CALL(archivesConnection, lastTimestamp(_)).RETURN(now - 10s);
      CHECK(archivesSanity->removeDeadArchives() == false);
    }
  }

  SECTION("migrateFilenames") {
    GIVEN("no migration needed") {
      using vector_t = decltype(std::declval<ArchivesStorage>().searchArchives());
      REQUIRE_CALL(archivesStorage, searchArchives())
          .RETURN(vector_t{std::pair{ArchiveId{"1111111111111111111111111111111100"s, Category::consumption},
                                     "/foo/1111111111111111111111111111111100.rrd"s},
                           std::pair{ArchiveId{"1111111111111111111111111111111100"s, Category::production},
                                     "/foo/1111111111111111111111111111111100-production.rrd"s}});
      archivesSanity->migrateFilenames();
    }

    GIVEN("legacy dsid archive") {
      dsid_t dsid{.id = {0x35, 0x04, 0x17, 0x5f, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x01, 0x6b, 0xe7}};

      using vector_t = decltype(std::declval<ArchivesStorage>().searchArchives());
      REQUIRE_CALL(archivesStorage, searchArchives())
          .RETURN(vector_t{std::pair{dsid, "/foo/3504175fe000000000016be7.rrd"s}});

      THEN("file is rename") {
        REQUIRE_CALL(archivesStorage, filepath(ArchiveId{"3504175fe00000000000000000016be700"s, Category::consumption}))
            .RETURN(ds::str("/storage-location/", _1, ".rrd"));
        REQUIRE_CALL(system, existsFile(_))
            .LR_SIDE_EFFECT(CHECK(_1 == "/storage-location/(3504175fe00000000000000000016be700,consumption).rrd"sv))
            .RETURN(false);
        REQUIRE_CALL(system, rename(_, _))
            .LR_SIDE_EFFECT(CHECK(_1 == "/foo/3504175fe000000000016be7.rrd"sv))
            .LR_SIDE_EFFECT(CHECK(_2 == "/storage-location/(3504175fe00000000000000000016be700,consumption).rrd"sv));
        archivesSanity->migrateFilenames();
      }
      WHEN("newer dsuid file exists") {
        REQUIRE_CALL(archivesStorage, filepath(ArchiveId{"3504175fe00000000000000000016be700"s, Category::consumption}))
            .RETURN(ds::str("/storage-location/", _1, ".rrd"));
        THEN("file is removed") {
          REQUIRE_CALL(system, existsFile(_))
              .LR_SIDE_EFFECT(CHECK(_1 == "/storage-location/(3504175fe00000000000000000016be700,consumption).rrd"sv))
              .RETURN(true);
          REQUIRE_CALL(system, remove("/foo/3504175fe000000000016be7.rrd"s));
          // new file will be found independently from the legacy dsid_t version
          archivesSanity->migrateFilenames();
        }
      }
    }
  }
  SECTION("restoreBackups") {
    THEN("restoring a backup") {
      REQUIRE_CALL(archivesStorage, searchRrdFiles())
          .RETURN(std::vector{
              std::filesystem::path{"/foo/3504175fe000000000016be7.rrd"s},
              std::filesystem::path{"/foo/3504175fe000000000016be7-backup.rrd"s},
              std::filesystem::path{"/foo/3504175fe000000000016be9-backup.rrd"s},
          });

      REQUIRE_CALL(system, rename("/foo/3504175fe000000000016be9-backup.rrd"s, "/foo/3504175fe000000000016be9.rrd"s));
      CHECK(archivesSanity->restoreBackups() == true);
    }
    THEN("no change") {
      REQUIRE_CALL(archivesStorage, searchRrdFiles())
          .RETURN(std::vector{
              std::filesystem::path{"/foo/3504175fe000000000016be7.rrd"s},
          });
      CHECK(archivesSanity->restoreBackups() == false);
    }
  }
}
} // namespace
