#include "controller-direction.h"

#include <digitalSTROM/dsuid.h>
#include <ds/log.h>
#include <ds/system.h>

#include <boost/range/algorithm/find.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/range/algorithm/replace_if.hpp>
#include <boost/thread/mutex.hpp>
#include <numeric>

#include "metering/archives-connection.h"
#include "metering/archives-storage.h"
#include "metering/archives/controller.h"
#include "metering/config.h"
#include "metering/metering.h"
#include "model/apartment.h"
#include "model/dsmeter.h"

static ds::LogChannel dsLogChannel("dssMeteringController");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

namespace dss::metering {

// genesis:
// in the begining there was only consumption metering
// since the extension to production metering we use two archives per meter, one
// for either direction. this was easier than extending an existing archive with
// production data sources. And, upon configuration change, their removal.
// Had we foreseen production metering from the start this class would not exist

namespace {

using system_clock = std::chrono::system_clock;

class DirectionImpl final : public ControllerDirection {
  using ControllerArchives = archives::ControllerArchives;

public:
  DirectionImpl(Apartment& apartment, std::unique_ptr<ControllerArchives>&& controllerArchives,
                ArchivesStorage& archivesStorage, Category category, ds::System& system)
      : m_apartment(apartment),
        m_controllerArchives(std::move(controllerArchives)),
        m_archivesStorage(archivesStorage),
        m_directionType(category),
        m_system(system),
        m_config(m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {})) {}

  std::vector<std::pair<Dsuid, dsModel::MeteringType>> meterings() const override;

  bool isActiveMeter(const dsuid_t& meterId) const {
    boost::mutex::scoped_lock lock(m_cachedSeries_mutex);
    return (boost::find(m_meters, meterId) != m_meters.end());
  }

  std::optional<ArchiveValue> getEnergy(const dsuid_t& meterId) const override;

  std::optional<LastReading> getLastReading(const Dsuid& dsuid, dsModel::MeteringType meteringType,
                                            EnergyUnit energyUnit) const override;
  std::optional<LastReading> getLastReadingApartment(dsModel::MeteringType meteringType,
                                                     EnergyUnit energyUnit) const override;

  std::optional<ArchiveValue> tryCreateCacheFromDB(const dsuid_t& meterId) const;

  bool tryGetValidArchiveFilename(const dsuid_t& meterId) const;
  void getOrCreateCachedSeries(const dsuid_t& meterId);

  void addEnergy(const dsuid_t& meterId, ElectricityMeasurement&& energyMeasurement) override;

  ValueRange getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                       bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) const override;

  unsigned int getApartmentPower() const override;
  double getInstantPower(const dsuid_t& meterId) const override;

  double getLastEnergy(const dsuid_t& meterId) const override {
    // keep latest energy value even when sensor value timeout
    auto value = getEnergy(meterId);
    return (value) ? value->energy : 0;
  }

  void pollTimestamp(ds::DateTime timepoint) override { m_lastPoll = timepoint; }

  double get15MinuteAverage(std::vector<Dsuid> meterIds, metering::SeriesTypes type, Unit unit) const override {
    auto range = getSeries(meterIds, 15min, type, unit == Unit::Wh, {}, 1);
    auto& values = range.values;
    if (values.size() != 1) {
      DS_NOTICE("unexpected value count received", values.size(), range);
      return 0;
    }
    DS_REQUIRE(values.size() == 1);
    DS_REQUIRE(std::isfinite(values[0].value));
    return values[0].value;
  };

  std::vector<std::string> needsFlushing(const ds::DateTime& sampledAt) const override {
    auto flushDb = [this, sampledAt](const ArchiveId& archiveId) {
      auto lastTimestamp = m_controllerArchives->lastTimestamp(archiveId);
      if (!lastTimestamp) {
        return true; // it's only a flush
      }
      auto deltaTime = sampledAt - *lastTimestamp;
      return (deltaTime > m_config.diskFlushInterval);
    };

    decltype(m_meters) tmp;
    std::copy_if(std::cbegin(m_meters), std::cend(m_meters), std::back_inserter(tmp), [&](const auto& dsuid) {
      return flushDb(ArchiveId{dsuid, m_directionType});
    });

    std::vector<std::string> ret{};
    std::transform(std::cbegin(tmp), std::cend(tmp), std::back_inserter(ret), [&](const auto& dsuid) {
      return m_archivesStorage.filepath(ArchiveId{dsuid, m_directionType});
    });
    return ret;
  }

  bool validateAndConvertArchive(const ArchiveId& archiveId) const override {
    DS_ASSERT(m_archivesStorage.hasArchive(archiveId));
    try {
      bool tunePowerMaxSetting = false;
      bool validate = m_controllerArchives->validateDBSeries(archiveId, tunePowerMaxSetting);
      if (!validate) {
        DS_NOTICE("remove invalid archive", archiveId)
        m_system.remove(m_archivesStorage.filepath(archiveId));
        return false;
      } else if (tunePowerMaxSetting) {
        DS_NOTICE("archive needs tuning", archiveId);
        m_controllerArchives->tuneDBPowerSettings(archiveId); // non-const
      }
      return true;
    } catch (const ds::Exception& e) {
      DS_WARNING(e.what());
      return false;
    }
    return true;
  }

private:
  mutable boost::mutex m_cachedSeries_mutex;
  mutable std::vector<dsuid_t> m_meters;

  std::optional<unsigned long> deltaLast(const dsuid_t& meterId, ElectricityMeasurement energyMeasurement);

  mutable std::vector<std::pair<dsuid_t, ArchiveValue>> m_energyCache;
  std::vector<std::pair<dsuid_t, ElectricityMeasurement>> m_lastMeasurements;

  Apartment& m_apartment;
  std::unique_ptr<ControllerArchives> m_controllerArchives;
  ArchivesStorage& m_archivesStorage;
  Category m_directionType;
  ds::System& m_system;
  const ElectricityMeteringConfig m_config;
  ds::DateTime m_lastPoll;
};

std::vector<std::pair<Dsuid, dsModel::MeteringType>> DirectionImpl::meterings() const {
  std::vector<std::pair<Dsuid, dsModel::MeteringType>> result;

  auto meterPresent = [&apartment = m_apartment](const dsuid_t dsuid) {
    // dsModel::metering technical name DS_REQUIRE meter to be present...
    if (auto meter = apartment.tryGetDSMeterByDSID(dsuid); meter && meter->isPresent()) {
      return meter;
    }
    return decltype(apartment.tryGetDSMeterByDSID(dsuid)){};
  };

  auto matchingMeteringDirection = [category = m_directionType](const auto& meter) {
    if (category == Category::consumption) {
      return meter.getCapability_HasMetering();
    } else {
      return meter.getCapability_HasMeteringProducerEnabled();
    }
  };

  for (const auto& meterId : m_meters) {
    // futeless check for HasMetering, if it sends metering data
    if (auto meter = meterPresent(meterId); !meter || !matchingMeteringDirection(*meter)) {
      continue;
    }
    if (m_directionType == Category::consumption) {
      result.push_back({meterId, dsModel::MeteringType::powerMetering});
      result.push_back({meterId, dsModel::MeteringType::energyMetering});
    } else {
      result.push_back({meterId, dsModel::MeteringType::powerProducedMetering});
      result.push_back({meterId, dsModel::MeteringType::energyProducedMetering});
    }
  }

  return result;
}

std::optional<unsigned long> DirectionImpl::deltaLast(const dsuid_t& meterId, ElectricityMeasurement newMeasurement) {
  auto it = boost::find_if(m_lastMeasurements, [&meterId](const auto& elt) { return elt.first == meterId; });
  if (it == m_lastMeasurements.end()) {
    // the first value sets only the baseline, e.g. startup
    DS_NOTICE("deltaEnergy: starting new baseline, assume system startup");
    m_lastMeasurements.push_back(std::make_pair(meterId, std::move(newMeasurement)));
    return {};
  }

  const auto [__, newEnergy, newTimestamp] = newMeasurement;
  auto [_, lastEnergy, lastTimestamp] = std::exchange(it->second, newMeasurement);

  // TODO check dsm reboot counter instead
  auto limit = [config = m_config.discontinuityDetection](Category category) -> unsigned {
    if (category == Category::consumption) {
      return config.maxConsumption;
    } else {
      return config.maxProduction;
    }
  };

  // sanity check, avoid big jumps: #10705
  if (newTimestamp <= lastTimestamp) {
    DS_NOTICE("deltaEnergy: timestamp not increasing, starting new baseline:", newTimestamp, lastTimestamp);
    return {};
  }
  if (newEnergy < lastEnergy) {
    DS_NOTICE("deltaEnergy: energy counter is going backwards:", newEnergy, lastEnergy);
    return {};
  }
  auto deltaTime = std::chrono::duration_cast<std::chrono::milliseconds>(newTimestamp - lastTimestamp);
  if (deltaTime > 5s) {
    DS_NOTICE("deltaEnergy: too long delay between updates, setting new baseline:", deltaTime);
    return {};
  }

  if (newEnergy - lastEnergy < std::ceil(deltaTime.count() / 1000.0) * limit(m_directionType)) {
    return (newEnergy - lastEnergy);
  }

  DS_NOTICE("deltaEnergy: sensor value marked as outlier:", newEnergy, lastEnergy, deltaTime, limit(m_directionType),
            "not adding to db");
  return {};
}

unsigned int DirectionImpl::getApartmentPower() const {
  auto expired = [sensorValueTimeout = m_config.sensorValueTimeout, now = system_clock::now()](
                     const ds::DateTime& lastChanged) { return (now - lastChanged) > sensorValueTimeout; };

  return std::accumulate(std::cbegin(m_energyCache), std::cend(m_energyCache), 0,
                         [&expired](unsigned int sum, const auto& pair) {
                           const auto& value = std::get<ArchiveValue>(pair);
                           if (!expired(value.lastChanged)) {
                             sum += gsl::narrow_cast<unsigned int>(value.power); // return double?
                           }
                           return sum;
                         });
}

double DirectionImpl::getInstantPower(const dsuid_t& meterId) const {
  const auto cache = boost::find_if(m_energyCache, [&meterId](auto& x) { return x.first == meterId; });
  if (cache == m_energyCache.end()) {
    return 0;
  }

  auto expired = [sensorValueTimeout = m_config.sensorValueTimeout, now = system_clock::now()](
                     const ds::DateTime& lastChanged) { return (now - lastChanged) > sensorValueTimeout; };

  const auto& energy = cache->second;
  if (expired(energy.lastChanged)) {
    DS_WARNING("no update from", meterId, "for", ds::hideArg(m_config.sensorValueTimeout));
    return 0;
  }

  return energy.power;
}

bool DirectionImpl::tryGetValidArchiveFilename(const dsuid_t& meterId) const {
  if (isActiveMeter(meterId)) {
    return true;
  }

  auto archiveId = ArchiveId{meterId, m_directionType};
  if (m_archivesStorage.hasArchive(archiveId)) {
    boost::mutex::scoped_lock lock(m_cachedSeries_mutex);
    DS_NOTICE("add known archive", archiveId);
    m_meters.push_back(meterId);
    return true;
  }
  return false;
}

void DirectionImpl::getOrCreateCachedSeries(const dsuid_t& meterId) {
  if (tryGetValidArchiveFilename(meterId)) {
    return;
  }

  auto archiveId = ArchiveId{meterId, m_directionType};
  m_controllerArchives->create(archiveId);
  {
    boost::mutex::scoped_lock lock(m_cachedSeries_mutex);
    DS_NOTICE("add known archive", archiveId);
    m_meters.push_back(meterId);
  }
}

std::optional<ArchiveValue> DirectionImpl::getEnergy(const dsuid_t& meterId) const {
  DS_CONTEXT("getEnergy", ds::hideArg(m_directionType), ds::hideArg(meterId));
  if (auto cache = boost::find_if(m_energyCache, [&meterId](auto& x) { return x.first == meterId; });
      cache != m_energyCache.end()) {
    return cache->second;
  }
  if (auto fromDB = tryCreateCacheFromDB(meterId)) {
    DS_NOTICE(*fromDB);
    m_energyCache.push_back(std::make_pair(meterId, *fromDB));
    return fromDB;
  }
  return {};
}

auto convertValue = [](const std::optional<ArchiveValue>& archiveValue, dsModel::MeteringType meteringType,
                       EnergyUnit energyUnit) -> std::optional<LastReading> {
  if (!archiveValue) {
    return std::nullopt;
  }

  LastReading result{};
  switch (meteringType) {
    case dsModel::MeteringType::powerMetering:
    case dsModel::MeteringType::powerProducedMetering:
      result = LastReading{meteringType, archiveValue->power, archiveValue->lastChanged};
      break;
    case dsModel::MeteringType::energyMetering:
    case dsModel::MeteringType::energyProducedMetering:
      result = LastReading{meteringType, archiveValue->energy, archiveValue->lastChanged};
      if (energyUnit == EnergyUnit::Wh) {
        result.value /= 3600;
      }
      break;
    case dsModel::MeteringType::energyConsumedTariff1Metering: // dsmeter knows nothing about tariffs
    case dsModel::MeteringType::energyConsumedTariff2Metering:
    case dsModel::MeteringType::energyProducedTariff1Metering:
    case dsModel::MeteringType::energyProducedTariff2Metering:
    case dsModel::MeteringType::none:
    default: break;
  }
  return result;
};

std::optional<LastReading> DirectionImpl::getLastReading(const Dsuid& dsuid, dsModel::MeteringType meteringType,
                                                         EnergyUnit energyUnit) const {
  return convertValue(getEnergy(Dsuid{dsuid}), meteringType, energyUnit);
}

std::optional<LastReading> DirectionImpl::getLastReadingApartment(dsModel::MeteringType meteringType,
                                                                  EnergyUnit energyUnit) const {
  auto meterIds = [meterings = meterings()](dsModel::MeteringType meteringType) {
    std::vector<Dsuid> result;
    for (auto metering : meterings) {
      if (metering.second != meteringType) {
        continue;
      }
      result.push_back(metering.first);
    }
    return result;
  };

  auto accumulateControllers = [&](auto&& meterIds) {
    ArchiveValue result{};

    for (const auto& dsuid : meterIds) {
      if (auto reading = getEnergy(dsuid)) {
        result.power += reading->power;
        result.energy += reading->energy;
        result.lastChanged = ds::max(result.lastChanged, reading->lastChanged);
      }
    }
    return result;
  };

  auto ids = meterIds(meteringType);
  auto tmp = accumulateControllers(ids);
  return convertValue(tmp, meteringType, energyUnit);
}

void DirectionImpl::addEnergy(const dsuid_t& meterId, ElectricityMeasurement&& energyMeasurement) {
  DS_CONTEXT("addEnergy", ds::hideArg(m_directionType), ds::hideArg(meterId));
  auto deltaEnergy = deltaLast(meterId, energyMeasurement);
  DS_DEBUG("sensor:", ds::hideArg(energyMeasurement), deltaEnergy);
  if (!deltaEnergy) {
    // the first value sets only the baseline or delta against previous is too high to be real
    return;
  }

  auto energyMetering = getEnergy(meterId);
  if (!energyMetering) {
    energyMetering.emplace(ArchiveValue{.power = gsl::narrow_cast<double>(energyMeasurement.power),
                                        .energy = gsl::narrow_cast<double>(*deltaEnergy),
                                        .lastChanged = energyMeasurement.timestamp});
    m_energyCache.push_back(std::make_pair(meterId, *energyMetering));
    DS_NOTICE("new entry", energyMetering);
  } else {
    energyMetering->energy += *deltaEnergy;
    energyMetering->power = energyMeasurement.power;
    energyMetering->lastChanged = energyMeasurement.timestamp;
    boost::replace_if(
        m_energyCache, [&meterId](auto& x) { return x.first == meterId; }, std::make_pair(meterId, *energyMetering));
  }

  getOrCreateCachedSeries(meterId);
  DS_ASSERT(boost::find(m_meters, meterId) != m_meters.end());

  // TODO why can't we leave it as double, why do we need ceil
  auto archiveValue = ArchiveValue{
      .power = gsl::narrow_cast<double>(energyMeasurement.power),
      .energy = std::ceil(energyMetering->energy),
      .lastChanged = energyMeasurement.timestamp,
  };

  try {
    m_controllerArchives->addValue({meterId, m_directionType}, std::move(archiveValue));
  } catch (const ds::Exception& e) {
    DS_WARNING(m_lastPoll, archiveValue.lastChanged, e.what());
  }
}

std::optional<ArchiveValue> DirectionImpl::tryCreateCacheFromDB(const dsuid_t& meterId) const {
  // after restart addConsumed energy will skip this and start from 0 [Ws] each time
  if (!tryGetValidArchiveFilename(meterId)) {
    DS_NOTICE("tryCreateCacheFromDB: no valid archive file for", meterId);
    return std::nullopt;
  }

  return m_controllerArchives->lastValue({meterId, m_directionType});
}

ValueRange DirectionImpl::getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                                    bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) const {
  std::vector<ArchiveId> archiveIds;
  for (auto it = std::begin(meterIds); it != std::end(meterIds);) {
    if (tryGetValidArchiveFilename(*it)) {
      archiveIds.push_back({*it, m_directionType});
      it++;
    } else {
      DS_DEBUG("no data file", *it);
      it = meterIds.erase(it);
    }
  }

  DS_DEBUG(meterIds);
  if (archiveIds.empty()) {
    return {};
  }

  auto& archiveConfigs = m_config.timeSeries;
  auto it = std::find_if(std::rbegin(archiveConfigs), std::rend(archiveConfigs),
                         [resolution](const auto& element) { return element.resolution == resolution; });
  DS_REQUIRE(it != std::rend(archiveConfigs), "unsupported resolution");
  int numberOfValues = it->numberOfValues;

  // adjust limits by start/end/values
  _valueCount = std::min(_valueCount, numberOfValues);

  long unsigned int step = resolution.count();
  auto align = [step](time_t value) { return value - value % step; };

  // supported range by rra at given resolution, can't go further into the past
  auto now = system_clock::now();
  time_t maxEnd = align(system_clock::to_time_t(now));
  time_t maxStart = align(system_clock::to_time_t(now) - (step * numberOfValues));
  time_t start{}, end{};

  if (timeDuration == TimeDuration{}) {
    end = maxEnd;
    if (_valueCount == 0) {
      start = maxStart;
    } else {
      start = end - (step * _valueCount);
    }
  } else if (!timeDuration.start) {
    end = align(system_clock::to_time_t(*timeDuration.end));
    if (_valueCount == 0) {
      start = maxStart;
    } else {
      start = end - (step * _valueCount);
    }
  } else if (!timeDuration.end) {
    start = align(system_clock::to_time_t(*timeDuration.start));
    if (_valueCount == 0) {
      end = maxEnd;
    } else {
      end = start + (step * _valueCount);
    }
  } else {
    start = align(system_clock::to_time_t(*timeDuration.start));
    end = align(system_clock::to_time_t(*timeDuration.end));
  }
  end += step;

  m_controllerArchives->flush(archiveIds);
  auto series = m_controllerArchives->xport(start, end, step, archiveIds, _type, _energyInWh);
  auto returnVector = std::move(series.values);

  bool lastValueEmpty = false;
  if (!returnVector.empty() && returnVector.back().value == 0) { // -> unknown
    DS_DEBUG("last value is unknown:", returnVector.back());
    lastValueEmpty = true;
  }

  if (_type == SeriesTypes::energy) {
    double currentCounter = 0.0;
    for (auto meterId : meterIds) {
      if (auto energyMetering = getEnergy(meterId)) {
        currentCounter += energyMetering->energy;
      }
    }
    if (_energyInWh) {
      currentCounter /= 3600;
    }
    auto lastZeroValue = Value{0, ds::DateTime{0s}};
    for (auto& data : returnVector) {
      if (data.value == 0) {
        lastZeroValue = data;
      } else {
        break;
      }
    }
    for (auto iter = returnVector.rbegin(); iter < returnVector.rend(); ++iter) {
      if (iter->timeStamp == lastZeroValue.timeStamp) {
        break;
      }
      double val = std::exchange(iter->value, currentCounter);
      currentCounter -= val;
    }
  }

  if (lastValueEmpty) {
    // delete the last value, if it is Unknown (zero)
    // assumption: avoid return to zero of graph if current value did not arrive yet
    returnVector.pop_back();
  }

  return {.start = series.start, .end = series.end, .step = series.step, .values = returnVector};
}

} // namespace

std::unique_ptr<ControllerDirection> makeControllerDirection(Apartment& apartment,
                                                             ArchivesConnection& archivesConnection,
                                                             ArchivesStorage& archivesStorage, Category category,
                                                             ds::System& system) {
  auto controllerArchives = archives::makeControllerArchive(archivesConnection, archivesStorage, system);
  return std::make_unique<DirectionImpl>(apartment, std::move(controllerArchives), archivesStorage, category, system);
}

std::unique_ptr<ControllerDirection> makeMockedControllerDirection(
    Apartment& apartment, std::unique_ptr<archives::ControllerArchives>&& controllerArchives,
    ArchivesStorage& archivesStorage, Category category, ds::System& system) {
  return std::make_unique<DirectionImpl>(apartment, std::move(controllerArchives), archivesStorage, category, system);
}

} // namespace dss::metering
