/*
    Copyright (c) 2009,2010,2011,2022 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "metering.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <digitalSTROM/dsuid.h>
#include <ds/asio/io-service.h>
#include <ds/asio/timer.h>
#include <ds/log.h>
#include <sys/types.h>
#include <time.h>

#include <algorithm>
#include <boost/range/algorithm.hpp>
#include <cmath>
#include <filesystem>
#include <functional>
#include <iterator>
#include <numeric>

#include "businterface.h"
#include "dss.h"
#include "metering/archives-connection.h"
#include "metering/archives-sanity.h"
#include "metering/archives-storage.h"
#include "metering/config.h"
#include "metering/controller-direction.h"
#include "metering/open-metering-system.h"
#include "metering/power-states.h"
#include "metering/readings/open-metering-system.h"
#include "metering/rrd-interface.h"
#include "metering/socket-server.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "propertysystem.h"
#include "security/security.h"

static ds::LogChannel dsLogChannel("dssMeteringMetering");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

using system_clock = std::chrono::system_clock;

namespace dss::metering {

namespace {

auto toCategory = [](dsModel::MeteringType meteringType) {
  switch (meteringType) {
    case dsModel::MeteringType::powerMetering:
    case dsModel::MeteringType::energyMetering:
    case dsModel::MeteringType::energyConsumedTariff1Metering:
    case dsModel::MeteringType::energyConsumedTariff2Metering: return Category::consumption;
    case dsModel::MeteringType::energyProducedMetering:
    case dsModel::MeteringType::powerProducedMetering:
    case dsModel::MeteringType::energyProducedTariff1Metering:
    case dsModel::MeteringType::energyProducedTariff2Metering: return Category::production;
    case dsModel::MeteringType::none:
    default: DS_FAIL_REQUIRE("bad metering type");
  }
};

class MeteringImpl : public MeteringSubsystem {
  Apartment& m_apartment;
  ds::System& m_system;
  ElectricityMeteringConfig m_electricityMeteringConfig;
  MeteringBusInterface* m_pMeteringBusInterface = NULL;

  std::shared_ptr<ArchivesConnection> m_archivesConnection;
  ArchivesStorage& m_archivesStorage;
  std::shared_ptr<ArchivesSanity> m_archivesSanity;
  std::shared_ptr<PowerStates> m_powerStates = makePowerStates(m_apartment);
  std::unique_ptr<ControllerDirection> m_consumption;
  std::unique_ptr<ControllerDirection> m_production;
  std::unique_ptr<metering::OpenMeteringSystem> m_openMeteringSystem;
  ds::Clock::time_point m_nextPollAt;

public:
  void flushDb(const ds::DateTime& now = std::chrono::system_clock::now()) const override;
  void validateAndConvertArchives() const override;

  ArchivesStorage& getArchivesStorage() const override { return m_archivesStorage; }
  ArchivesSanity& getArchivesSanity() const override { return *m_archivesSanity; }

  void doStart() override;
  void execute() override;
  void initialize() override;
  void pollLoop(boost::asio::io_service& io, ds::asio::Timer& timer);

public:
  MeteringImpl(DSS* _pDSS, Apartment& apartment, std::shared_ptr<ArchivesConnection> archivesConnection,
               std::unique_ptr<ArchivesSanity>&& archivesSanity, std::unique_ptr<ControllerDirection>&& consumption,
               std::unique_ptr<ControllerDirection>&& production, std::unique_ptr<metering::OpenMeteringSystem>&&,
               ds::System& system);
  ~MeteringImpl() override = default;

  const std::vector<TimeSeriesConfig> getConfig() const override { return m_electricityMeteringConfig.timeSeries; }
  void setMeteringBusInterface(MeteringBusInterface* _value) override { m_pMeteringBusInterface = _value; }

  std::optional<unsigned long> getConsumedPower(const dsuid_t& meterId) const override {
    return gsl::narrow<unsigned int>(m_consumption->getInstantPower(meterId));
  }
  unsigned int getConsumedPowerApartment() const override { return m_consumption->getApartmentPower(); }

  std::optional<unsigned long> getProducedPower(const dsuid_t& meterId) const override {
    return gsl::narrow<unsigned int>(m_production->getInstantPower(meterId));
  }
  unsigned int getProducedPowerApartment() const override { return m_production->getApartmentPower(); }

  std::optional<ArchiveValue> getEnergy(const dsuid_t& meterId) const override {
    return m_consumption->getEnergy(meterId);
  }
  std::optional<ArchiveValue> getProducedEnergy(const dsuid_t& meterId) const override {
    return m_production->getEnergy(meterId);
  }

  double get15minConsumed(std::vector<Dsuid> meterIds, SeriesTypes type) override;
  double get15minProduced(std::vector<Dsuid> meterIds, SeriesTypes type) override;

  TotalEnergy getLastEnergy(const dsuid_t& meterId) const override {
    return {m_production->getLastEnergy(meterId), m_consumption->getLastEnergy(meterId)};
  }

  void checkInitialState(MeteringSystem meteringSystem) {
    using dsModel::ApartmentMeteringOrigin;
    using dsModel::MeteringType;
    switch (meteringSystem) {
      case MeteringSystem::openMeteringSystem:
        if (m_openMeteringSystem->activeMeterId()) {
          auto consumed = m_openMeteringSystem->getLastReading(OpenMeteringSystemId{}, MeteringType::powerMetering,
                                                               EnergyUnit::Ws); // power not energy
          auto produced = m_openMeteringSystem->getLastReading(OpenMeteringSystemId{},
                                                               MeteringType::powerProducedMetering, EnergyUnit::Ws);
          DS_REQUIRE(consumed && produced);
          triggerState(MeteringSystem::openMeteringSystem,
                       TotalPower{.production = gsl::narrow_cast<unsigned int>(produced->value),
                                  .consumption = gsl::narrow_cast<unsigned int>(consumed->value)});
        }
        break;
      case MeteringSystem::apartmentSummation: {
        auto consumed = getLastReading(ApartmentMeteringOrigin{}, MeteringType::powerMetering,
                                       EnergyUnit::Ws); // power not energy
        auto produced = getLastReading(ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering, EnergyUnit::Ws);
        DS_REQUIRE(consumed && produced);
        triggerState(MeteringSystem::apartmentSummation,
                     TotalPower{.production = gsl::narrow_cast<unsigned int>(produced->value),
                                .consumption = gsl::narrow_cast<unsigned int>(consumed->value)});
      } break;

      case MeteringSystem::digitalStromMeter:
        // dsm power states are saved/retrieved on the dSM itself, createState is not involved
        DS_NOTICE("checkInitialState not implemented for", meteringSystem);
        break;
    }
  }

  std::shared_ptr<State> createState(MeteringSystem meteringSystem, PowerStateConfig&& powerStateConfig) override {
    auto state = m_powerStates->createState(meteringSystem, std::move(powerStateConfig));
    if (state) {
      checkInitialState(meteringSystem);
    }
    return state;
  }
  void removeState(MeteringSystem meteringSystem, const PowerStateConfig& powerStateConfig) override {
    return m_powerStates->removeState(meteringSystem, powerStateConfig);
  }
  void triggerState(MeteringSystem meteringSystem, TotalPower&& totalPower) override {
    m_powerStates->trigger(meteringSystem, std::move(totalPower));
  }

  void addConsumedEnergy(DSMeter& meter, ElectricityMeasurement&& energyEvent) override;
  void addProducedEnergy(DSMeter& meter, ElectricityMeasurement&& energyMeasurement) override {
    m_production->addEnergy(meter.getDSID(), std::move(energyMeasurement));
  }

  void addReading(OpenMeteringSystem::Reading&& p1Reading) override {
    m_openMeteringSystem->addReading(std::move(p1Reading));
  }

  ValueRange getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                       bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) override {
    return m_consumption->getSeries(meterIds, resolution, _type, _energyInWh, timeDuration, _valueCount);
  }

  ValueRange getProducedSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                               bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) override {
    return m_production->getSeries(meterIds, resolution, _type, _energyInWh, timeDuration, _valueCount);
  }

  std::optional<dsModel::OpenMeteringSystemOrigin> openMeteringSystemOrigin() const override {
    if (auto activeMeterId = m_openMeteringSystem->activeMeterId(); activeMeterId) {
      return dsModel::OpenMeteringSystemOrigin{.serialNumber = *activeMeterId};
    }
    return std::nullopt;
  }

  std::vector<std::pair<dsModel::MeteringOrigin, dsModel::MeteringType>> meterings() const override {
    std::vector<std::pair<dsModel::MeteringOrigin, dsModel::MeteringType>> result;

    auto toControllerMeteringOrigin = [](const dsuid_t& dsuid) {
      return dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{ds::str(dsuid)}};
    };

    bool hasConsumption{false};
    for (const auto& [meterId, meteringType] : m_consumption->meterings()) {
      result.push_back({toControllerMeteringOrigin(meterId), meteringType});
      hasConsumption = true;
    }

    bool hasProduction{false};
    for (const auto& [meterId, meteringType] : m_production->meterings()) {
      result.push_back({toControllerMeteringOrigin(meterId), meteringType});
      hasProduction = true;
    }

    for (const auto& [id, type] : m_openMeteringSystem->meterings()) {
      result.push_back({dsModel::OpenMeteringSystemOrigin{.serialNumber = id}, type}); // add technicalName
    }

    if (hasConsumption) {
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::powerMetering});
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::energyMetering});
    }

    if (hasProduction) {
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::powerProducedMetering});
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::energyProducedMetering});
    }

    return result;
  }

  std::optional<LastReading> getLastReading(dsModel::MeteringOrigin meteringOrigin, dsModel::MeteringType meteringType,
                                            EnergyUnit energyUnit) const override {
    std::optional<LastReading> result{};

    std::visit(::ds::Visitor{[&](const std::monostate& origin) { DS_FAIL_REQUIRE("monostate"); },
                             [&](dsModel::ApartmentMeteringOrigin& origin) {
                               if (toCategory(meteringType) == Category::production) {
                                 result = m_production->getLastReadingApartment(meteringType, energyUnit);
                               } else {
                                 result = m_consumption->getLastReadingApartment(meteringType, energyUnit);
                               }
                             },
                             [&, this](dsModel::ControllerMeteringOrigin& origin) {
                               if (toCategory(meteringType) == Category::production) {
                                 result = m_production->getLastReading(Dsuid{origin.id}, meteringType, energyUnit);
                               } else {
                                 result = m_consumption->getLastReading(Dsuid{origin.id}, meteringType, energyUnit);
                               }
                             },
                             [&, this](dsModel::OpenMeteringSystemOrigin origin) {
                               // serialNumber is optional from rest interface
                               result = m_openMeteringSystem->getLastReading(
                                   std::move(OpenMeteringSystemId{origin.serialNumber}), meteringType, energyUnit);
                             }},
               meteringOrigin);

    return result;
  }

  ValueRange getSeriesValue(dsModel::MeteringOrigin meteringOrigin, dsModel::MeteringType meteringType,
                            std::chrono::seconds resolution, int valueCount, EnergyUnit energyUnit) const override {
    ValueRange result{};

    auto toSeriesType = [](dsModel::MeteringType meteringType) {
      switch (meteringType) {
        case dsModel::MeteringType::powerMetering:
        case dsModel::MeteringType::powerProducedMetering: return SeriesTypes::power;
        case dsModel::MeteringType::energyMetering:
        case dsModel::MeteringType::energyProducedMetering:
        case dsModel::MeteringType::energyConsumedTariff1Metering:
        case dsModel::MeteringType::energyConsumedTariff2Metering:
        case dsModel::MeteringType::energyProducedTariff1Metering:
        case dsModel::MeteringType::energyProducedTariff2Metering: return SeriesTypes::energyDelta;
        case dsModel::MeteringType::none:
        default: DS_FAIL_REQUIRE("bad metering type");
      }
    };

    std::visit(
        ::ds::Visitor{[&](const std::monostate& origin) { DS_FAIL_REQUIRE("monostate"); },
                      [&](dsModel::ApartmentMeteringOrigin& origin) { DS_FAIL_REQUIRE("not implemented"); },
                      [&, this](dsModel::ControllerMeteringOrigin& origin) {
                        if (toCategory(meteringType) == Category::production) {
                          result = m_production->getSeries({Dsuid{origin.id}}, resolution, toSeriesType(meteringType),
                                                           energyUnit == EnergyUnit::Wh, TimeDuration{}, valueCount);
                        } else {
                          result = m_consumption->getSeries({Dsuid{origin.id}}, resolution, toSeriesType(meteringType),
                                                            energyUnit == EnergyUnit::Wh, TimeDuration{}, valueCount);
                        }
                      },
                      [&, this](dsModel::OpenMeteringSystemOrigin& origin) {
                        result = m_openMeteringSystem->getSeries(meteringType, toSeriesType(meteringType), resolution,
                                                                 TimeDuration{}, valueCount, energyUnit);
                      }},
        meteringOrigin);
    return result;
  }
}; // MeteringImpl

MeteringImpl::MeteringImpl(DSS* _pDSS, Apartment& apartment, std::shared_ptr<ArchivesConnection> archivesConnection,
                           std::unique_ptr<ArchivesSanity>&& archivesSanity,
                           std::unique_ptr<ControllerDirection>&& consumption,
                           std::unique_ptr<ControllerDirection>&& production,
                           std::unique_ptr<OpenMeteringSystem>&& openMeteringSystem, ds::System& system)
    : MeteringSubsystem(_pDSS),
      m_apartment(apartment),
      m_system(system),
      m_archivesConnection(std::move(archivesConnection)),
      m_archivesStorage(m_archivesConnection->getArchivesStorage()),
      m_archivesSanity(std::move(archivesSanity)),
      m_consumption(std::move(consumption)),
      m_production(std::move(production)),
      m_openMeteringSystem(std::move(openMeteringSystem)) {}

void MeteringImpl::addConsumedEnergy(DSMeter& meter, ElectricityMeasurement&& energyMeasurement) {
  auto meteringCache = MeteringCache{
      .power = gsl::narrow<int>(energyMeasurement.power), // property tree exported
      .lastSensorValue = gsl::narrow_cast<unsigned long>(energyMeasurement.energy),
      .lastChanged = dss::DateTime{energyMeasurement.timestamp},
  };

  m_consumption->addEnergy(meter.getDSID(), std::move(energyMeasurement));

  if (auto energyMetering = m_consumption->getEnergy(meter.getDSID())) {
    meteringCache.energyWs = energyMetering->energy;
    meteringCache.energyWh = energyMetering->energy / 3600;

    // backend data for property tree, etc
    meter.updateMeteringCache(std::move(meteringCache));
  }
}

double MeteringImpl::get15minConsumed(std::vector<Dsuid> meterIds, SeriesTypes type) {
  return m_consumption->get15MinuteAverage(meterIds, type, Unit::Wh);
}

double MeteringImpl::get15minProduced(std::vector<Dsuid> meterIds, SeriesTypes type) {
  return m_production->get15MinuteAverage(meterIds, type, Unit::Wh);
}

// subsystem interface
void MeteringImpl::initialize() {
  Subsystem::initialize();
  auto config = getDSS().getPropertySystem().createProperty(getConfigPropertyBasePath());
  m_archivesConnection->setRRDCachedPath(
      config->getOrCreateChildValue<std::string>("rrdDaemonAddress", "unix:/var/run/rrdcached.sock"));

  auto storageLocation =
      config->getOrCreateChildValue<std::string>("storageLocation", getDSS().getDataDirectory() + "metering/");
  storageLocation = addTrailingBackslash(storageLocation);
  DS_NOTICE(storageLocation);

  if (isEnabled()) {
    if (!std::filesystem::is_directory(storageLocation)) {
      throw std::runtime_error("Metering directory " + std::filesystem::canonical(storageLocation).string() +
                               " does not exist!");
    }

    if (!rwAccess(storageLocation)) {
      throw std::runtime_error("Metering directory " + std::filesystem::canonical(storageLocation).string() +
                               " is not readable/writable!");
    }
  }

  m_electricityMeteringConfig = m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {});
  DS_NOTICE("config", ds::hideArg(m_electricityMeteringConfig));
  m_archivesStorage.setMeteringStorageLocation(std::string(storageLocation));
}

void MeteringImpl::validateAndConvertArchives() const {
  for (auto element : m_archivesStorage.searchArchives()) {
    const auto& compoundId = std::get<CompoundId>(element);
    DS_NOTICE("checking:", compoundId);

    std::visit(ds::Visitor{[](const std::monostate&) { DS_FAIL_REQUIRE("monostate"); },
                           [&](const dsid_t& dsid) { DS_WARNING("unexpected", dsid); },
                           [&](const ArchiveId& archiveId) {
                             if (archiveId.second == Category::consumption) {
                               m_consumption->validateAndConvertArchive(archiveId);
                             } else {
                               m_production->validateAndConvertArchive(archiveId);
                             }
                           },
                           [&](const OpenMeteringSystemId& openMeteringSystemId) {
                             DS_NOTICE("not implemented", openMeteringSystemId);
                           }},
               compoundId);
  }
}

void MeteringImpl::flushDb(const ds::DateTime& now) const {
  std::vector<std::string> flushArchives = m_consumption->needsFlushing(now);
  boost::copy(m_production->needsFlushing(now), std::back_inserter(flushArchives));

  if (!flushArchives.empty()) {
    m_archivesConnection->flush(flushArchives);
  }
}

void MeteringImpl::pollLoop(boost::asio::io_service& io, ds::asio::Timer& timer) {
  if (m_Terminated) {
    DS_NOTICE("pollLoop: terminating");
    return;
  }

  auto startedAt = ds::Clock::now();
  if (m_nextPollAt != std::chrono::floor<std::chrono::seconds>(startedAt)) {
    DS_NOTICE("pollLoop: missed wakeup", startedAt, m_nextPollAt, ds::hideArg(startedAt - m_nextPollAt));
  }

  DS_ASSERT(m_pMeteringBusInterface != NULL, "");
  try {
    auto now = system_clock::now();
    m_consumption->pollTimestamp(now);
    m_production->pollTimestamp(now);
    m_pMeteringBusInterface->requestMeterData();
  } catch (dss::BusApiError& apiError) {
    log("Metering::execute: Couldn't get metering data: " + std::string(apiError.what()));
  }

  triggerState(MeteringSystem::apartmentSummation,
               {.production = m_production->getApartmentPower(), .consumption = m_consumption->getApartmentPower()});

  auto restartCached = m_archivesSanity->removeDeadArchives();
  if (restartCached) {
    // data loss, but rrdcached might be in busy loop
    m_archivesConnection->restartCached();
  } else {
    flushDb();
  }

  auto now = ds::Clock::now();
  if (now - startedAt > 1s) {
    DS_WARNING("pollLoop: exceeded 1s budget", now - startedAt);
  }

  m_nextPollAt = std::chrono::ceil<std::chrono::seconds>(now);
  timer.expiresFromNow(m_nextPollAt - now, [this, &io, &timer] { pollLoop(io, timer); });
}

void MeteringImpl::execute() {
  while (DSS::getInstance()->getModelMaintenance().isInitializing()) {
    sleepSeconds(1);
  }
  DS_NOTICE("pollLoop: modelReady received");

  m_archivesSanity->migrateFilenames();
  m_archivesSanity->removeDeadArchives();
  m_archivesSanity->restoreBackups(); // user uploaded configuration backup with included rrd
  DS_NOTICE("found", ds::hideArg(m_archivesStorage.searchArchives().size()), "archives");

  validateAndConvertArchives();

  ds::asio::IoService ioService; // `ds::asio::IoService` requires its construct to run in the same thread
  boost::asio::io_service::work work(ioService); // block in `run()` until `m_stopping`

  ds::asio::Timer timer(ioService);
  pollLoop(ioService, timer);
  auto socketServer = makeSocketServer(ioService, m_electricityMeteringConfig.socketPath, *this);
  socketServer->start();

  while (!m_Terminated) {
    try {
      ioService.run();
      break; // run() exited normally
    } catch (std::exception& e) {
      DS_ERROR("io_service", e.what());
    }
  }
}

void MeteringImpl::doStart() {
  DS_NOTICE("Writing files to:", ds::hideArg(m_archivesStorage.storageLocation()));
  run();
} // start

} // namespace

std::shared_ptr<MeteringSubsystem> makeMetering(DSS* dss, Apartment& apartment) {
  auto archivesConnection = makeArchivesConnection(makeRRDInterface(), ds::system());
  auto archivesSanity = makeArchivesSanity(*archivesConnection, archivesConnection->getArchivesStorage(), ds::system());

  auto& archivesStorage = archivesConnection->getArchivesStorage();
  auto consumption =
      makeControllerDirection(apartment, *archivesConnection, archivesStorage, Category::consumption, ds::system());
  auto production =
      makeControllerDirection(apartment, *archivesConnection, archivesStorage, Category::production, ds::system());

  auto openMeteringSystem = makeOpenMeteringSystem(*archivesConnection, archivesConnection->getArchivesStorage());

  return std::make_shared<MeteringImpl>(dss, apartment, archivesConnection, std::move(archivesSanity),
                                        std::move(consumption), std::move(production), std::move(openMeteringSystem),
                                        ds::system());
}

std::shared_ptr<MeteringSubsystem> makeMockedMetering(Apartment& apartment,
                                                      std::shared_ptr<ArchivesConnection> archivesConnection,
                                                      std::unique_ptr<ArchivesSanity>&& archivesSanity,
                                                      std::unique_ptr<ControllerDirection>&& consumption,
                                                      std::unique_ptr<ControllerDirection>&& production,
                                                      std::unique_ptr<OpenMeteringSystem>&& openMeteringSystem,
                                                      ds::System& system) {
  return std::make_shared<MeteringImpl>(nullptr, apartment, archivesConnection, std::move(archivesSanity),
                                        std::move(consumption), std::move(production), std::move(openMeteringSystem),
                                        system);
}

} // namespace dss::metering
