/*
    Copyright (c) 2009,2010,2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <digitalSTROM/dsuid.h>
#include <ds/reflect.h>
#include <ds/system-fwd.h>

#include <memory>
#include <string>
#include <vector>

#include "ds-model/meterings.h"
#include "dssfwd.h"
#include "metering/archives-fwd.h" // TODO rework unit tests
#include "metering/config.h"
#include "metering/metering-fwd.h"
#include "metering/power-states-fwd.h"
#include "subsystem.h"

DS_REFLECTED_ENUM((dss::metering), Unit, (W)(Ws)(Wh));

namespace dss {

struct Dsuid;
class DSMeter;
class MeteringBusInterface;

namespace metering {

namespace readings {
struct OpenMeteringSystem;
}

struct LastReading {
  dsModel::MeteringType type;
  double value;
  ds::DateTime lastChanged;
};

struct Value {
  double value;
  ds::DateTime timeStamp;
};

struct ValueRange {
  time_t start;
  time_t end;
  unsigned long step;
  std::vector<Value> values;
};

struct TimeDuration {
  std::optional<ds::DateTime> start;
  std::optional<ds::DateTime> end;
};

// rrd uses floating point internally
struct ArchiveValue {
  double power{0};
  double energy{0}; // [Ws]
  ds::DateTime lastChanged;
};

struct ElectricityMeasurement {
  double power{0};
  double energy{0}; // [Ws]
  ds::DateTime timestamp;
};

struct TotalEnergy {
  double production;
  double consumption;
  std::chrono::seconds resolution;
};

class Metering {
public:
  virtual ~Metering() = default;

  virtual const std::vector<TimeSeriesConfig> getConfig() const = 0;
  virtual void setMeteringBusInterface(MeteringBusInterface* _value) = 0;

  virtual void addConsumedEnergy(DSMeter& meter, ElectricityMeasurement&& energyMeasurement) = 0;
  virtual void addProducedEnergy(DSMeter& meter, ElectricityMeasurement&& energyMeasurement) = 0;

  virtual void addReading(metering::readings::OpenMeteringSystem&& p1Reading) = 0;

  // power is 0 if last value older then sensorValueTimeout
  virtual std::optional<unsigned long> getConsumedPower(const dsuid_t& dsuid) const = 0;
  virtual std::optional<unsigned long> getProducedPower(const dsuid_t& dsuid) const = 0;

  virtual unsigned int getConsumedPowerApartment() const = 0;
  virtual unsigned int getProducedPowerApartment() const = 0;

  virtual std::optional<ArchiveValue> getEnergy(const dsuid_t& dsuid) const = 0;
  virtual std::optional<ArchiveValue> getProducedEnergy(const dsuid_t& dsuid) const = 0;

  // populate the serial number of the open-metering-system
  virtual std::optional<dsModel::OpenMeteringSystemOrigin> openMeteringSystemOrigin() const = 0;

  virtual std::vector<std::pair<dsModel::MeteringOrigin, dsModel::MeteringType>> meterings() const = 0;

  // TODO EnergyUnit::Wh is what the UIs show
  virtual std::optional<LastReading> getLastReading(dsModel::MeteringOrigin meteringOrigin,
                                                    dsModel::MeteringType meteringType,
                                                    EnergyUnit = EnergyUnit::Ws) const = 0;
  virtual TotalEnergy getLastEnergy(const dsuid_t& meterId) const = 0;

  virtual ValueRange getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                               bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) = 0;
  virtual ValueRange getProducedSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                                       bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) = 0;

  virtual ValueRange getSeriesValue(dsModel::MeteringOrigin meteringOrigin, dsModel::MeteringType meteringType,
                                    std::chrono::seconds resolution, int valueCount = 0,
                                    EnergyUnit energyUnit = EnergyUnit::Wh) const = 0;

  virtual double get15minConsumed(std::vector<Dsuid> meterIds, SeriesTypes type) = 0;
  virtual double get15minProduced(std::vector<Dsuid> meterIds, SeriesTypes type) = 0;

  virtual std::shared_ptr<State> createState(MeteringSystem meteringSystem, PowerStateConfig&& powerStateConfig) = 0;
  virtual void removeState(MeteringSystem meteringSystem, const PowerStateConfig& powerStateConfig) = 0;
  virtual void triggerState(MeteringSystem meteringSystem, TotalPower&& totalPower) = 0;

  // unit tests
  virtual void validateAndConvertArchives() const = 0;
  virtual void flushDb(const ds::DateTime& now) const = 0;

  // dsstests
  virtual ArchivesStorage& getArchivesStorage() const = 0;
  virtual ArchivesSanity& getArchivesSanity() const = 0;
};

class MeteringSubsystem : public Metering, public ThreadedSubsystem {
public:
  MeteringSubsystem(DSS* dSS) : Metering(), ThreadedSubsystem(dSS, "Metering") {}
  Metering& getMetering() { return *this; }
};

std::shared_ptr<MeteringSubsystem> makeMetering(DSS* _pDSS, Apartment& m_apartment);
std::shared_ptr<MeteringSubsystem> makeMockedMetering(Apartment& apartment,
                                                      std::shared_ptr<ArchivesConnection> archivesConnection,
                                                      std::unique_ptr<ArchivesSanity>&& archivesSanity,
                                                      std::unique_ptr<ControllerDirection>&& consumption,
                                                      std::unique_ptr<ControllerDirection>&& production,
                                                      std::unique_ptr<OpenMeteringSystem>&& openMeteringSystem,
                                                      ds::System&);

} // namespace metering
} // namespace dss

DS_REFLECT_STRUCT((dss::metering), LastReading, (type)(value)(lastChanged));
DS_REFLECT_STRUCT((dss::metering), ArchiveValue, (power)(energy)(lastChanged));
DS_REFLECT_STRUCT((dss::metering), ElectricityMeasurement, (power)(energy)(timestamp));
DS_REFLECT_STRUCT((dss::metering), TimeDuration, (start)(end));
DS_REFLECT_STRUCT((dss::metering), Value, (value)(timeStamp));
DS_REFLECT_STRUCT((dss::metering), ValueRange, (start)(end)(step)(values));
DS_REFLECT_STRUCT((dss::metering), TotalEnergy, (production)(consumption)(resolution));
