/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "open-metering-system.h"

#include <ds/catch/catch.h>
#include <ds/log.h>
#include <ds/system-mock.h>

#include "metering/archives-storage-mock.h"
#include "metering/archives/open-metering-system-mock.h"
#include "metering/config.h"

static ds::LogChannel dsLogChannel("dssMeteringOpenMeteringSystemTest");
static const char* TAGS = "[dSS][metering][open-metering-system]";

namespace {

using namespace dss;
using namespace dss::metering;
using namespace dss::metering::archives;

using namespace trompeloeil;

using dsModel::MeteringType;
using MeterId = readings::MeterId;
using Reading = OpenMeteringSystem::Reading;

SCENARIO("OpenMeteringsSystem", TAGS) {
  ArchivesStorageMock storage{};
  auto openMeteringSystemArchive = std::make_unique<OpenMeteringSystemArchiveMock>();
  auto& archives{*openMeteringSystemArchive};
  ds::SystemMock system{};

  auto openMeteringSystem = makeOpenMeteringSystemMocked(std::move(openMeteringSystemArchive), storage, system);

  OpenMeteringSystemId openMeteringSystemId{"foobar3000"};

  SECTION("addReading") {
    WHEN("archive exists") {
      REQUIRE_CALL(storage, hasArchive(openMeteringSystemId)).RETURN(true);
      REQUIRE_CALL(archives, addValue(openMeteringSystemId, _));

      openMeteringSystem->addReading(Reading{.id = MeterId{openMeteringSystemId}});
    }
    WHEN("archive does not exist") {
      REQUIRE_CALL(storage, hasArchive(openMeteringSystemId)).RETURN(false);
      REQUIRE_CALL(archives, create(openMeteringSystemId, _));
      REQUIRE_CALL(archives, addValue(openMeteringSystemId, _));

      openMeteringSystem->addReading(Reading{.id = MeterId{openMeteringSystemId}});
    }
  }

  SECTION("meterings") {
    WHEN("archive exists") {
      REQUIRE_CALL(storage, searchOmsArchives()).RETURN(std::vector{openMeteringSystemId});
      CHECK(openMeteringSystem->meterings() ==
            std::vector{std::pair{openMeteringSystemId, MeteringType::powerMetering},
                        std::pair{openMeteringSystemId, MeteringType::energyMetering},
                        std::pair{openMeteringSystemId, MeteringType::energyConsumedTariff1Metering},
                        std::pair{openMeteringSystemId, MeteringType::energyConsumedTariff2Metering},
                        std::pair{openMeteringSystemId, MeteringType::powerProducedMetering},
                        std::pair{openMeteringSystemId, MeteringType::energyProducedMetering},
                        std::pair{openMeteringSystemId, MeteringType::energyProducedTariff1Metering},
                        std::pair{openMeteringSystemId, MeteringType::energyProducedTariff2Metering}});

      THEN("getLastReading works") {
        REQUIRE_CALL(archives, flush(_));
        REQUIRE_CALL(archives, lastValue(openMeteringSystemId)).RETURN(Reading{.lastChanged = ds::DateTime{300s}});
        CHECK(openMeteringSystem->getLastReading(OpenMeteringSystemId{}, MeteringType::powerMetering, EnergyUnit::Wh) ==
              LastReading{.type = MeteringType::powerMetering, .lastChanged = ds::DateTime{300s}});
      }

      THEN("activeMeterId matches") { CHECK(openMeteringSystem->activeMeterId() == openMeteringSystemId); }

      WHEN("meter is switched") {
        OpenMeteringSystemId openMeteringSystemId2{"foobar300x"};

        REQUIRE_CALL(storage, hasArchive(openMeteringSystemId2)).RETURN(false);
        REQUIRE_CALL(archives, create(openMeteringSystemId2, _));
        REQUIRE_CALL(archives, addValue(openMeteringSystemId2, _));
        openMeteringSystem->addReading(Reading{.id = MeterId{openMeteringSystemId2}});

        THEN("new meter becomes active meter") {
          CHECK(openMeteringSystem->activeMeterId() == openMeteringSystemId2);
          CHECK(openMeteringSystem->meterings() ==
                std::vector{std::pair{openMeteringSystemId2, MeteringType::powerMetering},
                            std::pair{openMeteringSystemId2, MeteringType::energyMetering},
                            std::pair{openMeteringSystemId2, MeteringType::energyConsumedTariff1Metering},
                            std::pair{openMeteringSystemId2, MeteringType::energyConsumedTariff2Metering},
                            std::pair{openMeteringSystemId2, MeteringType::powerProducedMetering},
                            std::pair{openMeteringSystemId2, MeteringType::energyProducedMetering},
                            std::pair{openMeteringSystemId2, MeteringType::energyProducedTariff1Metering},
                            std::pair{openMeteringSystemId2, MeteringType::energyProducedTariff2Metering}});
        }
      }
    }
    WHEN("archive does not exist") {
      REQUIRE_CALL(storage, searchOmsArchives()).RETURN(std::vector<OpenMeteringSystemId>{});
      CHECK(openMeteringSystem->meterings() == std::vector<std::pair<OpenMeteringSystemId, MeteringType>>{});

      THEN("id is unknown and getLastReading fails") {
        CHECK(openMeteringSystem->activeMeterId() == std::nullopt);
        CHECK_THROWS(
            openMeteringSystem->getLastReading(OpenMeteringSystemId{}, MeteringType::powerMetering, EnergyUnit::Wh));
      }

      WHEN("reading is added") {
        REQUIRE_CALL(storage, hasArchive(openMeteringSystemId)).RETURN(true);
        REQUIRE_CALL(archives, addValue(openMeteringSystemId, _));
        openMeteringSystem->addReading(Reading{.id = MeterId{openMeteringSystemId}});

        THEN("meterings are populated") {
          CHECK(openMeteringSystem->meterings() ==
                std::vector{std::pair{openMeteringSystemId, MeteringType::powerMetering},
                            std::pair{openMeteringSystemId, MeteringType::energyMetering},
                            std::pair{openMeteringSystemId, MeteringType::energyConsumedTariff1Metering},
                            std::pair{openMeteringSystemId, MeteringType::energyConsumedTariff2Metering},
                            std::pair{openMeteringSystemId, MeteringType::powerProducedMetering},
                            std::pair{openMeteringSystemId, MeteringType::energyProducedMetering},
                            std::pair{openMeteringSystemId, MeteringType::energyProducedTariff1Metering},
                            std::pair{openMeteringSystemId, MeteringType::energyProducedTariff2Metering}});
        }
        THEN("activeMeterId matches") { CHECK(openMeteringSystem->activeMeterId() == openMeteringSystemId); }
      }
    }
  }

  SECTION("getLastReading") {
    WHEN("archive exists") {
      const auto x =
          GENERATE(table<MeteringType, double>({{MeteringType::powerMetering, 55.0},
                                                {MeteringType::powerProducedMetering, 56.0},
                                                {MeteringType::energyMetering, 57.0 * 3600},
                                                {MeteringType::energyConsumedTariff1Metering, 59.0 * 3600},
                                                {MeteringType::energyConsumedTariff2Metering, 60.0 * 3600},
                                                {MeteringType::energyProducedMetering, 58.0 * 3600},
                                                {MeteringType::energyProducedTariff1Metering, 61.0 * 3600},
                                                {MeteringType::energyProducedTariff2Metering, 62.0 * 3600}}));
      INFO(x);

      auto toReturn = [type = std::get<MeteringType>(x), value = std::get<double>(x)]() -> Reading {
        switch (type) {
          case MeteringType::powerMetering: return Reading{.lastChanged = ds::DateTime{300s}, .powerConsumed = value};
          case MeteringType::energyMetering:
            return Reading{
                .lastChanged = ds::DateTime{300s}, .t1EnergyConsumed = value - 20.0, .t2EnergyConsumed = 20.};
          case MeteringType::energyConsumedTariff1Metering:
            return Reading{.lastChanged = ds::DateTime{300s}, .t1EnergyConsumed = value};
          case MeteringType::energyConsumedTariff2Metering:
            return Reading{.lastChanged = ds::DateTime{300s}, .t2EnergyConsumed = value};
          case MeteringType::powerProducedMetering:
            return Reading{.lastChanged = ds::DateTime{300s}, .powerProduced = value};
          case MeteringType::energyProducedMetering:
            return Reading{
                .lastChanged = ds::DateTime{300s}, .t1EnergyProduced = value - 20.0, .t2EnergyProduced = 20.};
          case MeteringType::energyProducedTariff1Metering:
            return Reading{.lastChanged = ds::DateTime{300s}, .t1EnergyProduced = value};
          case MeteringType::energyProducedTariff2Metering:
            return Reading{.lastChanged = ds::DateTime{300s}, .t2EnergyProduced = value};
          case MeteringType::none:
          default:;
        }
        DS_FAIL_REQUIRE(type);
      };

      const auto& [meteringType, value] = x;

      auto convertEnergyToWh = [meteringType = std::get<MeteringType>(x)](double value) {
        if (meteringType == MeteringType::powerMetering || meteringType == MeteringType::powerProducedMetering) {
          return value;
        } else {
          return value / 3600;
        };
      };

      REQUIRE_CALL(archives, flush(_));
      REQUIRE_CALL(archives, lastValue(openMeteringSystemId)).RETURN(toReturn());
      CHECK(openMeteringSystem->getLastReading(openMeteringSystemId, meteringType, EnergyUnit::Wh) ==
            LastReading{.type = meteringType, .value = convertEnergyToWh(value), .lastChanged = ds::DateTime{300s}});

      REQUIRE_CALL(archives, flush(_));
      REQUIRE_CALL(archives, lastValue(openMeteringSystemId)).RETURN(toReturn());
      CHECK(openMeteringSystem->getLastReading(openMeteringSystemId, meteringType, EnergyUnit::Ws) ==
            LastReading{.type = meteringType, .value = value, .lastChanged = ds::DateTime{300s}});
    }

    WHEN("queried without a serial number") {
      WHEN("an archive exists") {
        REQUIRE_CALL(storage, hasArchive(openMeteringSystemId)).RETURN(true);
        REQUIRE_CALL(archives, addValue(openMeteringSystemId, _));
        openMeteringSystem->addReading(Reading{.id = MeterId{openMeteringSystemId}});

        THEN("query succeeds") {
          REQUIRE_CALL(archives, flush(_));
          REQUIRE_CALL(archives, lastValue(openMeteringSystemId))
              .RETURN(Reading{.lastChanged = ds::DateTime{300s}, .powerConsumed = 20});
          CHECK(
              openMeteringSystem->getLastReading(OpenMeteringSystemId{}, MeteringType::powerMetering, EnergyUnit::Wh) ==
              LastReading{.type = MeteringType::powerMetering, .value = 20, .lastChanged = ds::DateTime{300s}});
        }
      }
      WHEN("no archive exists") {
        THEN("query fails") {
          CHECK_THROWS(
              openMeteringSystem->getLastReading(OpenMeteringSystemId{}, MeteringType::powerMetering, EnergyUnit::Wh));
        }
      }
    }
  }

  SECTION("getSeries") {
    openMeteringSystem->setActiveMeterId(openMeteringSystemId);

    WHEN("archive exists") {
      const auto x = GENERATE(table<MeteringType, SeriesTypes, Category, double>(
          {{MeteringType::powerMetering, SeriesTypes::power, Category::consumption, 55.0},
           {MeteringType::energyMetering, SeriesTypes::energyDelta, Category::consumption, 57.0 * 3600},
           {MeteringType::energyConsumedTariff1Metering, SeriesTypes::energyDelta, Category::consumption, 59.0 * 3600},
           {MeteringType::energyConsumedTariff2Metering, SeriesTypes::energyDelta, Category::consumption, 60.0 * 3600},
           {MeteringType::powerProducedMetering, SeriesTypes::power, Category::production, 56.0},
           {MeteringType::energyProducedMetering, SeriesTypes::energyDelta, Category::production, 58.0 * 3600},
           {MeteringType::energyProducedTariff1Metering, SeriesTypes::energyDelta, Category::production, 61.0 * 3600},
           {MeteringType::energyProducedTariff2Metering, SeriesTypes::energyDelta, Category::production,
            62.0 * 3600}}));
      INFO(x);

      REQUIRE_CALL(archives, lastTimestamp(openMeteringSystemId)).RETURN(ds::DateTime{2700000s}); // multiple of 15min
      REQUIRE_CALL(archives, flush(_));
      REQUIRE_CALL(archives, getSeries(openMeteringSystemId, std::get<Category>(x), std::get<SeriesTypes>(x), _,
                                       2'700'900, 900, EnergyUnit::Wh))
          .RETURN(ValueRange{});
      openMeteringSystem->getSeries(std::get<MeteringType>(x), std::get<SeriesTypes>(x), 15min, TimeDuration{}, 0,
                                    EnergyUnit::Wh);
    }
  }
}

} // namespace
