#include "open-metering-system.h"

#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include "archives-storage.h"
#include "archives/open-metering-system.h"
#include "ds-model/meterings.h"
#include "metering.h"
#include "metering/readings/open-metering-system.h"

static ds::LogChannel dsLogChannel("dssMeteringOpenMeteringSystem");

namespace dss::metering {

namespace {

using system_clock = std::chrono::system_clock;

auto openMeteringSystemId = [](const OpenMeteringSystem::Reading& reading) -> OpenMeteringSystemId {
  return OpenMeteringSystemId{reading.id};
};

auto emptyId = [](const OpenMeteringSystemId& id) { return id == OpenMeteringSystemId{}; };

class OpenMeteringSystemImpl final : public OpenMeteringSystem {
  using OpenMeteringSystemArchive = archives::OpenMeteringSystemArchive;
  std::unique_ptr<OpenMeteringSystemArchive> m_archives;
  ArchivesStorage& m_archivesStorage;
  ds::System& m_system;
  const ElectricityMeteringConfig m_config;

  mutable OpenMeteringSystemId m_activeMeter{}; // assume there is only one at the same time

public:
  OpenMeteringSystemImpl(std::unique_ptr<OpenMeteringSystemArchive>&& openMeteringSystemArchive,
                         ArchivesStorage& archivesStorage, ds::System& system)
      : m_archives(std::move(openMeteringSystemArchive)),
        m_archivesStorage(archivesStorage),
        m_system(system),
        m_config(m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {})) {}

  virtual ~OpenMeteringSystemImpl() override = default;

  void setActiveMeterId(const OpenMeteringSystemId& id) override {
    DS_NOTICE("initalize active meter", id);
    m_activeMeter = std::move(id);
  }

  virtual std::optional<OpenMeteringSystemId> activeMeterId() const override {
    if (!emptyId(m_activeMeter)) {
      return m_activeMeter;
    }
    return std::nullopt;
  }

  void addReading(Reading&& reading) override {
    auto oldId = std::exchange(m_activeMeter, openMeteringSystemId(reading));
    if (!emptyId(oldId) && m_activeMeter != oldId) {
      DS_WARNING("open-meter-system id change", oldId, "->", m_activeMeter);
    }

    if (!m_archivesStorage.hasArchive(openMeteringSystemId(reading))) {
      m_archives->create(openMeteringSystemId(reading));
    }
    m_archives->addValue(openMeteringSystemId(reading), std::move(reading));
  }

  std::vector<std::pair<OpenMeteringSystemId, dsModel::MeteringType>> meterings() const override {
    using dsModel::MeteringType;
    std::vector<std::pair<OpenMeteringSystemId, dsModel::MeteringType>> meterings{};

    auto addMeterings = [&meterings](const OpenMeteringSystemId& id) {
      meterings.push_back({id, MeteringType::powerMetering});
      meterings.push_back({id, MeteringType::energyMetering});
      meterings.push_back({id, MeteringType::energyConsumedTariff1Metering});
      meterings.push_back({id, MeteringType::energyConsumedTariff2Metering});
      meterings.push_back({id, MeteringType::powerProducedMetering});
      meterings.push_back({id, MeteringType::energyProducedMetering});
      meterings.push_back({id, MeteringType::energyProducedTariff1Metering});
      meterings.push_back({id, MeteringType::energyProducedTariff2Metering});
    };

    if (!emptyId(m_activeMeter)) {
      addMeterings(m_activeMeter);
    } else {
      for (auto id : m_archivesStorage.searchOmsArchives()) {
        // TODO pick newest one
        addMeterings(id);
      }
      if (!meterings.empty()) {
        m_activeMeter = meterings.front().first;
      }
    }
    return meterings;
  }

  std::optional<LastReading> getLastReading(OpenMeteringSystemId id, dsModel::MeteringType meteringType,
                                            EnergyUnit energyUnit) const override {
    DS_REQUIRE(!emptyId(id) || !emptyId(m_activeMeter), "no active OpenMeteringSystemId known");
    if (emptyId(id)) {
      id = m_activeMeter;
    }

    // expiry of data, archive removal

    auto convertEnergyUnit = [energyUnit](double value) {
      if (energyUnit == EnergyUnit::Wh) {
        return value / 3600;
      } else {
        return value;
      }
    };

    using dsModel::MeteringType;
    auto sumEnergyTariffs = [&convertEnergyUnit](const auto& lastValue, Category category) -> LastReading {
      if (category == Category::consumption) {
        return {MeteringType::energyMetering,
                convertEnergyUnit(lastValue.t1EnergyConsumed + lastValue.t2EnergyConsumed), lastValue.lastChanged};
      } else {
        return {MeteringType::energyProducedMetering,
                convertEnergyUnit(lastValue.t1EnergyProduced + lastValue.t2EnergyProduced), lastValue.lastChanged};
      }
    };

    m_archives->flush({id});
    if (auto lastValue = m_archives->lastValue(id); lastValue) {
      switch (meteringType) {
        case MeteringType::powerMetering:
          return LastReading{
              .type = meteringType, .value = lastValue->powerConsumed, .lastChanged = lastValue->lastChanged};
        case MeteringType::powerProducedMetering:
          return LastReading{meteringType, lastValue->powerProduced, lastValue->lastChanged};

        case MeteringType::energyMetering: return sumEnergyTariffs(*lastValue, Category::consumption);
        case MeteringType::energyConsumedTariff1Metering:
          return LastReading{meteringType, convertEnergyUnit(lastValue->t1EnergyConsumed), lastValue->lastChanged};
        case MeteringType::energyConsumedTariff2Metering:
          return LastReading{meteringType, convertEnergyUnit(lastValue->t2EnergyConsumed), lastValue->lastChanged};

        case MeteringType::energyProducedMetering: return sumEnergyTariffs(*lastValue, Category::production);
        case MeteringType::energyProducedTariff1Metering:
          return LastReading{meteringType, convertEnergyUnit(lastValue->t1EnergyProduced), lastValue->lastChanged};
        case MeteringType::energyProducedTariff2Metering:
          return LastReading{meteringType, convertEnergyUnit(lastValue->t2EnergyProduced), lastValue->lastChanged};
        case MeteringType::none:
        default: break;
      }
    }

    return {};
  }

  ValueRange getSeries(dsModel::MeteringType meteringType, SeriesTypes type, std::chrono::seconds resolution,
                       const TimeDuration& timeDuration, int valueCount, EnergyUnit energyUnit) const override {
    DS_REQUIRE(!emptyId(m_activeMeter), "open-metering-system not present");

    auto& archiveConfigs = m_config.timeSeries;
    auto it = std::find_if(std::rbegin(archiveConfigs), std::rend(archiveConfigs),
                           [resolution](const auto& element) { return element.resolution == resolution; });
    DS_REQUIRE(it != std::rend(archiveConfigs), "unsupported resolution");
    int numberOfValues = it->numberOfValues;

    long unsigned int step = resolution.count();
    auto now = system_clock::now();
    if (auto tmp = m_archives->lastTimestamp(m_activeMeter); tmp) {
      // PROBLEM: fix for distorted graphs, meter sends data only ever 10s, new data arrives with timestamp < now
      // BACKGROUND: e.g. due to interpolation rrd can send us values for t==now, but once new value come in then can have
      // a smaller timestamp. when the addon is polling those with its timestamp, the graph might go backwards
      // NOTE: new data should always have timestamps bigger than now, hence the graphs should never should go backward...
      // so something is still missing
      now = *tmp;
    }

    auto align = [step](time_t value) { return value - value % step; };
    time_t maxEnd = align(system_clock::to_time_t(now));
    time_t maxStart = align(system_clock::to_time_t(now) - (step * numberOfValues));
    time_t start{}, end{};

    if (timeDuration == TimeDuration{}) {
      end = maxEnd;
      if (valueCount == 0) {
        start = maxStart;
      } else {
        start = end - (step * valueCount);
      }
    } else if (!timeDuration.start) {
      end = align(system_clock::to_time_t(*timeDuration.end));
      if (valueCount == 0) {
        start = maxStart;
      } else {
        start = end - (step * valueCount);
      }
    } else if (!timeDuration.end) {
      start = align(system_clock::to_time_t(*timeDuration.start));
      if (valueCount == 0) {
        end = maxEnd;
      } else {
        end = start + (step * valueCount);
      }
    } else {
      start = align(system_clock::to_time_t(*timeDuration.start));
      end = align(system_clock::to_time_t(*timeDuration.end));
    }
    end += step;

    auto category = [](auto meteringType) {
      switch (meteringType) {
        case dsModel::MeteringType::powerMetering:
        case dsModel::MeteringType::energyMetering:
        case dsModel::MeteringType::energyConsumedTariff2Metering:
        case dsModel::MeteringType::energyConsumedTariff1Metering: return Category::consumption;
        case dsModel::MeteringType::powerProducedMetering:
        case dsModel::MeteringType::energyProducedMetering:
        case dsModel::MeteringType::energyProducedTariff1Metering:
        case dsModel::MeteringType::energyProducedTariff2Metering: return Category::production;
        case dsModel::MeteringType::none:
        default: DS_FAIL_REQUIRE("unhandled", meteringType);
      }
    };

    m_archives->flush({m_activeMeter});
    return m_archives->getSeries(m_activeMeter, category(meteringType), type, start, end, step, energyUnit);
  };
};

} // namespace

std::unique_ptr<OpenMeteringSystem> makeOpenMeteringSystem(ArchivesConnection& archivesConnection,
                                                           ArchivesStorage& archivesStorage) {
  auto openMeteringSystemArchive =
      archives::makeOpenMeteringSystemArchive(archivesConnection, archivesStorage, ds::system());
  return std::make_unique<OpenMeteringSystemImpl>(std::move(openMeteringSystemArchive), archivesStorage, ds::system());
}

std::unique_ptr<OpenMeteringSystem> makeOpenMeteringSystemMocked(
    std::unique_ptr<archives::OpenMeteringSystemArchive> openMeteringSysteArchive, ArchivesStorage& archivesStorage,
    ds::System& system) {
  return std::make_unique<OpenMeteringSystemImpl>(std::move(openMeteringSysteArchive), archivesStorage, system);
}

} // namespace dss::metering
