
#include "power-states.h"

#include <ds/catch/catch.h>
#include <metering/metering-mock.h>
#include <model/apartment-mock.h>
#include <model/state.h>

#include <trompeloeil.hpp>

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

static ds::LogChannel dsLogChannel("dssMeteringPowerStatesTest");

namespace {

constexpr auto TAGS = "[dSS][metering]";

SCENARIO("powerStates", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto powerStates = makePowerStates(apartment);
  auto config = PowerStateConfig{.category = Category::production, .setThreshold = 5, .resetThreshold = 7};

  WHEN("createState") {
    powerStates->createState(MeteringSystem::apartmentSummation, PowerStateConfig(config));
    THEN("state is created") {
      CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) != nullptr);
    }

    WHEN("removeState") {
      powerStates->removeState(MeteringSystem::apartmentSummation, config);
      THEN("state disappears") {
        CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) == nullptr);
      }
    }

    WHEN("2x createState") {
      powerStates->createState(MeteringSystem::apartmentSummation, PowerStateConfig(config));

      WHEN("removeState") {
        powerStates->removeState(MeteringSystem::apartmentSummation, config);
        THEN("state still exists") {
          CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) != nullptr);
        }
        WHEN("2x removeState") {
          powerStates->removeState(MeteringSystem::apartmentSummation, config);
          THEN("state disappears") {
            CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) == nullptr);
          }
        }
      }
    }

    WHEN("same config but different meteringSystem") {
      powerStates->createState(MeteringSystem::openMeteringSystem, PowerStateConfig(config));
      THEN("state is created") {
        CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) != nullptr);
        CHECK(apartment.tryGetState(StateType_Circuit, "openMeteringSystem-production-on5-off7"s) != nullptr);
      }
      WHEN("first state is removed") {
        powerStates->removeState(MeteringSystem::apartmentSummation, config);
        THEN("first state is removed") {
          CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) == nullptr);
        }
        THEN("2nd state remains") {
          CHECK(apartment.tryGetState(StateType_Circuit, "openMeteringSystem-production-on5-off7"s) != nullptr);
        }
      }
    }

    WHEN("removeState with incorrect meteringSystem") {
      CHECK_THROWS(powerStates->removeState(MeteringSystem::digitalStromMeter, config));
      THEN("state still exists") {
        CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) != nullptr);
      }
    }
  }

  SECTION("open-metering-system") {
    WHEN("createState") {
      powerStates->createState(MeteringSystem::openMeteringSystem, PowerStateConfig(config));
      THEN("state is created") {
        CHECK(apartment.tryGetState(StateType_Circuit, "openMeteringSystem-production-on5-off7"s) != nullptr);
      }
      WHEN("removeState") {
        powerStates->removeState(MeteringSystem::openMeteringSystem, config);
        THEN("state disappears") {
          CHECK(apartment.tryGetState(StateType_Circuit, "openMeteringSystem-production-on5-off7"s) == nullptr);
        }
      }
    }
  }
}

SCENARIO("trigger state changes", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto metering = MeteringMock{};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto powerStates = makePowerStates(apartment);

  WHEN("pollState no config") {
    powerStates->trigger(MeteringSystem::apartmentSummation, TotalPower{.production = 100, .consumption = 0});
  }

  SECTION("hysteresis power states ") {
    auto x = GENERATE(PowerStateConfig{.category = Category::production, .setThreshold = 15, .resetThreshold = 7},
                      PowerStateConfig{.category = Category::consumption, .setThreshold = 15, .resetThreshold = 7});

    CHECK(x.resetThreshold <= x.setThreshold); // lower smaller than higher limit

    powerStates->createState(MeteringSystem::apartmentSummation, PowerStateConfig(x));
    auto state1 = apartment.tryGetState(
        StateType_Circuit, ds::str("apartment-power-", x.category, "-on", x.setThreshold, "-off", x.resetThreshold));
    CHECK(state1 != nullptr);

    auto makeTotalPower = [](Category category, auto categoryLevel, auto otherLevel) -> TotalPower {
      if (category == Category::consumption) {
        return {.production = otherLevel, .consumption = categoryLevel};
      } else {
        return {.production = categoryLevel, .consumption = otherLevel};
      }
    };

    auto average = [](auto left, auto right) -> decltype(left) { return (left + right) / 2; };

    WHEN("state not active") {
      CHECK(state1->getState() == State_Inactive);
      WHEN("powerValue < setThreshold") {
        THEN("state remains unchanged") {
          powerStates->trigger(MeteringSystem::apartmentSummation, makeTotalPower(x.category, x.setThreshold - 1, 0u));
          CHECK(state1->getState() == State_Inactive);
        }
      }
      WHEN("powerValue > setThreshold") {
        THEN("state is active") {
          powerStates->trigger(MeteringSystem::apartmentSummation, makeTotalPower(x.category, x.setThreshold + 1, 0u));
          CHECK(state1->getState() == State_Active);
        }
        WHEN("incorrect meteringSystem") {
          THEN("state remains unchanged") {
            powerStates->trigger(MeteringSystem::digitalStromMeter, makeTotalPower(x.category, x.setThreshold + 1, 0u));
            CHECK(state1->getState() == State_Inactive);
          }
        }
      }
      WHEN("wrong category powerValue > setThreshold") {
        THEN("state remains unchanged") {
          powerStates->trigger(
              MeteringSystem::apartmentSummation,
              makeTotalPower(x.category, average(x.resetThreshold, x.setThreshold), x.setThreshold + 1));
          CHECK(state1->getState() == State_Inactive);
        }
      }
    }
    WHEN("state is active") {
      state1->setState(CallOrigin::system, State_Active);
      CHECK(state1->getState() == State_Active);
      WHEN("powerValue < resetThreshold") {
        THEN("state becomes inactive") {
          powerStates->trigger(MeteringSystem::apartmentSummation,
                               makeTotalPower(x.category, x.resetThreshold - 1, 0u));
          CHECK(state1->getState() == State_Inactive);
        }
        WHEN("incorrect meteringSystem") {
          THEN("state remains unchanged") {
            powerStates->trigger(MeteringSystem::digitalStromMeter,
                                 makeTotalPower(x.category, x.resetThreshold - 1, 0u));
            CHECK(state1->getState() == State_Active);
          }
        }
      }
      WHEN("powerValue between [resetThreshold,setThreshold]") {
        THEN("state remains active") {
          powerStates->trigger(MeteringSystem::apartmentSummation,
                               makeTotalPower(x.category, average(x.resetThreshold, x.setThreshold), 0u));
          CHECK(state1->getState() == State_Active);
        }
      }
      WHEN("wrong category < resetThreshold)") {
        THEN("state remains unchanged") {
          powerStates->trigger(
              MeteringSystem::apartmentSummation,
              makeTotalPower(x.category, average(x.resetThreshold, x.setThreshold), x.resetThreshold - 1));
        }
      }
    }
  }
}

} // namespace
