/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "busscanner.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "dsm-api-mock.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"

static ds::LogChannel dsLogChannel("modelBusScannerTest");
static const char* TAGS = "[model][busscanner]";

using namespace dss;
using namespace dss::model;
using namespace trompeloeil;
using trompeloeil::_;
using namespace std::literals;

namespace {

TEST_CASE("busscanner", TAGS) {
  auto busInterface = BusInterfaceMock{};
  auto propertySystem = dss::PropertySystem{};
  EventSinkMock eventSink{};
  auto apartment = makeApartment(nullptr, &propertySystem, &busInterface);
  auto modelMaintenance = ModelMaintenanceMock{propertySystem, eventSink};
  apartment->setModelMaintenance(&modelMaintenance);

  auto busScanner = BusScanner(busInterface.structureQuery, *apartment, modelMaintenance);

  auto dsm = apartment->allocateDSMeter(Dsuid{"0000000000000000000007575750000777"});
  auto zone = apartment->allocateZone(17);

  WHEN("initializeDevice") {
    REQUIRE_CALL(dsm_api_mock, DsmApiGetDeviceDescription(_, _, _, _, _, _, _)).RETURN(0);
    ALLOW_CALL(modelMaintenance, addModelEvent(_));
    ALLOW_CALL(busInterface, getConnectionURI()).RETURN(""s);
    REQUIRE_CALL(modelMaintenance, scheduleDeviceReadout(_, _));

    // copy of akm210 from device-mock
    auto deviceSpec = DeviceSpec{
        .ShortAddress = 0x12, //
        .FunctionID = functionID105(DeviceClasses_t::DEVICE_CLASS_SW, FID105::extraEquipment, FID105keys::_4),
        .ProductID = 0x20C8,                     //
        .VendorID = 2,                           //
        .revisionId = 0x356,                     //
        .Locked = false,                         //
        .ActiveState = true,                     //
        .OutputMode = 0,                         // bank3/0 -> 0:no output
        .LTMode = ButtonInputMode::AKM_STANDARD, //
        .Groups = std::vector{GroupIDBlack},     //
        .DSID = createMultiDsuid(4),             //
        .buttonID = 15,                          // if color:black -> app-taster
        .buttonActiveGroup = GroupIDBlack,       //
        .buttonSetsLocalPriority = true,         // taken from real device
        .buttonCallsPresent = true,              // taken from real device
        .Name = ""s,                             //
        .ZoneID = 17,                            //
        .binaryInputs = std::vector{DeviceBinaryInputSpec_t{
            .TargetGroup = GroupIDBlack,                     //
            .InputType = BinaryInputType::HeatingSystemMode, //
            .InputID = BinaryInputId::APP_MODE,              //
            .Independent = false,                            //
        }},
        .binaryInputsValid = true, //
        .sensorInputs =
            std::vector<DeviceSensorSpec_t>{{SensorType::Reserved2, 0, 0, 0}, {SensorType::NotUsed, 0, 0, 0}},
        .sensorInputsValid = true,
        .outputChannels = std::vector<ChannelType>{},
        .outputChannelsValid = false,
        .activeGroup = GroupIDBlack,
        .defaultGroup = GroupIDBlack};
    busScanner.initializeDeviceFromSpec(dsm, zone, deviceSpec);

    THEN("device with basic features is created") {
      auto device = apartment->getDeviceByDSID(deviceSpec.DSID);
      CHECK(device);
      CHECK(device->getButtonInputCount() == 4);
      CHECK(device->getButtonInputIndex() == 0); // dsuid is aligned by four
      CHECK(device->getButtonActiveGroup() == GroupIDBlack);
      CHECK(device->getButtonInputMode() == ButtonInputMode::AKM_STANDARD);
    }
  }
}

} // namespace
