/*
    Copyright (c) 2010, 2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>
             Michael Tross, aizo GmbH <michael.tross@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "modelmaintenance.h"

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <json.h>
#include <unistd.h>

#include <boost/filesystem.hpp>
#include <boost/range/adaptor/reversed.hpp>
#include <chrono>

#include "apartment.h"
#include "backend-vdcs.h"
#include "base.h"
#include "businterface.h"
#include "busscanner.h"
#include "cache/apartment-proxy.h"
#include "cluster.h"
#include "device.h"
#include "devicereference.h"
#include "ds/conf.h"
#include "ds485-device.h"
#include "dsmeter.h"
#include "dss.h"
#include "event.h"
#include "event/event_create.h"
#include "eventinterpreterplugins.h"
#include "group.h"
#include "hal/hardware-abstraction.h"
#include "handler/system_states.h"
#include "includerapidjson.h"
#include "internaleventrelaytarget.h"
#include "metering/metering.h"
#include "model/autoclustermaintenance.h"
#include "model/modelconst.h"
#include "modelpersistence.h"
#include "propertysystem.h"
#include "propertysystem_common_paths.h"
#include "scanner/metermaintenance.h"
#include "scanner/oemwebquery.h"
#include "scanner/websocketevent.h"
#include "scenehelper.h"
#include "security/security.h"
#include "set.h"
#include "state.h"
#include "structuremanipulator.h"
#include "util.h"
#include "zone.h"

static ds::LogChannel dsLogChannel("modelModelMaintenance");

namespace dss {

  //=============================================== ApartmentTreeListener

  /** Raises a ModelDirty event if something below the apartment node gets changed. */
  class ApartmentTreeListener : public PropertyListener {
  public:
    ApartmentTreeListener(ModelMaintenance* _pModelMaintenance, Apartment* _pApartment)
    : m_pModelMaintenance(_pModelMaintenance),
      m_pApartment(_pApartment)
    {
      if(m_pApartment->getPropertyNode() != NULL) {
        m_pApartment->getPropertyNode()->addListener(this);
      }
    }

  protected:
    virtual void propertyChanged(PropertyNodePtr _caller, PropertyNodePtr _changedNode) {
      if(_changedNode->hasFlag(PropertyNode::Archive)) {
        auto pEvent = std::make_shared<ModelEvent>(ModelEvent::etModelDirty);
        m_pModelMaintenance->addModelEvent(pEvent);
      }
      if((_changedNode->getName() == "setsLocalPriority") &&
         (_changedNode->getValueType() == vTypeBoolean)) {
        handleLocalPriority(_changedNode);
      }
      if((_changedNode->getName() == "callsPresent") &&
         (_changedNode->getValueType() == vTypeBoolean)) {
        handleCallsPresent(_changedNode);
      }
    }

  private:
    void handleLocalPriority(PropertyNodePtr _changedNode) {
      if((_changedNode->getParentNode() != NULL) &&
         (_changedNode->getParentNode()->getParentNode() != NULL)) {
        PropertyNode* deviceNode =
          _changedNode->getParentNode()->getParentNode();
        std::shared_ptr<Device> dev =
          m_pApartment->getDeviceByDSID(str2dsuid(deviceNode->getName()));
        bool value = _changedNode->getBoolValue();
        if(DSS::hasInstance()) {
          StructureModifyingBusInterface* pInterface;
          pInterface = DSS::getInstance()->getBusInterface().getStructureModifyingBusInterface();
          pInterface->setButtonSetsLocalPriority(dev->getDSMeterDSID(),
                                                 dev->getShortAddress(),
                                                 value);
        }
      }
    }
    void handleCallsPresent(PropertyNodePtr _changedNode) {
      if((_changedNode->getParentNode() != NULL) &&
         (_changedNode->getParentNode()->getParentNode() != NULL)) {
        PropertyNode* deviceNode =
          _changedNode->getParentNode()->getParentNode();
        std::shared_ptr<Device> dev =
          m_pApartment->getDeviceByDSID(str2dsuid(deviceNode->getName()));
        bool value = _changedNode->getBoolValue();
        if(DSS::hasInstance()) {
          StructureModifyingBusInterface* pInterface;
          pInterface = DSS::getInstance()->getBusInterface().getStructureModifyingBusInterface();
          pInterface->setButtonCallsPresent(dev->getDSMeterDSID(),
                                            dev->getShortAddress(),
                                            value);
        }
      }
    }
  private:
    ModelMaintenance* m_pModelMaintenance;
    Apartment* m_pApartment;
  }; // ApartmentTreeListener

 //=============================================== ModelMaintenance

  ModelMaintenance::ModelMaintenance(DSS* _pDSS, PropertySystem& propertySystem, EventSink& eventSink,
                                     const int _eventTimeoutMS)
      : ThreadedSubsystem(_pDSS, "Apartment"),
        m_IsInitializing(true),
        m_processedEvents(UINT_MAX - 2), // force early wrap around
        m_propertySystem(propertySystem),
        m_eventSink(eventSink),
        m_IsDirty(false),
        m_pendingSaveRequest(false),
        m_suppressSaveRequestNotify(m_processedEvents),
        m_pApartment(NULL),
        m_pMetering(NULL),
        m_EventTimeoutMS(_eventTimeoutMS),
        m_pStructureQueryBusInterface(NULL),
        m_pStructureModifyingBusInterface(NULL),
        m_taskProcessor(),
        m_taskProcessorMaySleep(),
        m_pMeterMaintenance(std::make_shared<MeterMaintenance>(_pDSS, "MeterMaintenance")) {}

  ModelMaintenance::~ModelMaintenance() = default;

  void ModelMaintenance::shutdown() {
    if (m_pendingSaveRequest) {
      writeConfiguration();
    }
    m_pMeterMaintenance->shutdown();
    ThreadedSubsystem::shutdown();
  }

  static void checkConfigFile(const boost::filesystem::path& _filename) {
    if (boost::filesystem::exists(_filename)) {
      DS_INFO("Found ", _filename.string());
      if (!rwAccess(_filename.string())) {
        throw std::runtime_error("Apartment file " + _filename.string() +
                                  " is not readable and writable!");
      }
    } else {
      DS_WARNING("Could not find config file.", _filename.string());
      boost::filesystem::path dir = _filename.parent_path();
      if (!boost::filesystem::is_directory(dir)) {
        throw std::runtime_error("Path " + dir.string() +
                                 " to apartment file is invalid!");
      }

      if (!rwAccess(dir.string())) {
        throw std::runtime_error("Directory " + dir.string() +
                                 " for apartment file is not readable and " +
                                 " writable!");
      }
    }
  }

  void ModelMaintenance::migrateHeating() {
    std::vector<std::pair<int, std::string> > opModes {
      {0, "OffTemperatur"},
      {1, "ComfortTemperatur"},
      {2, "EcoTemperatur"},
      {3, "notUsedTemperatur"},
      {4, "NightTemperatur"},
      {5, "HolidayTemperatur"},
      /* the passive cooling temperatures do not have to be migrated because of different implementation in new solution
      {6, "coolingTemperatur"},
      {7 , "coolingOffTemperatur"},
       */
      /* the following temperatures were not used in previous versions
      {9, "coolingTemperatur"},
      {10, "coolingOffTemperatur"},
      {11, "coolingEconomyTemperatur"},
      {12, "coolingNotUsedTemperatur"},
      {13, "coolingNightTemperatur"},
      {14, "coolingHolidayTemperatur"}*/ };
    // load heating-controller.xml config file
    PropertyNodePtr migration = m_propertySystem.createProperty("/migration/heating-controller");
    loadFromXML(DSS::getInstance()->getSavedPropsDirectory() + "heating-controller.xml", migration);
    // get /modes/<zoneId>/Temperature|Percent nodes to populate the dSS model
    // do heating config migration from add-on to dss
    auto zones = m_pApartment->getZones();
    bool migrationsDone = false;
    for (std::shared_ptr<Zone>& zone : zones) {
      if (!zone->isTemperatureConfigValid() || zone->isHeatingSetpointsValid()) {
        // reading the heating configuration from the apartment.xml makes the settings valid
        // if we have no configuration we can also skip reading the setpints
        //also if we already have valid setpoints we can skip
        continue;
      }
      int zoneId = zone->getID();
      {
        TemperatureConfig spec;
        PropertyNodePtr tempNode = migration->getProperty("modes/" + intToString(zoneId) + "/Temperature");
        if (tempNode) {
          for (auto& temp : opModes) {
            PropertyNodePtr node = tempNode->getProperty(temp.second);
            if (node) {
              try {
                spec.targetTemperatures[temp.first] = strToDouble(node->getAsString());
              } catch (const std::invalid_argument& e) {
                DS_WARNING("Heating Migration: Temperature value", ds::hideArg(node->getAsString()), "for", temp.second,
                           "in", zoneId, "could not be converted to double");
              }
            }
          }
          zone->setHeatingSetpoints(spec);
          DS_NOTICE("Migrate temperature setpoints for zone", ds::hideArg(zone->getName()), "(",
                    ds::hideArg(zone->getID()), ")");
          migrationsDone = true;
        }
      }
      {
        TemperatureConfig spec;
        PropertyNodePtr fixNode = migration->getProperty("modes/" + intToString(zoneId) + "/Percent");
        if (fixNode) {
          for (auto& temp : opModes) {
            PropertyNodePtr node = fixNode->getProperty(temp.second);
            if (node) {
              try {
                spec.fixedValues[temp.first] = strToDouble(node->getAsString());
              } catch (const std::invalid_argument& e) {
                DS_WARNING("Heating Migration: Percent value", ds::hideArg(node->getAsString()), "for", temp.second,
                           "in", zoneId, "could not be converted to double");
              }
            }
          }
          zone->setHeatingFixValues(spec);
          DS_NOTICE("Migrate fixed setpoints for zone", ds::hideArg(zone->getName()), "(", ds::hideArg(zone->getID()),
                    ")");
          migrationsDone = true;
        }
      }
    }
    /* we leave the migration node as a trigger for the heating/climate
     * addon for the migration. The addon will remove the node afterwards. */
    if (!migrationsDone) {
      m_propertySystem.getRootNode()->removeChild(m_propertySystem.getProperty("/migration"));
    }
  }

  void ModelMaintenance::initialize() {
    Subsystem::initialize();
    if(m_pApartment == NULL) {
      throw std::runtime_error("Need apartment to work...");
    }
    if(DSS::hasInstance()) {
      m_propertySystem.setStringValue(getConfigPropertyBasePath() + "configfile",
                                      m_propertySystem.getStringValue("/config/datadirectory") + "apartment.xml", true,
                                      false);

      m_propertySystem.setStringValue(
          getConfigPropertyBasePath() + "invalidbackup",
          m_propertySystem.getStringValue("/config/datadirectory") + "invalid-apartment.xml", true, false);

      boost::filesystem::path filename(m_propertySystem.getStringValue(getConfigPropertyBasePath() + "configfile"));

      m_propertySystem.setStringValue(getConfigPropertyBasePath() + "iconBasePath",
                                      m_propertySystem.getStringValue("/config/datadirectory") + "images/", true,
                                      false);

      checkConfigFile(filename);

      m_pStructureQueryBusInterface = DSS::getInstance()->getBusInterface().getStructureQueryBusInterface();
      m_pStructureModifyingBusInterface = DSS::getInstance()->getBusInterface().getStructureModifyingBusInterface();

      // load devices/dsMeters/etc. from a config-file
      readConfiguration();
      migrateHeating();
    }
  } // initialize

  void ModelMaintenance::notifyModelConsistent() {
    std::unique_lock<std::mutex> lock(m_rvMut);
    m_rvCond.notify_all();
  }

  void ModelMaintenance::doStart() {
    run();
    m_pMeterMaintenance->run();
  } // start

  void ModelMaintenance::execute() {
    DS_CONTEXT("modelMaintenance");
    if(DSS::hasInstance()) {
      DSS::getInstance()->getSecurity().loginAsSystemUser("ModelMaintenance needs system-rights");
      DSS::getInstance()->backendVdcs().reregister();
    }

    DS_NOTICE("ModelMaintenance::execute: Enumerating model");
    discoverDS485Devices();

    std::shared_ptr<ApartmentTreeListener> treeListener
      = std::shared_ptr<ApartmentTreeListener>(
          new ApartmentTreeListener(this, m_pApartment));

    while (!m_Terminated) {
      handleModelEvents();
      handleDeferredModelEvents();
      delayedConfigWrite();
    }
  } // execute

  void ModelMaintenance::discoverDS485Devices() {
    if (m_pStructureQueryBusInterface != NULL) {
      try {
        std::vector<DSMeterSpec_t> busMembers = m_pStructureQueryBusInterface->getBusMembers();
        BusScanner scanner(*m_pStructureQueryBusInterface, *m_pApartment, *this);
        for (DSMeterSpec_t& spec : busMembers) {
          std::shared_ptr<DSMeter> dsMeter;
          try{
            dsMeter = m_pApartment->getDSMeterByDSID(spec.DSID);
            scanner.synchronizeDSMeterData(dsMeter, spec);

            DS_DEBUG("dS485 Bus Device known:", spec.DSID, "type:", spec.DeviceType);
            if (dsMeter->isPresent() && !busMemberIsAnyDSM(dsMeter->getBusMemberType())) {
              m_pApartment->removeDSMeter(spec.DSID);
              DS_DEBUG("removing uninteresting Bus Device:", dsMeter->getDSID(), "type:", dsMeter->getBusMemberType());
            }
          } catch(ItemNotFoundException& e) {
            if (!busMemberIsAnyDSM(spec.DeviceType)) {
              // recover old dSM11 fw < 1.8.3
              if (!DsmApiIsdSM(spec.DSID)) {
                DS_WARNING("ignore dS485 Bus Device:", spec.DSID, spec.DeviceType);
                continue;
              }
            }
            dsMeter = m_pApartment->allocateDSMeter(spec.DSID);
            scanner.synchronizeDSMeterData(dsMeter, spec);
            DS_WARNING("dS485 Bus Device NEW:", spec.DSID, spec.DeviceType);
          }

          try {
            Set devices = m_pApartment->getDevices().getByLastKnownDSMeter(spec.DSID);
            for (int i = 0; i < devices.length(); i++) {
              devices[i].getDevice()->setIsValid(false);
            }
          } catch(ItemNotFoundException& e) {
            DS_DEBUG("discoverDS485Devices: error resetting device valid flags:", e.what());
          }

          // #7558 - force a rescan unconditionally
          if (true) {
            dsMeter->setIsPresent(true);
            dsMeter->setIsConnected(true);
            dsMeter->setIsValid(false);
            auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etDSMeterReady, spec.DSID);
            addModelEvent(pEvent);
          }
        }
      } catch(BusApiError& e) {
        DS_ERROR("Failed to discover ds485 devices");
      }
    }
  } // discoverDS485Devices

  void ModelMaintenance::scheduleConfigWrite()
  {
    std::unique_lock<std::mutex> lock(m_SaveRequestMutex);
    if (!m_pendingSaveRequest) {
      m_pendingSaveRequest = true;
      m_pendingSaveRequestTS = DateTime();
    }
  }

  void ModelMaintenance::delayedConfigWrite()
  {
    std::unique_lock<std::mutex> lock(m_SaveRequestMutex);
    if (!m_pendingSaveRequest ||
        DateTime().difference(m_pendingSaveRequestTS) < 30) {
      return;
    }

    writeConfiguration();
    m_pendingSaveRequest = false;
  }

  void ModelMaintenance::writeConfiguration() {
    if (!DSS::hasInstance()) {
      return;
    }

    PropertySystem& props(m_propertySystem);
    ModelPersistence persistence(*m_pApartment);
    persistence.writeConfigurationToXML(props.getStringValue(getConfigPropertyBasePath()
                                                             + "configfile"));
  } // writeConfiguration

  void ModelMaintenance::handleDeferredModelStateChanges(CallOrigin _origin, int _zoneID, int _groupID, int _sceneID) {
    std::vector<std::shared_ptr<Zone> > zonesToUpdate;
    if (_zoneID == 0) {
      zonesToUpdate = m_pApartment->getZones();
    } else {
      zonesToUpdate.push_back(m_pApartment->getZone(_zoneID));
    }
    for (std::shared_ptr<Zone> pZone : zonesToUpdate) {
      if (_groupID == 0) {
        for (std::shared_ptr<Group> pGroup : pZone->getGroups()) {
          pGroup->setOnState(_origin, _sceneID);
        }
      } else {
        std::shared_ptr<Group> pGroup = pZone->getGroup(_groupID);
        if(pGroup != NULL) {
          pGroup->setOnState(_origin, _sceneID);
        }
      }
    }
  }

  bool ModelMaintenance::handleDeferredModelEvents() {
    // TODO locking, seems events are only pushed from mainloop
    if (m_DeferredEvents.empty()) {
      return false;
    }

    m_DeferredEvents_t mDefEvents;
    mDefEvents.swap(m_DeferredEvents);

    for (std::shared_ptr<ModelDeferredEvent> evt : mDefEvents) {
      std::shared_ptr<ModelDeferredSceneEvent> mEvent = std::dynamic_pointer_cast <ModelDeferredSceneEvent> (evt);
      if (mEvent != NULL) {
        int sceneID = mEvent->getSceneID();
        int groupID = mEvent->getGroupID();
        int zoneID = mEvent->getZoneID();
        int originDeviceID = mEvent->getOriginDeviceID();

        try {
          std::shared_ptr<Zone> zone = m_pApartment->getZone(zoneID);
          std::shared_ptr<Group> group = zone->getGroup(groupID);
          dsuid_t originDSUID = mEvent->getSource();
          if ((originDSUID != DSUID_NULL) && (originDeviceID != 0)) {
            DeviceReference devRef = m_pApartment->getDevices().getByBusID(originDeviceID, mEvent->getSource());
            originDSUID = devRef.getDSID();
          }

          if (mEvent->isDue()) {
            if (!mEvent->isCalled()) {
              handleDeferredModelStateChanges(mEvent->getCallOrigin(), zoneID,
                                              groupID, sceneID);
              raiseEvent(createGroupCallSceneEvent(group, sceneID, groupID,
                                                   zoneID,
                                                   mEvent->getCallOrigin(),
                                                   originDSUID,
                                                   mEvent->getOriginToken(),
                                                   mEvent->getForcedFlag()));
            }
            // finished deferred processing of this event
          } else if (SceneHelper::isDimSequence(sceneID)) {
            if (!mEvent->isCalled()) {
              handleDeferredModelStateChanges(mEvent->getCallOrigin(), zoneID,
                                              groupID, sceneID);
              raiseEvent(createGroupCallSceneEvent(group, sceneID, groupID,
                                                   zoneID,
                                                   mEvent->getCallOrigin(),
                                                   originDSUID,
                                                   mEvent->getOriginToken(),
                                                   mEvent->getForcedFlag()));
              mEvent->setCalled();
            }
            m_DeferredEvents.push_back(mEvent);
          } else {
            m_DeferredEvents.push_back(mEvent);
          }

        } catch(ItemNotFoundException& e) {
          DS_ERROR("handleDeferredModelEvents: Could not find", groupID, "/", zoneID);
        }
        continue;
      }

      std::shared_ptr<ModelDeferredButtonEvent> bEvent = std::dynamic_pointer_cast <ModelDeferredButtonEvent> (evt);
      if (bEvent != NULL) {
        int deviceID = bEvent->getDeviceID();
        int buttonIndex = bEvent->getButtonIndex();
        int clickType = bEvent->getClickType();

        try {
          DeviceReference devRef = m_pApartment->getDevices().getByBusID(deviceID, bEvent->getSource());
          std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);

          if (bEvent->isDue() || (clickType == ClickTypeHE)) {
            std::shared_ptr<Event> event = std::make_shared<Event>(EventName::DeviceButtonClick, pDevRev);
            event->setProperty("clickType", intToString(clickType));
            event->setProperty("buttonIndex", intToString(buttonIndex));
            if (bEvent->getRepeatCount() > 0) {
              event->setProperty("holdCount", intToString(bEvent->getRepeatCount()));
            }
            raiseEvent(event);
            // finished deferred processing of this event
          } else {
            m_DeferredEvents.push_back(bEvent);
          }

        } catch(ItemNotFoundException& e) {
          DS_ERROR("handleDeferredModelEvents: Could not find device with dsid", ds::hideArg(bEvent->getSource()));
        }
        continue;
      }
    }
    mDefEvents.clear();
    return true;
  } // handleDeferredModelEvents

  bool ModelMaintenance::handleModelEvents() {
    decltype(m_ModelEvents)::value_type event;
    {
      std::unique_lock<std::mutex> lock(m_ModelEventsMutex);
      if (m_ModelEvents.empty() && (m_NewModelEvent.wait_for(lock, m_EventTimeoutMS) == std::cv_status::timeout)) {
        notifyModelConsistent(); // no pending changes
        return false;
      }

      if (m_ModelEvents.empty()) {
        // man pthread_cond_wait -> Condition Wait Semantics
        // condition needs to re-checked, really happens!
        // observed wait to return without a prior notify call
        return false;
      }

      DS_ASSERT(!m_ModelEvents.empty(), "");
      event = m_ModelEvents.front();
      m_ModelEvents.pop();
      m_processedEvents++;
    }

    auto pEventWithDSID = dynamic_cast<const ModelEventWithDSID*>(event.get());
    auto pEventWithStrings = dynamic_cast<const ModelEventWithStrings*>(event.get());

    const auto& eventType = event->getEventType();
    DS_CONTEXT(ds::hideArg(eventType));
    switch (event->getEventType()) {
    case ModelEvent::etNewDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 2) {
        DS_DEBUG("Expected exactly 2 parameter for ModelEvent::etNewDevice");
      } else {
        onAddDevice(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1));
      }
      break;
    case ModelEvent::etLostDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 2) {
        DS_DEBUG("Expected exactly 2 parameter for ModelEvent::etLostDevice");
      } else {
        onRemoveDevice(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1));
      }
      break;
    case ModelEvent::etDeviceChanged:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 1) {
        DS_DEBUG("Expected exactly 1 parameter for ModelEvent::etDeviceChanged");
      } else {
        rescanDevice(pEventWithDSID->getDSID(), event->getParameter(0));
      }
      break;
    case ModelEvent::etDeviceSetOutputValue:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 2) {
        DS_DEBUG("Expected exactly 2 parameters for ModelEvent::etDeviceSetOutputValue");
      } else {
        onDeviceSetOutputValue(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getTimestamp());
      }
      break;
    case ModelEvent::etDeviceConfigChanged:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 4) {
        DS_DEBUG("Expected exactly 4 parameter for ModelEvent::etDeviceConfigChanged");
      } else {
        onDeviceConfigChanged(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1),
                                                         event->getParameter(2), event->getParameter(3));
      }
      break;
    case ModelEvent::etDeviceConfigRead:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 4) {
        DS_DEBUG("Expected exactly 4 parameter for ModelEvent::etDeviceConfigRead");
      } else {
        onDeviceConfigRead(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1),
                                                         event->getParameter(2), event->getParameter(3));
      }
      break;

    case ModelEvent::etCallSceneDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() == 5) {
        onDeviceCallScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), (CallOrigin)event->getParameter(3), event->getParameter(4), event->getSingleStringParameter(), event->getTimestamp());
      } else {
        DS_DEBUG("Unexpected parameter count for ModelEvent::etCallSceneDevice");
      }
      break;
    case ModelEvent::etSaveSceneDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() == 4) {
        onDeviceSaveScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), (CallOrigin)event->getParameter(3), event->getSingleStringParameter());
      } else {
        DS_DEBUG("Unexpected parameter count for ModelEvent::etCallSceneDevice");
      }
      break;
    case ModelEvent::etBlinkDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() == 3) {
        onDeviceBlink(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), (CallOrigin)event->getParameter(2), event->getSingleStringParameter());
      } else {
        DS_DEBUG("Unexpected parameter count for ModelEvent::etBlinkDevice");
      }
      break;
    case ModelEvent::etCallSceneDeviceLocal:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() == 5) {
        onDeviceCallScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), (CallOrigin)event->getParameter(3), event->getParameter(4), event->getSingleStringParameter(), event->getTimestamp());
      } else {
        DS_DEBUG("Unexpected parameter count for ModelEvent::etCallSceneDeviceLocal");
      }
      break;
    case ModelEvent::etButtonClickDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 3) {
        DS_DEBUG("Expected exactly 3 parameter for ModelEvent::etButtonClickDevice");
      } else {
        onDeviceActionEvent(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2));
        onDeviceActionFiltered(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2));
      }
      break;
    case ModelEvent::etButtonDirectActionDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 2) {
        DS_DEBUG("Expected exactly 2 parameter for ModelEvent::etButtonDirectActionDevice");
      } else {
        onDeviceDirectActionEvent(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1));
      }
      break;
    case ModelEvent::etCallSceneGroup:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 7) {
        DS_DEBUG("Expected exactly 7 parameters for ModelEvent::etCallSceneGroup");
      } else {
        onGroupCallScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getParameter(3), (CallOrigin)event->getParameter(4), event->getParameter(5), event->getSingleStringParameter(), event->getParameter(6), event->getTimestamp());

        onGroupCallSceneFiltered(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getParameter(3), (CallOrigin)event->getParameter(4), event->getParameter(5), event->getSingleStringParameter());
      }
      break;
    case ModelEvent::etZoneGroupSaveScene:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 5) {
        DS_DEBUG("Expected exactly 5 parameters for ModelEvent::etZoneGroupSaveScene");
      } else {
        onGroupSaveScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getParameter(3), (CallOrigin)event->getParameter(4), event->getSingleStringParameter());
      }
      break;
    case ModelEvent::etUndoSceneGroup:
      if (event->getParameterCount() != 5) {
        DS_DEBUG("Expected 5 parameters for ModelEvent::etUndoSceneGroup");
      } else {
        onGroupUndoScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getParameter(3), (CallOrigin)event->getParameter(4), event->getSingleStringParameter(), event->getTimestamp());
      }
      break;
    case ModelEvent::etZoneGroupStopDimming:
      if (event->getParameterCount() != 4) {
        DS_DEBUG("Expected 4 parameters for ModelEvent::etZoneGroupStopDimming");
      } else {
        onGroupStopDimming(event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getParameter(3), event->getTimestamp());
      }
      break;
    case ModelEvent::etBlinkGroup:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 4) {
        DS_DEBUG("Expected at least 3 parameter for ModelEvent::etBlinkGroup");
      } else {
        onGroupBlink(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), (CallOrigin)event->getParameter(3), event->getSingleStringParameter());
      }
      break;
    case ModelEvent::etModelDirty:
      if (DSS::hasInstance()) {
        // hack: if difference > INT_MAX, than difference is negative
        if (static_cast<int>(m_processedEvents - m_suppressSaveRequestNotify) > 0) {
          raiseEvent(ModelChangedEvent::createApartmentChanged()); /* raiseTimedEvent */
          // skip etModelDirty events already on the queue
          m_suppressSaveRequestNotify = m_processedEvents;
        }
      }
      if (m_pApartment) {
        m_pApartment->emitStructureChanged();
      }
      notifyModelConsistent(); //< some transaction completed
      scheduleConfigWrite();
      break;
    case ModelEvent::etModelOperationModeChanged:
      notifyModelConsistent(); //< some transaction completed
      scheduleConfigWrite();
      break;
    case ModelEvent::etLostDSMeter:
      DS_ASSERT(pEventWithDSID != NULL, "");
      onLostDSMeter(pEventWithDSID->getDSID());
      break;
    case ModelEvent::etDS485DeviceDiscovered:
      DS_ASSERT(pEventWithDSID != NULL, "");
      onJoinedDSMeter(pEventWithDSID->getDSID());
      break;
    case ModelEvent::etDSMeterReady:
      if (event->getParameterCount() != 0) {
        DS_DEBUG("Expected exactly 0 parameter for ModelEvent::etDSMeterReady");
      } else {
        DS_ASSERT(pEventWithDSID != NULL, "");
        dsuid_t meterID = pEventWithDSID->getDSID();
        try{
          std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(meterID);
          mod->setIsConnected(true);
          mod->setIsValid(false);
        } catch (ItemNotFoundException& e) {
          DS_DEBUG("dSM is ready, but it is not yet known, re-discovering devices");
          discoverDS485Devices();
        }
      }
      break;
    case ModelEvent::etBusReady:
      DS_INFO("Got bus ready event.");
      discoverDS485Devices();
      break;
    case ModelEvent::etBusDown:
      DS_INFO("Got bus down event.");
      try {
        std::vector<std::shared_ptr<DSMeter> > meters = m_pApartment->getDSMeters();
        for (std::shared_ptr<DSMeter> meter : meters) {
          onLostDSMeter(meter->getDSID());
        }
      } catch (ItemNotFoundException& _e) {
      }
      break;
    case ModelEvent::etMeteringValues:
      if (event->getParameterCount() != 4) {
        DS_DEBUG("Expected exactly 4 parameter for ModelEvent::etMeteringValues");
      } else {
        DS_ASSERT(pEventWithDSID != NULL, "");
        dsuid_t meterID = pEventWithDSID->getDSID();
        unsigned int power = event->getParameter(0);
        unsigned int energy = event->getParameter(1);
        try {
          auto meter = m_pApartment->getDSMeterByDSID(meterID);
          m_pMetering->addConsumedEnergy(
              *meter, {gsl::narrow_cast<double>(power), gsl::narrow_cast<double>(energy), event->getTimestamp()});
          if (meter->getCapability_HasMeteringProducerEnabled()) {
            m_pMetering->addProducedEnergy(*meter, {.power = gsl::narrow_cast<double>(event->getParameter(2)),
                                                    .energy = gsl::narrow_cast<double>(event->getParameter(3)),
                                                    .timestamp = event->getTimestamp()});
          }
        } catch (ItemNotFoundException& _e) {
        }
      }
      break;
    case ModelEvent::etDeviceSensorEvent:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 2) {
        DS_DEBUG("Expected at least 2 parameter for ModelEvent::etDeviceSensorEvent");
      } else {
        onSensorEvent(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1));
      }
      break;
    case ModelEvent::etDeviceBinaryStateEvent:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 4) {
        DS_DEBUG("Expected at least 4 parameter for ModelEvent::etDeviceBinaryStateEvent");
      } else {
        onBinaryInputEvent(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1),
            static_cast<BinaryInputType>(event->getParameter(2)),
            static_cast<BinaryInputStateValue>(event->getParameter(3)));
      }
      break;
    case ModelEvent::etDeviceSensorValue:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 3) {
        DS_DEBUG("Expected at least 3 parameter for ModelEvent::etDeviceSensorValue");
      } else {
        onSensorValue(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2));
      }
      break;
    case ModelEvent::etDeviceSensorValueEx: {
      auto evtSensorEx = dynamic_cast<const ModelEventWithSensorEx*>(event.get());
      DS_ASSERT(evtSensorEx != 0, "");
      onSensorValueEx(evtSensorEx->m_deviceDSID, evtSensorEx->getParameter(0),
          /* sensorIndex */ evtSensorEx->getParameter(1),
          /* sensorValue */ evtSensorEx->m_sensorValue,
          /* Age */         evtSensorEx->getParameter(2),
          /* Context ID */  evtSensorEx->getParameter(3),
          /* Context Msg */ evtSensorEx->getSingleStringParameter());
      }
      break;
    case ModelEvent::etCircuitPowerStateChange:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 3) {
        DS_DEBUG("Expected at least 3 parameter for ModelEvent::etCircuitPowerStateChange");
      } else {
        onCircuitPowerStateChange(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2));
      }
      break;
    case ModelEvent::etZoneSensorValue:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 5) {
        DS_DEBUG("Expected at least 5 parameter for ModelEvent::etZoneSensorValue");
      } else {
        onZoneSensorValue(pEventWithDSID->getDSID(),
                          str2dsuid(event->getSingleStringParameter()),
                          event->getParameter(0),
                          event->getParameter(1),
                          static_cast<SensorType>(event->getParameter(2)),
                          event->getParameter(3),
                          event->getParameter(4));
      }
      break;
    case ModelEvent::etDeviceEANReady:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 11) {
        DS_DEBUG("Expected 8 parameters for ModelEvent::etDeviceEANReady");
      } else {
        onEANReady(pEventWithDSID->getDSID(),
                   event->getParameter(0),
                   (const DeviceOEMState_t)event->getParameter(1),
                   (const DeviceOEMInetState_t)event->getParameter(2),
                   ((unsigned long long)event->getParameter(3)) << 32 | ((unsigned long long)event->getParameter(4) & 0xFFFFFFFF),
                   event->getParameter(5),
                   event->getParameter(6),
                   event->getParameter(7),
                   event->getParameter(8),
                   event->getParameter(9),
                   event->getParameter(10));
      }
      break;
    case ModelEvent::etDeviceOEMDataReady:
      DS_ASSERT(pEventWithStrings != NULL, "");
      if ((event->getParameterCount() != 1) && (pEventWithStrings->getStringParameterCount() != 5)) {
        DS_DEBUG("Expected 6 parameters for ModelEvent::etDeviceOEMDataReady");
      } else {
        onOEMDataReady(pEventWithDSID->getDSID(),
                       (DeviceOEMState_t)event->getParameter(0),
                       pEventWithStrings->getStringParameter(0),
                       pEventWithStrings->getStringParameter(1),
                       pEventWithStrings->getStringParameter(2),
                       pEventWithStrings->getStringParameter(3),
                       pEventWithStrings->getStringParameter(4));
      }
      break;
    case ModelEvent::etDeviceOEMDataUpdateProductInfoState:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() != 1) {
        DS_DEBUG("Expected 1 parameter for ModelEvent::etDeviceOEMDataUpdateProductInfoState");
      } else {
        onDeviceOEMDataUpdateProductInfoState(pEventWithDSID->getDSID(),
                       (DeviceOEMState_t)event->getParameter(0));
      }
      break;

    case ModelEvent::etControllerConfig:
      DS_ASSERT(pEventWithDSID != NULL, "");
      onHeatingControllerConfig(
          pEventWithDSID->getDSID(),
          event->getParameter(0),
          event->getSingleObjectParameter());
      break;
    case ModelEvent::etControllerState:
      DS_ASSERT(pEventWithDSID != NULL, "");
      onHeatingControllerState(
          pEventWithDSID->getDSID(),
          event->getParameter(0),
          event->getParameter(1));
      break;
    case ModelEvent::etControllerValues:
      DS_ASSERT(pEventWithDSID != NULL, "");
      onHeatingControllerValues(
          pEventWithDSID->getDSID(),
          event->getParameter(0),
          event->getSingleObjectParameter());
      break;
    case ModelEvent::etClusterConfigLock:
      onClusterConfigLock(event->getParameter(0), event->getParameter(1));
      break;
    case ModelEvent::etClusterLockedScenes:
    {
      uint8_t lockedScenes[16];
      for (int i = 0; i < 16; ++i) {
        lockedScenes[i] = event->getParameter(i + 1);
      }
      std::vector<int> pLockedScenes = parseBitfield(lockedScenes, 128);
      onClusterLockedScenes(event->getParameter(0), pLockedScenes);
      break;
    }
    case ModelEvent::etGenericEvent:
    {
      unsigned origin = event->getParameter(1);
      if (!ds::isValidEnum(static_cast<CallOrigin>(origin))) {
        DS_NOTICE("etGenericEvent: invalid origin", origin);
        break;
      }
      onGenericEvent(static_cast<GenericEventType_t> (event->getParameter(0)),
                     std::static_pointer_cast<GenericEventPayload_t> (event->getSingleObjectParameter()),
                     static_cast<CallOrigin> (event->getParameter(1)));
      break;
    }
    case ModelEvent::etOperationLock:
    {
      unsigned origin = event->getParameter(3);
      if (!ds::isValidEnum(static_cast<CallOrigin>(origin))) {
        DS_NOTICE("etOperationLock: invalid origin", origin);
        break;
      }

      // parm> 0:zoneId 1:groupId 2:lock 3:origin
      onOperationLock(event->getParameter(0), event->getParameter(1), event->getParameter(2),
                      static_cast<CallOrigin>(event->getParameter(3)));
      break;
    }
    case ModelEvent::etZoneGroupSetOutputValue:
      if (event->getParameterCount() != 3) {
        DS_DEBUG("Expected exactly 3 parameters for ModelEvent::etZoneGroupSetOutputValue");
      } else {
        onZoneGroupSetOutputValue(event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getTimestamp());
      }
      break;
    case ModelEvent::etZoneGroupSetSceneLocalPrio:
      if (event->getParameterCount() != 3) {
        DS_DEBUG("Expected exactly 3 parameters for ModelEvent::etZoneGroupSetSceneLocalPrio");
      } else {
        onZoneGroupSetSceneLocalPrio(event->getParameter(0), event->getParameter(1), event->getParameter(2));
      }
      break;
    case ModelEvent::etZoneGroupCallSceneMin:
      if (event->getParameterCount() != 3) {
        DS_DEBUG("Expected exactly 3 parameters for ModelEvent::etZoneGroupCallSceneMin");
      } else {
        onZoneGroupCallSceneMin(event->getParameter(0), event->getParameter(1), event->getParameter(2), event->getTimestamp());
      }
      break;
    case ModelEvent::etDeviceDirty:
      DS_ASSERT(pEventWithDSID != NULL, "");
      onAutoClusterMaintenance(pEventWithDSID->getDSID());
      break;
    case ModelEvent::etDummyEvent:
      // unit test
      break;
    case ModelEvent::etClusterCleanup:
      onAutoClusterCleanup();
      break;
    case ModelEvent::etMeterReady:
      onMeterReady();
      break;
    case ModelEvent::etDeviceDataReady:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 3) {
        DS_DEBUG("Expected at least 3 parameter for ModelEvent::etDeviceDataReady");
      } else {
        onDeviceDataReady(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2));
      }
      break;
    case ModelEvent::etDsmStateChange:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() < 1) {
        DS_DEBUG("Expected at least 1 parameter for ModelEvent::etDsmStateChange");
      } else {
        onDsmStateChange(pEventWithDSID->getDSID(), event->getParameter(0));
      }
      break;
    case ModelEvent::etVdceEvent:
    {
      auto vdcModelEvent = dynamic_cast<const VdceModelEvent*>(event.get());
      DS_ASSERT(vdcModelEvent != 0, "");
      onVdceEvent(*vdcModelEvent);
      break;
    }
    case ModelEvent::etUpstreamShort:
      assert(pEventWithDSID != NULL);
      if (pEventWithDSID->getParameterCount() == 5) {
        onUpstreamShort(pEventWithDSID->getDSID(), static_cast<uint16_t>(pEventWithDSID->getParameter(0)),
            static_cast<uint8_t>(pEventWithDSID->getParameter(1)),
            static_cast<uint8_t>(pEventWithDSID->getParameter(2)),
            static_cast<uint8_t>(pEventWithDSID->getParameter(3)),
            static_cast<uint8_t>(pEventWithDSID->getParameter(4)));
      } else {
        DS_DEBUG("Unexpected parameter count for ModelEvent::etUpstreamShort");
      }
      break;
    case ModelEvent::etUndoSceneDevice:
      DS_ASSERT(pEventWithDSID != NULL, "");
      if (event->getParameterCount() == 4) {
        onDeviceUndoScene(pEventWithDSID->getDSID(), event->getParameter(0), event->getParameter(1), event->getParameter(2), (CallOrigin)event->getParameter(3), event->getSingleStringParameter(), event->getTimestamp());
      } else {
        DS_DEBUG("Unexpected parameter count for ModelEvent::etCallSceneDevice");
      }
      break;
    default:
      DS_ASSERT(false, "");
      break;
    }

    return true;
  } // handleModelEvents

  void ModelMaintenance::addModelEvent(std::shared_ptr<const ModelEvent> _pEvent) {
    // filter out dirty events, as this will rewrite apartment.xml
    if (m_IsInitializing && (_pEvent->getEventType() == ModelEvent::etModelDirty)) {
      m_IsDirty = true;
      // notify_one not necessary, since event not added to m_ModelEvents
    } else {
      std::unique_lock<std::mutex> lock(m_ModelEventsMutex);
      m_ModelEvents.push(_pEvent);
      m_NewModelEvent.notify_one(); // trigger m_NewModelEvent.wait_for
    }
  } // addModelEvent

  void ModelMaintenance::readConfiguration() {
    if(DSS::hasInstance()) {
      std::string configFileName = m_propertySystem.getStringValue(getConfigPropertyBasePath() + "configfile");
      std::string backupFileName = m_propertySystem.getStringValue(getConfigPropertyBasePath() + "invalidbackup");
      ModelPersistence persistence(*m_pApartment);
      if (boost::filesystem::exists(configFileName)) {
        persistence.readConfigurationFromXML(configFileName, backupFileName);
      } else {
        m_pApartment->setName("dSS");
      }
      DS_NOTICE("processed apartment.xml");
    }
  } // readConfiguration

  void ModelMaintenance::raiseEvent(const std::shared_ptr<Event> &event) {
    m_eventSink.pushEvent(event);
  }

  void ModelMaintenance::synchronizeHeatingAssignment(const dsuid_t& _dSMeterID) {
    std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::UpdateAutoselect);
    pEvent->setProperty("meterDSUID", dsuid2str(_dSMeterID));
    DS_NOTICE("synchronizeHeatingAssignment");
    m_eventSink.pushEvent(pEvent);
  } // synchronizeHeatingAssignment

  class SetLastCalledSceneAction : public IDeviceAction {
  protected:
    int m_SceneID;
  public:
    SetLastCalledSceneAction(const int _sceneID)
    : m_SceneID(_sceneID) {}
    virtual ~SetLastCalledSceneAction() {}

    virtual bool perform(std::shared_ptr<Device> _device) {
      _device->setLastCalledScene(m_SceneID);
      return true;
    }
  };

  class SetLastButOneCalledSceneAction : public IDeviceAction {
  protected:
    int m_SceneID;
  public:
    SetLastButOneCalledSceneAction(const int _sceneID)
    : m_SceneID(_sceneID) {}
    SetLastButOneCalledSceneAction()
    : m_SceneID(-1) {}
    virtual ~SetLastButOneCalledSceneAction() {}

    virtual bool perform(std::shared_ptr<Device> _device) {
      if (m_SceneID >= 0) {
        _device->setLastButOneCalledScene(m_SceneID);
      } else {
        _device->setLastButOneCalledScene();
      }
      return true;
    }
  };

  void ModelMaintenance::onGroupCallScene(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const bool _forced, std::string _token, bool callSceneMin, ds::DateTime timeStamp) {
    try {
      if(_sceneID < 0 || _sceneID > MaxSceneNumber) {
        DS_ERROR("onGroupCallScene: Scene number is out of bounds.", _zoneID, _groupID, _sceneID);
        return;
      }
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      std::shared_ptr<Group> group = zone->getGroup(_groupID);
      if(group != NULL) {
        DS_DEBUG("OnGroupCallScene:", _groupID, _zoneID, _sceneID);
        Set s = zone->getDevices().getByGroup(_groupID);
        if(SceneHelper::rememberScene(_sceneID & 0x00ff)) {
          SetLastCalledSceneAction act(_sceneID & 0x00ff);
          s.perform(act);

          std::shared_ptr<Zone> pZone = m_pApartment->getZone(_zoneID);
          if(_groupID == GroupIDBroadcast) {
            for (std::shared_ptr<Group> pGroup : pZone->getGroups()) {
              pGroup->setLastCalledScene(_sceneID & 0x00ff);
            }
          } else {
            std::shared_ptr<Group> pGroup = pZone->getGroup(_groupID);
            if(pGroup != NULL) {
              pGroup->setLastCalledScene(_sceneID & 0x00ff);
            }
          }
        }

        if (callSceneMin) {
          for (int i = 0; i < s.length(); i++) {
            m_pApartment->proxy().callSceneMin(s.get(i).getDSID(), _sceneID, timeStamp);
          }
        } else {
          m_pApartment->proxy().callScene(_zoneID, _groupID, _sceneID, _forced, timeStamp);
        }

        std::shared_ptr<Event> pEvent;
        pEvent.reset(new Event(EventName::CallSceneBus, group));
        pEvent->setProperty("sceneID", intToString(_sceneID));
        pEvent->setProperty("groupID", intToString(_groupID));
        pEvent->setProperty("zoneID", intToString(_zoneID));
        dsuid_t originDSUID = _source;
        if ((_source != DSUID_NULL) && (_originDeviceID != 0)) {
          DeviceReference devRef = m_pApartment->getDevices().getByBusID(_originDeviceID, _source);
          originDSUID = devRef.getDSID();
        }
        if (_forced) {
          pEvent->setProperty("forced", "true");
        }
        pEvent->setProperty("callOrigin", _origin);
        pEvent->setProperty("originDSUID", dsuid2str(originDSUID));
        pEvent->setProperty("originToken", _token);
        raiseEvent(pEvent);
      } else {
        DS_ERROR("OnGroupCallScene: Could not find group with", _groupID, _zoneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnGroupCallScene: Could not find zone with", _zoneID);
    }
  } // onGroupCallScene

  void ModelMaintenance::onGroupSaveScene(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, std::string _token) {
    try {
      if(_sceneID < 0 || _sceneID > MaxSceneNumber) {
        DS_ERROR("OnGroupSaveScene: Scene number is out of bounds.", _zoneID, _groupID, _sceneID);
        return;
      }
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      std::shared_ptr<Group> group = zone->getGroup(_groupID);
      if(group != NULL) {
        DS_DEBUG("OnGroupSaveScene:", _groupID, _zoneID, _sceneID);
        Set s = zone->getDevices().getByGroup(_groupID);

        for (int i = 0; i < s.length(); i++) {
          m_pApartment->proxy().saveScene(s.get(i).getDSID(), _sceneID);
        }

      } else {
        DS_ERROR("OnGroupSaveScene: Could not find group with", _groupID, _zoneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnGroupSaveScene: Could not find zone with", _zoneID);
    }
  }

  void ModelMaintenance::onGroupUndoScene(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, CallOrigin _origin, const std::string _token, ds::DateTime timeStamp) {
    try {
      if(_sceneID < -1 || _sceneID > MaxSceneNumber) {
        DS_ERROR("onGroupUndoScene: Scene number is out of bounds", _zoneID, _groupID, _sceneID);
        return;
      }
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      std::shared_ptr<Group> group = zone->getGroup(_groupID);
      Set s = zone->getDevices().getByGroup(_groupID);
      if(group != NULL) {
        DS_DEBUG("OnGroupUndoScene:", _groupID, _zoneID, _sceneID);
        SetLastButOneCalledSceneAction act(_sceneID);
        s.perform(act);

        std::vector<std::shared_ptr<Zone> > zonesToUpdate;
        if(_zoneID == 0) {
          zonesToUpdate = m_pApartment->getZones();
        } else {
          zonesToUpdate.push_back(zone);
        }
        for (std::shared_ptr<Zone> pZone : zonesToUpdate) {
          if(_groupID == 0) {
            for (std::shared_ptr<Group> pGroup : pZone->getGroups()) {
              if (_sceneID >= 0) {
                pGroup->setLastButOneCalledScene(_sceneID);
              } else {
                pGroup->setLastButOneCalledScene();
              }
              pGroup->setOnState(_origin, pGroup->getLastCalledScene());
            }
          } else {
            std::shared_ptr<Group> pGroup = pZone->getGroup(_groupID);
            if(pGroup != NULL) {
              if (_sceneID >= 0) {
                pGroup->setLastButOneCalledScene(_sceneID);
              } else {
                pGroup->setLastButOneCalledScene();
              }
              pGroup->setOnState(_origin, pGroup->getLastCalledScene());
            }
          }
        }

        m_pApartment->proxy().undoScene(_zoneID, _groupID, _sceneID, timeStamp);

        dsuid_t originDSUID = _source;
        if ((_source != DSUID_NULL) && (_originDeviceID != 0)) {
          DeviceReference devRef = m_pApartment->getDevices().
            getByBusID(_originDeviceID, _source);
          originDSUID = devRef.getDSID();
        }
        raiseEvent(createGroupUndoSceneEvent(group, _sceneID, _groupID,
                                             _zoneID, _origin, originDSUID,
                                             _token));
      } else {
        DS_ERROR("OnGroupUndoScene: Could not find group with", _groupID, _zoneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnGroupUndoScene: Could not find zone with", _zoneID);
    }
  } // onGroupUndoScene

  void ModelMaintenance::onGroupStopDimming(const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, ds::DateTime timeStamp) {
    try {
      if(_sceneID < -1 || _sceneID > MaxSceneNumber) {
        DS_ERROR("onGroupStopDimming: Scene number is out of bounds", _zoneID, _groupID, _sceneID);
        return;
      }
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      std::shared_ptr<Group> group = zone->getGroup(_groupID);
      Set s = zone->getDevices().getByGroup(_groupID);
      if(group != NULL) {
        m_pApartment->proxy().stopDimming(_zoneID, _groupID, _sceneID, timeStamp);
      } else {
        DS_ERROR("onGroupStopDimming: Could not find group with", _groupID, _zoneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("onGroupStopDimming: Could not find zone with", _zoneID);
    }
  } // onGroupUndoScene

  void ModelMaintenance::onGroupCallSceneFiltered(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const bool _forced, std::string _token) {
    // Filter Strategy:
    // Check for Source != 0 and per Zone and Group
    // - delayed On-Scene processing, for Scene1/Scene2/Scene3/Scene4
    // - report only the first Inc/Dec

    bool passThrough = true;

    if (SceneHelper::isMultiTipSequence(_sceneID)) {
      // do not filter calls from myself
      if (_source == DSUID_NULL) {
        passThrough = true;
      }
      // do not filter calls to broadcast
      else if (_groupID == 0) {
        passThrough = true;
      }
      else {
        passThrough = false;
      }
    } else if (SceneHelper::isDimSequence(_sceneID)) {
      // filter dimming sequences to only report the first dimming event
      passThrough = false;
    }

    if (passThrough) {
      DS_DEBUG("CallSceneFilter: pass through, from", _source, _zoneID, _groupID, _originDeviceID, _origin, _token,
               _sceneID);

      // flush all pending scene call events from same origin
      for (std::shared_ptr<ModelDeferredEvent> evt : m_DeferredEvents) {
        std::shared_ptr<ModelDeferredSceneEvent> pEvent = std::dynamic_pointer_cast <ModelDeferredSceneEvent> (evt);
        // object may be NULL in case of ModelDeferredButtonEvent's
        if (pEvent == NULL) {
          continue;
        }
        if ((pEvent->getSource() == _source) &&
            (pEvent->getOriginDeviceID() == _originDeviceID)) {
          pEvent->clearTimestamp();
        }
      }

      std::shared_ptr<ModelDeferredSceneEvent> mEvent = std::make_shared<ModelDeferredSceneEvent>(_source, _zoneID, _groupID, _originDeviceID, _sceneID, _origin, _forced, _token);
      mEvent->clearTimestamp();  // force immediate event processing
      m_DeferredEvents.push_back(mEvent);
      return;
    }

    for (std::shared_ptr<ModelDeferredEvent> evt : m_DeferredEvents) {
      std::shared_ptr<ModelDeferredSceneEvent> pEvent = std::dynamic_pointer_cast <ModelDeferredSceneEvent> (evt);
      if (pEvent == NULL) {
        continue;
      }
      if (pEvent->getZoneID() == _zoneID && pEvent->getGroupID() == _groupID) {
        if ((pEvent->getSource() == _source) &&
            ((pEvent->getOriginDeviceID() == _originDeviceID) ||
             (_originDeviceID == 0))) {
          // dimming, adjust the old event's timestamp to keep it active
          if (SceneHelper::isDimSequence(_sceneID) && ((pEvent->getSceneID() == _sceneID) || (_sceneID == SceneDimArea))) {
            pEvent->setTimestamp();
            return;
          }
          // going through SceneOff
          if (((pEvent->getSceneID() == SceneOff) && (_sceneID == Scene2)) ||
              ((pEvent->getSceneID() == SceneOffE1) && (_sceneID == Scene12)) ||
              ((pEvent->getSceneID() == SceneOffE2) && (_sceneID == Scene22)) ||
              ((pEvent->getSceneID() == SceneOffE3) && (_sceneID == Scene32)) ||
              ((pEvent->getSceneID() == SceneOffE4) && (_sceneID == Scene42))) {
            DS_DEBUG("CallSceneFilter: update sceneID from", pEvent->getSceneID(), _zoneID, _groupID, _originDeviceID,
                     _origin, _sceneID);

            pEvent->setScene(_sceneID);
            return;
          }
          // area-on-2-3-4 sequence
          if (((pEvent->getSceneID() == SceneA11) || (pEvent->getSceneID() == SceneA21) ||
               (pEvent->getSceneID() == SceneA31) || (pEvent->getSceneID() == SceneA41)) &&
              ((_sceneID == Scene2) || (_sceneID == Scene12) || (_sceneID == Scene22) ||
               (_sceneID == Scene32) || (_sceneID == Scene42))) {
            DS_DEBUG("CallSceneFilter: update sceneID from ", pEvent->getSceneID(), _zoneID, _groupID, _originDeviceID,
                     _origin, _sceneID);

            pEvent->setScene(_sceneID);
            return;
          }
          // area-off-2-3-4 sequence, check for triple- or 4x-click
          if (((pEvent->getSceneID() == SceneOffA1) || (pEvent->getSceneID() == SceneOffA2) ||
               (pEvent->getSceneID() == SceneOffA3) || (pEvent->getSceneID() == SceneOffA4)) &&
              (_sceneID >= Scene2) && (_sceneID <= Scene44)) {
            DS_DEBUG("CallSceneFilter: update sceneID from", pEvent->getSceneID(), _zoneID, _groupID, _originDeviceID,
                     _origin, _sceneID);

            pEvent->setScene(_sceneID);
            return;
          }
          // previous state match
          if ((int) SceneHelper::getPreviousScene(_sceneID) == pEvent->getSceneID()) {
            DS_DEBUG("CallSceneFilter: update sceneID from", pEvent->getSceneID(), _zoneID, _groupID, _originDeviceID,
                     _origin, _sceneID);

            pEvent->setScene(_sceneID);
            return;
          }
          // next state match
          if ((int) SceneHelper::getNextScene((unsigned int) pEvent->getSceneID()) == _sceneID) {
            DS_DEBUG("CallSceneFilter: update sceneID from", pEvent->getSceneID(), _zoneID, _groupID, _originDeviceID,
                     _origin, _sceneID);

            pEvent->setScene(_sceneID);
            return;
          }
        }
      }
    }

    DS_DEBUG("CallSceneFilter: deferred processing, from source", _source, _zoneID, _groupID, _originDeviceID, _origin,
             _token, _sceneID);

    std::shared_ptr<ModelDeferredSceneEvent> mEvent = std::make_shared<ModelDeferredSceneEvent>(_source, _zoneID, _groupID, _originDeviceID, _sceneID, _origin, _forced, _token);
    m_DeferredEvents.push_back(mEvent);
  } // onGroupCallSceneFiltered

  void ModelMaintenance::onGroupBlink(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const CallOrigin _origin, const std::string _token) {
    try {
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      std::shared_ptr<Group> group = zone->getGroup(_groupID);
      if(group != NULL) {
        DS_DEBUG("OnGroupBlink:", _groupID, _zoneID);
        std::shared_ptr<Event> pEvent;
        pEvent.reset(new Event(EventName::IdentifyBlink, group));
        pEvent->setProperty("groupID", intToString(_groupID));
        pEvent->setProperty("zoneID", intToString(_zoneID));
        dsuid_t originDSUID = _source;
        if ((_source != DSUID_NULL) && (_originDeviceID != 0)) {
          DeviceReference devRef = m_pApartment->getDevices().getByBusID(_originDeviceID, _source);
          originDSUID = devRef.getDSID();
        }
        pEvent->setProperty("callOrigin", _origin);
        pEvent->setProperty("originDSUID", dsuid2str(originDSUID));
        pEvent->setProperty("originToken", _token);
        raiseEvent(pEvent);
      } else {
        DS_ERROR("OnGroupBlink: Could not find group with", _groupID, _zoneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnGroupBlink: Could not find zone with", _zoneID);
    }
  } // onGroupBlink

  void ModelMaintenance::onDeviceActionFiltered(dsuid_t _source, const int _deviceID, const int _buttonNr, const int _clickType) {

    // Filter Strategy:
    // Check for Source != 0 and per DeviceID
    // - delayed Tipp processing, for 1T..4T
    // - generate final hold event after HE

    bool passThrough = false;

    // do not filter calls from myself
    if (_source == DSUID_NULL) {
      passThrough = true;
    }

    try {
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_source);
      std::shared_ptr<Device> dev = mod->getDevices().getByBusID(_deviceID, _source).getDevice();
      // ButtonInputMode Turbo (1) or Switched (2) do not generate multi-tip-sequences
      if (dev->getButtonInputMode() == ButtonInputMode::TURBO
          || dev->getButtonInputMode() == ButtonInputMode::SWITCHED) {
        passThrough = true;
      }
    } catch (ItemNotFoundException& ex) {
    }

    if (passThrough) {
      DS_DEBUG("DeviceActionFilter: pass through, from source", _source, _deviceID, "buttonIndex=", _buttonNr,
               _clickType);

      std::shared_ptr<ModelDeferredButtonEvent> mEvent = std::make_shared<ModelDeferredButtonEvent>(_source, _deviceID, _buttonNr, _clickType);
      mEvent->clearTimestamp();  // force immediate event processing
      m_DeferredEvents.push_back(mEvent);
      return;
    }

    for (std::shared_ptr<ModelDeferredEvent> evt : m_DeferredEvents) {
      std::shared_ptr<ModelDeferredButtonEvent> pEvent = std::dynamic_pointer_cast <ModelDeferredButtonEvent> (evt);
      if (pEvent == NULL) {
        continue;
      }
      if ((pEvent->getDeviceID() == _deviceID) &&
          (pEvent->getButtonIndex() == _buttonNr) &&
          (pEvent->getSource() == _source)) {
        // holding, adjust the old event's timestamp to keep it active
        if ((_clickType == ClickTypeHR) &&
            ((pEvent->getClickType() == ClickTypeHS) || (pEvent->getClickType() == ClickTypeHR))) {
          DS_DEBUG("ActionDeviceFilter: hold repeat, update ClickType from", pEvent->getClickType(), "to", _clickType);

          pEvent->setTimestamp();
          pEvent->incRepeatCount();
          pEvent->setClickType(_clickType);
          return;
        }
        if ((_clickType == ClickTypeHE) &&
            ((pEvent->getClickType() == ClickTypeHR) || (pEvent->getClickType() == ClickTypeHS)))  {
          DS_DEBUG("ActionDeviceFilter: hold end, update ClickType from", pEvent->getClickType(), "to", _clickType);

          pEvent->incRepeatCount();
          pEvent->setClickType(_clickType);
          return;
        }
        // going through 1T..4T
        if (((pEvent->getClickType() == ClickType1T) && (_clickType == ClickType2T)) ||
            ((pEvent->getClickType() == ClickType2T) && (_clickType == ClickType3T)) ||
            ((pEvent->getClickType() == ClickType3T) && (_clickType == ClickType4T))) {
          DS_DEBUG("ActionDeviceFilter: update ClickType from", pEvent->getClickType(), "to", _clickType);

          pEvent->setClickType(_clickType);
          return;
        }
        // going through 1C..3C/3T..4T
        if (((pEvent->getClickType() == ClickType1C) && (_clickType == ClickType3C)) ||
            ((pEvent->getClickType() == ClickType2C) && (_clickType == ClickType3C)) ||
            ((pEvent->getClickType() == ClickType1C) && (_clickType == ClickType2T)) ||
            ((pEvent->getClickType() == ClickType1C) && (_clickType == ClickType3T)) ||
            ((pEvent->getClickType() == ClickType2C) && (_clickType == ClickType3T)) ||
            ((pEvent->getClickType() == ClickType3C) && (_clickType == ClickType4T))) {
          DS_DEBUG("ActionDeviceFilter: update ClickType from", pEvent->getClickType(), "to", _clickType);

          pEvent->setClickType(_clickType);
          return;
        }
      }
    }

    DS_DEBUG("DeviceActionFilter: deferred processing, from source ", _source, _deviceID, "buttonIndex=", _buttonNr,
             _clickType);

    std::shared_ptr<ModelDeferredButtonEvent> mEvent = std::make_shared<ModelDeferredButtonEvent>(_source, _deviceID, _buttonNr, _clickType);
    m_DeferredEvents.push_back(mEvent);
  } // onDeviceActionFiltered

  void ModelMaintenance::onDeviceNameChanged(dsuid_t _meterID,
                                             const devid_t _deviceID,
                                             const std::string& _name) {
    DS_INFO("Device name changed on the bus.", _meterID, "bus-id:", _deviceID);
    try {
      std::shared_ptr<DSMeter> pMeter =
        m_pApartment->getDSMeterByDSID(_meterID);
      DeviceReference ref = pMeter->getDevices().getByBusID(_deviceID, pMeter);
      ref.getDevice()->setName(_name);
    } catch(ItemNotFoundException& e) {
      DS_DEBUG("onDeviceNameChanged: Item not found:", e.what());
    }
  } // onDeviceNameChanged

  void ModelMaintenance::onDsmNameChanged(dsuid_t _meterID,
                                          const std::string& _name)
  {
    DS_INFO("dSM name changed on the bus.", _meterID);
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      pMeter->setName(_name);

    } catch (ItemNotFoundException& e) {
      DS_DEBUG("onDsmNameChanged: Item not found:", e.what());
    }
  }

  void ModelMaintenance::onDsmFlagsChanged(dsuid_t dsm, const std::bitset<8> flags) {
    DS_INFO_ENTER(dsm, flags);
    try {
      m_pApartment->getDSMeterByDSID(dsm)->setPropertyFlags(flags);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to apply dsm flags", e, dsm, flags);
    }
  }

  void ModelMaintenance::muteFire(CallOrigin origin) {
    // TODO(someday): Once signal concept is implemented,
    // this logic shall be go into mute fire signal handler
    {
      auto fireMuteEnabled = m_pApartment->getNonScriptState(StateName::FireMuteEnabled)->getState();
      DS_REQUIRE(fireMuteEnabled == State_Active, "Mute fire requires active fireMuteEnabled.", fireMuteEnabled);
    }
    auto state = m_pApartment->getNonScriptState(StateName::Fire);
    if (auto oldValue = state->getEnumState<FireOptions>()) {
      switch (*oldValue) {
        case FireOptions::active:
          DS_NOTICE("Muting active fire.", state, state->getType());
          state->setEnumState(origin, FireOptions::activeMuted);

          try {
            m_pApartment->getBusInterface()->getActionRequestInterface()->muteFire();
          } catch (const std::exception& e) {
            DS_ERROR("Failed to broadcast mute fire to ds485 bus.", e);
          }
          break;
        case FireOptions::inactive:
          DS_NOTICE("Ignoring mute fire in inactive fire.");
          break;
        case FireOptions::activeMuted:
          DS_NOTICE("Ignoring mute fire in activeMuted fire.");
          break;
        default:
          DS_FAIL_REQUIRE("Invalid fire state.", int(*oldValue));
      }
    } else {
      DS_FAIL_REQUIRE("Invalid fire state.");
    }
  }

  void ModelMaintenance::onDeviceCallScene(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const bool _forced, const std::string _token, ds::DateTime timeStamp) {
    try {
      if(_sceneID < 0 || _sceneID > MaxSceneNumber) {
        DS_ERROR("onDeviceCallScene: _sceneID is out of bounds", _dsMeterID, _deviceID, _sceneID);
        return;
      }
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_dsMeterID);
      try {
        DS_DEBUG("OnDeviceCallScene:", _dsMeterID, "for", _deviceID, _sceneID);
        DeviceReference devRef = mod->getDevices().getByBusID(_deviceID, _dsMeterID);
        if(SceneHelper::rememberScene(_sceneID & 0x00ff)) {
          devRef.getDevice()->setLastCalledScene(_sceneID & 0x00ff);
        }

        m_pApartment->proxy().callScene(devRef.getDSID(), _sceneID, _forced, timeStamp);

        std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> event = std::make_shared<Event>(EventName::CallScene, pDevRev);
        event->setProperty("sceneID", intToString(_sceneID));
        event->setProperty("callOrigin", _origin);
        event->setProperty("originToken", _token);
        if (_forced) {
          event->setProperty("forced", "true");
        }
        raiseEvent(event);
      } catch(ItemNotFoundException& e) {
        DS_ERROR("OnDeviceCallScene: Could not find device with", _deviceID, "on dsMeter", _dsMeterID, _sceneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnDeviceCallScene: Could not find dsMeter with", _dsMeterID);
    }
  } // onDeviceCallScene

  void ModelMaintenance::onDeviceSaveScene(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const std::string _token) {
    try {
      if(_sceneID < 0 || _sceneID > MaxSceneNumber) {
        DS_ERROR("OnDeviceSaveScene: _sceneID is out of bounds", _dsMeterID, "device ", _deviceID, _sceneID);
        return;
      }
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_dsMeterID);
      try {
        DS_DEBUG("OnDeviceSaveScene:", _dsMeterID, "for", _deviceID, _sceneID);
        DeviceReference devRef = mod->getDevices().getByBusID(_deviceID, _dsMeterID);

        m_pApartment->proxy().saveScene(devRef.getDSID(), _sceneID);
      } catch(ItemNotFoundException& e) {
        DS_ERROR("OnDeviceSaveScene: Could not find device with", _deviceID, _dsMeterID, _sceneID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnDeviceSaveScene: Could not find dsMeter", _dsMeterID);
    }
  }

  void ModelMaintenance::onDeviceUndoScene(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const std::string _token, ds::DateTime timeStamp)
  {
    try {
      if(_sceneID < -1 || _sceneID > MaxSceneNumber) {
        DS_ERROR("onDeviceUndoScene: Scene number is out of bounds. device short address:", _deviceID, _sceneID);
        return;
      }
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_dsMeterID);
      DS_INFO("OnDeviceUndoScene:", _dsMeterID, _deviceID, _sceneID);
      DeviceReference devRef = mod->getDevices().getByBusID(_deviceID, _dsMeterID);
      SetLastButOneCalledSceneAction act(_sceneID);
      act.perform(devRef.getDevice());

      m_pApartment->proxy().undoScene(devRef.getDSID(), _sceneID, timeStamp);

      dsuid_t originDSUID = _dsMeterID;
      if((_dsMeterID != DSUID_NULL) && (_originDeviceID != 0)) {
        DeviceReference originDevRef = m_pApartment->getDevices().
          getByBusID(_originDeviceID, _dsMeterID);
        originDSUID = originDevRef.getDSID();
      }

      std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(devRef);
      std::shared_ptr<Event> event = std::make_shared<Event>(EventName::UndoScene, pDevRef);
      event->setProperty("sceneID", intToString(_sceneID));
      event->setProperty("callOrigin", _origin);
      event->setProperty("originToken", _token);
      raiseEvent(event);
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnDeviceUndoScene: Could not find DSMeter", _dsMeterID);
    }
  }

  void ModelMaintenance::onDeviceBlink(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const CallOrigin _origin, const std::string _token) {
    try {
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_dsMeterID);
      try {
        DS_DEBUG("OnDeviceBlink:", _dsMeterID, "for", _deviceID);
        DeviceReference devRef = mod->getDevices().getByBusID(_deviceID, _dsMeterID);
        std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> event = std::make_shared<Event>(EventName::IdentifyBlink, pDevRev);
        event->setProperty("callOrigin", _origin);
        event->setProperty("originToken", _token);
        raiseEvent(event);
      } catch(ItemNotFoundException& e) {
        DS_ERROR("OnDeviceBlink: Could not find device with bus-id", _deviceID, _dsMeterID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("OnDeviceBlink: Could not find dsMeter with bus-id", _dsMeterID);
    }
  } // onDeviceBlink

  void ModelMaintenance::onDeviceActionEvent(const dsuid_t& _dsMeterID, const int _deviceID, const int _buttonNr, const int _clickType) {
    try {
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_dsMeterID);
      DS_DEBUG("onDeviceActionEvent:", _dsMeterID, "for", _deviceID, "button", _buttonNr, "click:", _clickType);
      try {
        DeviceReference devRef = mod->getDevices().getByBusID(_deviceID, _dsMeterID);
        std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> event = std::make_shared<Event>(EventName::ButtonClickBus, pDevRev);
        event->setProperty("clickType", intToString(_clickType));
        event->setProperty("buttonIndex", intToString(_buttonNr));
        raiseEvent(event);
      } catch(ItemNotFoundException& e) {
        DS_ERROR("onDeviceActionEvent: Could not find device with bus-id '", _deviceID, _dsMeterID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("onDeviceActionEvent: Could not find dsMeter with bus-id '", _dsMeterID);
    }
  } // onDeviceActionEvent

  void ModelMaintenance::onDeviceDirectActionEvent(const dsuid_t& _dsMeterID, const int _deviceID, const int _actionID) {
    try {
      std::shared_ptr<DSMeter> mod = m_pApartment->getDSMeterByDSID(_dsMeterID);
      DS_DEBUG("onDeviceDirectActionEvent:", _dsMeterID, "for", _deviceID, "action:", _actionID);
      try {
        DeviceReference devRef = mod->getDevices().getByBusID(_deviceID, _dsMeterID);
        std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> event = std::make_shared<Event>(EventName::ButtonDeviceAction, pDevRev);
        event->setProperty("actionID", intToString(_actionID));
        raiseEvent(event);
      } catch(ItemNotFoundException& e) {
        DS_ERROR("onDeviceDirectActionEvent: Could not find device with bus-id '", _deviceID, _dsMeterID);
      }
    } catch(ItemNotFoundException& e) {
      DS_ERROR("onDeviceDirectActionEvent: Could not find dsMeter with bus-id '", _dsMeterID);
    }
  } // onDeviceDirectActionEvent

  void ModelMaintenance::onAddDevice(const dsuid_t& _dsMeterID, const int _zoneID, const int _devID) {
    DS_DEBUG("Device discovered:");
    try {
      DeviceReference devRef = m_pApartment->getDSMeterByDSID(_dsMeterID)->getDevices().getByBusID(_devID, _dsMeterID);
      DS_DEBUG("  DSID   : ", ds::hideArg(devRef.getDSID()));
      {
        std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
        mEvent->setProperty("action", "added");
        m_eventSink.pushEvent(mEvent);
      }
    } catch(ItemNotFoundException& e) {}
    DS_DEBUG("  DSMeter: ", ds::hideArg(_dsMeterID));
    DS_DEBUG("  Zone:      ", ds::hideArg(_zoneID));
    DS_DEBUG("  BusID:     ", ds::hideArg(_devID));

    rescanDevice(_dsMeterID, _devID);
    // model_ready sets initializing flag to false and we want to perform
    // this check only if a new device was discovered after the model ready
    // event
    if (m_IsInitializing) {
      return;
    }

    try {
      const auto devRef = m_pApartment->getDSMeterByDSID(_dsMeterID)->getDevices().getByBusID(_devID, _dsMeterID);
      const auto pDevRef = std::make_shared<DeviceReference>(devRef);
      const auto device = devRef.getDevice();

      // if the newly added devices provides any sensors that are not
      // yet assigned in the zone: assign it automatically, UC 1.1, including
      // UC 8.1 check
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      StructureManipulator manipulator(*m_pApartment);
      manipulator.autoAssignZoneSensors(*zone);

      pollSensors(pDevRef);
    } catch (const std::exception& ex) {
      DS_ERROR("Could not check sensor assignments for device", _devID, _dsMeterID, _zoneID, ". ", ex);
    }
  } // onAddDevice

  void ModelMaintenance::onRemoveDevice(const dsuid_t& _dsMeterID, const int _zoneID, const int _devID) {
    DS_DEBUG("Device disappeared:");
    try {
      DeviceReference devRef = m_pApartment->getDSMeterByDSID(_dsMeterID)->getDevices().getByBusID(_devID, _dsMeterID);
      DS_DEBUG("  DSID   : ", ds::hideArg(devRef.getDSID()));
      {
        std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
        mEvent->setProperty("action", "removed");
        m_eventSink.pushEvent(mEvent);
      }
    } catch(ItemNotFoundException& e) {}
    DS_DEBUG("  DSMeter: ", ds::hideArg(_dsMeterID));
    DS_DEBUG("  Zone:      ", ds::hideArg(_zoneID));
    DS_DEBUG("  BusID:     ", ds::hideArg(_devID));

    try {
      std::shared_ptr<DSMeter> dsMeter = m_pApartment->getDSMeterByDSID(_dsMeterID);
      std::shared_ptr<Device> device = dsMeter->getDevices().getByBusID(_devID, _dsMeterID).getDevice();
      device->setIsPresent(false);

      // save time when device went inactive to apartment.xml
      addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));

    } catch(ItemNotFoundException& e) {
      DS_DEBUG("Removed device", _dsMeterID, " is not known to us");
    }
  } // onRemoveDevice

  void ModelMaintenance::onJoinedDSMeter(const dsuid_t& _dSMeterID) {
    std::shared_ptr<DSMeter> meter;

    DS_DEBUG("onJoinedDSMeter: dSM Connected:", _dSMeterID);
    try {
      meter = m_pApartment->getDSMeterByDSID(_dSMeterID);
    } catch (ItemNotFoundException& e) {
      DS_DEBUG("dSM is ready, but it is not yet known, re-discovering devices");
      discoverDS485Devices();
      return;
    }
    meter->setIsPresent(true);
    meter->setIsConnected(true);
    meter->setIsValid(false);

    // synchronize dss with dsm zone sensors
    // in case a dsm crashed during zone sensor assignment.
    m_pMeterMaintenance->triggerMeterSynchronization();

    // synchronize vdc heating assignment in case of vdsm migration
    if (BusMember_vDC == meter->getBusMemberType()) {
      synchronizeHeatingAssignment(_dSMeterID);
    }

    try {
      Set devices = m_pApartment->getDevices().getByLastKnownDSMeter(_dSMeterID);
      for (int i = 0; i < devices.length(); i++) {
        devices[i].getDevice()->setIsValid(false);
      }
    } catch(ItemNotFoundException& e) {
      DS_DEBUG("onJoinedDSMeter: error resetting device valid flags:", e.what());
    }
  } // onJoinedDSMeter

  void ModelMaintenance::onLostDSMeter(const dsuid_t& _dSMeterID) {
    DS_DEBUG("dSM disconnected:", _dSMeterID);
    try {
      std::shared_ptr<DSMeter> meter =
                              m_pApartment->getDSMeterByDSID(_dSMeterID);
      meter->setIsPresent(false);
      meter->setIsConnected(false);
      Set devices = meter->getDevices();
      for (int i = 0; i < devices.length(); i++) {
        devices[i].getDevice()->setIsConnected(false);
      }
    } catch(ItemNotFoundException& e) {
      DS_DEBUG("Lost dSM", _dSMeterID, " was not in our list");
    }
  }

  void ModelMaintenance::onDeviceConfigChanged(const dsuid_t& _dsMeterID, int _deviceID,
                                               int _configClass, int _configIndex, int _value) {
    try {
      DeviceReference devRef = m_pApartment->getDevices().getByBusID(_deviceID, _dsMeterID);
      std::shared_ptr<Device> device = devRef.getDevice();

      m_pApartment->proxy().updateDeviceConfig(device->getDSID(), _configClass, _configIndex, _value);
      if(_configClass == CfgClassFunction) {
        if (_configIndex == CfgFunction_Mode) {
          device->setOutputMode(_value);
        } else if (_configIndex == CfgFunction_ButtonMode) {
          device->setButtonID(_value & 0xf);
        } else if (_configIndex == CfgFunction_PbGroup) {
          device->setButtonActiveGroup(_value);
        } else if (_configIndex == CfgFunction_LTMode) {
          device->setButtonInputMode(static_cast<ButtonInputMode>(_value));
        }
      } else if (_configClass == CfgClassDevice) {
        if(_configIndex >= 0x40 && _configIndex <= 0x40 + 8 * 3) {
          uint8_t inputIndex = (_configIndex - 0x40) / 3;
          uint8_t offset = (_configIndex - 0x40) % 3;
          switch (offset) {
          case 0:
            device->setBinaryInputTargetId(inputIndex, _value);
            break;
          case 1:
            device->setBinaryInputType(inputIndex, static_cast<BinaryInputType>(_value & 0xff));
            break;
          case 2:
            device->setBinaryInputId(inputIndex, static_cast<BinaryInputId>(_value >> 4));
            break;
          }
        }
      }
      {
        std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(devRef);
        std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
        mEvent->setProperty("action", "configure");
        mEvent->setProperty("configClass", intToString(_configClass));
        mEvent->setProperty("configIndex", intToString(_configIndex));
        mEvent->setProperty("configValue", intToString(_value));
        m_eventSink.pushEvent(mEvent);
      }
    } catch(std::runtime_error& e) {
      DS_ERROR("Error updating config of device: ", e.what());
    }
  } // onDeviceConfigChanged

  void ModelMaintenance::onDeviceConfigRead(const dsuid_t& _dsMeterID, int _deviceID,
                                               int _configClass, int _configIndex, int _value) {
    try {
      DeviceReference devRef = m_pApartment->getDevices().getByBusID(
                                                    _deviceID, _dsMeterID);
      std::shared_ptr<Device> device = devRef.getDevice();

      m_pApartment->proxy().updateDeviceConfig(device->getDSID(), _configClass, _configIndex, _value);
    } catch(std::runtime_error& e) {
      DS_ERROR(std::string("Error handling config read event of device: ") + e.what());
    }
  } // onDeviceConfigChanged

  void ModelMaintenance::onSensorEvent(dsuid_t _meterID,
                                       const devid_t _deviceID,
                                       const int& _eventIndex) {
    try {
      std::shared_ptr<DSMeter> pMeter =
        m_pApartment->getDSMeterByDSID(_meterID);
      DeviceReference devRef = pMeter->getDevices().getByBusID(_deviceID, pMeter);
      std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);

      std::string eventName = pDevRev->getDevice()->getSensorEventName(_eventIndex);
      if (eventName.empty()) {
        eventName = "event" + intToString(_eventIndex);
      }
      std::shared_ptr<Event> pEvent;
      pEvent.reset(new Event("deviceSensorEvent", pDevRev));
      pEvent->setProperty("sensorEvent", eventName);
      pEvent->setProperty("sensorIndex", "event" + intToString(_eventIndex));
      raiseEvent(pEvent);
    } catch(ItemNotFoundException& e) {
      DS_WARNING("onSensorEvent: Datamodel failure:", e.what());
    }
  } // onSensorEvent

  void ModelMaintenance::onBinaryInputEvent(dsuid_t _meterID,
      const devid_t _deviceID, const int& _eventIndex, BinaryInputType _eventType, BinaryInputStateValue _state) {
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      DeviceReference devRef = pMeter->getDevices().getByBusID(_deviceID, pMeter);
      std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);

      raiseEvent(createDeviceBinaryInputEvent(pDevRev, _eventIndex, _eventType,
                                              _state));

      devRef.getDevice()->handleBinaryInputEvent(_eventIndex, _state);

      // increment event counter as last step to catch possible
      // data model exceptions in above sequence
      pMeter->incrementBinaryInputEventCount();
    } catch(const std::exception& e) {
      DS_WARNING("onBinaryInputEvent:", e.what());
    }
  } // onBinaryInputEvent

  void ModelMaintenance::onCircuitPowerStateChange(dsuid_t _meterID,
                                                   const int& _baseIndex,
                                                   const int& _stateMask,
                                                   const int& _stateValue) {
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      for (int i = 0; i < 16; i++) {
        if ((_stateMask & (1 << i)) == 0) {
          continue;
        }
        std::shared_ptr<State> pState = pMeter->getPowerState(i);
        pState->setState(CallOrigin::system, ((_stateValue & (1 << i)) > 0) ? State_Active : State_Inactive);
      }
    } catch(ItemNotFoundException& e) {
      DS_WARNING("onCircuitPowerStateChange: Datamodel failure:", e.what());
    }
  } // onCircuitPowerStateChange

  void ModelMaintenance::onSensorValue(dsuid_t _meterID,
                                       const devid_t _deviceID,
                                       const int& _sensorIndex,
                                       const int& _sensorValue) {
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      DeviceReference devRef = pMeter->getDevices().getByBusID(_deviceID, pMeter);
      std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
      std::shared_ptr<Device> pDev = devRef.getDevice();

      // binary input state synchronization event
      if (_sensorIndex == 32) {
        int index = 0;
        for (auto&& binaryInput : pDev->getBinaryInputs()) {
          auto&& state = binaryInput->getState();
          int oldState = state.getState();
          auto binaryInputState =
              (_sensorValue & (1 << index)) ? BinaryInputStateValue::Active : BinaryInputStateValue::Inactive;
          pDev->handleBinaryInputEvent(index, binaryInputState);
          int newState = state.getState();
          if (newState != oldState) {
            raiseEvent(createDeviceBinaryInputEvent(pDevRev, index,
                                                    binaryInput->m_inputType,
                                                    binaryInputState));
          }
          index++;
        }

      // device status and error event
      } else if (_sensorIndex <= 31 && _sensorIndex >= 16) {
        raiseEvent(createDeviceStatusEvent(pDevRev, _sensorIndex,
                                           _sensorValue));

      // regular sensor value event
      } else if (_sensorIndex <= 15) {
        auto sensorType = SensorType::UnknownType;
        try {
          pDev->setSensorValue(_sensorIndex, (const unsigned int) _sensorValue);
          const auto& sensor = pDev->getSensor(_sensorIndex);
          sensorType = sensor.m_sensorType;
        } catch (ItemNotFoundException& e) {
          DS_NOTICE("onSensorValue:", e.what());
        }

        m_pApartment->aggregateSensorValue(sensorType, pDev->getDSID(), sensorValueToDouble(sensorType, _sensorValue));
        raiseEvent(createDeviceSensorValueEvent(pDevRev, _sensorIndex,
                                              sensorType, _sensorValue));
      }
    } catch(ItemNotFoundException& e) {
      DS_WARNING("onSensorValue: Datamodel failure:", e.what());
    }
  } // onSensorValue

  void ModelMaintenance::onSensorValueEx(dsuid_t _meterID, const devid_t _deviceID,
      const int& _sensorIndex, const double& _sensorValue, const uint32_t& _sensorAge,
      const uint32_t& _contextId, const std::string& _contextMsg) {
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      DeviceReference devRef = pMeter->getDevices().getByBusID(_deviceID, pMeter);
      std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
      std::shared_ptr<Device> pDev = devRef.getDevice();

      auto sensorType = SensorType::UnknownType;
      try {
        pDev->setSensorValue(_sensorIndex, _sensorValue, _sensorAge);
        const auto& sensor = pDev->getSensor(_sensorIndex);
        sensorType = sensor.m_sensorType;
      } catch (ItemNotFoundException& e) {
        DS_NOTICE("onSensorValueEx: sensor index failure:", e.what());
      }
      m_pApartment->aggregateSensorValue(sensorType, pDev->getDSID(), _sensorValue);
      raiseEvent(createDeviceSensorValueExEvent(pDevRev, _sensorIndex, sensorType,
          _sensorValue, _sensorAge, _contextId, _contextMsg));

    } catch(ItemNotFoundException& e) {
      DS_WARNING("onSensorValueEx: Datamodel failure:", e.what());
    }
  }

  void ModelMaintenance::onZoneSensorValue(const dsuid_t& _meterID,
                                           const dsuid_t& _sourceDevice,
                                           const int& _zoneID,
                                           const int& _groupID,
                                           SensorType _sensorType,
                                           const int& _sensorValue,
                                           const int& _precision) {
    try {
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      std::shared_ptr<Group> group = zone->getGroup(_groupID);

      double fValue = sensorValueToDouble(_sensorType, _sensorValue);
      group->sensorPush(_sourceDevice, _sensorType, fValue);

      // check for a valid dsuid
      dsuid_t devdsuid = _sourceDevice;
      if (_sourceDevice.id[0] == 0x00) {
        uint32_t dsuidType =
            (_sourceDevice.id[10] << 24) |
            (_sourceDevice.id[9] << 16) |
            (_sourceDevice.id[8] << 8) |
            (_sourceDevice.id[7]);
        // the dsm sends a type "0" dsuid but with the serial number only,
        // need to replace with the real sgtin, #10887
        if (dsuidType == 0) {
          try {
            uint32_t serialNumber;
            if (dsuid_get_serial_number(&devdsuid, &serialNumber) == 0) {
              DeviceReference devRef = m_pApartment->getDevices().getBySerial(serialNumber);
              devdsuid = devRef.getDSID();
            }
          } catch (ItemNotFoundException& e) {}
        }
      }
      raiseEvent(createZoneSensorValueEvent(group, _sensorType, _sensorValue, devdsuid));
    } catch(ItemNotFoundException& e) {
      DS_WARNING("onZoneSensorValue: Datamodel failure:", e.what());
    }
  } // onZoneSensorValue

  void ModelMaintenance::onEANReady(dsuid_t _dsMeterID,
                                        const devid_t _deviceID,
                                        const DeviceOEMState_t _state,
                                        const DeviceOEMInetState_t _iNetState,
                                        const unsigned long long _eanNumber,
                                        const int _serialNumber,
                                        const int _partNumber,
                                        const bool _isIndependent,
                                        const bool _isConfigLocked,
                                        const int _pairedDevices,
                                        const bool _isVisible) {
    try {
      DeviceReference devRef = m_pApartment->getDevices().getByBusID(_deviceID, _dsMeterID);
      if (_state == DEVICE_OEM_VALID) {
        devRef.getDevice()->setOemInfo(_eanNumber, _serialNumber, _partNumber, _iNetState, _isIndependent);
        if ((_iNetState == DEVICE_OEM_EAN_INTERNET_ACCESS_OPTIONAL) ||
            (_iNetState == DEVICE_OEM_EAN_INTERNET_ACCESS_MANDATORY)) {
          // query Webservice
          getTaskProcessor().addEvent(
              std::make_shared<OEMWebQuery>(WebserviceConnection::getInstanceMsHub(), devRef.getDevice(),
                                            devRef.getDevice()->getOemProductInfoState(), m_propertySystem));
          devRef.getDevice()->setOemProductInfoState(DEVICE_OEM_LOADING);
        } else {
          devRef.getDevice()->setOemProductInfoState(DEVICE_OEM_NONE);
        }
      }
      devRef.getDevice()->setOemInfoState(_state);
      devRef.getDevice()->setConfigLock(_isConfigLocked);
      devRef.getDevice()->setVisibility(_isVisible);
      // special case BL SK-204, firmware does not provide the number
      // of paired devices so we had to hardcode it in the busscanner;
      // at the same time the TNY reader does read out the visibility bit
      // which we will use below, but we will keep our hardcoded pairing
      // setting
      if (!((devRef.getDevice()->getDeviceType() == DEVICE_TYPE_SK) && (devRef.getDevice()->getDeviceNumber() == 204) &&
            (devRef.getDevice()->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL))) {
        devRef.getDevice()->setPairedDevices(_pairedDevices);
      }
    } catch(std::runtime_error& e) {
      DS_WARNING("Error updating OEM data of device:", e.what());
    }
  } // onEANReady

  void ModelMaintenance::onOEMDataReady(dsuid_t _deviceID,
                                             const DeviceOEMState_t _state,
                                             const std::string& _productName,
                                             const std::string& _iconPath,
                                             const std::string& _productURL,
                                             const std::string& _defaultName,
                                             const std::string& _configLink) {
    try {
      std::shared_ptr<Device> pDevice = m_pApartment->getDeviceByDSID(_deviceID);
      if (_state == DEVICE_OEM_VALID) {
        pDevice->setOemProductInfo(_productName, _iconPath, _productURL, _configLink);
        if (pDevice->getName().empty()) {
          pDevice->setName(_defaultName);
        }
      }
      pDevice->setOemProductInfoState(_state);
      {
        std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(pDevice, m_pApartment);
        std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
        mEvent->setProperty("action", "oemDataChanged");
        m_eventSink.pushEvent(mEvent);
      }
    } catch(std::runtime_error& e) {
      DS_WARNING("Error updating OEM data of device:", e.what());
    }
  } // onEANReady

  void ModelMaintenance::onDeviceOEMDataUpdateProductInfoState(
                                             dsuid_t _deviceID,
                                             const DeviceOEMState_t _state) {
    try {
      std::shared_ptr<Device> pDevice = m_pApartment->getDeviceByDSID(_deviceID);
      pDevice->setOemProductInfoState(_state);
      // do not queue the "oemDataChanged" event here, because the oem data
      // has actually not changed in this case
    } catch(std::runtime_error& e) {
      DS_WARNING("Error updating OEM data state of device:", e.what());
    }
  } // onEANReady


  void ModelMaintenance::onHeatingControllerConfig(dsuid_t _dsMeterID, const int _zoneID, std::shared_ptr<void> _spec) {
    std::shared_ptr<ZoneHeatingConfigSpec_t> config =
        std::static_pointer_cast<ZoneHeatingConfigSpec_t> (_spec);

    try {
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      zone->setHeatingConfigSpec(*config.get());
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error on heating control config event, item not found:", e.what());
      return;
    }

    DS_INFO("onHeatingControllerConfig:", _dsMeterID);
    raiseEvent(createHeatingControllerConfig(_zoneID, _dsMeterID, *config));
  } // onHeatingControllerConfig

  void ModelMaintenance::onHeatingControllerValues(dsuid_t _dsMeterID, const int _zoneID, std::shared_ptr<void> _spec) {
    std::shared_ptr<ZoneHeatingOperationModeSpec_t> values =
        std::static_pointer_cast<ZoneHeatingOperationModeSpec_t> (_spec);

    std::shared_ptr<Zone> zone;
    try {
      zone = m_pApartment->getZone(_zoneID);
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error on heating control value event, item not found:", e.what());
      return;
    }

    DS_INFO("onHeatingControllerValues:", _dsMeterID);

    const TemperatureConfig& hProp = zone->temperatureConfig();
    (void)zone->getHeatingStatus();

    raiseEvent(createHeatingControllerValue(_zoneID, _dsMeterID, hProp, *values));
  } // onHeatingControllerValues

  void ModelMaintenance::onHeatingControllerState(dsuid_t _dsMeterID, const int _zoneID, const int _State) {

    try {
      std::shared_ptr<Zone> zone = m_pApartment->getZone(_zoneID);
      zone->setHeatingControlState(_State);
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error on heating state event, item not found:", e.what());
    }

    raiseEvent(createHeatingControllerState(_zoneID, _dsMeterID, _State));
  } // onHeatingControllerState

  void ModelMaintenance::onClusterConfigLock(const int _clusterID, const bool _configurationLock) {
    try {
      std::shared_ptr<Cluster> cluster = m_pApartment->getCluster(_clusterID);
      cluster->setConfigurationLocked(_configurationLock);
      raiseEvent(createClusterConfigLockEvent(cluster, 0, _clusterID, _configurationLock, CallOrigin::dsmApi));
      addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error on cluster config lock change, cluster not found:", e.what());
    }
  } // onClusterConfigLock

  void ModelMaintenance::onClusterLockedScenes(const int _clusterID, const std::vector<int> &_lockedScenes) {
    try {
      std::shared_ptr<Cluster> cluster = m_pApartment->getCluster(_clusterID);
      cluster->setLockedScenes(_lockedScenes);
      addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error on cluster locked scenes change, cluster not found:", e.what());
    }
  } // onClusterConfigLock

  void ModelMaintenance::onGenericEvent(const GenericEventType_t _eventType, const std::shared_ptr<GenericEventPayload_t> &_pPayload, const CallOrigin _origin) {
    /*
     * Generic Events
        0 no event
        1 Sun         {active=1, inactive=2} {direction}
        2 Frost       {active=1, inactive=2}
        3 HeatingMode {Off=0, Heat=1, Cold=2, Auto=3} sent by the heating system, indicates type of energy delivered into rooms
        4 Service     {active=1, inactive=2}
     */
    switch (_eventType) {
    case ge_sun:
      {
        if (_pPayload->length != 2) {
          break;
        }

        unsigned value = _pPayload->payload[0];
        CardinalDirection_t direction =
          static_cast<CardinalDirection_t>(_pPayload->payload[1]);

        // arg0: value {active=1, inactive=2}
        // arg1: {direction}
        if (!valid(direction) || value == 0 || value > 2) {
          DS_INFO("generic sunshine: invalid value:", value, "direction:", direction);
          break;
        }

        raiseEvent(createGenericSignalSunshine(value, direction, _origin));
      }
      break;

    case ge_frost:
      {
        if (_pPayload->length != 1) {
          break;
        }

        unsigned value = _pPayload->payload[0];

        // arg0: {active=1, inactive=2}
        if (value != 1 && value != 2) {
          DS_INFO("generic frost: invalid", value);
          break;
        }

        raiseEvent(createGenericSignalFrostProtection(value, _origin));
      }
      break;

    case ge_heating_mode:
      {
        if (_pPayload->length != 1) {
          break;
        }

        unsigned value = _pPayload->payload[0];

        // sent by the heating system, indicates type of energy delivered into rooms
        // arg0: {Off=0, Heat=1, Cold=2, Auto=3}
        if (value > 3) {
          DS_INFO("generic heating mode: invalid", value);
          break;
        }
        if (value == 3) {
          DS_WARNING("generic heating mode: prevent auto value");
          break;
        }

        raiseEvent(createGenericSignalHeatingModeSwitch(static_cast<HeatingModeSwitchValue>(value), _origin));
      }
      break;

    case ge_service:
      {
        if (_pPayload->length != 1) {
          break;
        }

        unsigned value = _pPayload->payload[0];

        // arg0: {active=1, inactive=2}
        if (value != 1 && value != 2) {
          DS_INFO("generic building service: invalid", value);
          break;
        }

        raiseEvent(createGenericSignalBuildingService(_pPayload->payload[0], _origin));
      }
      break;
    case ge_none: //no break
    default: DS_DEBUG("unknown generic event:", _eventType); break;
    }
  } // onGenericEvent

  void ModelMaintenance::onOperationLock(const int _zoneID, const int _groupID, bool _lock, CallOrigin _callOrigin) {
    if (!isCluster(_groupID)) {
      return;
    }
    try {
      std::shared_ptr<Cluster> cluster = m_pApartment->getCluster(_groupID);
      cluster->setOperationLock(_lock, _callOrigin);

      Set s = cluster->getDevices();

      for (int i = 0; i < s.length(); i++) {
        m_pApartment->proxy().setOperationLock(s.get(i).getDSID(), _lock);
      }

      raiseEvent(createOperationLockEvent(cluster, _zoneID, _groupID, _lock, _callOrigin));
    } catch (ItemNotFoundException& e) {
      DS_WARNING("Unknown cluster:", e.what());
    }
  }

  void ModelMaintenance::onAutoClusterMaintenance(dsuid_t _deviceID)
  {
    try {
      std::shared_ptr<Device> device = m_pApartment->getDeviceByDSID(_deviceID);
      AutoClusterMaintenance maintenance(m_pApartment);
      maintenance.consistencyCheck(*device.get());
      addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error cluster consistency check, item not found:", e.what());
    }
  }

  void ModelMaintenance::onAutoClusterCleanup() {
    try {
      AutoClusterMaintenance maintenance(m_pApartment);
      maintenance.cleanupEmptyCluster();
      addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error cluster consistency check, item not found:", e.what());
    }
  }

  void ModelMaintenance::onMeterReady() {
    if (m_IsInitializing) {
      setupWebUpdateEvent();
      m_IsInitializing = false;

      // handle delayed model dirty
      if (m_IsDirty) {
        m_IsDirty = false;
        addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
      }
    }
  }

  void ModelMaintenance::onDeviceDataReady(dsuid_t _meterID,
                                       const devid_t _deviceID,
                                       const int& _pairedDevices,
                                       const bool& _visible) {
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      DeviceReference devRef = pMeter->getDevices().getByBusID(_deviceID, pMeter);
      std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(devRef);
      std::shared_ptr<Device> pDev = devRef.getDevice();
      // special case BL SK-204, firmware does not provide the number
      // of paired devices so we had to hardcode it in the busscanner;
      // at the same time the TNY reader does read out the visibility bit
      // which we will use below, but we will keep our hardcoded pairing
      // setting
      if (!((pDev->getDeviceType() == DEVICE_TYPE_SK) && (pDev->getDeviceNumber() == 204) &&
            (pDev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL))) {
        pDev->setPairedDevices(_pairedDevices);
      }

      pDev->setVisibility(_visible);
      addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
    } catch(ItemNotFoundException& e) {
      DS_WARNING("onSensorValue: Datamodel failure:", e.what());
    }
  } // onDeviceDataReady

  void ModelMaintenance::onDsmStateChange(dsuid_t _meterID,
                                          const uint8_t& _state) {
    DS_NOTICE("dSM line state change:", _meterID, _state);
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      pMeter->setState(_state);
      if (_state == DSM_STATE_IDLE) {
        std::unique_lock<std::mutex> lock(m_readoutTasksMutex);
        std::vector<std::pair<dsuid_t, std::shared_ptr<Task> > >::iterator it;
        for (it = m_deviceReadoutTasks.begin(); it != m_deviceReadoutTasks.end();) {
          dsuid_t id = (*it).first;
          if (dsuid_equal(&id, &_meterID)) {
            DS_DEBUG("onDsmStateChange: scheduling device readout task on dSM", _meterID);
            m_taskProcessorMaySleep.addEvent((*it).second);
            it = m_deviceReadoutTasks.erase(it);
          } else {
            it++;
          }
        }
      }
    } catch(ItemNotFoundException& e) {
      DS_WARNING("onDsmStateChange: Datamodel failure:", e.what());
    }
  } // onDsmStateChange

  void ModelMaintenance::onVdceEvent(const VdceModelEvent& event)
  {
    try {
      std::shared_ptr<Device> pDev = m_pApartment->getDeviceByDSID(event.m_deviceDSID);
      std::shared_ptr<DeviceReference> pDevRev = std::make_shared<DeviceReference>(pDev->getDSID(), &pDev->getApartment());

      std::vector<dsModel::StateStatus> stateStatuses{};
      for (auto&& state : event.m_states.getContainer()) {
        const std::string& name = state.first;
        const std::string& value = state.second;
        stateStatuses.push_back(dsModel::StateStatus{.id = dsModel::StateId{name}, .value = dsModel::StateOptionId{value}});

        raiseEvent(createDeviceStateEvent(pDevRev, name, value));
        pDev->setStateValue(name, value);
      }
      pDev->updateStateStatuses(std::move(stateStatuses));
      for (auto&& dynactions : event.m_dynamicActions.getContainer()) {
        const std::string& name = dynactions.first;
        const std::string& value = dynactions.second;
        raiseEvent(createDeviceActionChangedEvent(pDevRev, name, value));
      }
      for (auto it = event.m_events.begin(); it != event.m_events.end(); it++) {
        const auto& deviceEvent = *it;
        raiseEvent(createDeviceEventEvent(pDevRev, deviceEvent));
      }
      std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::IdentifyDevice, pDevRev);
      mEvent->setContext("identifyDevice");
      m_eventSink.pushEvent(mEvent);

    } catch(ItemNotFoundException& e) {
      DS_WARNING("onVdceEvent: Datamodel failure:", e.what());
    }
  }

  void ModelMaintenance::onUpstreamShort(dsuid_t _meterID, uint16_t _deviceID, uint8_t _circuitID, uint8_t _isSensor,
      uint8_t _keyType, uint8_t _clickType) {
    try {
      const auto& pMeter = m_pApartment->getDSMeterByDSID(_meterID);
      const auto& pDevRef = std::make_shared<DeviceReference>(pMeter->getDevices().getByBusID(_deviceID, pMeter));

      std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::IdentifyDevice, pDevRef);
      mEvent->setContext("identifyDevice");
      mEvent->setProperty("circuitID", _circuitID);
      mEvent->setProperty("isSensor", _isSensor);
      mEvent->setProperty("keyType", _keyType);
      mEvent->setProperty("clickType", _clickType);
      m_eventSink.pushEvent(mEvent);
    } catch (const std::exception& e) {
      DS_ERROR(e);
    }
  }

  void ModelMaintenance::rescanDevice(const dsuid_t& _dsMeterID, const int _deviceID) {
    BusScanner
      scanner(
        *m_pStructureQueryBusInterface,
        *m_pApartment,
        *this);
    try {
      std::shared_ptr<DSMeter> dsMeter = m_pApartment->getDSMeterByDSID(_dsMeterID);
      scanner.scanDeviceOnBus(dsMeter, _deviceID);
      std::shared_ptr<Device> device = dsMeter->getDevices().getByBusID(_deviceID, dsMeter).getDevice();
      if (device->getBinaryInputCount() > 0) {
        scanner.syncBinaryInputStates(dsMeter, device);
      }
     if (device->getZoneID() == ZONE_ID_BROADCAST) {
       DS_WARNING("Device", ds::hideArg(device->getDSID()), "assigned to zone 0 after rescan");
      }
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error scanning device, item not found:", e.what());
    } catch(std::runtime_error& e) {
      DS_WARNING("Error scanning device:", e.what());
    }
  } // rescanDevice

  void ModelMaintenance::onDeviceSetOutputValue(const dsuid_t& dsuidMeter, int deviceID, uint8_t value, ds::DateTime timeStamp) {
    try {
      std::shared_ptr<DSMeter> dsMeter = m_pApartment->getDSMeterByDSID(dsuidMeter);
      std::shared_ptr<Device> device = dsMeter->getDevices().getByBusID(deviceID, dsMeter).getDevice();
      m_pApartment->proxy().setOutputValue(device->getDSID(), value, timeStamp);
    } catch(ItemNotFoundException& e) {
      DS_WARNING("Error intercepting setOutputValue, device not found:", e.what());
    } catch(std::runtime_error& e) {
      DS_WARNING("Error intercepting setOutputValue:", e.what());
    }
  } // onDeviceSetOutputValue

  void ModelMaintenance::onZoneGroupSetOutputValue(int zoneID, int groupID, uint8_t value, ds::DateTime timeStamp) {
    try {
      m_pApartment->proxy().setOutputValue(zoneID, groupID, value, timeStamp);
    } catch (std::runtime_error& e) {
      DS_WARNING("Error intercepting setZoneGroupOutputValue:", e.what());
    }
  } // onZoneGroupSetOutputValue

  void ModelMaintenance::onZoneGroupSetSceneLocalPrio(uint16_t zoneID, uint8_t groupID, uint8_t sceneID) {
    try {
      m_pApartment->proxy().setSceneLocalPrio(zoneID, groupID, sceneID);
    } catch(std::runtime_error& e) {
      DS_WARNING("Error intercepting onZoneGroupSetSceneLocalPrio:", e.what());
    }
  } // onZoneGroupSetSceneLocalPrio

  void ModelMaintenance::onZoneGroupCallSceneMin(uint16_t zoneID, uint8_t groupID, uint8_t sceneID, ds::DateTime timeStamp) {
    try {
      m_pApartment->proxy().callSceneMin(zoneID, groupID, sceneID, timeStamp);
    } catch (const std::exception& e) {
      DS_WARNING("Error intercepting onZoneGroupSetSceneMin:", e.what());
    }
  } // onZoneGroupSetSceneMin


  void ModelMaintenance::setApartment(Apartment* _value) {
    m_pApartment = _value;
    if(m_pApartment != NULL) {
      m_pApartment->setModelMaintenance(this);
    }
  } // setApartment

  void ModelMaintenance::setMetering(metering::Metering* _value) {
    m_pMetering = _value;
  } // setMetering

  void ModelMaintenance::setStructureQueryBusInterface(StructureQueryBusInterface* _value) {
    m_pStructureQueryBusInterface = _value;
  } // setStructureQueryBusInterface

  void ModelMaintenance::setStructureModifyingBusInterface(StructureModifyingBusInterface* _value) {
    m_pStructureModifyingBusInterface = _value;
  } // setStructureModifyingBusInterface


  const std::string ModelMaintenance::kWebUpdateEventName = "ModelMaintenace_updateWebData";

  void ModelMaintenance::setupWebUpdateEvent() {
    if (!DSS::hasInstance()) {
      DS_DEBUG("ModelMaintenance::setupWebUpdateEvent: no DSS instance available; not starting WebUpdateProcess");
      return;
    }
    EventInterpreterInternalRelay* pRelay =
      dynamic_cast<EventInterpreterInternalRelay*>(DSS::getInstance()->getEventInterpreter().getPluginByName(EventInterpreterInternalRelay::getPluginName()));
    m_pRelayTarget = std::make_shared<InternalEventRelayTarget>(boost::ref<EventInterpreterInternalRelay>(*pRelay));

    std::shared_ptr<EventSubscription> updateWebSubscription(
            new dss::EventSubscription(
                kWebUpdateEventName,
                EventInterpreterInternalRelay::getPluginName(),
                DSS::getInstance()->getEventInterpreter(),
                std::shared_ptr<SubscriptionOptions>())
    );
    m_pRelayTarget->subscribeTo(updateWebSubscription);
    m_pRelayTarget->setCallback(boost::bind(&ModelMaintenance::updateWebData, this, _1, _2));
    sendWebUpdateEvent(5);
  } // setupCleanupEvent

  void ModelMaintenance::updateWebData(Event& _event, const EventSubscription& _subscription) {
    std::vector<std::shared_ptr<Device> > deviceVec = m_pApartment->getDevicesVector();
    for (std::shared_ptr<Device> device : deviceVec) {
      if ((device->getOemInfoState() == DEVICE_OEM_VALID) &&
           ((device->getOemProductInfoState() == DEVICE_OEM_VALID) ||
            (device->getOemProductInfoState() == DEVICE_OEM_UNKNOWN))) {
        // query Webservice
        getTaskProcessor().addEvent(std::make_shared<OEMWebQuery>(WebserviceConnection::getInstanceMsHub(), device,
                                                                  device->getOemProductInfoState(), m_propertySystem));
        device->setOemProductInfoState(DEVICE_OEM_LOADING);
      }
    }
    sendWebUpdateEvent();
  } // cleanupTerminatedScripts

  void ModelMaintenance::sendWebUpdateEvent(int _interval) {
    std::shared_ptr<Event> pEvent = std::make_shared<Event>(kWebUpdateEventName);
    pEvent->setProperty("time", "+" + intToString(_interval));
    m_eventSink.pushEvent(pEvent);
  } // sendCleanupEvent

  void ModelMaintenance::pollSensors(std::shared_ptr<DeviceReference> pDevRef) {
    if (!pDevRef) {
      return;
    }

    try {
      std::shared_ptr<Device> device = pDevRef->getDevice();
      DS_REQUIRE(device, "Could not get the device");
      std::shared_ptr<Zone> zone = m_pApartment->getZone(device->getZoneID());
      DS_REQUIRE(zone, "Could not get the zone");

      if (device->getSensorCount()) {
        DeviceBusInterface* busItf = m_pApartment->getBusInterface()->getDeviceBusInterface();

        for (const auto& sensor : device->getSensors()) {
          if (sensor->m_sensorPollInterval == 0) {
            continue;
          }
            if (zone->isZoneSensor(*device, sensor->m_sensorType)) {
              busItf->getSensorValueEx(*device.get(), sensor->m_sensorIndex);
            }
        }
      }
    } catch (const std::exception& e) {
      DS_WARNING("ModelMaintenance::pollSensors: Could not query sensor. Message:", e);
    }
  } // pollSensors

  void ModelMaintenance::scheduleDeviceReadout(const dsuid_t &_dSMeterID,
                                               std::shared_ptr<Task> task) {
    try {
      std::shared_ptr<DSMeter> pMeter = m_pApartment->getDSMeterByDSID(_dSMeterID);
      if (pMeter->getState() == DSM_STATE_IDLE) {
        m_taskProcessorMaySleep.addEvent(task);
      } else {
        std::unique_lock<std::mutex> lock(m_readoutTasksMutex);
        m_deviceReadoutTasks.push_back(std::make_pair(_dSMeterID, task));
      }
    } catch(ItemNotFoundException& e) {
      DS_WARNING("scheduleDeviceReadou: Datamodel failure:", e.what());
    }
  }

  void ModelMaintenance::publishWebSocketEvent(std::shared_ptr<Event> _event) {
    getTaskProcessor().addEvent(std::make_shared<WebSocketEvent>(_event));
  }
} // namespace dss
