/*
    Copyright (c) 2021 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "shademanipulator.h"
#include "test-device-presence.h"
#include "dss.h"

#include <ds/log.h>

#include <string>

#include "businterface.h"

using namespace dss;
using namespace dss::model;
using namespace std::literals;

static ds::LogChannel dsLogChannel("modelShadeManipulator");

namespace dss::model {

std::optional<dsModel::ShadeType> tryDefaultShadeType(const Device& device) {
  switch (device.hardwareType()) {
    case HardwareType::GR_KL200: return dsModel::ShadeType::rollerShutter;
    case HardwareType::GR_KL210: return dsModel::ShadeType::awningTerrace;
    case HardwareType::GR_KL220: return dsModel::ShadeType::windowBlind;
    case HardwareType::GR_HKL230: return dsModel::ShadeType::windowBlind;
    case HardwareType::GR_KL300:
    case HardwareType::GR_HKL330:
      // all shade types are supported, explicitly return none to indicate that active type
      // is unknown, we might change handling of it upon introducing group 13
      return dsModel::ShadeType::none;
    case HardwareType::GE_KM200:
    case HardwareType::GE_TKM210:
    case HardwareType::GE_SDM200:
    case HardwareType::GE_SDM200JS:
    case HardwareType::GE_SDM200CS:
    case HardwareType::GE_SDM300W:
    case HardwareType::GE_SDM300S:
    case HardwareType::GE_SDS200CW:
    case HardwareType::GE_SDS200CS:
    case HardwareType::GE_SDS220CT:
    case HardwareType::GE_TKM220:
    case HardwareType::GE_TKM230:
    case HardwareType::GE_KL200:
    case HardwareType::GE_UMV200:
    case HardwareType::GE_UMV210:
    case HardwareType::GE_KM300:
    case HardwareType::GE_TKM300:
    case HardwareType::GN_KM200:
    case HardwareType::GN_TKM200:
    case HardwareType::GN_TKM210:
    case HardwareType::GR_TKM200:
    case HardwareType::GR_TKM210:
    case HardwareType::BL_KM200:
    case HardwareType::BL_KM300:
    case HardwareType::BL_SDS200:
    case HardwareType::BL_UMV200:
    case HardwareType::RT_TKM200:
    case HardwareType::RT_SDM200:
    case HardwareType::SW_TKM200:
    case HardwareType::SW_TKM210:
    case HardwareType::SW_AKM200:
    case HardwareType::SW_AKM210:
    case HardwareType::SW_AKM220:
    case HardwareType::SW_KL200:
    case HardwareType::SW_ZWS200:
    case HardwareType::SW_ZWS200F:
    case HardwareType::SW_ZWS200EF:
    case HardwareType::SW_SSL200JS:
    case HardwareType::SW_SSL200FS:
    case HardwareType::SW_SDS200CW:
    case HardwareType::SW_SDS200CS:
    case HardwareType::SW_SDS220CT:
    case HardwareType::SW_SKM300: // no output, no shade attached
    case HardwareType::SW_TNY200:
    case HardwareType::SW_UMR200:
    case HardwareType::SW_SK200:
    case HardwareType::SW_SK201:
    case HardwareType::SW_SK202:
    case HardwareType::BL_SK204:
    case HardwareType::SW_ZWS205:
    case HardwareType::GE_ZWD300W:
    case HardwareType::unknown:
    default: break;
  }
  DS_DEBUG("Unrecognized shade type.", device.getProductID(), device.hardwareType());
  return std::nullopt;
}

namespace {

struct ShadeManipulatorImpl : public ShadeManipulator {
  Apartment& m_apartment;
  DeviceBusInterface& m_deviceBusInterface;

  ShadeManipulatorImpl(Apartment& apartment, DeviceBusInterface& deviceBusInterface)
      : m_apartment(apartment), m_deviceBusInterface(deviceBusInterface) {}

  dsModel::ShadeType shadeType(const Device& device) override {
    if (device.hardwareType() == HardwareType::GR_HKL330 || device.hardwareType() == HardwareType::GR_KL300) {
      if (!device.hasOutput()) {
        return dsModel::ShadeType{};  
      }
    }

    if (auto shadeType = tryDefaultShadeType(device); !shadeType || shadeType != dsModel::ShadeType::none) {
      // not a shade device or only supports fixed shade type
      return shadeType.value_or(dsModel::ShadeType{});
    }

    if (auto shadeType = device.getShadeType(); shadeType) {
      return *shadeType;
    }

    //TODO - Potentially move to proxy cache
    if (device.isPresent()) {
      try {
        DS_NOTICE("shadeType lookup", device.hardwareType(), defaultShadeType(device));
        auto bankValue = m_deviceBusInterface.getDeviceConfig(device, 1, 0x5a);

        // https://myesmart.atlassian.net/wiki/spaces/DSD/pages/8829272143/GR-KL300#Attached-Device
        auto fromBank = [](const uint8_t& bankValue) -> dsModel::ShadeType {
          switch (bankValue) {
            case 0: return dsModel::ShadeType::none;
            case 1: return dsModel::ShadeType::windowBlind;
            case 2: return dsModel::ShadeType::rollerShutter;
            case 3: return dsModel::ShadeType::awningFacade;
            case 4: return dsModel::ShadeType::awningTerrace;
            case 5: return dsModel::ShadeType::awningFacadeHorizontal;
            case 7: return dsModel::ShadeType::curtain;
            case 8: return dsModel::ShadeType::internalBlind;
          }
          return dsModel::ShadeType::none;
        };

        auto shadeType = fromBank(bankValue);
        const_cast<Device&>(device).setShadeType(shadeType);
        return shadeType;
      } catch (const BusApiError& e) {
        if (std::string(e.what()).find("[-20]") != std::string::npos) {
            auto&& dss = *DSS::getInstance();
            auto dsa = makeDsaHttpClient();
            auto& backendVdcs = dss.backendVdcs();
            DS_WARNING("Device failed to react shade type request, pinging device", device.getDSID());
            // side effect: if device is offline dsa will update its presence status
            testDevicePresence(std::locale(""), device, *dsa, backendVdcs);
        } else {
          DS_WARNING("Could not scan DSUID for shade type: ", device.getDSID(), e.what());
        }
        return dsModel::ShadeType::none;
      } catch (const ds::Exception& e) {
        DS_NOTICE("Could not scan DSUID for shade type: ", device.getDSID(), e.what());
        return dsModel::ShadeType::none;
      }
    } else {
      return dsModel::ShadeType::none;
    }

  }

  void setShadeType(Device& device, dsModel::ShadeType shadeType) override {
    DS_REQUIRE(tryDefaultShadeType(device), "non shade device", device.hardwareType());
    DS_REQUIRE(defaultShadeType(device) == dsModel::ShadeType::none, "fixed shade device type");

    if (auto cache = device.getShadeType(); cache && cache == shadeType) {
      DS_NOTICE("setShadeType already set", device.hardwareType(), shadeType);
      return;
    }

    // https://myesmart.atlassian.net/wiki/spaces/DSD/pages/8829272143/GR-KL300#Attached-Device
    auto dssModelShadeType = [](const dsModel::ShadeType& shadeType) -> uint8_t {
      switch (shadeType) {
        case dsModel::ShadeType::none: return 0;
        case dsModel::ShadeType::rollerShutter: return 2;
        case dsModel::ShadeType::windowBlind: return 1;
        case dsModel::ShadeType::awningFacade: return 3;
        case dsModel::ShadeType::awningFacadeHorizontal: return 5;
        case dsModel::ShadeType::awningTerrace: return 4;
        case dsModel::ShadeType::curtain: return 7;
        case dsModel::ShadeType::internalBlind: return 8;
      }
      DS_FAIL_REQUIRE("invalid", shadeType);
    };
    m_deviceBusInterface.setDeviceConfig(device, 1, 0x5a, dssModelShadeType(shadeType));
    device.setShadeType(shadeType);
  }
};

} // namespace

dsModel::ShadeType defaultShadeType(const Device& device) {
  if (auto maybeShadeType = tryDefaultShadeType(device); maybeShadeType) {
    return maybeShadeType.value();
  }
  return {}; // dsModel::ShadeType::none
}

std::unique_ptr<ShadeManipulator> makeShadeManipulator(Apartment& apartment, DeviceBusInterface& deviceBusInterface) {
  return std::make_unique<ShadeManipulatorImpl>(apartment, deviceBusInterface);
}

} // namespace dss::model
