/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "sceneaccess.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "handler/system_states.h"
#include "model/apartment-mock.h"
#include "model/cluster.h"
#include "model/device-mock.h"
#include "model/state.h"

static ds::LogChannel dsLogChannel("sceneAccessTest");
static const char* TAGS = "[sceneaccess]";

using namespace dss;
using namespace std::literals;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

SCENARIO("SceneAccess", TAGS) {
  SECTION("SceneAccessCategory type") {
    SECTION("mold string conversion") {
      struct X {
        SceneAccessCategory sceneAccessCategory;
        std::string printable;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{{SceneAccessCategory::unknown, "unknown"s},
                                                               {SceneAccessCategory::manual, "manual"s},
                                                               {SceneAccessCategory::timer, "timer"s},
                                                               {SceneAccessCategory::algorithm, "algorithm"s}}));
      INFO(x.printable);
      CHECK(SceneAccess::stringToCategory(x.printable) == x.sceneAccessCategory);
    }
    SECTION("invalid conversion") {
      CHECK(SceneAccess::stringToCategory("foo") == SceneAccessCategory::unknown);
    }
  }

  SECTION("checkAccess") {
    ApartmentMock apartment{};
    auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});
    auto& cluster16 = *apartment.getCluster(16);

    WHEN("manual access") {
      CHECK(SceneAccess::checkAccess(apartment, &device, SceneAccessCategory::manual) == true);
      CHECK(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::manual) == true);
    }

    auto fire = apartment.allocateState(StateType_Apartment, dss::StateName::Fire, "");
    auto wind = apartment.allocateState(StateType_Apartment, dss::StateName::Wind, "");

    GIVEN("fire or wind active ") {

      WHEN("fire active") {
        fire->setState(CallOrigin::system, State_Active);

        THEN("blind/shade devices are blocked") {
          device.addToGroup(GroupIDGray);
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &device, SceneAccessCategory::algorithm),
                            "Execution blocked: fire is active");
        }
        THEN("blind/shade groups are block") {
          cluster16.setApplicationType(ApplicationType::blinds);
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm),
                            "Execution blocked: fire is active");
        }
        THEN("grey group is blocked") {
          auto& groupGray = *apartment.getGroup(GroupIDGray);
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &groupGray, SceneAccessCategory::algorithm),
                            "Execution blocked: fire is active");
        }
      }
      WHEN("wind active") {
        wind->setState(CallOrigin::system, State_Active);

        THEN("blind/shade devices are blocked") {
          device.addToGroup(GroupIDGray);
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &device, SceneAccessCategory::algorithm),
                            "Execution blocked: wind is active");
        }
        THEN("blind/shade groups are block") {
          cluster16.setApplicationType(ApplicationType::blinds);
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm),
                            "Execution blocked: wind is active");
        }
        THEN("grey group is blocked") {
          auto& groupGray = *apartment.getGroup(GroupIDGray);
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &groupGray, SceneAccessCategory::algorithm),
                            "Execution blocked: wind is active");
        }
      }

      WHEN("both active") {
        wind->setState(CallOrigin::system, State_Active);
        fire->setState(CallOrigin::system, State_Active);

        THEN("non shade devices are not affected") {
          auto& light = *createDevice(apartment, {.hardwareType = HardwareType::GE_KL200});
          CHECK(SceneAccess::checkAccess(apartment, &light, SceneAccessCategory::algorithm) == true);
        }
        THEN("non shade groups are not affected") {
          cluster16.setApplicationType(ApplicationType::lights);
          CHECK(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm) == true);
        }
      }
    }

    WHEN("partial wind state") {
      cluster16.setApplicationType(ApplicationType::blinds);
      cluster16.setIsValid(true); // mark configured and usable

      auto windCluster16 = apartment.allocateState(StateType_Apartment, "wind.group16", "");

      WHEN("wind in cluster is active") {
        windCluster16->setState(CallOrigin::system, State_Active);
        THEN("cluster is blocked") {
          CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm),
                            "Execution blocked: wind is active");
        }
        GIVEN("device in cluster") {
          device.addToGroup(cluster16.getID());
          THEN("devices in cluster16 are blocked") {
            CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &device, SceneAccessCategory::algorithm),
                              "Execution blocked: wind is active");
          }
          WHEN("grey group shares device with blocked cluster") {
            auto& groupGray = *apartment.getGroup(GroupIDGray);
            device.addToGroup(groupGray.getID());
            THEN("group grey is also blocked") {
              CHECK_THROWS_FIND(SceneAccess::checkAccess(apartment, &groupGray, SceneAccessCategory::algorithm),
                                "Execution blocked: wind is active");
            }
          }
        }
        WHEN("cluster16 is no longer a blind application") {
          cluster16.setApplicationType(ApplicationType::lights);
          THEN("group is unblocked") {
            CHECK(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm));
          }
        }
        WHEN("cluster16 is inactive") {
          cluster16.setIsValid(false);
          THEN("group is unblocked") {
            CHECK(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm));
          }
        }
      }
      WHEN("wind in cluster becomes inactive") {
        windCluster16->setState(CallOrigin::system, State_Inactive);
        THEN("group is unblocked") {
          CHECK(SceneAccess::checkAccess(apartment, &cluster16, SceneAccessCategory::algorithm));
        }
        GIVEN("device in cluster") {
          device.addToGroup(cluster16.getID());
          THEN("devices is unblocked") {
            CHECK(SceneAccess::checkAccess(apartment, &device, SceneAccessCategory::algorithm));
          }
        }
      }
    }
  }

  SECTION("checkStates") {
    ApartmentMock apartment{};
    auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});

    auto panic = apartment.allocateState(StateType_Apartment, dss::StateName::Panic, "");
    auto fire = apartment.allocateState(StateType_Apartment, dss::StateName::Fire, "");
    auto alarm = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm, "");
    auto alarm2 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm2, "");
    auto alarm3 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm3, "");
    auto alarm4 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm4, "");
    auto wind = apartment.allocateState(StateType_Apartment, dss::StateName::Wind, "");
    auto rain = apartment.allocateState(StateType_Apartment, dss::StateName::Rain, "");
    auto hail = apartment.allocateState(StateType_Apartment, dss::StateName::Hail, "");

    WHEN("apartment scene call") {
      auto x =
          GENERATE(values(std::initializer_list<uint8_t>{SceneWindActive, SceneRainActive, SceneHailActive, SceneFire,
                                                         SceneAlarm, SceneAlarm2, SceneAlarm3, SceneAlarm4}));

      auto stateNameFromScene = [](auto sceneNumber) {
        switch (sceneNumber) {
          case SceneWindActive: return "wind"; break;
          case SceneRainActive: return "rain"; break;
          case SceneHailActive: return "hail"; break;
          case SceneFire: return "fire"; break;
          case SceneAlarm: return "alarm"; break;
          case SceneAlarm2: return "alarm2"; break;
          case SceneAlarm3: return "alarm3"; break;
          case SceneAlarm4: return "alarm4"; break;
          default: throw std::runtime_error(ds::str(sceneNumber, " does not map to a state"));
        }
      };

      WHEN("state inactive") {
        THEN("pass") { CHECK(SceneAccess::checkStates(apartment, &device, x) == true); }
      }
      WHEN("state already active") {
        for (auto state : std::vector{panic, fire, alarm, alarm2, alarm3, alarm4, wind, rain, hail}) {
          state->setState(CallOrigin::system, State_Active);
        }
        THEN("refused") {
          CHECK_THROWS_FIND(SceneAccess::checkStates(apartment, &device, x),
                            ds::str("Execution blocked: state ", stateNameFromScene(x), " is already active"));
        }
      }
    }
    WHEN("normal scnee") {
      THEN("pass") { CHECK(SceneAccess::checkStates(apartment, &device, Scene1) == true); }
    }
  }
}

}; // namespace
