#include "umr200-config.h"

#include <ds/enum.h>
#include <ds/log.h>

#include "device-config.capnp.dss.h"
#include "dss.h"
#include "includerapidjson.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "structuremanipulator.h"
#include "web/handler/jsonhelper.h"
#include "web/webrequests.h"

static ds::LogChannel dsLogChannel("viewUmr200Config");

namespace dss {
namespace view {
namespace umr200Config {
using namespace dss::view::deviceConfig::dssCapnp;

static Application toApplication(dss::ApplicationType applicationType) {
  switch (applicationType) {
    case dss::ApplicationType::lights: return Application::LIGHTS;
    case dss::ApplicationType::blinds: return Application::BLINDS;
    case dss::ApplicationType::heating: return Application::HEATING;
    case dss::ApplicationType::audio: return Application::AUDIO;
    case dss::ApplicationType::video: return Application::VIDEO;
    case dss::ApplicationType::security: return Application::SECURITY;
    case dss::ApplicationType::access: return Application::ACCESS;
    case dss::ApplicationType::joker: return Application::JOKER;
    case dss::ApplicationType::window: return Application::WINDOW;
    case dss::ApplicationType::recirculation: return Application::RECIRCULATION;
    case dss::ApplicationType::temperature: return Application::TEMPERATURE_CONTROL;
    case dss::ApplicationType::apartmentVentilation: return Application::APARTMENT_VENTILATION;
    case dss::ApplicationType::none: return Application::NONE;
    case dss::ApplicationType::cooling: return Application::COOLING;
    case dss::ApplicationType::ventilation: return Application::VENTILATION;
    case dss::ApplicationType::apartmentRecirculation: return Application::APARTMENT_RECIRCULATION;
    case dss::ApplicationType::awnings: return Application::AWNINGS;
  }
  DS_ERROR("Unrecognized application type", applicationType);
  return Application::NONE;
}

Args getArgs(const Device& device) {
  Args args;

  auto&& deviceClusters = device.getClusters();
  args.memberOfClusters.reserve(deviceClusters.size());
  for (auto&& deviceCluster : deviceClusters) {
    deviceConfig::dssCapnp::Cluster x;
    x.id = deviceCluster->getID();
    x.title = deviceCluster->getName();
    x.application = toApplication(deviceCluster->getApplicationType());
    args.memberOfClusters.push_back(std::move(x));
  }

  args.showApartmentApplications = device.supportsApartmentApplications();
  args.showInput = true;
  args.inputTypeMode = Args::InputTypeMode::SHOWN_PUSH_BUTTON_AND_SENSOR;
  args.showPushButtonApplication = true;
  args.showPushButtonMode = true;
  DS_DEBUG_LEAVE(args);
  return args;
}

static void fillPushButton(Value& value, const Args& args, const Device& device) {
  auto toApplication = [&](dss::ApplicationType applicationType) {
    DS_DEBUG(applicationType);
    switch (applicationType) {
      case ApplicationType::lights: return Value::Application::LIGHTS;
      case ApplicationType::blinds: return Value::Application::BLINDS;
      case ApplicationType::heating: return Value::Application::HEATING;
      case ApplicationType::audio: return Value::Application::AUDIO;
      case ApplicationType::video: return Value::Application::VIDEO;
      case ApplicationType::security: return Value::Application::SECURITY;
      case ApplicationType::access: return Value::Application::ACCESS;
      case ApplicationType::joker: return Value::Application::JOKER;
      case ApplicationType::window: return Value::Application::WINDOW;
      case ApplicationType::ventilation: return Value::Application::VENTILATION;
      case ApplicationType::recirculation: return Value::Application::RECIRCULATION;
      case ApplicationType::temperature:
        if (args.showTemperatureControlApplication) {
          return Value::Application::TEMPERATURE_CONTROL;
        }
        break;
      case ApplicationType::awnings: return Value::Application::AWNINGS;
      case ApplicationType::apartmentVentilation:
        if (args.showApartmentApplications) {
          return Value::Application::APARTMENT_VENTILATION;
        }
        break;
      case ApplicationType::none:
      case ApplicationType::cooling:
      case ApplicationType::apartmentRecirculation: break;
    }
    DS_ERROR("Unrecognized application type", applicationType, args.showTemperatureControlApplication,
        args.showApartmentApplications);
    return Value::Application::LIGHTS;
  };
  std::optional<int> targetClusterId;
  value.application = [&] {
    auto&& targetGroupId = device.getButtonActiveGroup();
    DS_DEBUG(targetGroupId);
    if (isApplicationGroup(targetGroupId)) {
      return toApplication(static_cast<ApplicationType>(targetGroupId));
    }
    if (isCluster(targetGroupId)) {
      if (auto&& cluster = device.tryGetGroup(targetGroupId).lock()) {
        targetClusterId = targetGroupId;
        DS_DEBUG(targetClusterId);
        return toApplication(cluster->getApplicationType());
      } else {
        DS_ERROR("Failed to get group", targetGroupId, device);
      }
    }
    DS_ERROR("Unrecognized", targetGroupId, args);
    return Value::Application::JOKER;
  }();

  if (value.application != Value::Application::JOKER) {
    auto&& buttonId = device.getButtonID();
    DS_DEBUG(buttonId);
    auto target = Value::PushButtonTarget::ZONE;
    auto area = Value::PushButtonTargetArea::AREA1;
    auto presets = Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4;
    [&] {
      switch (buttonId) {
        case ButtonId_Area1:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA1;
          presets = Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4;
          return;
        case ButtonId_Area2:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA2;
          presets = Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4;
          return;
        case ButtonId_Area3:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA3;
          presets = Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4;
          return;
        case ButtonId_Area4:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA4;
          presets = Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4;
          return;
        case ButtonId_Zone:
          target = Value::PushButtonTarget::ZONE;
          presets = Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4;
          return;
        case ButtonId_Zone_Extended1:
          target = Value::PushButtonTarget::ZONE;
          presets = Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14;
          return;
        case ButtonId_Zone_Extended2:
          target = Value::PushButtonTarget::ZONE;
          presets = Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24;
          return;
        case ButtonId_Zone_Extended3:
          target = Value::PushButtonTarget::ZONE;
          presets = Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34;
          return;
        case ButtonId_Zone_Extended4:
          target = Value::PushButtonTarget::ZONE;
          presets = Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44;
          return;
        case ButtonId_Area1_Extended:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA1;
          presets = Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14;
          return;
        case ButtonId_Area2_Extended:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA2;
          presets = Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24;
          return;
        case ButtonId_Area3_Extended:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA3;
          presets = Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34;
          return;
        case ButtonId_Area4_Extended:
          target = Value::PushButtonTarget::AREA;
          area = Value::PushButtonTargetArea::AREA4;
          presets = Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44;
          return;
      }
      DS_ERROR("Unrecognized non-joker", buttonId);
    }();
    if (targetClusterId) {
      if (target == Value::PushButtonTarget::ZONE) {
        target = Value::PushButtonTarget::CLUSTER;
      } else {
        DS_ERROR("TargetGroupId is cluster but buttonId is not set to zone", buttonId, targetClusterId, target);
      }
    }

    value.pushButtonTarget = target;
    value.pushButtonTargetArea = area;
    value.pushButtonTargetClusterId = targetClusterId.value_or(0);
    value.pushButtonTargetPresets = presets;
    value.pushButtonLocalPriorityFor1xClick = device.getButtonSetsLocalPriority();
    value.pushButtonCallsPresent = device.getButtonCallsPresent();

    if (args.showPushButtonMode) {
      value.pushButtonMode = [&] {
        auto&& buttonInputMode = device.getButtonInputMode();
        DS_DEBUG(buttonInputMode);
        switch (buttonInputMode) {
          case ButtonInputMode::STANDARD:
            return Value::PushButtonMode::MODE1WAY;
          case ButtonInputMode::TWO_WAY_DW_WITH_INPUT2:
            return Value::PushButtonMode::MODE2WAY_DOWN;
          case ButtonInputMode::TWO_WAY_UP_WITH_INPUT2:
            return Value::PushButtonMode::MODE2WAY_UP;
          // not applicable options
          case ButtonInputMode::TURBO:
          case ButtonInputMode::SWITCHED:
          case ButtonInputMode::TWO_WAY_DW_WITH_INPUT1:
          case ButtonInputMode::TWO_WAY_DW_WITH_INPUT3:
          case ButtonInputMode::TWO_WAY_DW_WITH_INPUT4:
          case ButtonInputMode::TWO_WAY_UP_WITH_INPUT1:
          case ButtonInputMode::TWO_WAY_UP_WITH_INPUT3:
          case ButtonInputMode::TWO_WAY_UP_WITH_INPUT4:
          case ButtonInputMode::TWO_WAY:
          case ButtonInputMode::ONE_WAY:
          case ButtonInputMode::AKM_STANDARD:
          case ButtonInputMode::AKM_INVERTED:
          case ButtonInputMode::AKM_ON_RISING_EDGE:
          case ButtonInputMode::AKM_ON_FALLING_EDGE:
          case ButtonInputMode::AKM_OFF_RISING_EDGE:
          case ButtonInputMode::AKM_OFF_FALLING_EDGE:
          case ButtonInputMode::AKM_RISING_EDGE:
          case ButtonInputMode::AKM_FALLING_EDGE:
          case ButtonInputMode::DEACTIVATED:
          case ButtonInputMode::HEATING_PUSHBUTTON:
            break;
        }
        DS_ERROR("Unrecognized push button", buttonInputMode);
        return Value::PushButtonMode::MODE1WAY;
      }();
    }
  }
  if (value.application == Value::Application::JOKER) {
    value.pushButtonTarget = [&] {
      auto&& buttonId = device.getButtonID();
      DS_DEBUG(buttonId);
      switch (buttonId) {
        case 1:
          return Value::PushButtonTarget::ALARM;
        case 2:
          return Value::PushButtonTarget::PANIC;
        case 3:
          return Value::PushButtonTarget::LEAVING_HOME;
        case 5:
          return Value::PushButtonTarget::DOOR_BELL;
        case 15:
          return Value::PushButtonTarget::APP;
      }
      DS_ERROR("Unrecognized joker", buttonId);
      return Value::PushButtonTarget::APP;
    }();
  }
  DS_DEBUG_LEAVE(value);
}

static void fillState(Value& value, const Args& args, const Device& device) {
  value.stateType = [&] {
    auto inputType = device.getBinaryInputs().at(0)->m_inputType;
    switch (inputType) {
      case BinaryInputType::AppMode:
        return Value::StateType::GENERIC;
      case BinaryInputType::Presence:
        return Value::StateType::PRESENCE;
      case BinaryInputType::TwilightExternal:
        return Value::StateType::TWILIGHT;
      case BinaryInputType::Movement:
        return Value::StateType::MOTION;
      case BinaryInputType::SmokeDetector:
        return Value::StateType::SMOKE;
      case BinaryInputType::WindDetector:
        return Value::StateType::WIND_STRENGTH_ABOVE_LIMIT;
      case BinaryInputType::RainDetector:
        return Value::StateType::RAIN;
      case BinaryInputType::SunRadiation:
        return Value::StateType::SUN_RADIATION;
      case BinaryInputType::SunProtection:
        return Value::StateType::SUN_PROTECTION;
      case BinaryInputType::FrostDetector:
        return Value::StateType::FROST;
      case BinaryInputType::RoomThermostat:
        return Value::StateType::TEMPERATURE_BELOW_LIMIT;
      case BinaryInputType::WindowContact:
        return Value::StateType::WINDOW_IS_OPEN;
      case BinaryInputType::DoorContact:
        return Value::StateType::DOOR_IS_OPEN;
      case BinaryInputType::GarageDoorContact:
        return Value::StateType::GARAGE_DOOR_IS_OPEN;
      case BinaryInputType::HeatingSystem:
        return Value::StateType::HEATING_SYSTEM_ENABLED;
      case BinaryInputType::HeatingSystemMode:
        return Value::StateType::CHANGE_OVER_SIGNAL;
      case BinaryInputType::Malfunction:
        return Value::StateType::MALFUNCTION;
      case BinaryInputType::Service:
        return Value::StateType::SERVICE;
      // not applicable options
      case BinaryInputType::RoomBrightness:
      case BinaryInputType::PresenceInDarkness:
      case BinaryInputType::MovementInDarkness:
      case BinaryInputType::BatteryLow:
      case BinaryInputType::WindowTilt:
      case BinaryInputType::PowerUp:
        break;
    };
    DS_ERROR("Unrecognized", inputType);
    return Value::StateType::GENERIC;
  }();

  value.stateTarget = [&] {
    auto&& targetGroupId = device.getBinaryInputs().at(0)->m_targetGroupId;
    switch (targetGroupId) {
      case GroupIDBlack:
        return Value::StateTarget::WHOLE_APARTMENT;
      case GroupIDGlobalAppDsVentilation: {
        if (args.showApartmentApplications &&
            (value.stateType == Value::StateType::MALFUNCTION || value.stateType == Value::StateType::SERVICE)) {
          return Value::StateTarget::APARTMENT_VENTILATION;
        }
        break;
      }
      case GroupIDVentilation: {
        if (value.stateType == Value::StateType::MALFUNCTION || value.stateType == Value::StateType::SERVICE) {
          return Value::StateTarget::VENTILATION;
        }
        break;
      }
      case GroupIDRecirculation: {
        if (value.stateType == Value::StateType::MALFUNCTION || value.stateType == Value::StateType::SERVICE) {
          return Value::StateTarget::RECIRCULATION;
        }
        break;
      }
    }
    DS_ERROR("Unrecognized", targetGroupId, value.stateType, args.showApartmentApplications);
    return Value::StateTarget::WHOLE_APARTMENT;
  }();
  value.stateMode = [&] {
    auto&& buttonInputMode = device.getButtonInputMode();
    switch (buttonInputMode) {
      case ButtonInputMode::AKM_STANDARD:
        return Value::StateMode::STANDARD;
      case ButtonInputMode::AKM_INVERTED:
        return Value::StateMode::INVERTED;
      case ButtonInputMode::AKM_ON_RISING_EDGE:
        return Value::StateMode::RISING_EDGE1;
      case ButtonInputMode::AKM_ON_FALLING_EDGE:
        return Value::StateMode::FALLING_EDGE1;
      case ButtonInputMode::AKM_OFF_RISING_EDGE:
        return Value::StateMode::RISING_EDGE0;
      case ButtonInputMode::AKM_OFF_FALLING_EDGE:
        return Value::StateMode::FALLING_EDGE0;
      case ButtonInputMode::AKM_RISING_EDGE:
        return Value::StateMode::TOGGLE_MODE_ON_RISING_EDGE;
      case ButtonInputMode::AKM_FALLING_EDGE:
        return Value::StateMode::TOGGLE_MODE_ON_FALLING_EDGE;
      // not applicable options
      case ButtonInputMode::STANDARD:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT2:
      case ButtonInputMode::TURBO:
      case ButtonInputMode::SWITCHED:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY:
      case ButtonInputMode::ONE_WAY:
      case ButtonInputMode::DEACTIVATED:
      case ButtonInputMode::HEATING_PUSHBUTTON:
        break;
    }
    DS_ERROR("Unrecognized state", buttonInputMode);
    return Value::StateMode::STANDARD;
  }();
  value.application = Value::Application::JOKER;
}

void fillInput(Value& value, const Args& args, const Device& device) {
  auto type = [&] {
    auto&& buttonInputMode = device.getButtonInputMode();
    switch (buttonInputMode) {
      case ButtonInputMode::STANDARD:
      case ButtonInputMode::TURBO:
      case ButtonInputMode::SWITCHED:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY:
      case ButtonInputMode::ONE_WAY:
        return Value::InputType::PUSH_BUTTON;
      case ButtonInputMode::AKM_STANDARD:
      case ButtonInputMode::AKM_INVERTED:
      case ButtonInputMode::AKM_ON_RISING_EDGE:
      case ButtonInputMode::AKM_ON_FALLING_EDGE:
      case ButtonInputMode::AKM_OFF_RISING_EDGE:
      case ButtonInputMode::AKM_OFF_FALLING_EDGE:
      case ButtonInputMode::AKM_RISING_EDGE:
      case ButtonInputMode::AKM_FALLING_EDGE:
        return Value::InputType::STATE;
      // non application options
      case ButtonInputMode::DEACTIVATED: // TODO(someday): really not applicable?
      case ButtonInputMode::HEATING_PUSHBUTTON:
        break;
    }
    DS_ERROR("Unrecognized input", buttonInputMode);
    return Value::InputType::PUSH_BUTTON;
  }();
  value.inputType = type;

  switch (type) {
    case Value::InputType::PUSH_BUTTON:
      fillPushButton(value, args, device);
      break;
    case Value::InputType::STATE:
      fillState(value, args, device);
      break;
  }
}

void get(JSONWriter& json, const Device& device) {
  DS_REQUIRE((device.getDeviceType() == DEVICE_TYPE_UMR && device.getDeviceNumber() == 200) ||
             (device.getDeviceType() == DEVICE_TYPE_TKM && device.getDeviceNumber() == 300),
      "Request not supported on this device");

  Data data;
  data.args = getArgs(device);
  fillInput(data.value, data.args, device);

  data.addFieldsToJson(json);
  DS_NOTICE("Get", data);
}

static void setPushButton(Device& device, StructureManipulator& manipulator, const Value& value, const Args& args, JSONWriter& updatedDevicesJSON, const Apartment& apartment) {
  auto supportsApartmentApplications = device.supportsApartmentApplications();
  auto multiDeviceIndex = device.multiDeviceIndex();
  bool wasSlave = false;
  bool wasMaster = false;

  if (device.getDeviceFeatures().pairing) {
    wasMaster = device.is2WayMaster();
    wasSlave = device.getPartnerDevice()->is2WaySlave();
  }

  auto application = [&] {
    switch (value.application) {
      case Value::Application::LIGHTS:
        return GroupIDYellow;
      case Value::Application::BLINDS:
        return GroupIDGray;
      case Value::Application::HEATING:
        return GroupIDHeating;
      case Value::Application::AUDIO:
        return GroupIDCyan;
      case Value::Application::VIDEO:
        return GroupIDViolet;
      case Value::Application::JOKER:
        return GroupIDBlack;
      case Value::Application::VENTILATION:
        return GroupIDVentilation;
      case Value::Application::RECIRCULATION:
        return GroupIDRecirculation;
      case Value::Application::ACCESS:
        return GroupIDGreen;
      case Value::Application::SECURITY:
        return GroupIDRed;
      case Value::Application::WINDOW:
        return GroupIDWindow;
      case Value::Application::APARTMENT_VENTILATION:
        if (!supportsApartmentApplications) {
          break;
        }
        return GroupIDGlobalAppDsVentilation;
      case Value::Application::AWNINGS:
        return GroupIDGlobalAppDsAwnings;
      case Value::Application::TEMPERATURE_CONTROL:
        // TODO(someday): is it good idea to configure TEMPERATURE_CONTROL as targetGroupId?
        // Temperature control has very special (incompatible) scene numbers
        return GroupIDControlTemperature;
    }
    DS_FAIL_REQUIRE("Unrecognized", value.application);
  }();
  std::optional<int> targetClusterId;
  std::vector<std::shared_ptr<Device> > addedDevices;
  std::vector<std::shared_ptr<Device> > updatedDevices;
  std::vector<std::shared_ptr<Device> > removedDevices;

  if (value.application != Value::Application::JOKER) {
    auto buttonId = [&] {
      switch (value.pushButtonTarget) {
        case Value::PushButtonTarget::DEVICE:
          switch (value.pushButtonTargetPresets) {
            case Value::PushButtonTargetPresets::DEVICE_PRESETS2_TO4:
              return ButtonId_Device;
            case Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14:
            case Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24:
            case Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34:
            case Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44:
            case Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14:
            case Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24:
            case Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34:
            case Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44:
              break;
          }
          break;
        case Value::PushButtonTarget::CLUSTER: // fall through
        case Value::PushButtonTarget::ZONE:
          if (value.pushButtonTarget == Value::PushButtonTarget::CLUSTER) {
            DS_REQUIRE(isCluster(value.pushButtonTargetClusterId));
            targetClusterId = value.pushButtonTargetClusterId;
          }
          switch (value.pushButtonTargetPresets) {
            case Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4:
              return ButtonId_Zone;
            case Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14:
              return ButtonId_Zone_Extended1;
            case Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24:
              return ButtonId_Zone_Extended2;
            case Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34:
              return ButtonId_Zone_Extended3;
            case Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44:
              return ButtonId_Zone_Extended4;
            case Value::PushButtonTargetPresets::DEVICE_PRESETS2_TO4:
            case Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4:
            case Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14:
            case Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24:
            case Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34:
            case Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44:
              break;
          }
          break;
        case Value::PushButtonTarget::AREA:
          switch (value.pushButtonTargetArea) {
            case Value::PushButtonTargetArea::AREA1:
              switch (value.pushButtonTargetPresets) {
                case Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4:
                  return ButtonId_Area1;
                case Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14:
                  return ButtonId_Area1_Extended;
                case Value::PushButtonTargetPresets::DEVICE_PRESETS2_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14:
                case Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24:
                case Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34:
                case Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44:
                case Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24:
                case Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34:
                case Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44:
                  break;
              }
              break;
            case Value::PushButtonTargetArea::AREA2:
              switch (value.pushButtonTargetPresets) {
                case Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4:
                  return ButtonId_Area2;
                case Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24:
                  return ButtonId_Area2_Extended;
                case Value::PushButtonTargetPresets::DEVICE_PRESETS2_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14:
                case Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24:
                case Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34:
                case Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44:
                case Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14:
                case Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34:
                case Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44:
                  break;
              }
              break;
            case Value::PushButtonTargetArea::AREA3:
              switch (value.pushButtonTargetPresets) {
                case Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4:
                  return ButtonId_Area3;
                case Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34:
                  return ButtonId_Area3_Extended;
                case Value::PushButtonTargetPresets::DEVICE_PRESETS2_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14:
                case Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24:
                case Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34:
                case Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44:
                case Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14:
                case Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24:
                case Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44:
                  break;
              }
              break;
            case Value::PushButtonTargetArea::AREA4:
              switch (value.pushButtonTargetPresets) {
                case Value::PushButtonTargetPresets::AREA4_PRESETS1_TO4:
                  return ButtonId_Area4;
                case Value::PushButtonTargetPresets::AREA4_PRESETS42_TO44:
                  return ButtonId_Area4_Extended;
                case Value::PushButtonTargetPresets::DEVICE_PRESETS2_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::ZONE_PRESETS10_TO14:
                case Value::PushButtonTargetPresets::ZONE_PRESETS20_TO24:
                case Value::PushButtonTargetPresets::ZONE_PRESETS30_TO34:
                case Value::PushButtonTargetPresets::ZONE_PRESETS40_TO44:
                case Value::PushButtonTargetPresets::AREA1_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA2_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA3_PRESETS1_TO4:
                case Value::PushButtonTargetPresets::AREA1_PRESETS12_TO14:
                case Value::PushButtonTargetPresets::AREA2_PRESETS22_TO24:
                case Value::PushButtonTargetPresets::AREA3_PRESETS32_TO34:
                  break;
              }
              break;
          }
          break;
        case Value::PushButtonTarget::DISABLED:
          DS_FAIL_REQUIRE("PushButtonTarget::DISABLED not implemented");
          break;
        case Value::PushButtonTarget::APP:
        case Value::PushButtonTarget::ALARM:
        case Value::PushButtonTarget::PANIC:
        case Value::PushButtonTarget::LEAVING_HOME:
        case Value::PushButtonTarget::DOOR_BELL:
          break;
      }
      DS_FAIL_REQUIRE(
          "Unrecognized", value.pushButtonTarget, value.pushButtonTargetArea, value.pushButtonTargetPresets);
    }();

    DS_REQUIRE(ds::isValidEnum(value.pushButtonMode));
    DS_REQUIRE(
        multiDeviceIndex == 0 ||
            (multiDeviceIndex == 2 && device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300) ||
            value.pushButtonMode == Value::PushButtonMode::MODE1WAY,
        "Only first device can be !mode1way.", multiDeviceIndex, value.pushButtonMode);

    // Validation passed, apply to device. No recovery on error (now).
    DS_INFO("Set", application, buttonId, targetClusterId, value.pushButtonLocalPriorityFor1xClick,
        value.pushButtonCallsPresent, multiDeviceIndex, value.pushButtonMode);
    auto setDeviceBoolProperty = [&](Device& x, const char* path, bool value) {
      auto&& propertySystem = DSS::getInstance()->getPropertySystem();
      auto&& propertyPath = ds::str("/apartment/zones/zone0/devices/", dsuid2str(x.getDSID()), path);
      auto node = propertySystem.getProperty(propertyPath);
      DS_REQUIRE(node, "Device property does not exist", propertyPath);
      node->setBooleanValue(value);
    };
    auto setDevice = [&](Device& x) {
      manipulator.setJokerGroup(x.sharedFromThis(), static_cast<int>(application));
      x.setDeviceButtonID(buttonId);
      if (targetClusterId) {
        x.setDeviceButtonActiveGroup(*targetClusterId);
      }
      if (!x.is2WaySlave()) {
        setDeviceBoolProperty(x, "/button/setsLocalPriority", value.pushButtonLocalPriorityFor1xClick);
        setDeviceBoolProperty(x, "/button/callsPresent", value.pushButtonCallsPresent);
      }
      updatedDevices.push_back(x.sharedFromThis());
    };
    setDevice(device);

    if (multiDeviceIndex == 0 ||
        (multiDeviceIndex == 2 && device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300)) {
      auto partnerDevice = device.tryGetPartnerDevice();
      [&] {
        switch (value.pushButtonMode) {
          case Value::PushButtonMode::MODE1WAY:
            device.setDeviceButtonInputMode(ButtonInputMode::STANDARD);
            if (wasSlave && partnerDevice) {
              partnerDevice->setDeviceButtonInputMode(ButtonInputMode::STANDARD);
              addedDevices.push_back(partnerDevice->sharedFromThis());
            }
            return;
          case Value::PushButtonMode::MODE2WAY_DOWN:
            if (multiDeviceIndex == 2 && device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300) {
              device.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT4);
              if (partnerDevice) {
                partnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT3);
                removedDevices.push_back(partnerDevice->sharedFromThis());
              }
            } else {
              device.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT2);
              if (partnerDevice) {
                partnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT1);
                removedDevices.push_back(partnerDevice->sharedFromThis());
              }
            }
            break;
          case Value::PushButtonMode::MODE2WAY_UP:
            if (multiDeviceIndex == 2 && device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300) {
              device.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT4);
              if (partnerDevice) {
                partnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT3);
                removedDevices.push_back(partnerDevice->sharedFromThis());
              }
            } else {
              device.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2);
              if (partnerDevice) {
                partnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1);
                removedDevices.push_back(partnerDevice->sharedFromThis());
              }
            }
            break;
        }
        setDevice(*partnerDevice);
      }();
    }
  } // value.application != Value::Application::JOKER

  if (value.application == Value::Application::JOKER) {
    auto buttonId = [&] {
      switch (value.pushButtonTarget) {
        case Value::PushButtonTarget::ALARM:
          return 1;
        case Value::PushButtonTarget::PANIC:
          return 2;
        case Value::PushButtonTarget::LEAVING_HOME:
          return 3;
        case Value::PushButtonTarget::DOOR_BELL:
          return 5;
        case Value::PushButtonTarget::APP:
          return 15;
        case Value::PushButtonTarget::DISABLED:
          DS_FAIL_REQUIRE("PushButtonTarget::DISABLED not implemented");
          break;
        case Value::PushButtonTarget::DEVICE:
        case Value::PushButtonTarget::ZONE:
        case Value::PushButtonTarget::AREA:
        case Value::PushButtonTarget::CLUSTER:
          break;
      }
      DS_FAIL_REQUIRE("Unrecognized", value.pushButtonTarget);
    }();

    // Validation passed, apply to device. No recovery on error (now).
    manipulator.setJokerGroup(device.sharedFromThis(), GroupIDBlack);
    device.setDeviceButtonID(buttonId);
    updatedDevices.push_back(device.sharedFromThis());
    if (value.pushButtonTarget == Value::PushButtonTarget::DOOR_BELL) {
      device.setDeviceButtonInputMode(ButtonInputMode::TURBO);
    } else {
      device.setDeviceButtonInputMode(ButtonInputMode::STANDARD);
    }
    if (device.getDeviceFeatures().pairing && wasMaster) {  // app button is always 1way
      auto& partnerDevice = *device.getPartnerDevice();
      partnerDevice.setDeviceButtonInputMode(ButtonInputMode::STANDARD);
      addedDevices.push_back(partnerDevice.sharedFromThis());
    }

  } // value.application == Value::Application::JOKER

  updatedDevicesJSON.startObject("actions");
  updatedDevicesJSON.startArray("add");
  if (!addedDevices.empty()) {
    for (const std::shared_ptr<Device>& device : addedDevices) {
      const DeviceReference deviceRef(device, &apartment);
      toJSON(deviceRef, updatedDevicesJSON);
    }
  }
  updatedDevicesJSON.endArray();
  updatedDevicesJSON.startArray("remove");
  if (!removedDevices.empty()) {
    for (const std::shared_ptr<Device>& device : removedDevices) {
      const DeviceReference deviceRef(device, &apartment);
      toJSON(deviceRef, updatedDevicesJSON);
    }
  }
  updatedDevicesJSON.endArray();
  updatedDevicesJSON.startArray("update");
  if (!updatedDevices.empty()) {
    for (const std::shared_ptr<Device>& device : updatedDevices) {
      const DeviceReference deviceRef(device, &apartment);
      toJSON(deviceRef, updatedDevicesJSON);
    }
  }
  updatedDevicesJSON.endArray();
  updatedDevicesJSON.endObject();
}

static void setState(Device& device, StructureManipulator& manipulator, const Value& value, const Args& args, JSONWriter& updatedDevicesJSON, const Apartment& apartment) {
  bool supportsApartmentApplications = device.supportsApartmentApplications();
  std::vector<std::shared_ptr<Device> > addedDevices;
  std::vector<std::shared_ptr<Device> > updatedDevices;
  std::vector<std::shared_ptr<Device> > removedDevices;

  auto&& binaryInputType = [&] {
    switch (value.stateType) {
      case Value::StateType::GENERIC:
        return BinaryInputType::AppMode;
      case Value::StateType::PRESENCE:
        return BinaryInputType::Presence;
      case Value::StateType::TWILIGHT:
        return BinaryInputType::TwilightExternal;
      case Value::StateType::MOTION:
        return BinaryInputType::Movement;
      case Value::StateType::SMOKE:
        return BinaryInputType::SmokeDetector;
      case Value::StateType::WIND_STRENGTH_ABOVE_LIMIT:
        return BinaryInputType::WindDetector;
      case Value::StateType::RAIN:
        return BinaryInputType::RainDetector;
      case Value::StateType::SUN_RADIATION:
        return BinaryInputType::SunRadiation;
      case Value::StateType::SUN_PROTECTION:
        return BinaryInputType::SunProtection;
      case Value::StateType::FROST:
        return BinaryInputType::FrostDetector;
      case Value::StateType::TEMPERATURE_BELOW_LIMIT:
        return BinaryInputType::RoomThermostat;
      case Value::StateType::WINDOW_IS_OPEN:
        return BinaryInputType::WindowContact;
      case Value::StateType::DOOR_IS_OPEN:
        return BinaryInputType::DoorContact;
      case Value::StateType::GARAGE_DOOR_IS_OPEN:
        return BinaryInputType::GarageDoorContact;
      case Value::StateType::HEATING_SYSTEM_ENABLED:
        return BinaryInputType::HeatingSystem;
      case Value::StateType::CHANGE_OVER_SIGNAL:
        return BinaryInputType::HeatingSystemMode;
      case Value::StateType::MALFUNCTION:
        return BinaryInputType::Malfunction;
      case Value::StateType::SERVICE:
        return BinaryInputType::Service;
    }
    DS_FAIL_REQUIRE("Unrecognized", value.stateType);
  }();
  auto&& targetGroupId = [&] {
    if (supportsApartmentApplications &&
        (value.stateType == Value::StateType::MALFUNCTION || value.stateType == Value::StateType::SERVICE)) {
      switch (value.stateTarget) {
        case Value::StateTarget::WHOLE_APARTMENT:
          return GroupIDBlack; // TODO(requirements): do not touch at all? How does this work with other groups?
        case Value::StateTarget::APARTMENT_VENTILATION:
          return GroupIDGlobalAppDsVentilation;
        case Value::StateTarget::VENTILATION:
          return GroupIDVentilation;
        case Value::StateTarget::RECIRCULATION:
          return GroupIDRecirculation;
      }
      DS_FAIL_REQUIRE("Unrecognized", value.stateTarget);
    } else {
      // ignore target
      return GroupIDBlack; // TODO(requirements): do not touch at all? How does this work with other groups?
    }
  }();
  auto buttonInputMode = [&] {
    switch (value.stateMode) {
      case Value::StateMode::STANDARD:
        return ButtonInputMode::AKM_STANDARD;
      case Value::StateMode::INVERTED:
        return ButtonInputMode::AKM_INVERTED;
      case Value::StateMode::RISING_EDGE1:
        return ButtonInputMode::AKM_ON_RISING_EDGE;
      case Value::StateMode::FALLING_EDGE1:
        return ButtonInputMode::AKM_ON_FALLING_EDGE;
      case Value::StateMode::RISING_EDGE0:
        return ButtonInputMode::AKM_OFF_RISING_EDGE;
      case Value::StateMode::FALLING_EDGE0:
        return ButtonInputMode::AKM_OFF_FALLING_EDGE;
      case Value::StateMode::TOGGLE_MODE_ON_RISING_EDGE:
        return ButtonInputMode::AKM_RISING_EDGE;
      case Value::StateMode::TOGGLE_MODE_ON_FALLING_EDGE:
        return ButtonInputMode::AKM_FALLING_EDGE;
    }
    DS_FAIL_REQUIRE("Unrecognized", value.stateMode);
  }();

  bool wasSlave = false;
  if (device.getDeviceFeatures().pairing) {
    wasSlave = device.getPartnerDevice()->is2WaySlave();
  }

  // Validation passed, apply to device. No recovery on error (now).
  if (device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW) {
    manipulator.setJokerGroup(device.sharedFromThis(), GroupIDBlack);
  }
  device.setDeviceButtonInputMode(buttonInputMode); // TODO(requirements): device.setDeviceBinaryInputType()?
  device.setDeviceBinaryInputTargetId(0, targetGroupId);
  device.setDeviceBinaryInputType(0, binaryInputType);

  if (device.getDeviceFeatures().pairing && wasSlave) {
    device.getPartnerDevice()->setDeviceButtonInputMode(ButtonInputMode::STANDARD);
    addedDevices.push_back(device.getPartnerDevice()->sharedFromThis());
  }
  updatedDevices.push_back(device.sharedFromThis());

  updatedDevicesJSON.startObject("actions");
  updatedDevicesJSON.startArray("update");
  if (!updatedDevices.empty()) {
    for (const std::shared_ptr<Device>& device : updatedDevices) {
      const DeviceReference deviceRef(device, &apartment);
      toJSON(deviceRef, updatedDevicesJSON);
    }
  }
  updatedDevicesJSON.endArray();
  updatedDevicesJSON.startArray("remove");
  updatedDevicesJSON.endArray();
  updatedDevicesJSON.startArray("add");
  if (!addedDevices.empty()) {
    for (const std::shared_ptr<Device>& device : addedDevices) {
      const DeviceReference deviceRef(device, &apartment);
      toJSON(deviceRef, updatedDevicesJSON);
    }
  }
  updatedDevicesJSON.endArray();
  updatedDevicesJSON.endObject();
}

void setValue(Device& device, StructureManipulator& manipulator, const std::string& json, JSONWriter& updatedDevices, const Apartment& apartment) {
  DS_REQUIRE((device.getDeviceType() == DEVICE_TYPE_UMR && device.getDeviceNumber() == 200) ||
                 (device.getDeviceType() == DEVICE_TYPE_TKM && device.getDeviceNumber() == 300) ||
                 (device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300),
             "Request not supported on this device");
  //Read, Modify and then Write
  Data data;
  data.args = getArgs(device);
  fillInput(data.value, data.args, device);
  JSONWriter writer;
  data.addFieldsToJson(writer);
  rapidjson::Document currentValues;
  currentValues.Parse(writer.successJSON().c_str());

  rapidjson::Document valueJSON;
  valueJSON.Parse(json.c_str());
  rapidjson::Document unifiedJSONDoc;
  unifiedJSONDoc.SetObject();
  auto& allocator = unifiedJSONDoc.GetAllocator();
  const rapidjson::Value& valueStruct = currentValues["result"]["value"];
  for (rapidjson::Value::ConstMemberIterator itr = valueStruct.MemberBegin(); itr != valueStruct.MemberEnd(); ++itr) {
    std::string memberName = itr->name.GetString();
    if (valueJSON.HasMember(memberName.c_str())) {
      rapidjson::Value& value = valueJSON[memberName.c_str()];
      rapidjson::Value copiedVal(value, allocator);
      if (value.IsString() || (value.IsNumber() && !value.IsDouble()) || value.IsBool()) {
        unifiedJSONDoc.AddMember(rapidjson::Value(memberName.c_str(), allocator).Move(), copiedVal, allocator);
      } else {
        DS_ERROR("Unrecognized type in value object field:", itr->name.GetString());
      }
    } else {
      const rapidjson::Value& value = itr->value;
      rapidjson::Value copiedVal(itr->value, allocator);
      if (value.IsString() || (value.IsNumber() && !value.IsDouble()) || value.IsBool()) {
        unifiedJSONDoc.AddMember(rapidjson::Value(memberName.c_str(), allocator).Move(), copiedVal, allocator);
      } else {
        DS_ERROR("Unrecognized type in value object field:", itr->name.GetString());
      }
    }
  }
  rapidjson::StringBuffer buffer;
  rapidjson::Writer<StringBuffer> rapidWriter(buffer);
  unifiedJSONDoc.Accept(rapidWriter);

  auto value = Value::fromJson(buffer.GetString());
  auto args = getArgs(device);
  DS_NOTICE("Apply value", value, args);
  switch (value.inputType) {
    case Value::InputType::PUSH_BUTTON:
      setPushButton(device, manipulator, value, args, updatedDevices, apartment);
      break;
    case Value::InputType::STATE:
      setState(device, manipulator, value, args, updatedDevices, apartment);
      break;
  }
}

} // namespace umr200Config
} // namespace view
} // namespace dss
