/*
    Copyright (c) 2017 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include "internalrequesthandler.h"

#include <digitalSTROM/dsuid.h>
#include <json-c/json.h>
#include <limits.h>

#include <algorithm>
#include <boost/filesystem.hpp>
#include <fstream>

#include "businterface.h"
#include "ds485types.h"
#include "dss.h"
#include "event.h"
#include "eventcollector.h"
#include "eventinterpreterplugins.h"
#include "jsonhelper.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/modelmaintenance.h"
#include "model/scenehelper.h"
#include "model/set.h"
#include "model/zone.h"
#include "util.h"
#include "vdc-element-reader.h"
#include "vdc-messages.pb.h"
#include "view/umr200-config.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("jsonApiInternal");

namespace dss {

 /* local helper function */
static bool getByteFromJson(json_object *obj, const char *key, uint8_t *value) {
  json_object *jval;
  if (!json_object_object_get_ex(obj, key, &jval)) {
    return false;
  }

  errno = 0;
  int32_t i = json_object_get_int(jval);
  if ((i == 0) && (errno == EINVAL)) {
    DS_ERROR("Parsing an int value failed", errno);
    return false;
  }

  if ((i < 0) || (i > UCHAR_MAX)) {
    return false;
  }

  if (value != NULL) {
    *value = (uint8_t)i;
  }

  return true;
}


  //=========================================== InternalRequestHandler

InternalRequestHandler::InternalRequestHandler(Apartment &_apartment, EventInterpreter &interpreter)
    : m_Apartment(_apartment), m_manipulator(m_Apartment) {}

 // check if product family matches the list in the spec that is due for name
 // removal: GE-SDS, GE-SDM, SW-ZWS, SW-SSL
  bool InternalRequestHandler::needToClearName(std::shared_ptr<const Device> device) {
    return ((((device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE) &&
              ((device->getDeviceType() == DEVICE_TYPE_SDS) || (device->getDeviceType() == DEVICE_TYPE_SDM))) ||
             ((device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW) &&
              ((device->getDeviceType() == DEVICE_TYPE_ZWS) || (device->getProductID() == ProductID_SSL_200_JS) ||
               (device->getProductID() == ProductID_SSL_200_FS)))) &&
            !device->getName().empty());
  }

  void InternalRequestHandler::removeCustomActions(std::shared_ptr<Device> device) {
    if (!DSS::hasInstance()) {
      return;
    }

    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> query;
    vdcapi::PropertyElement* e1 = query.Add();
    e1->set_name("customActions");
    vdcapi::Message message = device->getVdcProperty(query);
    vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();

    if (response.properties_size() == 0) {
      return;
    }

    for (int i = 0; i < response.properties_size(); i++) {
      vdcapi::PropertyElement el = response.properties(i);

      if (el.has_name() && el.name() == "customActions") {
        for (int j = 0; j < el.elements_size(); j++) {
          const vdcapi::PropertyElement &action = el.elements(j);
          if (action.has_name()) {
            vdcapi::PropertyElement params;
            device->setCustomAction(action.name(), "", "", params);
          }
        } // custom actions loop
      } // custom actions element
    } // response props
  }

  std::string InternalRequestHandler::getRentalResetJSON() {
    std::string ret;

    if (!DSS::hasInstance()) {
      return ret;
    }

    PropertyNodePtr node = DSS::getInstance()->getPropertySystem().getProperty("/config/rentalreset");
    if (node == NULL) {
      DS_NOTICE("No rental reset configuration file available, nothing to do");
      return ret;
    }

    std::string filename = node->getStringValue();
    if (!boost::filesystem::exists(filename)) {
      DS_ERROR("Rental reset configuration file ", filename, " does not exist!");
      return ret;
    }

    std::ifstream in(filename.c_str(), std::ios::in | std::ios::binary);
    if (!in) {
      DS_ERROR("Rental reset configuration file ", filename,
               " could not be read!");

      return ret;
    }

    in.seekg(0, std::ios::end);
    ret.resize(in.tellg());
    in.seekg(0, std::ios::beg);
    in.read(&ret[0], ret.size());
    in.close();

    return ret;
  }

  std::string InternalRequestHandler::resetSceneConfiguration() {
    std::string ret;

    if (!DSS::hasInstance()) {
      // shouldn't be possible under normal circumstances
      return "No dSS instance;";
    }

    PropertyNodePtr node = DSS::getInstance()->getPropertySystem().getProperty("/config/rentalreset");
    if (node == NULL) {
      DS_NOTICE("No rental reset configuration file available, nothing to do");
      return "";
    }

    std::string filename = node->getStringValue();
    if (!boost::filesystem::exists(filename)) {
      return "Rental reset configuration file " + filename + "does not exist!";
    }

    json_object *parsed = json_object_from_file(filename.c_str());
    if (parsed == NULL) {
      return "Error parsing rental reset json file;";
    }

    json_object *jdevices;
    if (!json_object_object_get_ex(parsed, "devices", &jdevices)) {
      return "Rental reset JSON has no devices object;";
    }

    std::shared_ptr<Zone> zone0 = m_Apartment.getZone(0);
    Set devices = zone0->getDevices();
    for (int d = 0; d < devices.length(); d++) {
      std::shared_ptr<Device> device = devices[d].getDevice();

      std::string key =
          device->getDeviceClassString(device->getDeviceClass()) +
          device->getDeviceTypeString(device->getDeviceType()) +
          intToString(device->getDeviceNumber());

      json_object *jdevice;
      if (!json_object_object_get_ex(jdevices, key.c_str(), &jdevice)) {
        continue;
      }
      
      json_object *jproduct;
      if (!json_object_object_get_ex(jdevice, "productid", &jproduct)) {
        continue;
      }

      errno = 0;
      int32_t productid = json_object_get_int(jproduct);
      if ((productid == 0) && (errno == EINVAL)) {
        DS_ERROR("productid is not a number", errno);
        continue;
      }

      if (productid != device->getProductID()) {
        continue;
      }

      if (device->hasOutput() && (device->getOutputMode() == 0)) {
        uint8_t outputmode = 0;
        if (getByteFromJson(jdevice, "outputmode", &outputmode)) {
          try {
            if (outputmode != 0) {
              device->setDeviceOutputMode(outputmode);
            }
          } catch (const std::exception &ex) {
            ret += "Could not reset output mode on device " +
                   dsuid2str(device->getDSID()) + " " + ex.what() + ";";
          }
        }
      }

      // reset SW devices only if they are in the Yellow group
      if ((device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW) && !device->isInGroup(GroupIDYellow)) {
        continue;
      }
      json_object *jconfig;
      if (!json_object_object_get_ex(jdevice, "config", &jconfig)) {
        continue;
      }

      if (json_object_get_type(jconfig) != json_type_array) {
        continue;
      }

      // skip the device if we get a certain number of ds485 failures in a row
      #define RR_MAX_FAILURES 5
      int failures = 0;

      for (int i = 0; i < (int)json_object_array_length(jconfig); i++) {
        if (failures >= RR_MAX_FAILURES) {
          ret += "Aborting configuration reset of device " +
                 dsuid2str(device->getDSID()) +
                 " due to consecutive dS485 timeouts;";
          break;
        }

        json_object *line = json_object_array_get_idx(jconfig, i);

        uint8_t bank = 0;
        if (!getByteFromJson(line, "bank", &bank)) {
          continue;
        }

        uint8_t offset = 0;
        if (!getByteFromJson(line, "offset", &offset)) {
          continue;
        }

        uint8_t value = 0;
        if (!getByteFromJson(line, "value", &value)) {
          continue;
        }

        uint8_t bitmask = 0;
        getByteFromJson(line, "bitmask", &bitmask);

        std::vector<int> scenes;
        json_object *jscenes;
        bool locked = false;

        // If the json part contains a "scenes" array, then it means that the
        // bank/offset/value combination affects the given scenes, we need
        // to check if those scenes are locked. If they are - we are not
        // supposed to change their configurations.
        if (json_object_object_get_ex(jconfig, "scenes", &jscenes)) {
          if (json_object_get_type(jscenes) == json_type_array) {
            for (int s = 0; s < (int)json_object_array_length(jscenes); s++) {
              json_object *jnum = json_object_array_get_idx(jscenes, s);
              errno = 0;
              int num = json_object_get_int(jnum);
              if ((num == 0) && (errno == EINVAL)) {
                DS_ERROR("scene is not a number", errno);
                continue;
              }
              scenes.push_back(num);
            }
          }
        }

        if (scenes.size() > 0) {
          const std::vector<int> lockedScenes = device->getLockedScenes();
          for (size_t q = 0; q < scenes.size(); q++) {
            if (std::find(lockedScenes.begin(), lockedScenes.end(),
                          scenes.at(q)) != lockedScenes.end()) {
              locked = true;
              break;
            }
          }
        }

        // the scene that we are about to modify is locked, skip it
        if (locked) {
          continue;
        }

        // apply bitmask to vlaue
        try {
          if (bitmask == 0) {
            device->setDeviceConfig(bank, offset, value);
          } else {
            // If the bitmask is set, we need to perform a read-modify-write
            // to make sure that we do not overwrite the bits specified in the
            // mask. This is needed to fulfill the requirement to not overwrite
            // "dont care" values of area scenes.
            uint8_t old = device->getDeviceConfig(bank, offset);

            // The bitmask identifies the bits that we are not allowed to
            // change. Get the old value, but clear the rest:
            old = old & bitmask;

            // The new value may have the masked bits set, but we do not
            // want to keep the bits in the old value, so clear them from the
            // new one.
            value = value & (~bitmask);

            // Now apply everything and write the value to device.
            value = value | old;
            device->setDeviceConfig(bank, offset, value);
          }
          failures = 0;
        } catch (const std::exception &ex) {
          ret += "Could not reset scene configuration for device " +
                 dsuid2str(device->getDSID()) + " " + ex.what() + ";";
          failures++;
        }
      }
    }

    return ret;
  }

  WebServerResponse InternalRequestHandler::jsonHandleRequest(const RestfulRequest& _request, std::shared_ptr<Session> _session, const struct mg_connection* _connection) {
    if (_request.getMethod() == "rentalReset") {
      // this call must block until the reset is complete, the calling
      // script depends on it

      // notify add-ons that they need to reset themselves, we'll check their
      // status later
      std::shared_ptr<Event> event = std::make_shared<Event>(EventName::AddonReset);
      event->setProperty("type", "rental-reset");
      if (DSS::hasInstance()) {
        DSS::getInstance()->raiseEvent(event);
      }

      // spec says to give the add-ons at least a minute, before polling for
      // their reset status, so remember the time when the event was senta
      DateTime addonEventSent;

      std::string jsonMessage = resetSceneConfiguration();

      // caller may provide a list of vdc implementation IDs, these vdcs
      // must be deleted if they are inactive
      std::vector<std::string> vdclist = splitString(
              _request.getParameter("vdcsToRemove"), ',', true);

      // delete all scene names: since we do not know which scene name has
      // been set, iterate thorugh all groups in each zone and check
      std::vector<std::shared_ptr<Zone> > zones = m_Apartment.getZones();
      for (size_t z = 0; z < zones.size(); z++) {
        std::shared_ptr<Zone> zone = zones.at(z);

        // reset devices
        if (zone->getID() == 0) {
          Set devices = zone->getDevices();
          for (int d = 0; d < devices.length(); d++) {
            std::shared_ptr<Device> device = devices[d].getDevice();

            // clearnames of devices by specified product families
            if (needToClearName(device)) {
              try {
                device->setName("");
                m_manipulator.deviceSetName(device, "");
              } catch (const std::exception &ex) {
                jsonMessage += "Could not reset name for device  " +
                               dsuid2str(device->getDSID()) + " " +
                               ex.what() + ";";
              }
            }
            // reset custom actions on all vDCs
            if (device->isVdcDevice()) {
              try {
                removeCustomActions(device);
              } catch (const std::exception &ex) {
                jsonMessage += "Could not remove custom action on device " +
                               dsuid2str(device->getDSID()) + " " +
                               ex.what() + ";";
              }
            }
          }
          continue;
        }

        std::vector<std::shared_ptr<Group> > groups = zone->getGroups();
        for (size_t g = 0; g < groups.size(); g++) {
          std::shared_ptr<Group> group = groups.at(g);
          for (int s = 0; s <= MaxSceneNumberOutsideZoneZero; s++) {
            // we are only interested in non area scenes
            if (!SceneHelper::isAreaScene(s)) {
              if (!group->getSceneName(s).empty()) {
                try {
                  m_manipulator.sceneSetName(group, s, "");
                } catch (const std::exception &ex) {
                  jsonMessage += "Could not reset name of scene  " +
                                 intToString(s) + " " + ex.what() + ";";
                }
              }
            }
          } // scenes
        } // groups
      } // zones

      if (!vdclist.empty()) {
        // iterate through meters and remove vDCs match the type supplied in the
        // list as parameter
        std::vector<std::shared_ptr<DSMeter> > vdsms;
        for (std::shared_ptr<DSMeter> dsMeter : m_Apartment.getDSMetersAll()) {
          if (dsMeter->getBusMemberType() == BusMember_vDSM && dsMeter->isPresent()) {
            vdsms.push_back(dsMeter);
          }
        }
        std::vector<std::shared_ptr<DSMeter> > meters = m_Apartment.getDSMeters();
        for (size_t m = 0; m < meters.size(); m++) {
          std::shared_ptr<DSMeter> meter = meters.at(m);
          std::string id = meter->getVdcImplementationId();
          if (!meter->isPresent() && !id.empty()) {
            if (std::find(vdclist.begin(), vdclist.end(), id) != vdclist.end()) {
              for (std::shared_ptr<DSMeter> vdsm : vdsms) {
                // we don't know on which vdsm the vdc is connected, so we try all of them
                try {
                  m_Apartment.getBusInterface()->getStructureModifyingBusInterface()
                      ->removeVdc(vdsm->getDSID(), meter->getDSID());
                } catch  (const std::exception &ex) {
                  DS_NOTICE("Could not delete vdc.", ex);
                }
              }
              std::shared_ptr<Zone> zone0 = m_Apartment.getZone(0);
              Set devices = zone0->getDevices().getByLastKnownDSMeter(meter->getDSID());
              for (int devId = 0; devId < devices.length(); ++devId) {
                auto& device = devices[devId];
                m_Apartment.removeDevice(device.getDSID());
              }
              m_Apartment.removeDSMeter(meter->getDSID());
            }
          }
        } // meters loop
      }


      // according to the spec, we should poll for the addon status
      // not earlier than a minute after the reset event went out
      while (DateTime().difference(addonEventSent) < 60) {
        sleep(5);
      }

      if (DSS::hasInstance()) {
        bool addonsDone;
        do {
          addonsDone = true;
          // now loop over script rental reset nodes until there are no nodes
          // set to "false"
          PropertyNodePtr scripts = DSS::getInstance()->getPropertySystem().getProperty("/scripts");
          if (scripts != NULL) {
            for (int n = 0; n < scripts->getChildCount(); n++) {
              PropertyNodePtr script = scripts->getChild(n);
              if ((script == NULL) || (script->getValueType() != vTypeNone)) {
                continue;
              }
              PropertyNodePtr resetted = script->getPropertyByName("rental-reset");
              if (resetted != NULL) {
                if (resetted->getBoolValue() == false) {
                  sleep(5);
                  addonsDone = false;
                  break;
                }
              }
            }
          }
        } while (!addonsDone);
      }

	  // raise ModelDirty to force rewrite of model data to apartment.xml
      if (DSS::hasInstance()) {
        DSS::getInstance()->getModelMaintenance().addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
      }

      if (jsonMessage.empty()) {
        return JSONWriter::success();
      } else {
        return JSONWriter::failure(jsonMessage);
      }
    }

    JSONWriter json;
    // only the last two calls need a valid device
    dsuid_t dsuid = str2dsuid(_request.getParameter("dsuid"));
    // will throw if not found
    std::shared_ptr<Device> pDevice = m_Apartment.getDeviceByDSID(dsuid);

    if (_request.getMethod() == "getUMR200Config") {
      view::umr200Config::get(json, *pDevice);
      return json.successJSON();
    } else if (_request.getMethod() == "setUMR200Config") {
      const auto& value = _request.getRequiredParameter("value");

      DS_NOTICE("Set umr200 config.", *pDevice, value);

      view::umr200Config::setValue(*pDevice, m_manipulator, value, json, m_Apartment);
      return json.successJSON();
    } else {
      throw std::runtime_error("Unhandled function");
    }
  } // jsonHandleRequest
} // namespace dss
