/*
    Copyright (c) 2015 digitalSTROM.org, Zurich, Switzerland

    Author: Remy Mahler, digitalSTROM AG <remy.mahler@digitalstrom.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN

#include <ds/log.h>

#include <boost/test/unit_test.hpp>
#include <memory>
#include <string>

#include "base.h"
#include "businterface-mock.h"
#include "ds485types.h"
#include "event.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/group.h"
#include "scripting/jscluster.h"
#include "structuremanipulator.h"
#include "util/modelmaintenance-mockup.h"

static ds::LogChannel dsLogChannel("testsJsClusterTests");

using namespace dss;
using namespace std::literals;
using trompeloeil::_;

namespace {

class InstanceHelper
{
public:
  InstanceHelper(Apartment* _apartment, PropertySystem& propertySystem, EventSink& eventSink)
      : m_propertySystem(propertySystem),
        m_eventSink(eventSink),
        modelMaintenance(new ModelMaintenanceMock(m_propertySystem, eventSink)),
        m_apartment(_apartment) {
    m_apartment->setBusInterface(&busInterface);
    modelMaintenance->setApartment(m_apartment);
    modelMaintenance->setStructureModifyingBusInterface(busInterface.getStructureModifyingBusInterface());
    modelMaintenance->initialize();
  };

  ~InstanceHelper() {
    delete modelMaintenance;
  };

public:
  PropertySystem& m_propertySystem;
  EventSink& m_eventSink;
  BusInterfaceMock busInterface{};
  ModelMaintenanceMock* modelMaintenance;

  Apartment* m_apartment;
};

static const int MAX_CLUSTERS = GroupIDAppUserMax - GroupIDAppUserMin + 1;
DSUID_DEFINE(dsuid1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10);

} // namespace

BOOST_AUTO_TEST_SUITE(JSCluster)

BOOST_AUTO_TEST_CASE(testJSClusterCreate_single) {
  PropertySystem propertySystem{};
  EventQueue eventQueue{nullptr};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt1 = *pApartment;
  InstanceHelper helper(&apt1, propertySystem, eventQueue);
  auto& busInterface = helper.busInterface;

  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();

  ScriptExtension* ext = new ClusterScriptExtension(apt1);
  env->addExtension(ext);
  std::string output = "test";

  boost::scoped_ptr<ScriptContext> ctx(env->getContext());
  REQUIRE_CALL(busInterface.structureModify, clusterSetName(_, "test"s)).SIDE_EFFECT(DS_DEBUG(_1));
  REQUIRE_CALL(busInterface.structureModify, clusterSetApplication(_, _, _)).SIDE_EFFECT(DS_DEBUG(_1, _2, _3));
  REQUIRE_CALL(busInterface.structureModify, clusterSetProperties(_, _, _, _)).SIDE_EFFECT(DS_DEBUG(_1, _2, _3, _4));
  REQUIRE_CALL(busInterface.structureModify, clusterSetLockedScenes(_, _)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  int clusterId = ctx->evaluate<int>("Cluster.createCluster(2,'" + output + "')");
  BOOST_CHECK(clusterId != -1);

  std::vector<std::shared_ptr<Cluster> > clusters = apt1.getClusters();
  std::shared_ptr<Cluster> assignedCluster;
  for (std::shared_ptr<Cluster> cluster : clusters) {
    if (cluster->getApplicationType() != ApplicationType::none) {
      assignedCluster = cluster;
    }
  }
  BOOST_CHECK(assignedCluster != NULL);
  BOOST_CHECK(output.compare(assignedCluster->getName()) == 0);
  BOOST_CHECK_EQUAL(assignedCluster->getApplicationType(), ApplicationType::blinds);
}

BOOST_AUTO_TEST_CASE(testJSClusterCreate_use_all_clusters) {
  PropertySystem propertySystem{};
  EventQueue eventQueue{nullptr};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt1 = *pApartment;
  InstanceHelper helper(&apt1, propertySystem, eventQueue);
  auto& busInterface = helper.busInterface;

  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();

  ScriptExtension* ext = new ClusterScriptExtension(apt1);
  env->addExtension(ext);
  std::string output = "test";

  boost::scoped_ptr<ScriptContext> ctx(env->getContext());

  for (int i = 0; i <= MAX_CLUSTERS; ++i) {
    int clusterId{};
    if (i < MAX_CLUSTERS) {
      REQUIRE_CALL(busInterface.structureModify, clusterSetName(_, "test"s))
          .SIDE_EFFECT(DS_DEBUG("clusterSetName", _1));
      REQUIRE_CALL(busInterface.structureModify, clusterSetApplication(_, _, _))
          .SIDE_EFFECT(DS_DEBUG("clusterSetApplication", _1, _2, _3));
      REQUIRE_CALL(busInterface.structureModify, clusterSetProperties(_, _, _, _))
          .SIDE_EFFECT(DS_DEBUG("clusterSetProperties", _1, _2, _3, _4));
      REQUIRE_CALL(busInterface.structureModify, clusterSetLockedScenes(_, _))
          .SIDE_EFFECT(DS_DEBUG("clusterSetLockedScenes", _1, _2));
      clusterId = ctx->evaluate<int>("Cluster.createCluster(2, '" + output + "')");
    } else {
      clusterId = ctx->evaluate<int>("Cluster.createCluster(2, '" + output + "')");
    }
    if (i == MAX_CLUSTERS) {
      BOOST_CHECK_EQUAL(clusterId, -1);
    } else {
      BOOST_CHECK(clusterId != -1);
    }
  }

  std::vector<std::shared_ptr<Cluster> > clusters = apt1.getClusters();

  std::shared_ptr<Cluster> assignedCluster;
  for (std::shared_ptr<Cluster> cluster : clusters) {
    BOOST_CHECK(output.compare(cluster->getName()) == 0);
    BOOST_CHECK_EQUAL(cluster->getApplicationType(), ApplicationType::blinds);
  }
}

BOOST_AUTO_TEST_CASE(testJSClusterRelease) {
  PropertySystem propSystem;
  EventQueue eventQueue{nullptr};
  auto pApartment = makeApartment(nullptr, &propSystem);
  auto& apt1 = *pApartment;
  InstanceHelper helper(&apt1, propSystem, eventQueue);
  auto& busInterface = helper.busInterface;
  StructureManipulator structureManipulator(apt1);

  // create cluster
  std::string output = "test";
  std::shared_ptr<Cluster> cluster = apt1.getEmptyCluster();
  cluster->setApplicationType(ApplicationType::blinds);
  cluster->setName(output);

  // create device and add to cluster
  std::shared_ptr<Device> dev = apt1.allocateDevice(dsuid1);
  dev->setShortAddress(1);
  dev->setName("dev");
  dev->addToGroup(cluster->getID());

  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();

  ScriptExtension* ext = new ClusterScriptExtension(apt1);
  env->addExtension(ext);

  // cluster can not be removed. not empty
  boost::scoped_ptr<ScriptContext> ctx(env->getContext());
  bool success = ctx->evaluate<bool>("Cluster.removeCluster(" + intToString(cluster->getID()) + ")");
  BOOST_CHECK_EQUAL(success, false);

  // remove device from cluster
  // TODO -> structuremanipulator-test, use mock here
  REQUIRE_CALL(busInterface.structureModify, removeFromGroup(_, _, _));
  structureManipulator.deviceRemoveFromGroup(dev, cluster);

  // cluster can be removed
  REQUIRE_CALL(busInterface.structureModify, clusterSetName(_, _)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  REQUIRE_CALL(busInterface.structureModify, clusterSetApplication(_, _, _)).SIDE_EFFECT(DS_DEBUG(_1, _2, _3));
  REQUIRE_CALL(busInterface.structureModify, clusterSetProperties(_, _, _, _)).SIDE_EFFECT(DS_DEBUG(_1, _2, _3, _4));
  REQUIRE_CALL(busInterface.structureModify, clusterSetLockedScenes(_, _)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  success = ctx->evaluate<bool>("Cluster.removeCluster(" + intToString(cluster->getID()) + ")");
  BOOST_CHECK_EQUAL(success, true);
}

BOOST_AUTO_TEST_CASE(testJSClusterAddRemoveDevice) {
  PropertySystem propertySystem;
  EventQueue eventQueue{nullptr};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt1 = *pApartment;
  InstanceHelper helper(&apt1, propertySystem, eventQueue);
  auto& busInterface = helper.busInterface;

  // create cluster
  std::string output = "test";
  std::shared_ptr<Cluster> cluster = apt1.getEmptyCluster();
  cluster->setApplicationType(ApplicationType::blinds);
  cluster->setName(output);

  // create device and add to cluster
  std::shared_ptr<Device> dev = apt1.allocateDevice(dsuid1);
  dev->setShortAddress(1);
  dev->setName("dev");

  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();

  ScriptExtension* ext = new ClusterScriptExtension(apt1);
  env->addExtension(ext);

  BOOST_CHECK_EQUAL(dev->isInGroup(cluster->getID()), false);
  boost::scoped_ptr<ScriptContext> ctx(env->getContext());

  // uses StructureManipulator.deviceAddToGroup(_device, cluster);
  REQUIRE_CALL(busInterface.structureModify, addToGroup(_, _, 1));
  REQUIRE_CALL(busInterface.device, setDeviceButtonActiveGroup(_, _));
  ctx->evaluate<void>("Cluster.clusterAddDevice(" + intToString(cluster->getID()) + ", '" + dsuid2str(dsuid1)+ "')");
  BOOST_CHECK_EQUAL(dev->isInGroup(cluster->getID()), true);

  // TODO -> structuremanipulator-test, use mock here
  REQUIRE_CALL(busInterface.structureModify, removeFromGroup(_, _, _));
  ctx->evaluate<void>("Cluster.clusterRemoveDevice(" + intToString(cluster->getID()) + ", '" + dsuid2str(dsuid1)+ "')");
  BOOST_CHECK_EQUAL(dev->isInGroup(cluster->getID()), false);
}

BOOST_AUTO_TEST_CASE(testJSClusterLockUnlock) {
  PropertySystem propertySystem;
  EventQueue eventQueue{nullptr};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt1 = *pApartment;
  InstanceHelper helper(&apt1, propertySystem, eventQueue);
  auto& busInterface = helper.busInterface;

  // create cluster
  std::string output = "test";
  std::shared_ptr<Cluster> cluster = apt1.getEmptyCluster();
  cluster->setApplicationType(ApplicationType::blinds);
  cluster->setName(output);

  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();

  ScriptExtension* ext = new ClusterScriptExtension(apt1);
  env->addExtension(ext);

  BOOST_CHECK_EQUAL(cluster->isConfigurationLocked(), false);
  boost::scoped_ptr<ScriptContext> ctx(env->getContext());
  REQUIRE_CALL(busInterface.structureModify, clusterSetConfigurationLock(_, true)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  ctx->evaluate<void>("Cluster.clusterConfigurationLock(" + intToString(cluster->getID()) + ", true)");
  BOOST_CHECK_EQUAL(cluster->isConfigurationLocked(), true);

  REQUIRE_CALL(busInterface.structureModify, clusterSetConfigurationLock(_, false)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  ctx->evaluate<void>("Cluster.clusterConfigurationLock(" + intToString(cluster->getID()) + ", false)");
  BOOST_CHECK_EQUAL(cluster->isConfigurationLocked(), false);
}

BOOST_AUTO_TEST_CASE(testJSClusterSetGetName) {
  PropertySystem propertySystem;
  EventQueue eventQueue{nullptr};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt1 = *pApartment;
  InstanceHelper helper(&apt1, propertySystem, eventQueue);
  auto& busInterface = helper.busInterface;

  // create cluster
  std::string output = "test";
  std::shared_ptr<Cluster> cluster = apt1.getEmptyCluster();
  cluster->setApplicationType(ApplicationType::blinds);
  cluster->setName(output);

  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();

  ScriptExtension* ext = new ClusterScriptExtension(apt1);
  env->addExtension(ext);

  boost::scoped_ptr<ScriptContext> ctx(env->getContext());

  output = "firstNameTest";
  REQUIRE_CALL(busInterface.structureModify, clusterSetName(_, "firstNameTest"s)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  REQUIRE_CALL(busInterface.structureModify, clusterSetApplication(_, _, _)).SIDE_EFFECT(DS_DEBUG(_1, _2, _3));
  REQUIRE_CALL(busInterface.structureModify, clusterSetProperties(_, _, _, _)).SIDE_EFFECT(DS_DEBUG(_1, _2, _3, _4));
  REQUIRE_CALL(busInterface.structureModify, clusterSetLockedScenes(_, _)).SIDE_EFFECT(DS_DEBUG(_1, _2));
  ctx->evaluate<void>("Cluster.clusterSetName(" + intToString(cluster->getID()) + ", '" + output + "')");
  BOOST_CHECK(output.compare(cluster->getName()) == 0);

  output = "secondNameTest";
  cluster->setName(output);

  std::string name = ctx->evaluate<std::string>("Cluster.clusterGetName(" + intToString(cluster->getID()) + ")");
  BOOST_CHECK(output.compare(name) == 0);
}

BOOST_AUTO_TEST_SUITE_END()
