/*
    Copyright (c) 2012 digitalSTROM.org, Zurich, Switzerland

    Author: Christian Hitz, aizo ag <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN
#define BOOST_TEST_IGNORE_NON_ZERO_CHILD_CODE
#include <ds/log.h>
#include <rrd.h>

#include <boost/test/unit_test.hpp>
#include <filesystem>
#include <memory>

#include "event.h"
#include "eventinterpreterplugins.h"
#include "metering/archives-sanity.h"
#include "metering/archives-storage.h"
#include "metering/metering.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/modelconst.h"
#include "propertysystem.h"
#include "util/dss_instance_fixture.h"

using namespace std;
using namespace dss;
using namespace dss::metering;

namespace chrono = std::chrono;
using system_clock = std::chrono::system_clock;

DSUID_DEFINE(dsuid, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13);
DSUID_DEFINE(dsmeterDSID, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10);

static ds::LogChannel dsLogChannel("meteringTests");

auto& setupStorage(Metering& metering) {
  auto& storage = metering.getArchivesStorage();
  storage.setMeteringStorageLocation("."s); // use better tmp
  return storage;
}

auto createEmptyArchive = [](Metering& metering, DSMeter& meter, Category category) {
  DS_REQUIRE(category == Category::consumption);
  metering.addConsumedEnergy(meter, ElectricityMeasurement{0, 0, ds::DateTime{1s}});   // baseline
  metering.addConsumedEnergy(meter, ElectricityMeasurement{10, 10, ds::DateTime{2s}}); // create archive
};

BOOST_AUTO_TEST_SUITE(Metering)

BOOST_AUTO_TEST_CASE(seriesSizes) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  std::shared_ptr<DSMeter> pMeter = apt.allocateDSMeter(dsuid);
  pMeter->setCapability_HasMetering(true);
  auto pMeters = std::vector<Dsuid>{pMeter->getDSID()};

  auto pMetering = makeMetering(NULL, apt);
  auto& metering(*pMetering);

  auto archiveId = ArchiveId{pMeter->getDSID(), Category::consumption};

  auto& storage = setupStorage(metering);
  DS_DEFER(std::filesystem::remove(storage.filepath(archiveId)));
  createEmptyArchive(*pMetering, *pMeter, Category::consumption);

  {
    auto pValues = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {}, 0);
    BOOST_CHECK_EQUAL(pValues.values.size(), 600 - 1);
    BOOST_CHECK(std::chrono::seconds{pValues.step} == 1s);
  }
  {
    auto pValues = metering.getSeries(pMeters, 60s, SeriesTypes::power, false, {}, 0);
    BOOST_CHECK_EQUAL(pValues.values.size(), 720 - 1);
    BOOST_CHECK(std::chrono::seconds{pValues.step} == 60s);
  }
  {
    auto pValues = metering.getSeries(pMeters, 15min, SeriesTypes::power, false, {}, 0);
    BOOST_CHECK_EQUAL(pValues.values.size(), 2976 - 1);
    BOOST_CHECK(std::chrono::seconds{pValues.step} == 15min);
  }
  {
    auto pValues = metering.getSeries(pMeters, 24h, SeriesTypes::power, false, {}, 0);
    BOOST_CHECK_EQUAL(pValues.values.size(), 370 - 1);
    BOOST_CHECK(std::chrono::seconds{pValues.step} == 24h);
  }
  {
    auto pValues = metering.getSeries(pMeters, 7 * 24h, SeriesTypes::power, false, {}, 0);
    BOOST_CHECK_EQUAL(pValues.values.size(), 260 - 1);
    BOOST_CHECK(std::chrono::seconds{pValues.step} == 7 * 24h);
  }
  {
    auto pValues = metering.getSeries(pMeters, 30 * 24h, SeriesTypes::power, false, {}, 0);
    BOOST_CHECK_EQUAL(pValues.values.size(), 60 - 1);
    BOOST_CHECK(std::chrono::seconds{pValues.step} == 30 * 24h);
  }
} // seriesSizes

BOOST_AUTO_TEST_CASE(seriesRanges) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  std::shared_ptr<DSMeter> pMeter = apt.allocateDSMeter(dsuid);
  pMeter->setCapability_HasMetering(true);
  auto pMeters = std::vector<Dsuid>{pMeter->getDSID()};

  auto pMetering = makeMetering(NULL, apt);
  auto& metering(*pMetering);

  auto archiveId = ArchiveId{pMeter->getDSID(), Category::consumption};

  auto& storage = setupStorage(metering);
  DS_DEFER(std::filesystem::remove(storage.filepath(archiveId)));
  createEmptyArchive(*pMetering, *pMeter, Category::consumption);

  {
    auto now = system_clock::now();
    auto range = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {now - 20s, std::nullopt}, 0);
    auto& pValues = range.values;
    BOOST_CHECK_EQUAL(pValues.size(), 20 - 1);
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(0).timeStamp), system_clock::to_time_t(now - 19s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(1).timeStamp), system_clock::to_time_t(now - 18s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(18).timeStamp), system_clock::to_time_t(now - 1s));
  }

  {
    auto now = system_clock::now();
    auto range = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {now - 20s, std::nullopt}, 10);
    auto& pValues = range.values;
    BOOST_CHECK_EQUAL(pValues.size(), 10 - 1);
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(0).timeStamp), system_clock::to_time_t(now - 19s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(1).timeStamp), system_clock::to_time_t(now - 18s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(8).timeStamp), system_clock::to_time_t(now - 11s));
  }

  {
    auto now = system_clock::now();
    auto range = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {std::nullopt, now - 580s}, 0);
    auto& pValues = range.values;
    BOOST_CHECK_EQUAL(pValues.size(), 20 - 1);
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(0).timeStamp), system_clock::to_time_t(now - 599s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(1).timeStamp), system_clock::to_time_t(now - 598s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(18).timeStamp), system_clock::to_time_t(now - 581s));
  }

  {
    auto now = system_clock::now();
    auto range = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {std::nullopt, now - 580s}, 10);
    auto& pValues = range.values;
    BOOST_CHECK_EQUAL(pValues.size(), 10 - 1);
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(0).timeStamp), system_clock::to_time_t(now - 589s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(1).timeStamp), system_clock::to_time_t(now - 588s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(8).timeStamp), system_clock::to_time_t(now - 581s));
  }

  {
    auto now = system_clock::now();
    auto range = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {now - 580s, now - 560s}, 0);
    auto& pValues = range.values;
    BOOST_CHECK_EQUAL(pValues.size(), 20 - 1);
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(0).timeStamp), system_clock::to_time_t(now - 579s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(1).timeStamp), system_clock::to_time_t(now - 578s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(18).timeStamp), system_clock::to_time_t(now - 561s));
  }

  {
    auto now = system_clock::now();
    auto range = metering.getSeries(pMeters, 1s, SeriesTypes::power, false, {now - 580s, now - 560s}, 10);
    auto& pValues = range.values;
    BOOST_CHECK_EQUAL(pValues.size(), 20 - 1);
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(0).timeStamp), system_clock::to_time_t(now - 579s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(1).timeStamp), system_clock::to_time_t(now - 578s));
    BOOST_CHECK_EQUAL(system_clock::to_time_t(pValues.at(18).timeStamp), system_clock::to_time_t(now - 561s));
  }

  {
    auto _15MinuteAverage = metering.get15minConsumed(pMeters, SeriesTypes::energyDelta);
    BOOST_CHECK_EQUAL(_15MinuteAverage, 0); // no data present
  }
} // seriesSizes

BOOST_AUTO_TEST_CASE(seriesDateForward) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  std::shared_ptr<DSMeter> pMeter = apt.allocateDSMeter(dsuid);
  pMeter->setCapability_HasMetering(true);

  std::vector<std::shared_ptr<DSMeter> > pMeters;
  pMeters.push_back(pMeter);
  auto metering = makeMetering(NULL, apt);

  auto& storage = setupStorage(*metering);
  DS_DEFER(std::filesystem::remove(storage.filepath({pMeter->getDSID(), Category::consumption})));
  auto fileName = storage.filepath({pMeter->getDSID(), Category::consumption});

  {
    // add some entry to db.
    // check that timestamp of last entry is in range of latest entry
    DateTime now;
    for (int ctr = 0; ctr < 10; ++ctr) {
      metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{now}});
      now = now.addSeconds(1);
    }

    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();

    DateTime lastDbEntry(rrd_last_r(fileName.c_str()));
    int delta = lastDbEntry.difference(now);

    bool success = (delta > -2) && (delta < 2);
    BOOST_CHECK_EQUAL(success, true);
  }

  {
    // db roll forward
    // check that timestamp of last entry is in range of latest entry
    DateTime now;
    now = now.addDay(6);
    metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{now}});
    metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{now.addSeconds(1)}});

    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();

    DateTime lastDbEntry(rrd_last_r(fileName.c_str()));
    int delta = lastDbEntry.difference(now);
    bool success = (delta > -2) && (delta < 2);
    BOOST_CHECK_EQUAL(success, true);
  }

  {
    // timestamp of data too far in the future.
    // remove database
    DateTime now;
    DateTime future = now.addDay(15);
    metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{future}});
    metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{future.addSeconds(1)}});

    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();

    BOOST_CHECK(!std::filesystem::exists(fileName));
  }
}

BOOST_AUTO_TEST_CASE(seriesDateReverse) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  std::shared_ptr<DSMeter> pMeter = apt.allocateDSMeter(dsuid);
  pMeter->setCapability_HasMetering(true);
  auto metering = makeMetering(NULL, apt);
  DateTime recentTS;

  auto& storage = setupStorage(*metering);
  DS_DEFER(std::filesystem::remove(storage.filepath({pMeter->getDSID(), Category::consumption})));
  auto fileName = storage.filepath({pMeter->getDSID(), Category::consumption});

  {
    // insert some data
    for (int ctr = 0; ctr < 10; ++ctr) {
      metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{recentTS}});
      recentTS = recentTS.addSeconds(1);
    }

    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();

    recentTS = rrd_last_r(fileName.c_str());
  }

  {
    // insert earlier data
    // db should not change
    DateTime past = recentTS.addDay(-5);
    metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{past}});

    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();

    DateTime lastDbEntry(rrd_last_r(fileName.c_str()));
    int delta = lastDbEntry.difference(recentTS);
    BOOST_CHECK_EQUAL(delta, 0);
  }

  {
    // insert earlier data
    // db should not change, since system date is used as a reference
    DateTime past = recentTS.addDay(-15);
    metering->addConsumedEnergy(*pMeter, {22, 44, ds::DateTime{past}});

    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();

    DateTime lastDbEntry(rrd_last_r(fileName.c_str()));
    int delta = lastDbEntry.difference(recentTS);
    BOOST_CHECK_EQUAL(delta, 0);
  }
}

BOOST_AUTO_TEST_CASE(seriesDateCheckDbNoReset) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  std::shared_ptr<DSMeter> pMeter = apt.allocateDSMeter(dsuid);
  pMeter->setCapability_HasMetering(true);
  auto metering = makeMetering(NULL, apt);

  auto& storage = setupStorage(*metering);
  DS_DEFER(std::filesystem::remove(storage.filepath({pMeter->getDSID(), Category::consumption})));
  auto fileName = storage.filepath({pMeter->getDSID(), Category::consumption});

  // newly created archive alway has lastTimstamp == now, refuses to add values from the past
  auto recentTS = system_clock::now() + 6 * 24h;

  {
    // insert some data
    for (int ctr = 0; ctr < 10; ++ctr) {
      metering->addConsumedEnergy(*pMeter, {22, 44, recentTS});
      recentTS += 1s;
    }
    recentTS = system_clock::from_time_t(rrd_last_r(fileName.c_str()));
  }

  {
    // check DB reset: timestamp within threshold
    // db not modified
    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();
    auto lastDbEntry = system_clock::from_time_t(rrd_last_r(fileName.c_str()));
    BOOST_CHECK(lastDbEntry == recentTS);
  }
}

BOOST_AUTO_TEST_CASE(seriesDateCheckDbReset) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  std::shared_ptr<DSMeter> pMeter = apt.allocateDSMeter(dsuid);
  pMeter->setCapability_HasMetering(true);
  auto metering = makeMetering(NULL, apt);

  auto& storage = setupStorage(*metering);
  DS_DEFER(std::filesystem::remove(storage.filepath({pMeter->getDSID(), Category::consumption})));
  auto fileName = storage.filepath({pMeter->getDSID(), Category::consumption});

  // newly created archive alway has lastTimstamp == now, refuses to add values from the past
  auto recentTS = system_clock::now() + 10 * 24h;

  {
    // insert some data
    for (int ctr = 0; ctr < 10; ++ctr) {
      metering->addConsumedEnergy(*pMeter, {22, 44, recentTS});
      recentTS += 1s;
    }
    recentTS = system_clock::from_time_t(rrd_last_r(fileName.c_str()));
  }

  {
    // check DB reset: timestamp outside threshold
    // db should be removed
    auto& sanity = metering->getArchivesSanity();
    sanity.removeDeadArchives();
    BOOST_CHECK(!std::filesystem::exists(fileName));
  }
}

BOOST_FIXTURE_TEST_CASE(testEnergyMeterRestoreData, DSSInstanceFixture) {
  auto& apt = DSS::getInstance()->getApartment();
  auto pMetering = &apt.getMetering();

  auto& storage = setupStorage(*pMetering);
  DS_DEFER(std::filesystem::remove(storage.filepath({dsmeterDSID, Category::consumption})));

  {
    std::shared_ptr<DSMeter> meter = apt.allocateDSMeter(dsmeterDSID);
    meter->setCapability_HasMetering(true);
    pMetering->addConsumedEnergy(*meter, {0, 100, ds::DateTime{300s}});
    pMetering->addConsumedEnergy(*meter, {0, 120, ds::DateTime{301s}});
    BOOST_CHECK_EQUAL(20, pMetering->getEnergy(dsmeterDSID)->energy);
  }
  apt.removeDSMeter(dsmeterDSID);
  {
    std::shared_ptr<DSMeter> meter = apt.allocateDSMeter(dsmeterDSID);
    meter->setCapability_HasMetering(true);
    pMetering->addConsumedEnergy(*meter, {0, 10, ds::DateTime{302s}});
    pMetering->addConsumedEnergy(*meter, {0, 20, ds::DateTime{303s}});
    BOOST_CHECK_EQUAL(30, pMetering->getEnergy(dsmeterDSID)->energy);
    // TODO check for floating point values
  }
}

BOOST_AUTO_TEST_SUITE_END()
