/*
 *  Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */


#define BOOST_TEST_NO_MAIN

#include <boost/algorithm/string/predicate.hpp>
#include <boost/range/algorithm.hpp>
#include <memory>
#include <boost/test/unit_test.hpp>

#include <ds/log.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sqlite3_wrapper.h"
#include "model/device.h"
#include "vdc-db.h"
#include "propertysystem.h"
#include "vdc-connection.h"
#include "web/webrequests.h"
#include "web/handler/devicerequesthandler.h"
#include "tests/util/dss_instance_fixture.h"

using namespace dss;

// http://db.aizo.net/vdc-db.php

BOOST_AUTO_TEST_SUITE(VDC_DB)

namespace {
struct Recreate {
  Recreate() { vdcDb::recreate(*DSS::getInstance()); }
};

class Fixture : public DSSInstanceFixture {
public:
  Recreate recreate;
  std::unique_ptr<VdcDb> dbPtr;
  VdcDb& db;

  Fixture() : dbPtr(vdcDb::make(*DSS::getInstance())), db(*dbPtr) {}
};
} // namespace

#if 0
static void dumpStates(std::vector<DeviceStateSpec_t> states) {
  for (const DeviceStateSpec_t &state: states) {
    std::string values;
    for (const std::string val: state.Values) {
      if (values.size() > 0) {
        values += ", ";
      }
      values += val;
    }
    Logger::getInstance()->log(state.Name + ": [" + values + "]", lsWarning);
  }
}
#endif

static const char *gtin = "1234567890123";

BOOST_FIXTURE_TEST_CASE(getSpec, Fixture) {
  const auto specs = db.getSpec(gtin);
  const auto dummyNodeFound =
      boost::find_if<boost::return_found_end>(specs, [&](const VdcDb::SpecDesc& x) { return x.id == "dummyNode"; });
  BOOST_REQUIRE(dummyNodeFound);
  BOOST_CHECK_EQUAL(dummyNodeFound.front().value, "dummy value223");
}

BOOST_FIXTURE_TEST_CASE(getStates, Fixture) {
  std::vector<DeviceStateSpec_t> states_s;
  states_s = db.getStatesLegacy(gtin);
  //dumpStates(states_s);

  BOOST_REQUIRE_EQUAL(states_s.size(), 1);
  BOOST_CHECK_EQUAL(states_s[0].Name, "dummyState");
  BOOST_REQUIRE_EQUAL(states_s[0].Values.size(), 4);
  BOOST_CHECK_EQUAL(states_s[0].Values[0], "d");

  std::vector<VdcDb::StateDesc> states_i;
  states_i = db.getStates(gtin);
  BOOST_REQUIRE_EQUAL(states_i.size(), 1);
  BOOST_CHECK_EQUAL(states_i[0].id, "dummyState");
  BOOST_REQUIRE_EQUAL(states_i[0].values.size(), 4);
  std::vector<std::string> units { "d", "u", "mm", "y" };
  for (int i = 0; i < 4; ++i) {
    auto match = std::find(units.begin(), units.end(), states_i[0].values[i]);
    BOOST_CHECK(match != units.end());
    units.erase(match);
  }

  states_i = db.getStates(gtin);
  BOOST_REQUIRE_EQUAL(states_i.size(), 1);
  BOOST_CHECK_EQUAL(states_i[0].id, "dummyState");
  BOOST_REQUIRE_EQUAL(states_i[0].values.size(), 4);
  BOOST_CHECK_EQUAL(states_i[0].values[0], "d");
}

#if 0
static void dumpProperties(const std::vector<VdcDb::PropertyDesc> &props) {
  std::string out;
  for (const VdcDb::PropertyDesc &prop: props) {
    out += "\tname: " + prop.id + ", type: " + intToString(prop.typeId) + ", default: " + prop.defaultValue + "\n";
  }
  Logger::getInstance()->log("properties: \n" + out, lsWarning);
}
#endif

BOOST_FIXTURE_TEST_CASE(lookupProperties, Fixture) {
  std::vector<VdcDb::PropertyDesc> props;
  props = db.getProperties(gtin);
  BOOST_REQUIRE_EQUAL(props.size(), 1);
  BOOST_CHECK_EQUAL(props[0].id, "dummyProperty");
  //dumpProperties(props);

  props = db.getProperties(gtin);
  BOOST_REQUIRE_EQUAL(props.size(), 1);
  BOOST_CHECK_EQUAL(props[0].id, "dummyProperty");
  //dumpProperties(props);
}

#if 0
static void dumpOutputs(const std::vector<VdcDb::OutputDesc> &outputs) {
  std::string out;
  for (const VdcDb::OutputDesc &o: outputs) {
    out += "\tname: " + o.id
        + ", typeId: " + intToString(o.typeId) +
        + ", typePostfix: " + o.typePostfix +
        + ", dsType: " + o.outputType +
        + ", dsIndex: " + o.outputIndex +
        ", default: " + o.defaultValue + "\n";
  }
  Logger::getInstance()->log("outputs: \n" + out, lsWarning);
}
#endif

BOOST_FIXTURE_TEST_CASE(lookupOutputs, Fixture) {
  std::vector<VdcDb::OutputDesc> outputs;
  outputs = db.getOutputs(gtin);
  //dumpOutputs(outputs);

  BOOST_REQUIRE_EQUAL(outputs.size(), 5);
  BOOST_CHECK_EQUAL(outputs[0].id, "flowRateHand");
  BOOST_CHECK_EQUAL(outputs[4].id, "waterTemperatureTop");
}

#if 0
static void dumpActionDesc(const std::vector<VdcDb::ActionDesc> &actions) {
  std::string out;
  for (const VdcDb::ActionDesc &action: actions) {
    out += "name: " + action.id + "\n";
    for (const auto &p: action.params) { out += "\t name: " + p.id + ", default: " + p.defaultValue + "\n"; }
  }
  Logger::getInstance()->log("actions: \n" + out, lsWarning);
}
#endif

BOOST_FIXTURE_TEST_CASE(lookupActions, Fixture) {
  const std::vector<VdcDb::ActionDesc>& actions = db.getActions(gtin);
  //dumpActionDesc(actions);

  BOOST_REQUIRE_EQUAL(actions.size(), 4);
  BOOST_CHECK_EQUAL(actions[0].id, "dummyAction1");
  BOOST_CHECK_EQUAL(actions[0].categoryId, "catEvent1");
  BOOST_CHECK_EQUAL(actions[1].id, "dummyAction2");
  BOOST_CHECK_EQUAL(actions[1].categoryId, "catEvent2");
  BOOST_CHECK_EQUAL(actions[2].id, "waterFlowAction");
  BOOST_CHECK_EQUAL(actions[2].categoryId, "");
  BOOST_CHECK_EQUAL(actions[3].id, "waterStop");
  BOOST_CHECK_EQUAL(actions[3].categoryId, "");
  BOOST_REQUIRE_EQUAL(actions[0].params.size(), 4);
  BOOST_CHECK_EQUAL(actions[0].params[0].id, "dummyActionParam1");
  BOOST_CHECK_EQUAL(actions[0].params[1].id, "dummyActionParam2");
  BOOST_CHECK_EQUAL(actions[0].params[2].id, "flowRateHand");
  BOOST_CHECK_EQUAL(actions[0].params[3].id, "waterTemperatureTop");
  BOOST_REQUIRE_EQUAL(actions[2].params.size(), 4);
  BOOST_CHECK_EQUAL(actions[2].params[0].id, "flowRateMain");
  BOOST_CHECK_EQUAL(actions[2].params[1].id, "flowRateTop");
  BOOST_CHECK_EQUAL(actions[2].params[2].id, "waterTemperatureMain");
  BOOST_CHECK_EQUAL(actions[2].params[3].id, "waterTemperatureTop");
}

#if 0
static void dumpStandardActionDesc(const std::vector<VdcDb::StandardActionDesc> &actions) {
  std::string out;
  for (auto action: actions) {
    out += "name: " + action.id + ", action_name: " + action.action + "\n";
    for (auto arg: action.params) {
        out += "\t " + arg.first + ": " + arg.second + "\n";
    }
  }
  Logger::getInstance()->log("standard actions: \n" + out, lsWarning);
}
#endif

BOOST_FIXTURE_TEST_CASE(lookupStandardActions, Fixture) {
  std::vector<VdcDb::StandardActionDesc> stdActions;
  stdActions = db.getStandardActions(gtin);
  //dumpStandardActionDesc(stdActions);

  BOOST_REQUIRE_EQUAL(stdActions.size(), 2);
  BOOST_CHECK_EQUAL(stdActions[0].id, "std.dummy");
  BOOST_REQUIRE_EQUAL(stdActions[0].params.size(), 4);
  BOOST_CHECK_EQUAL(stdActions[0].params[2].first, "flowRateHand");
  BOOST_CHECK_EQUAL(stdActions[0].params[2].second, "1");
  BOOST_CHECK_EQUAL(stdActions[0].params[3].first, "waterTemperatureTop");
  BOOST_CHECK_EQUAL(stdActions[0].params[3].second, "42");
  BOOST_CHECK_EQUAL(stdActions[1].id, "std.moreDummy");

  BOOST_CHECK_NO_THROW(stdActions = db.getStandardActions(gtin));
  //dumpStandardActionDesc(stdActions);
  BOOST_REQUIRE_EQUAL(stdActions.size(), 2);
  BOOST_CHECK_EQUAL(stdActions[0].id, "std.dummy");
  BOOST_CHECK_EQUAL(stdActions[1].id, "std.moreDummy");
}

#if 0
static void dumpSensorDesc(const std::vector<VdcDb::SensorDesc> &sensors) {
  std::string out;
  for (const VdcDb::SensorDesc &sensor: sensors) {
    out += "index: " + sensor.sensorIndex + "\n";
    out += "dsType: " + sensorTypeName(sensor.dsType) + "(" + intToString((int)sensor.dsType) + ")\n";
    out += "\tname: " + sensor.prop.id + ", type: " + intToString(sensor.prop.typeId) + ", default: " + sensor.prop.defaultValue + "\n";
  }
  Logger::getInstance()->log("sensors: \n" + out, lsWarning);
}
#endif

BOOST_FIXTURE_TEST_CASE(lookupSensors, Fixture) {
  std::vector<VdcDb::SensorDesc> sensors;
  sensors = db.getSensors(gtin);
//  dumpSensorDesc(sensors);

  BOOST_REQUIRE_EQUAL(sensors.size(), 1);
  BOOST_CHECK_EQUAL(sensors[0].sensorIndex, "0");
  BOOST_CHECK_EQUAL(sensors[0].prop.id, "dummySensor");
  BOOST_CHECK_EQUAL(sensors[0].prop.typeId, VdcDb::integer);
  BOOST_CHECK_EQUAL(sensors[0].prop.maxValue, "100");
  BOOST_CHECK_EQUAL(sensors[0].prop.minValue, "0");
  BOOST_CHECK_EQUAL(sensors[0].prop.defaultValue, "50");
}

#if 0
static void dumpCategories(const VdcDb::Categories &categories) {
  std::string out;
  for (const VdcDb::Category &category: categories) {
    out += "id: " + category.id + ", order: " + category.orderId;
    out += ", alwaysVisible: " + std::string(category.alwaysVisible ? "true" : "false");
    out += ", compact: " + std::string(category.compact ? "true" : "false");
    out += "\n";
  }
  Logger::getInstance()->log("categories: \n" + out, lsWarning);
}
#endif

BOOST_FIXTURE_TEST_CASE(lookupCategories, Fixture) {
  VdcDb::Categories categories;
  categories = db.getCategories(gtin);
//  dumpCategories(categories);

  BOOST_REQUIRE_EQUAL(categories.size(), 8);
  BOOST_CHECK_EQUAL(categories[0].id, "HandShower");
  BOOST_CHECK_EQUAL(categories[0].orderId, "12");
  BOOST_CHECK_EQUAL(categories[0].alwaysVisible, false);
  BOOST_CHECK_EQUAL(categories[6].id, "catEvent1");
  BOOST_CHECK_EQUAL(categories[6].orderId, "100");
  BOOST_CHECK_EQUAL(categories[6].alwaysVisible, true);
  BOOST_CHECK_EQUAL(categories[7].id, "catEvent2");
  BOOST_CHECK_EQUAL(categories[7].orderId, "100");
  BOOST_CHECK_EQUAL(categories[7].alwaysVisible, false);
}

BOOST_FIXTURE_TEST_CASE(checkNotFound, Fixture) {
  auto invalidGtin = "0000000000000";

  BOOST_CHECK(db.getStates(invalidGtin).empty());
  BOOST_CHECK(db.getSensors(invalidGtin).empty());
  BOOST_CHECK(db.getProperties(invalidGtin).empty());
  BOOST_CHECK(db.getOutputs(invalidGtin).empty());
  BOOST_CHECK(db.getActions(invalidGtin).empty());
  BOOST_CHECK(db.getEvents(invalidGtin).empty());
  BOOST_CHECK(db.getStandardActions(invalidGtin).empty());
}

BOOST_AUTO_TEST_SUITE_END()
